!=====================================================================
!
!          S p e c f e m 3 D  G l o b e  V e r s i o n  7 . 0
!          --------------------------------------------------
!
!     Main historical authors: Dimitri Komatitsch and Jeroen Tromp
!                        Princeton University, USA
!                and CNRS / University of Marseille, France
!                 (there are currently many more authors!)
! (c) Princeton University and CNRS / University of Marseille, April 2014
!
! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 3 of the License, or
! (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
!
!=====================================================================

!--------------------------------------------------------------------------------------------------
! SEA 1D
!
! used as 1-D reference model for SEA 99, Vs model by Lebedev & Nolet 2003
!--------------------------------------------------------------------------------------------------

  module model_sea1d_par

  ! number of layers in DATA/Lebedev_sea99 1D model
  integer, parameter :: NR_SEA1D = 163

  ! model_sea1d_variables
  double precision, dimension(:), allocatable :: &
    SEA1DM_V_radius_sea1d,SEA1DM_V_density_sea1d, &
    SEA1DM_V_vp_sea1d,SEA1DM_V_vs_sea1d, &
    SEA1DM_V_Qkappa_sea1d,SEA1DM_V_Qmu_sea1d

  end module model_sea1d_par

!
!--------------------------------------------------------------------------------------------------
!

  subroutine model_sea1d_broadcast(CRUSTAL)

! standard routine to setup model
  use constants, only: myrank
  use model_sea1d_par

  implicit none

  logical :: CRUSTAL

  ! local parameters
  integer :: ier

  ! allocates arrays
  allocate(SEA1DM_V_radius_sea1d(NR_SEA1D), &
           SEA1DM_V_density_sea1d(NR_SEA1D), &
           SEA1DM_V_vp_sea1d(NR_SEA1D), &
           SEA1DM_V_vs_sea1d(NR_SEA1D), &
           SEA1DM_V_Qkappa_sea1d(NR_SEA1D), &
           SEA1DM_V_Qmu_sea1d(NR_SEA1D), &
           stat=ier)
  if (ier /= 0 ) call exit_MPI(myrank,'Error allocating SEA1DM_V arrays')

  ! all processes will define same parameters
  call define_model_sea1d(CRUSTAL)

  end subroutine model_sea1d_broadcast

!
!-------------------------------------------------------------------------------------------------
!

  subroutine model_sea1d(x,rho,vp,vs,Qkappa,Qmu,iregion_code)

  use constants
  use shared_parameters, only: R_PLANET,RHOAV

  use model_sea1d_par

  implicit none

! input:
! radius r: meters

! output:
! density rho: kg/m^3
! compressional wave speed vp: km/s
! shear wave speed vs: km/s

  double precision :: x,rho,vp,vs,Qmu,Qkappa
  integer :: iregion_code

  ! local parameters
  double precision :: r,frac,scaleval
  integer :: i

!! DK DK implementation of model sea1d below and its radii has not been thoroughly
!! DK DK checked yet

! compute real physical radius in meters
  r = x * R_PLANET

  i = 1
  do while(r >= SEA1DM_V_radius_sea1d(i) .and. i /= NR_SEA1D)
    i = i + 1
  enddo

! make sure we stay in the right region
  if (iregion_code == IREGION_INNER_CORE .and. i > 13) i = 13
  if (iregion_code == IREGION_OUTER_CORE .and. i < 15) i = 15
  if (iregion_code == IREGION_OUTER_CORE .and. i > 37) i = 37
  if (iregion_code == IREGION_CRUST_MANTLE .and. i < 39) i = 39

  if (i == 1) then
    rho = SEA1DM_V_density_sea1d(i)
    vp = SEA1DM_V_vp_sea1d(i)
    vs = SEA1DM_V_vs_sea1d(i)
    Qmu = SEA1DM_V_Qmu_sea1d(i)
    Qkappa = SEA1DM_V_Qkappa_sea1d(i)
  else

! interpolate from SEA1DM_V_radius_sea1d(i-1) to r using the values at i-1 and i
    frac = (r-SEA1DM_V_radius_sea1d(i-1))/(SEA1DM_V_radius_sea1d(i)-SEA1DM_V_radius_sea1d(i-1))

    rho = SEA1DM_V_density_sea1d(i-1) + frac * (SEA1DM_V_density_sea1d(i)-SEA1DM_V_density_sea1d(i-1))
    vp = SEA1DM_V_vp_sea1d(i-1) + frac * (SEA1DM_V_vp_sea1d(i)-SEA1DM_V_vp_sea1d(i-1))
    vs = SEA1DM_V_vs_sea1d(i-1) + frac * (SEA1DM_V_vs_sea1d(i)-SEA1DM_V_vs_sea1d(i-1))
    Qmu = SEA1DM_V_Qmu_sea1d(i-1) + frac * (SEA1DM_V_Qmu_sea1d(i)-SEA1DM_V_Qmu_sea1d(i-1))
    Qkappa = SEA1DM_V_Qkappa_sea1d(i-1) + frac * (SEA1DM_V_Qkappa_sea1d(i)-SEA1DM_V_Qkappa_sea1d(i-1))

  endif

! make sure Vs is zero in the outer core even if roundoff errors on depth
! also set fictitious attenuation to a very high value (attenuation is not used in the fluid)
  if (iregion_code == IREGION_OUTER_CORE) then
    vs = 0.d0
    Qkappa = 3000.d0
    Qmu = 3000.d0
  endif

! non-dimensionalize
! time scaling (s^{-1}) is done with scaleval
  scaleval=dsqrt(PI*GRAV*RHOAV)
  rho=rho*1000.0d0/RHOAV
  vp=vp*1000.0d0/(R_PLANET*scaleval)
  vs=vs*1000.0d0/(R_PLANET*scaleval)

  end subroutine model_sea1d

!-------------------

  subroutine define_model_sea1d(USE_EXTERNAL_CRUSTAL_MODEL)

  use constants
  use model_sea1d_par

  implicit none

  logical :: USE_EXTERNAL_CRUSTAL_MODEL

  ! local parameters
  integer :: i

! define all the values in the model

  SEA1DM_V_radius_sea1d(1)= 0.0000000000
  SEA1DM_V_radius_sea1d(2)= 101425.0000000000
  SEA1DM_V_radius_sea1d(3)= 202850.0000000000
  SEA1DM_V_radius_sea1d(4)= 304275.0000000000
  SEA1DM_V_radius_sea1d(5)= 405700.0000000000
  SEA1DM_V_radius_sea1d(6)= 507125.0000000000
  SEA1DM_V_radius_sea1d(7)= 608550.0000000000
  SEA1DM_V_radius_sea1d(8)= 709975.0000000000
  SEA1DM_V_radius_sea1d(9)= 811400.0000000000
  SEA1DM_V_radius_sea1d(10)= 912825.0000000000
  SEA1DM_V_radius_sea1d(11)= 1014250.0000000000
  SEA1DM_V_radius_sea1d(12)= 1115675.0000000000
  SEA1DM_V_radius_sea1d(13)= 1217100.0000000000
  SEA1DM_V_radius_sea1d(14)= 1217100.0000000000
  SEA1DM_V_radius_sea1d(15)= 1315735.0000000000
  SEA1DM_V_radius_sea1d(16)= 1414370.0000000000
  SEA1DM_V_radius_sea1d(17)= 1513004.0000000000
  SEA1DM_V_radius_sea1d(18)= 1611639.0000000000
  SEA1DM_V_radius_sea1d(19)= 1710274.0000000000
  SEA1DM_V_radius_sea1d(20)= 1808909.0000000000
  SEA1DM_V_radius_sea1d(21)= 1907544.0000000000
  SEA1DM_V_radius_sea1d(22)= 2006178.0000000000
  SEA1DM_V_radius_sea1d(23)= 2104813.0000000000
  SEA1DM_V_radius_sea1d(24)= 2203448.0000000000
  SEA1DM_V_radius_sea1d(25)= 2302082.0000000000
  SEA1DM_V_radius_sea1d(26)= 2400717.0000000000
  SEA1DM_V_radius_sea1d(27)= 2499352.0000000000
  SEA1DM_V_radius_sea1d(28)= 2597987.0000000000
  SEA1DM_V_radius_sea1d(29)= 2696622.0000000000
  SEA1DM_V_radius_sea1d(30)= 2795256.0000000000
  SEA1DM_V_radius_sea1d(31)= 2893891.0000000000
  SEA1DM_V_radius_sea1d(32)= 2992526.0000000000
  SEA1DM_V_radius_sea1d(33)= 3091161.0000000000
  SEA1DM_V_radius_sea1d(34)= 3189796.0000000000
  SEA1DM_V_radius_sea1d(35)= 3288431.0000000000
  SEA1DM_V_radius_sea1d(36)= 3387066.0000000000
  SEA1DM_V_radius_sea1d(37)= 3485700.0000000000
  SEA1DM_V_radius_sea1d(38)= 3485700.0000000000
  SEA1DM_V_radius_sea1d(39)= 3536048.0000000000
  SEA1DM_V_radius_sea1d(40)= 3586396.0000000000
  SEA1DM_V_radius_sea1d(41)= 3636743.0000000000
  SEA1DM_V_radius_sea1d(42)= 3687091.0000000000
  SEA1DM_V_radius_sea1d(43)= 3737438.0000000000
  SEA1DM_V_radius_sea1d(44)= 3787786.0000000000
  SEA1DM_V_radius_sea1d(45)= 3838134.0000000000
  SEA1DM_V_radius_sea1d(46)= 3888482.0000000000
  SEA1DM_V_radius_sea1d(47)= 3938830.0000000000
  SEA1DM_V_radius_sea1d(48)= 3989177.0000000000
  SEA1DM_V_radius_sea1d(49)= 4039525.0000000000
  SEA1DM_V_radius_sea1d(50)= 4089872.0000000000
  SEA1DM_V_radius_sea1d(51)= 4140220.0000000000
  SEA1DM_V_radius_sea1d(52)= 4190568.0000000000
  SEA1DM_V_radius_sea1d(53)= 4240916.0000000000
  SEA1DM_V_radius_sea1d(54)= 4291264.0000000000
  SEA1DM_V_radius_sea1d(55)= 4341612.0000000000
  SEA1DM_V_radius_sea1d(56)= 4391959.0000000000
  SEA1DM_V_radius_sea1d(57)= 4442306.0000000000
  SEA1DM_V_radius_sea1d(58)= 4492654.0000000000
  SEA1DM_V_radius_sea1d(59)= 4543002.0000000000
  SEA1DM_V_radius_sea1d(60)= 4593350.0000000000
  SEA1DM_V_radius_sea1d(61)= 4643698.0000000000
  SEA1DM_V_radius_sea1d(62)= 4694046.0000000000
  SEA1DM_V_radius_sea1d(63)= 4744393.0000000000
  SEA1DM_V_radius_sea1d(64)= 4794740.0000000000
  SEA1DM_V_radius_sea1d(65)= 4845089.0000000000
  SEA1DM_V_radius_sea1d(66)= 4895436.0000000000
  SEA1DM_V_radius_sea1d(67)= 4945784.0000000000
  SEA1DM_V_radius_sea1d(68)= 4996132.0000000000
  SEA1DM_V_radius_sea1d(69)= 5046480.0000000000
  SEA1DM_V_radius_sea1d(70)= 5096827.0000000000
  SEA1DM_V_radius_sea1d(71)= 5147175.0000000000
  SEA1DM_V_radius_sea1d(72)= 5197522.0000000000
  SEA1DM_V_radius_sea1d(73)= 5247870.0000000000
  SEA1DM_V_radius_sea1d(74)= 5298218.0000000000
  SEA1DM_V_radius_sea1d(75)= 5348566.0000000000
  SEA1DM_V_radius_sea1d(76)= 5398914.0000000000
  SEA1DM_V_radius_sea1d(77)= 5449261.0000000000
  SEA1DM_V_radius_sea1d(78)= 5499610.0000000000
  SEA1DM_V_radius_sea1d(79)= 5549957.0000000000
  SEA1DM_V_radius_sea1d(80)= 5600304.0000000000
  SEA1DM_V_radius_sea1d(81)= 5650652.0000000000
  SEA1DM_V_radius_sea1d(82)= 5701000.0000000000
  SEA1DM_V_radius_sea1d(83)= 5711000.0000000000
  SEA1DM_V_radius_sea1d(84)= 5711000.0000000000
  SEA1DM_V_radius_sea1d(85)= 5721000.0000000000
  SEA1DM_V_radius_sea1d(86)= 5731000.0000000000
  SEA1DM_V_radius_sea1d(87)= 5741000.0000000000
  SEA1DM_V_radius_sea1d(88)= 5751000.0000000000
  SEA1DM_V_radius_sea1d(89)= 5761000.0000000000
  SEA1DM_V_radius_sea1d(90)= 5771000.0000000000
  SEA1DM_V_radius_sea1d(91)= 5781000.0000000000
  SEA1DM_V_radius_sea1d(92)= 5791000.0000000000
  SEA1DM_V_radius_sea1d(93)= 5801000.0000000000
  SEA1DM_V_radius_sea1d(94)= 5811000.0000000000
  SEA1DM_V_radius_sea1d(95)= 5821000.0000000000
  SEA1DM_V_radius_sea1d(96)= 5831000.0000000000
  SEA1DM_V_radius_sea1d(97)= 5841000.0000000000
  SEA1DM_V_radius_sea1d(98)= 5851000.0000000000
  SEA1DM_V_radius_sea1d(99)= 5861000.0000000000
  SEA1DM_V_radius_sea1d(100)= 5871000.0000000000
  SEA1DM_V_radius_sea1d(101)= 5881000.0000000000
  SEA1DM_V_radius_sea1d(102)= 5891000.0000000000
  SEA1DM_V_radius_sea1d(103)= 5901000.0000000000
  SEA1DM_V_radius_sea1d(104)= 5911000.0000000000
  SEA1DM_V_radius_sea1d(105)= 5921000.0000000000
  SEA1DM_V_radius_sea1d(106)= 5931000.0000000000
  SEA1DM_V_radius_sea1d(107)= 5941000.0000000000
  SEA1DM_V_radius_sea1d(108)= 5951000.0000000000
  SEA1DM_V_radius_sea1d(109)= 5961000.0000000000
  SEA1DM_V_radius_sea1d(110)= 5961000.0000000000
  SEA1DM_V_radius_sea1d(111)= 5971000.0000000000
  SEA1DM_V_radius_sea1d(112)= 5981000.0000000000
  SEA1DM_V_radius_sea1d(113)= 5991000.0000000000
  SEA1DM_V_radius_sea1d(114)= 6001000.0000000000
  SEA1DM_V_radius_sea1d(115)= 6011000.0000000000
  SEA1DM_V_radius_sea1d(116)= 6021000.0000000000
  SEA1DM_V_radius_sea1d(117)= 6031000.0000000000
  SEA1DM_V_radius_sea1d(118)= 6041000.0000000000
  SEA1DM_V_radius_sea1d(119)= 6051000.0000000000
  SEA1DM_V_radius_sea1d(120)= 6061000.0000000000
  SEA1DM_V_radius_sea1d(121)= 6071000.0000000000
  SEA1DM_V_radius_sea1d(122)= 6081000.0000000000
  SEA1DM_V_radius_sea1d(123)= 6091000.0000000000
  SEA1DM_V_radius_sea1d(124)= 6101000.0000000000
  SEA1DM_V_radius_sea1d(125)= 6111000.0000000000
  SEA1DM_V_radius_sea1d(126)= 6121000.0000000000
  SEA1DM_V_radius_sea1d(127)= 6131000.0000000000
  SEA1DM_V_radius_sea1d(128)= 6141000.0000000000
  SEA1DM_V_radius_sea1d(129)= 6151000.0000000000
  SEA1DM_V_radius_sea1d(130)= 6161000.0000000000
  SEA1DM_V_radius_sea1d(131)= 6171000.0000000000
  SEA1DM_V_radius_sea1d(132)= 6181000.0000000000
  SEA1DM_V_radius_sea1d(133)= 6191000.0000000000
  SEA1DM_V_radius_sea1d(134)= 6201000.0000000000
  SEA1DM_V_radius_sea1d(135)= 6211000.0000000000
  SEA1DM_V_radius_sea1d(136)= 6221000.0000000000
  SEA1DM_V_radius_sea1d(137)= 6231000.0000000000
  SEA1DM_V_radius_sea1d(138)= 6241000.0000000000
  SEA1DM_V_radius_sea1d(139)= 6251000.0000000000
  SEA1DM_V_radius_sea1d(140)= 6261000.0000000000
  SEA1DM_V_radius_sea1d(141)= 6271000.0000000000
  SEA1DM_V_radius_sea1d(142)= 6281000.0000000000
  SEA1DM_V_radius_sea1d(143)= 6291000.0000000000
  SEA1DM_V_radius_sea1d(144)= 6301000.0000000000
  SEA1DM_V_radius_sea1d(145)= 6311000.0000000000
  SEA1DM_V_radius_sea1d(146)= 6321000.0000000000
  SEA1DM_V_radius_sea1d(147)= 6326000.0000000000
  SEA1DM_V_radius_sea1d(148)= 6331000.0000000000
  SEA1DM_V_radius_sea1d(149)= 6336000.0000000000
  SEA1DM_V_radius_sea1d(150)= 6341000.0000000000
  SEA1DM_V_radius_sea1d(151)= 6346000.0000000000
  SEA1DM_V_radius_sea1d(152)= 6346000.0000000000
  SEA1DM_V_radius_sea1d(153)= 6351000.0000000000
  SEA1DM_V_radius_sea1d(154)= 6353800.0000000000
  SEA1DM_V_radius_sea1d(155)= 6356600.0000000000
  SEA1DM_V_radius_sea1d(156)= 6360000.0000000000
  SEA1DM_V_radius_sea1d(157)= 6363000.0000000000
  SEA1DM_V_radius_sea1d(158)= 6365000.0000000000
  SEA1DM_V_radius_sea1d(159)= 6366000.0000000000
  SEA1DM_V_radius_sea1d(160)= 6366000.0000000000
  SEA1DM_V_radius_sea1d(161)= 6368000.0000000000
  SEA1DM_V_radius_sea1d(162)= 6368000.0000000000
  SEA1DM_V_radius_sea1d(163)= 6371000.0000000000

  SEA1DM_V_density_sea1d(1)= 13.0121900000000
  SEA1DM_V_density_sea1d(2)= 13.0100200000000
  SEA1DM_V_density_sea1d(3)= 13.0035600000000
  SEA1DM_V_density_sea1d(4)= 12.9928300000000
  SEA1DM_V_density_sea1d(5)= 12.9778000000000
  SEA1DM_V_density_sea1d(6)= 12.9585000000000
  SEA1DM_V_density_sea1d(7)= 12.9349100000000
  SEA1DM_V_density_sea1d(8)= 12.9070300000000
  SEA1DM_V_density_sea1d(9)= 12.8748700000000
  SEA1DM_V_density_sea1d(10)= 12.8384300000000
  SEA1DM_V_density_sea1d(11)= 12.7977100000000
  SEA1DM_V_density_sea1d(12)= 12.7526900000000
  SEA1DM_V_density_sea1d(13)= 12.7037000000000
  SEA1DM_V_density_sea1d(14)= 12.1391000000000
  SEA1DM_V_density_sea1d(15)= 12.0877600000000
  SEA1DM_V_density_sea1d(16)= 12.0333900000000
  SEA1DM_V_density_sea1d(17)= 11.9757900000000
  SEA1DM_V_density_sea1d(18)= 11.9148500000000
  SEA1DM_V_density_sea1d(19)= 11.8503900000000
  SEA1DM_V_density_sea1d(20)= 11.7822500000000
  SEA1DM_V_density_sea1d(21)= 11.7102700000000
  SEA1DM_V_density_sea1d(22)= 11.6343000000000
  SEA1DM_V_density_sea1d(23)= 11.5541800000000
  SEA1DM_V_density_sea1d(24)= 11.4697400000000
  SEA1DM_V_density_sea1d(25)= 11.3808400000000
  SEA1DM_V_density_sea1d(26)= 11.2873100000000
  SEA1DM_V_density_sea1d(27)= 11.1890000000000
  SEA1DM_V_density_sea1d(28)= 11.0857400000000
  SEA1DM_V_density_sea1d(29)= 10.9773800000000
  SEA1DM_V_density_sea1d(30)= 10.8637600000000
  SEA1DM_V_density_sea1d(31)= 10.7447200000000
  SEA1DM_V_density_sea1d(32)= 10.6201000000000
  SEA1DM_V_density_sea1d(33)= 10.4897500000000
  SEA1DM_V_density_sea1d(34)= 10.3535000000000
  SEA1DM_V_density_sea1d(35)= 10.2112100000000
  SEA1DM_V_density_sea1d(36)= 10.0627000000000
  SEA1DM_V_density_sea1d(37)= 9.9085500000000
  SEA1DM_V_density_sea1d(38)= 5.5497800000000
  SEA1DM_V_density_sea1d(39)= 5.5263200000000
  SEA1DM_V_density_sea1d(40)= 5.5027000000000
  SEA1DM_V_density_sea1d(41)= 5.4789400000000
  SEA1DM_V_density_sea1d(42)= 5.4550400000000
  SEA1DM_V_density_sea1d(43)= 5.4309700000000
  SEA1DM_V_density_sea1d(44)= 5.4067700000000
  SEA1DM_V_density_sea1d(45)= 5.3824200000000
  SEA1DM_V_density_sea1d(46)= 5.3579200000000
  SEA1DM_V_density_sea1d(47)= 5.3332700000000
  SEA1DM_V_density_sea1d(48)= 5.3084700000000
  SEA1DM_V_density_sea1d(49)= 5.2835200000000
  SEA1DM_V_density_sea1d(50)= 5.2584400000000
  SEA1DM_V_density_sea1d(51)= 5.2331900000000
  SEA1DM_V_density_sea1d(52)= 5.2078000000000
  SEA1DM_V_density_sea1d(53)= 5.1822700000000
  SEA1DM_V_density_sea1d(54)= 5.1565900000000
  SEA1DM_V_density_sea1d(55)= 5.1307500000000
  SEA1DM_V_density_sea1d(56)= 5.1047600000000
  SEA1DM_V_density_sea1d(57)= 5.0786400000000
  SEA1DM_V_density_sea1d(58)= 5.0523600000000
  SEA1DM_V_density_sea1d(59)= 5.0259400000000
  SEA1DM_V_density_sea1d(60)= 4.9993600000000
  SEA1DM_V_density_sea1d(61)= 4.9726500000000
  SEA1DM_V_density_sea1d(62)= 4.9457800000000
  SEA1DM_V_density_sea1d(63)= 4.9187500000000
  SEA1DM_V_density_sea1d(64)= 4.8915900000000
  SEA1DM_V_density_sea1d(65)= 4.8642700000000
  SEA1DM_V_density_sea1d(66)= 4.8368200000000
  SEA1DM_V_density_sea1d(67)= 4.8092100000000
  SEA1DM_V_density_sea1d(68)= 4.7814400000000
  SEA1DM_V_density_sea1d(69)= 4.7535400000000
  SEA1DM_V_density_sea1d(70)= 4.7254900000000
  SEA1DM_V_density_sea1d(71)= 4.6972900000000
  SEA1DM_V_density_sea1d(72)= 4.6689400000000
  SEA1DM_V_density_sea1d(73)= 4.6404400000000
  SEA1DM_V_density_sea1d(74)= 4.6117900000000
  SEA1DM_V_density_sea1d(75)= 4.5830000000000
  SEA1DM_V_density_sea1d(76)= 4.5540600000000
  SEA1DM_V_density_sea1d(77)= 4.5249700000000
  SEA1DM_V_density_sea1d(78)= 4.4957300000000
  SEA1DM_V_density_sea1d(79)= 4.4663500000000
  SEA1DM_V_density_sea1d(80)= 4.4368100000000
  SEA1DM_V_density_sea1d(81)= 4.4071300000000
  SEA1DM_V_density_sea1d(82)= 4.3773100000000
  SEA1DM_V_density_sea1d(83)= 4.3713900000000
  SEA1DM_V_density_sea1d(84)= 4.0645800000000
  SEA1DM_V_density_sea1d(85)= 4.0522200000000
  SEA1DM_V_density_sea1d(86)= 4.0398700000000
  SEA1DM_V_density_sea1d(87)= 4.0275200000000
  SEA1DM_V_density_sea1d(88)= 4.0151600000000
  SEA1DM_V_density_sea1d(89)= 4.0028100000000
  SEA1DM_V_density_sea1d(90)= 3.9904500000000
  SEA1DM_V_density_sea1d(91)= 3.9781000000000
  SEA1DM_V_density_sea1d(92)= 3.9657500000000
  SEA1DM_V_density_sea1d(93)= 3.9533900000000
  SEA1DM_V_density_sea1d(94)= 3.9410400000000
  SEA1DM_V_density_sea1d(95)= 3.9286900000000
  SEA1DM_V_density_sea1d(96)= 3.9163300000000
  SEA1DM_V_density_sea1d(97)= 3.9039800000000
  SEA1DM_V_density_sea1d(98)= 3.8916200000000
  SEA1DM_V_density_sea1d(99)= 3.8792700000000
  SEA1DM_V_density_sea1d(100)= 3.8669200000000
  SEA1DM_V_density_sea1d(101)= 3.8545600000000
  SEA1DM_V_density_sea1d(102)= 3.8422100000000
  SEA1DM_V_density_sea1d(103)= 3.8298600000000
  SEA1DM_V_density_sea1d(104)= 3.8175000000000
  SEA1DM_V_density_sea1d(105)= 3.8051500000000
  SEA1DM_V_density_sea1d(106)= 3.7928000000000
  SEA1DM_V_density_sea1d(107)= 3.7804400000000
  SEA1DM_V_density_sea1d(108)= 3.7680900000000
  SEA1DM_V_density_sea1d(109)= 3.7557300000000
  SEA1DM_V_density_sea1d(110)= 3.5469600000000
  SEA1DM_V_density_sea1d(111)= 3.5409000000000
  SEA1DM_V_density_sea1d(112)= 3.5348400000000
  SEA1DM_V_density_sea1d(113)= 3.5287900000000
  SEA1DM_V_density_sea1d(114)= 3.5227300000000
  SEA1DM_V_density_sea1d(115)= 3.5166700000000
  SEA1DM_V_density_sea1d(116)= 3.5106100000000
  SEA1DM_V_density_sea1d(117)= 3.5045500000000
  SEA1DM_V_density_sea1d(118)= 3.4984900000000
  SEA1DM_V_density_sea1d(119)= 3.4924300000000
  SEA1DM_V_density_sea1d(120)= 3.4863800000000
  SEA1DM_V_density_sea1d(121)= 3.4803200000000
  SEA1DM_V_density_sea1d(122)= 3.4742600000000
  SEA1DM_V_density_sea1d(123)= 3.4682000000000
  SEA1DM_V_density_sea1d(124)= 3.4621400000000
  SEA1DM_V_density_sea1d(125)= 3.4560800000000
  SEA1DM_V_density_sea1d(126)= 3.4500200000000
  SEA1DM_V_density_sea1d(127)= 3.4439700000000
  SEA1DM_V_density_sea1d(128)= 3.4379100000000
  SEA1DM_V_density_sea1d(129)= 3.4318500000000
  SEA1DM_V_density_sea1d(130)= 3.4257900000000
  SEA1DM_V_density_sea1d(131)= 3.4197300000000
  SEA1DM_V_density_sea1d(132)= 3.4136800000000
  SEA1DM_V_density_sea1d(133)= 3.4076200000000
  SEA1DM_V_density_sea1d(134)= 3.4015600000000
  SEA1DM_V_density_sea1d(135)= 3.3955000000000
  SEA1DM_V_density_sea1d(136)= 3.3894400000000
  SEA1DM_V_density_sea1d(137)= 3.3833800000000
  SEA1DM_V_density_sea1d(138)= 3.3773200000000
  SEA1DM_V_density_sea1d(139)= 3.3712600000000
  SEA1DM_V_density_sea1d(140)= 3.3652100000000
  SEA1DM_V_density_sea1d(141)= 3.3591500000000
  SEA1DM_V_density_sea1d(142)= 3.3530900000000
  SEA1DM_V_density_sea1d(143)= 3.3470300000000
  SEA1DM_V_density_sea1d(144)= 3.3409700000000
  SEA1DM_V_density_sea1d(145)= 3.3349100000000
  SEA1DM_V_density_sea1d(146)= 3.3288500000000
  SEA1DM_V_density_sea1d(147)= 3.3288500000000
  SEA1DM_V_density_sea1d(148)= 3.3227900000000
  SEA1DM_V_density_sea1d(149)= 3.3227900000000
  SEA1DM_V_density_sea1d(150)= 3.3227900000000
  SEA1DM_V_density_sea1d(151)= 3.3227900000000
  SEA1DM_V_density_sea1d(152)= 2.8500000000000
  SEA1DM_V_density_sea1d(153)= 2.8500000000000
  SEA1DM_V_density_sea1d(154)= 2.8500000000000
  SEA1DM_V_density_sea1d(155)= 2.8500000000000
  SEA1DM_V_density_sea1d(156)= 2.8500000000000
  SEA1DM_V_density_sea1d(157)= 2.8500000000000
  SEA1DM_V_density_sea1d(158)= 2.8500000000000
  SEA1DM_V_density_sea1d(159)= 2.8500000000000
  SEA1DM_V_density_sea1d(160)= 2.8500000000000
  SEA1DM_V_density_sea1d(161)= 2.8500000000000
  SEA1DM_V_density_sea1d(162)= 2.8500000000000
  SEA1DM_V_density_sea1d(163)= 2.8500000000000

  SEA1DM_V_vp_sea1d(1)= 11.2409400000000
  SEA1DM_V_vp_sea1d(2)= 11.2398900000000
  SEA1DM_V_vp_sea1d(3)= 11.2367600000000
  SEA1DM_V_vp_sea1d(4)= 11.2315600000000
  SEA1DM_V_vp_sea1d(5)= 11.2242700000000
  SEA1DM_V_vp_sea1d(6)= 11.2149200000000
  SEA1DM_V_vp_sea1d(7)= 11.2034800000000
  SEA1DM_V_vp_sea1d(8)= 11.1899700000000
  SEA1DM_V_vp_sea1d(9)= 11.1743800000000
  SEA1DM_V_vp_sea1d(10)= 11.1567200000000
  SEA1DM_V_vp_sea1d(11)= 11.1369900000000
  SEA1DM_V_vp_sea1d(12)= 11.1151700000000
  SEA1DM_V_vp_sea1d(13)= 11.0914200000000
  SEA1DM_V_vp_sea1d(14)= 10.2577900000000
  SEA1DM_V_vp_sea1d(15)= 10.2317700000000
  SEA1DM_V_vp_sea1d(16)= 10.1991900000000
  SEA1DM_V_vp_sea1d(17)= 10.1600600000000
  SEA1DM_V_vp_sea1d(18)= 10.1143700000000
  SEA1DM_V_vp_sea1d(19)= 10.0621400000000
  SEA1DM_V_vp_sea1d(20)= 10.0033600000000
  SEA1DM_V_vp_sea1d(21)= 9.9380100000000
  SEA1DM_V_vp_sea1d(22)= 9.8661300000000
  SEA1DM_V_vp_sea1d(23)= 9.7876800000000
  SEA1DM_V_vp_sea1d(24)= 9.7026900000000
  SEA1DM_V_vp_sea1d(25)= 9.6111500000000
  SEA1DM_V_vp_sea1d(26)= 9.5130500000000
  SEA1DM_V_vp_sea1d(27)= 9.4084000000000
  SEA1DM_V_vp_sea1d(28)= 9.2972000000000
  SEA1DM_V_vp_sea1d(29)= 9.1794500000000
  SEA1DM_V_vp_sea1d(30)= 9.0551400000000
  SEA1DM_V_vp_sea1d(31)= 8.9242800000000
  SEA1DM_V_vp_sea1d(32)= 8.7868700000000
  SEA1DM_V_vp_sea1d(33)= 8.6429000000000
  SEA1DM_V_vp_sea1d(34)= 8.4923900000000
  SEA1DM_V_vp_sea1d(35)= 8.3353300000000
  SEA1DM_V_vp_sea1d(36)= 8.1717000000000
  SEA1DM_V_vp_sea1d(37)= 8.0022600000000
  SEA1DM_V_vp_sea1d(38)= 13.7318200000000
  SEA1DM_V_vp_sea1d(39)= 13.6839600000000
  SEA1DM_V_vp_sea1d(40)= 13.6355700000000
  SEA1DM_V_vp_sea1d(41)= 13.5866700000000
  SEA1DM_V_vp_sea1d(42)= 13.5372000000000
  SEA1DM_V_vp_sea1d(43)= 13.4871700000000
  SEA1DM_V_vp_sea1d(44)= 13.4365700000000
  SEA1DM_V_vp_sea1d(45)= 13.3853700000000
  SEA1DM_V_vp_sea1d(46)= 13.3335400000000
  SEA1DM_V_vp_sea1d(47)= 13.2811000000000
  SEA1DM_V_vp_sea1d(48)= 13.2280100000000
  SEA1DM_V_vp_sea1d(49)= 13.1742700000000
  SEA1DM_V_vp_sea1d(50)= 13.1198500000000
  SEA1DM_V_vp_sea1d(51)= 13.0647300000000
  SEA1DM_V_vp_sea1d(52)= 13.0089100000000
  SEA1DM_V_vp_sea1d(53)= 12.9523700000000
  SEA1DM_V_vp_sea1d(54)= 12.8951000000000
  SEA1DM_V_vp_sea1d(55)= 12.8370600000000
  SEA1DM_V_vp_sea1d(56)= 12.7782600000000
  SEA1DM_V_vp_sea1d(57)= 12.7186700000000
  SEA1DM_V_vp_sea1d(58)= 12.6582800000000
  SEA1DM_V_vp_sea1d(59)= 12.5970700000000
  SEA1DM_V_vp_sea1d(60)= 12.5350400000000
  SEA1DM_V_vp_sea1d(61)= 12.4721600000000
  SEA1DM_V_vp_sea1d(62)= 12.4084000000000
  SEA1DM_V_vp_sea1d(63)= 12.3437700000000
  SEA1DM_V_vp_sea1d(64)= 12.2782500000000
  SEA1DM_V_vp_sea1d(65)= 12.2118200000000
  SEA1DM_V_vp_sea1d(66)= 12.1444600000000
  SEA1DM_V_vp_sea1d(67)= 12.0761600000000
  SEA1DM_V_vp_sea1d(68)= 12.0069000000000
  SEA1DM_V_vp_sea1d(69)= 11.9366700000000
  SEA1DM_V_vp_sea1d(70)= 11.8654400000000
  SEA1DM_V_vp_sea1d(71)= 11.7932100000000
  SEA1DM_V_vp_sea1d(72)= 11.7199700000000
  SEA1DM_V_vp_sea1d(73)= 11.6456800000000
  SEA1DM_V_vp_sea1d(74)= 11.5703400000000
  SEA1DM_V_vp_sea1d(75)= 11.4939400000000
  SEA1DM_V_vp_sea1d(76)= 11.4164500000000
  SEA1DM_V_vp_sea1d(77)= 11.3378700000000
  SEA1DM_V_vp_sea1d(78)= 11.2581700000000
  SEA1DM_V_vp_sea1d(79)= 11.1773300000000
  SEA1DM_V_vp_sea1d(80)= 11.0953600000000
  SEA1DM_V_vp_sea1d(81)= 11.0122200000000
  SEA1DM_V_vp_sea1d(82)= 10.9280200000000
  SEA1DM_V_vp_sea1d(83)= 10.9113000000000
  SEA1DM_V_vp_sea1d(84)= 10.0182900000000
  SEA1DM_V_vp_sea1d(85)= 9.9989600000000
  SEA1DM_V_vp_sea1d(86)= 9.9796300000000
  SEA1DM_V_vp_sea1d(87)= 9.9603000000000
  SEA1DM_V_vp_sea1d(88)= 9.9409700000000
  SEA1DM_V_vp_sea1d(89)= 9.9216400000000
  SEA1DM_V_vp_sea1d(90)= 9.9023100000000
  SEA1DM_V_vp_sea1d(91)= 9.8829800000000
  SEA1DM_V_vp_sea1d(92)= 9.8636600000000
  SEA1DM_V_vp_sea1d(93)= 9.8443300000000
  SEA1DM_V_vp_sea1d(94)= 9.8250000000000
  SEA1DM_V_vp_sea1d(95)= 9.8056700000000
  SEA1DM_V_vp_sea1d(96)= 9.7863400000000
  SEA1DM_V_vp_sea1d(97)= 9.7670100000000
  SEA1DM_V_vp_sea1d(98)= 9.7476800000000
  SEA1DM_V_vp_sea1d(99)= 9.7283500000000
  SEA1DM_V_vp_sea1d(100)= 9.7090300000000
  SEA1DM_V_vp_sea1d(101)= 9.6897000000000
  SEA1DM_V_vp_sea1d(102)= 9.6703700000000
  SEA1DM_V_vp_sea1d(103)= 9.6510400000000
  SEA1DM_V_vp_sea1d(104)= 9.6317100000000
  SEA1DM_V_vp_sea1d(105)= 9.6123800000000
  SEA1DM_V_vp_sea1d(106)= 9.5930500000000
  SEA1DM_V_vp_sea1d(107)= 9.5737200000000
  SEA1DM_V_vp_sea1d(108)= 9.5543900000000
  SEA1DM_V_vp_sea1d(109)= 9.5350600000000
  SEA1DM_V_vp_sea1d(110)= 9.0766800000000
  SEA1DM_V_vp_sea1d(111)= 9.0188500000000
  SEA1DM_V_vp_sea1d(112)= 8.9610200000000
  SEA1DM_V_vp_sea1d(113)= 8.9031800000000
  SEA1DM_V_vp_sea1d(114)= 8.8453500000000
  SEA1DM_V_vp_sea1d(115)= 8.7875100000000
  SEA1DM_V_vp_sea1d(116)= 8.7296800000000
  SEA1DM_V_vp_sea1d(117)= 8.6718500000000
  SEA1DM_V_vp_sea1d(118)= 8.6140100000000
  SEA1DM_V_vp_sea1d(119)= 8.5561800000000
  SEA1DM_V_vp_sea1d(120)= 8.4983400000000
  SEA1DM_V_vp_sea1d(121)= 8.4405100000000
  SEA1DM_V_vp_sea1d(122)= 8.3826700000000
  SEA1DM_V_vp_sea1d(123)= 8.3248400000000
  SEA1DM_V_vp_sea1d(124)= 8.2670100000000
  SEA1DM_V_vp_sea1d(125)= 8.2091700000000
  SEA1DM_V_vp_sea1d(126)= 8.1513400000000
  SEA1DM_V_vp_sea1d(127)= 8.0935000000000
  SEA1DM_V_vp_sea1d(128)= 8.0356700000000
  SEA1DM_V_vp_sea1d(129)= 7.9778300000000
  SEA1DM_V_vp_sea1d(130)= 7.9200000000000
  SEA1DM_V_vp_sea1d(131)= 7.9200000000000
  SEA1DM_V_vp_sea1d(132)= 7.9200000000000
  SEA1DM_V_vp_sea1d(133)= 7.9200000000000
  SEA1DM_V_vp_sea1d(134)= 7.9200000000000
  SEA1DM_V_vp_sea1d(135)= 7.9200000000000
  SEA1DM_V_vp_sea1d(136)= 7.9200000000000
  SEA1DM_V_vp_sea1d(137)= 7.9200000000000
  SEA1DM_V_vp_sea1d(138)= 7.9200000000000
  SEA1DM_V_vp_sea1d(139)= 7.9200000000000
  SEA1DM_V_vp_sea1d(140)= 7.9200000000000
  SEA1DM_V_vp_sea1d(141)= 7.9200000000000
  SEA1DM_V_vp_sea1d(142)= 7.9200000000000
  SEA1DM_V_vp_sea1d(143)= 7.9200000000000
  SEA1DM_V_vp_sea1d(144)= 7.9200000000000
  SEA1DM_V_vp_sea1d(145)= 7.9200000000000
  SEA1DM_V_vp_sea1d(146)= 7.9200000000000
  SEA1DM_V_vp_sea1d(147)= 7.9200000000000
  SEA1DM_V_vp_sea1d(148)= 7.9200000000000
  SEA1DM_V_vp_sea1d(149)= 7.9200000000000
  SEA1DM_V_vp_sea1d(150)= 7.9200000000000
  SEA1DM_V_vp_sea1d(151)= 7.9200000000000
  SEA1DM_V_vp_sea1d(152)= 6.4000000000000
  SEA1DM_V_vp_sea1d(153)= 6.4000000000000
  SEA1DM_V_vp_sea1d(154)= 6.4000000000000
  SEA1DM_V_vp_sea1d(155)= 6.4000000000000
  SEA1DM_V_vp_sea1d(156)= 6.4000000000000
  SEA1DM_V_vp_sea1d(157)= 6.4000000000000
  SEA1DM_V_vp_sea1d(158)= 6.4000000000000
  SEA1DM_V_vp_sea1d(159)= 6.4000000000000
  SEA1DM_V_vp_sea1d(160)= 6.4000000000000
  SEA1DM_V_vp_sea1d(161)= 6.4000000000000
  SEA1DM_V_vp_sea1d(162)= 6.4000000000000
  SEA1DM_V_vp_sea1d(163)= 6.4000000000000

  SEA1DM_V_vs_sea1d(1)= 3.5645400000000
  SEA1DM_V_vs_sea1d(2)= 3.5636500000000
  SEA1DM_V_vs_sea1d(3)= 3.5610200000000
  SEA1DM_V_vs_sea1d(4)= 3.5566300000000
  SEA1DM_V_vs_sea1d(5)= 3.5504900000000
  SEA1DM_V_vs_sea1d(6)= 3.5426100000000
  SEA1DM_V_vs_sea1d(7)= 3.5329700000000
  SEA1DM_V_vs_sea1d(8)= 3.5215900000000
  SEA1DM_V_vs_sea1d(9)= 3.5084500000000
  SEA1DM_V_vs_sea1d(10)= 3.4935700000000
  SEA1DM_V_vs_sea1d(11)= 3.4769300000000
  SEA1DM_V_vs_sea1d(12)= 3.4585500000000
  SEA1DM_V_vs_sea1d(13)= 3.4385400000000
  SEA1DM_V_vs_sea1d(14)= 0.0000000000000
  SEA1DM_V_vs_sea1d(15)= 0.0000000000000
  SEA1DM_V_vs_sea1d(16)= 0.0000000000000
  SEA1DM_V_vs_sea1d(17)= 0.0000000000000
  SEA1DM_V_vs_sea1d(18)= 0.0000000000000
  SEA1DM_V_vs_sea1d(19)= 0.0000000000000
  SEA1DM_V_vs_sea1d(20)= 0.0000000000000
  SEA1DM_V_vs_sea1d(21)= 0.0000000000000
  SEA1DM_V_vs_sea1d(22)= 0.0000000000000
  SEA1DM_V_vs_sea1d(23)= 0.0000000000000
  SEA1DM_V_vs_sea1d(24)= 0.0000000000000
  SEA1DM_V_vs_sea1d(25)= 0.0000000000000
  SEA1DM_V_vs_sea1d(26)= 0.0000000000000
  SEA1DM_V_vs_sea1d(27)= 0.0000000000000
  SEA1DM_V_vs_sea1d(28)= 0.0000000000000
  SEA1DM_V_vs_sea1d(29)= 0.0000000000000
  SEA1DM_V_vs_sea1d(30)= 0.0000000000000
  SEA1DM_V_vs_sea1d(31)= 0.0000000000000
  SEA1DM_V_vs_sea1d(32)= 0.0000000000000
  SEA1DM_V_vs_sea1d(33)= 0.0000000000000
  SEA1DM_V_vs_sea1d(34)= 0.0000000000000
  SEA1DM_V_vs_sea1d(35)= 0.0000000000000
  SEA1DM_V_vs_sea1d(36)= 0.0000000000000
  SEA1DM_V_vs_sea1d(37)= 0.0000000000000
  SEA1DM_V_vs_sea1d(38)= 7.2433800000000
  SEA1DM_V_vs_sea1d(39)= 7.2260300000000
  SEA1DM_V_vs_sea1d(40)= 7.2085500000000
  SEA1DM_V_vs_sea1d(41)= 7.1909200000000
  SEA1DM_V_vs_sea1d(42)= 7.1731300000000
  SEA1DM_V_vs_sea1d(43)= 7.1551600000000
  SEA1DM_V_vs_sea1d(44)= 7.1370000000000
  SEA1DM_V_vs_sea1d(45)= 7.1186000000000
  SEA1DM_V_vs_sea1d(46)= 7.0999800000000
  SEA1DM_V_vs_sea1d(47)= 7.0810900000000
  SEA1DM_V_vs_sea1d(48)= 7.0619300000000
  SEA1DM_V_vs_sea1d(49)= 7.0424700000000
  SEA1DM_V_vs_sea1d(50)= 7.0227000000000
  SEA1DM_V_vs_sea1d(51)= 7.0026000000000
  SEA1DM_V_vs_sea1d(52)= 6.9821500000000
  SEA1DM_V_vs_sea1d(53)= 6.9613400000000
  SEA1DM_V_vs_sea1d(54)= 6.9401300000000
  SEA1DM_V_vs_sea1d(55)= 6.9185200000000
  SEA1DM_V_vs_sea1d(56)= 6.8964900000000
  SEA1DM_V_vs_sea1d(57)= 6.8740200000000
  SEA1DM_V_vs_sea1d(58)= 6.8510900000000
  SEA1DM_V_vs_sea1d(59)= 6.8276700000000
  SEA1DM_V_vs_sea1d(60)= 6.8037600000000
  SEA1DM_V_vs_sea1d(61)= 6.7793300000000
  SEA1DM_V_vs_sea1d(62)= 6.7543700000000
  SEA1DM_V_vs_sea1d(63)= 6.7288500000000
  SEA1DM_V_vs_sea1d(64)= 6.7027700000000
  SEA1DM_V_vs_sea1d(65)= 6.6760900000000
  SEA1DM_V_vs_sea1d(66)= 6.6488100000000
  SEA1DM_V_vs_sea1d(67)= 6.6208900000000
  SEA1DM_V_vs_sea1d(68)= 6.5923300000000
  SEA1DM_V_vs_sea1d(69)= 6.5631100000000
  SEA1DM_V_vs_sea1d(70)= 6.5332000000000
  SEA1DM_V_vs_sea1d(71)= 6.5026000000000
  SEA1DM_V_vs_sea1d(72)= 6.4712600000000
  SEA1DM_V_vs_sea1d(73)= 6.4392000000000
  SEA1DM_V_vs_sea1d(74)= 6.4063800000000
  SEA1DM_V_vs_sea1d(75)= 6.3727800000000
  SEA1DM_V_vs_sea1d(76)= 6.3383900000000
  SEA1DM_V_vs_sea1d(77)= 6.3031900000000
  SEA1DM_V_vs_sea1d(78)= 6.2671500000000
  SEA1DM_V_vs_sea1d(79)= 6.2302600000000
  SEA1DM_V_vs_sea1d(80)= 6.1925100000000
  SEA1DM_V_vs_sea1d(81)= 6.1538700000000
  SEA1DM_V_vs_sea1d(82)= 6.1144200000000
  SEA1DM_V_vs_sea1d(83)= 6.1065800000000
  SEA1DM_V_vs_sea1d(84)= 5.4546300000000
  SEA1DM_V_vs_sea1d(85)= 5.4378400000000
  SEA1DM_V_vs_sea1d(86)= 5.4210500000000
  SEA1DM_V_vs_sea1d(87)= 5.4042500000000
  SEA1DM_V_vs_sea1d(88)= 5.3874600000000
  SEA1DM_V_vs_sea1d(89)= 5.3706700000000
  SEA1DM_V_vs_sea1d(90)= 5.3538800000000
  SEA1DM_V_vs_sea1d(91)= 5.3370900000000
  SEA1DM_V_vs_sea1d(92)= 5.3203000000000
  SEA1DM_V_vs_sea1d(93)= 5.3035100000000
  SEA1DM_V_vs_sea1d(94)= 5.2867200000000
  SEA1DM_V_vs_sea1d(95)= 5.2699300000000
  SEA1DM_V_vs_sea1d(96)= 5.2531400000000
  SEA1DM_V_vs_sea1d(97)= 5.2363500000000
  SEA1DM_V_vs_sea1d(98)= 5.2195600000000
  SEA1DM_V_vs_sea1d(99)= 5.2027700000000
  SEA1DM_V_vs_sea1d(100)= 5.1859800000000
  SEA1DM_V_vs_sea1d(101)= 5.1691900000000
  SEA1DM_V_vs_sea1d(102)= 5.1524000000000
  SEA1DM_V_vs_sea1d(103)= 5.1356100000000
  SEA1DM_V_vs_sea1d(104)= 5.1188200000000
  SEA1DM_V_vs_sea1d(105)= 5.1020200000000
  SEA1DM_V_vs_sea1d(106)= 5.0852300000000
  SEA1DM_V_vs_sea1d(107)= 5.0684400000000
  SEA1DM_V_vs_sea1d(108)= 5.0516500000000
  SEA1DM_V_vs_sea1d(109)= 5.0348600000000
  SEA1DM_V_vs_sea1d(110)= 4.7959100000000
  SEA1DM_V_vs_sea1d(111)= 4.7761200000000
  SEA1DM_V_vs_sea1d(112)= 4.7563200000000
  SEA1DM_V_vs_sea1d(113)= 4.7365300000000
  SEA1DM_V_vs_sea1d(114)= 4.7167300000000
  SEA1DM_V_vs_sea1d(115)= 4.6969400000000
  SEA1DM_V_vs_sea1d(116)= 4.6771400000000
  SEA1DM_V_vs_sea1d(117)= 4.6573400000000
  SEA1DM_V_vs_sea1d(118)= 4.6375500000000
  SEA1DM_V_vs_sea1d(119)= 4.6177500000000
  SEA1DM_V_vs_sea1d(120)= 4.5979600000000
  SEA1DM_V_vs_sea1d(121)= 4.5781600000000
  SEA1DM_V_vs_sea1d(122)= 4.5583700000000
  SEA1DM_V_vs_sea1d(123)= 4.5385700000000
  SEA1DM_V_vs_sea1d(124)= 4.5187700000000
  SEA1DM_V_vs_sea1d(125)= 4.4989800000000
  SEA1DM_V_vs_sea1d(126)= 4.4791800000000
  SEA1DM_V_vs_sea1d(127)= 4.4593900000000
  SEA1DM_V_vs_sea1d(128)= 4.4395900000000
  SEA1DM_V_vs_sea1d(129)= 4.4198000000000
  SEA1DM_V_vs_sea1d(130)= 4.4000000000000
  SEA1DM_V_vs_sea1d(131)= 4.4000000000000
  SEA1DM_V_vs_sea1d(132)= 4.4000000000000
  SEA1DM_V_vs_sea1d(133)= 4.4000000000000
  SEA1DM_V_vs_sea1d(134)= 4.4000000000000
  SEA1DM_V_vs_sea1d(135)= 4.4000000000000
  SEA1DM_V_vs_sea1d(136)= 4.4000000000000
  SEA1DM_V_vs_sea1d(137)= 4.4000000000000
  SEA1DM_V_vs_sea1d(138)= 4.4000000000000
  SEA1DM_V_vs_sea1d(139)= 4.4000000000000
  SEA1DM_V_vs_sea1d(140)= 4.4000000000000
  SEA1DM_V_vs_sea1d(141)= 4.4000000000000
  SEA1DM_V_vs_sea1d(142)= 4.4000000000000
  SEA1DM_V_vs_sea1d(143)= 4.4000000000000
  SEA1DM_V_vs_sea1d(144)= 4.4000000000000
  SEA1DM_V_vs_sea1d(145)= 4.4000000000000
  SEA1DM_V_vs_sea1d(146)= 4.4000000000000
  SEA1DM_V_vs_sea1d(147)= 4.4000000000000
  SEA1DM_V_vs_sea1d(148)= 4.4000000000000
  SEA1DM_V_vs_sea1d(149)= 4.4000000000000
  SEA1DM_V_vs_sea1d(150)= 4.4000000000000
  SEA1DM_V_vs_sea1d(151)= 4.4000000000000
  SEA1DM_V_vs_sea1d(152)= 3.4500000000000
  SEA1DM_V_vs_sea1d(153)= 3.4500000000000
  SEA1DM_V_vs_sea1d(154)= 3.4500000000000
  SEA1DM_V_vs_sea1d(155)= 3.4500000000000
  SEA1DM_V_vs_sea1d(156)= 3.4500000000000
  SEA1DM_V_vs_sea1d(157)= 3.4500000000000
  SEA1DM_V_vs_sea1d(158)= 3.4500000000000
  SEA1DM_V_vs_sea1d(159)= 3.4500000000000
  SEA1DM_V_vs_sea1d(160)= 3.4500000000000
  SEA1DM_V_vs_sea1d(161)= 3.4500000000000
  SEA1DM_V_vs_sea1d(162)= 3.4500000000000
  SEA1DM_V_vs_sea1d(163)= 3.4500000000000

  SEA1DM_V_Qkappa_sea1d(1)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(2)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(3)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(4)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(5)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(6)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(7)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(8)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(9)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(10)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(11)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(12)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(13)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(14)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(15)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(16)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(17)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(18)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(19)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(20)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(21)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(22)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(23)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(24)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(25)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(26)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(27)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(28)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(29)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(30)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(31)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(32)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(33)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(34)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(35)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(36)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(37)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(38)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(39)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(40)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(41)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(42)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(43)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(44)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(45)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(46)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(47)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(48)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(49)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(50)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(51)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(52)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(53)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(54)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(55)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(56)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(57)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(58)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(59)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(60)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(61)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(62)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(63)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(64)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(65)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(66)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(67)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(68)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(69)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(70)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(71)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(72)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(73)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(74)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(75)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(76)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(77)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(78)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(79)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(80)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(81)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(82)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(83)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(84)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(85)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(86)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(87)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(88)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(89)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(90)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(91)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(92)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(93)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(94)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(95)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(96)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(97)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(98)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(99)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(100)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(101)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(102)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(103)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(104)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(105)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(106)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(107)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(108)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(109)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(110)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(111)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(112)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(113)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(114)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(115)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(116)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(117)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(118)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(119)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(120)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(121)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(122)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(123)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(124)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(125)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(126)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(127)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(128)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(129)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(130)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(131)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(132)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(133)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(134)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(135)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(136)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(137)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(138)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(139)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(140)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(141)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(142)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(143)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(144)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(145)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(146)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(147)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(148)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(149)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(150)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(151)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(152)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(153)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(154)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(155)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(156)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(157)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(158)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(159)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(160)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(161)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(162)= 99999.0000000000000
  SEA1DM_V_Qkappa_sea1d(163)= 99999.0000000000000

  SEA1DM_V_Qmu_sea1d(1)= 84.6000000000000
  SEA1DM_V_Qmu_sea1d(2)= 84.6000000000000
  SEA1DM_V_Qmu_sea1d(3)= 84.6000000000000
  SEA1DM_V_Qmu_sea1d(4)= 84.6000000000000
  SEA1DM_V_Qmu_sea1d(5)= 84.6000000000000
  SEA1DM_V_Qmu_sea1d(6)= 84.6000000000000
  SEA1DM_V_Qmu_sea1d(7)= 84.6000000000000
  SEA1DM_V_Qmu_sea1d(8)= 84.6000000000000
  SEA1DM_V_Qmu_sea1d(9)= 84.6000000000000
  SEA1DM_V_Qmu_sea1d(10)= 84.6000000000000
  SEA1DM_V_Qmu_sea1d(11)= 84.6000000000000
  SEA1DM_V_Qmu_sea1d(12)= 84.6000000000000
  SEA1DM_V_Qmu_sea1d(13)= 84.6000000000000
  SEA1DM_V_Qmu_sea1d(14)= 0.0000000000000
  SEA1DM_V_Qmu_sea1d(15)= 0.0000000000000
  SEA1DM_V_Qmu_sea1d(16)= 0.0000000000000
  SEA1DM_V_Qmu_sea1d(17)= 0.0000000000000
  SEA1DM_V_Qmu_sea1d(18)= 0.0000000000000
  SEA1DM_V_Qmu_sea1d(19)= 0.0000000000000
  SEA1DM_V_Qmu_sea1d(20)= 0.0000000000000
  SEA1DM_V_Qmu_sea1d(21)= 0.0000000000000
  SEA1DM_V_Qmu_sea1d(22)= 0.0000000000000
  SEA1DM_V_Qmu_sea1d(23)= 0.0000000000000
  SEA1DM_V_Qmu_sea1d(24)= 0.0000000000000
  SEA1DM_V_Qmu_sea1d(25)= 0.0000000000000
  SEA1DM_V_Qmu_sea1d(26)= 0.0000000000000
  SEA1DM_V_Qmu_sea1d(27)= 0.0000000000000
  SEA1DM_V_Qmu_sea1d(28)= 0.0000000000000
  SEA1DM_V_Qmu_sea1d(29)= 0.0000000000000
  SEA1DM_V_Qmu_sea1d(30)= 0.0000000000000
  SEA1DM_V_Qmu_sea1d(31)= 0.0000000000000
  SEA1DM_V_Qmu_sea1d(32)= 0.0000000000000
  SEA1DM_V_Qmu_sea1d(33)= 0.0000000000000
  SEA1DM_V_Qmu_sea1d(34)= 0.0000000000000
  SEA1DM_V_Qmu_sea1d(35)= 0.0000000000000
  SEA1DM_V_Qmu_sea1d(36)= 0.0000000000000
  SEA1DM_V_Qmu_sea1d(37)= 0.0000000000000
  SEA1DM_V_Qmu_sea1d(38)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(39)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(40)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(41)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(42)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(43)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(44)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(45)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(46)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(47)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(48)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(49)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(50)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(51)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(52)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(53)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(54)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(55)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(56)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(57)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(58)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(59)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(60)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(61)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(62)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(63)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(64)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(65)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(66)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(67)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(68)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(69)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(70)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(71)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(72)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(73)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(74)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(75)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(76)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(77)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(78)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(79)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(80)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(81)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(82)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(83)= 312.0000000000000
  SEA1DM_V_Qmu_sea1d(84)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(85)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(86)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(87)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(88)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(89)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(90)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(91)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(92)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(93)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(94)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(95)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(96)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(97)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(98)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(99)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(100)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(101)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(102)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(103)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(104)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(105)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(106)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(107)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(108)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(109)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(110)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(111)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(112)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(113)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(114)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(115)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(116)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(117)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(118)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(119)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(120)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(121)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(122)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(123)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(124)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(125)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(126)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(127)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(128)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(129)= 143.0000000000000
  SEA1DM_V_Qmu_sea1d(130)= 110.0000000000000
  SEA1DM_V_Qmu_sea1d(131)= 80.0000000000000
  SEA1DM_V_Qmu_sea1d(132)= 50.0000000000000
  SEA1DM_V_Qmu_sea1d(133)= 50.0000000000000
  SEA1DM_V_Qmu_sea1d(134)= 50.0000000000000
  SEA1DM_V_Qmu_sea1d(135)= 50.0000000000000
  SEA1DM_V_Qmu_sea1d(136)= 50.0000000000000
  SEA1DM_V_Qmu_sea1d(137)= 50.0000000000000
  SEA1DM_V_Qmu_sea1d(138)= 50.0000000000000
  SEA1DM_V_Qmu_sea1d(139)= 50.0000000000000
  SEA1DM_V_Qmu_sea1d(140)= 50.0000000000000
  SEA1DM_V_Qmu_sea1d(141)= 50.0000000000000
  SEA1DM_V_Qmu_sea1d(142)= 50.0000000000000
  SEA1DM_V_Qmu_sea1d(143)= 50.0000000000000
  SEA1DM_V_Qmu_sea1d(144)= 50.0000000000000
  SEA1DM_V_Qmu_sea1d(145)= 100.0000000000000
  SEA1DM_V_Qmu_sea1d(146)= 150.0000000000000
  SEA1DM_V_Qmu_sea1d(147)= 150.0000000000000
  SEA1DM_V_Qmu_sea1d(148)= 150.0000000000000
  SEA1DM_V_Qmu_sea1d(149)= 150.0000000000000
  SEA1DM_V_Qmu_sea1d(150)= 150.0000000000000
  SEA1DM_V_Qmu_sea1d(151)= 150.0000000000000
  SEA1DM_V_Qmu_sea1d(152)= 300.0000000000000
  SEA1DM_V_Qmu_sea1d(153)= 300.0000000000000
  SEA1DM_V_Qmu_sea1d(154)= 300.0000000000000
  SEA1DM_V_Qmu_sea1d(155)= 300.0000000000000
  SEA1DM_V_Qmu_sea1d(156)= 300.0000000000000
  SEA1DM_V_Qmu_sea1d(157)= 300.0000000000000
  SEA1DM_V_Qmu_sea1d(158)= 300.0000000000000
  SEA1DM_V_Qmu_sea1d(159)= 300.0000000000000
  SEA1DM_V_Qmu_sea1d(160)= 300.0000000000000
  SEA1DM_V_Qmu_sea1d(161)= 300.0000000000000
  SEA1DM_V_Qmu_sea1d(162)= 300.0000000000000
  SEA1DM_V_Qmu_sea1d(163)= 300.0000000000000

! strip the crust and replace it by mantle
  if (SUPPRESS_CRUSTAL_MESH .or. USE_EXTERNAL_CRUSTAL_MODEL) then
    do i=NR_SEA1D-12,NR_SEA1D
      SEA1DM_V_density_sea1d(i) = SEA1DM_V_density_sea1d(NR_SEA1D-13)
      SEA1DM_V_vp_sea1d(i) = SEA1DM_V_vp_sea1d(NR_SEA1D-13)
      SEA1DM_V_vs_sea1d(i) = SEA1DM_V_vs_sea1d(NR_SEA1D-13)
      SEA1DM_V_Qkappa_sea1d(i) = SEA1DM_V_Qkappa_sea1d(NR_SEA1D-13)
      SEA1DM_V_Qmu_sea1d(i) = SEA1DM_V_Qmu_sea1d(NR_SEA1D-13)
    enddo
  endif

  end subroutine define_model_sea1d

