!=====================================================================
!
!          S p e c f e m 3 D  G l o b e  V e r s i o n  7 . 0
!          --------------------------------------------------
!
!     Main historical authors: Dimitri Komatitsch and Jeroen Tromp
!                        Princeton University, USA
!                and CNRS / University of Marseille, France
!                 (there are currently many more authors!)
! (c) Princeton University and CNRS / University of Marseille, April 2014
!
! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 3 of the License, or
! (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
!
!=====================================================================

!--------------------------------------------------------------------------------------------------
!
! 1D REF model of Kustowski et al. (2008)
!
! this is STW105 - new reference model, also known as REF
!
! A recent 1D Earth model developed by Kustowski et al. This model is the 1D background
! model for the 3D models s362ani, s362wmani, s362ani_prem, and s29ea.
!
! see chapter 3, in:
! Kustowski, B, Ekstr"om, G., and A. M. Dziewonski, 2008,
! Anisotropic shear-wave velocity structure of the Earth's mantle: A global model,
! J. Geophys. Res., 113, B06306, doi:10.1029/2007JB005169.
!
! model is identical to PREM at crustal depths, between 220 and 400km
! and below 670km.
!
! attenuation structure is taken from model QL6:
! Durek, J. J. and G. Ekstr"om, 1996.
! A radial model of anelasticity consistent with long period surface wave attenuation,
! Bull. Seism. Soc. Am., 86, 144-158
!
! 1-D profiles: see http://www.iris.edu/ds/products/emc-stw105/
!               internal discontinuities: contains a 410-km and 650-km discontinuity
!--------------------------------------------------------------------------------------------------

  module model_1dref_par

  ! number of layers in DATA/s362ani/REF
  integer, parameter :: NR_REF = 750

  ! model_1dref_variables
  double precision, dimension(:),allocatable :: Mref_V_radius_ref,Mref_V_density_ref, &
                                                Mref_V_vpv_ref,Mref_V_vph_ref, &
                                                Mref_V_vsv_ref,Mref_V_vsh_ref, &
                                                Mref_V_eta_ref,Mref_V_Qkappa_ref,Mref_V_Qmu_ref

  end module model_1dref_par

!
!-------------------------------------------------------------------------------------------------
!

  subroutine model_1dref_broadcast(CRUSTAL)

! standard routine to setup model

  use model_1dref_par

  implicit none

  logical :: CRUSTAL

  ! local parameters
  integer :: ier

  ! allocates arrays
  allocate(Mref_V_radius_ref(NR_REF), &
           Mref_V_density_ref(NR_REF), &
           Mref_V_vpv_ref(NR_REF), &
           Mref_V_vph_ref(NR_REF), &
           Mref_V_vsv_ref(NR_REF), &
           Mref_V_vsh_ref(NR_REF), &
           Mref_V_eta_ref(NR_REF), &
           Mref_V_Qkappa_ref(NR_REF), &
           Mref_V_Qmu_ref(NR_REF), &
           stat=ier)
  if (ier /= 0 ) stop 'Error allocating 1d_ref arrays'

  ! all processes will define same parameters
  call define_model_1dref(CRUSTAL)

  end subroutine model_1dref_broadcast

!
!-------------------------------------------------------------------------------------------------
!

  subroutine model_1dref(x,rho,vpv,vph,vsv,vsh,eta,Qkappa,Qmu,iregion_code,CRUSTAL)

  use constants
  use shared_parameters, only: R_PLANET,RHOAV

  use model_1dref_par

  implicit none

! model_1dref_variables

! input:
! dimensionless radius x

! output: non-dimensionalized
! mass density rho
! compressional wave speed vpv
! compressional wave speed vph
! shear wave speed vsv
! shear wave speed vsh
! dimensionless parameter eta
! shear quality factor Qmu
! bulk quality factor Qkappa

  double precision :: x,rho,vpv,vph,vsv,vsh,eta,Qmu,Qkappa
  integer :: iregion_code
  logical :: CRUSTAL

  ! local parameters
  double precision :: r,frac,scaleval
  integer :: i

  ! compute real physical radius in meters
  r = x * R_PLANET

  i = 1
  do while(r >= Mref_V_radius_ref(i) .and. i /= NR_REF)
    i = i + 1
  enddo

! make sure we stay in the right region
  if (iregion_code == IREGION_INNER_CORE .and. i > 180) i = 180

  if (iregion_code == IREGION_OUTER_CORE .and. i < 182) i = 182
  if (iregion_code == IREGION_OUTER_CORE .and. i > 358) i = 358

  if (iregion_code == IREGION_CRUST_MANTLE .and. i < 360) i = 360

  ! if crustal model is used, mantle gets expanded up to surface
  ! for any depth less than 24.4 km, values from mantle below moho are taken
  if (CRUSTAL .and. i > 717) i = 717


  if (i == 1) then
    ! first layer in inner core
    rho = Mref_V_density_ref(i)
    vpv = Mref_V_vpv_ref(i)
    vph = Mref_V_vph_ref(i)
    vsv = Mref_V_vsv_ref(i)
    vsh = Mref_V_vsh_ref(i)
    eta = Mref_V_eta_ref(i)
    Qkappa = Mref_V_Qkappa_ref(i)
    Qmu = Mref_V_Qmu_ref(i)
  else
    ! interpolates between one layer below to actual radius layer,
    ! that is from radius_ref(i-1) to r using the values at i-1 and i
    frac = (r-Mref_V_radius_ref(i-1))/(Mref_V_radius_ref(i)-Mref_V_radius_ref(i-1))
    ! interpolated model parameters
    rho = Mref_V_density_ref(i-1) + frac * (Mref_V_density_ref(i)-Mref_V_density_ref(i-1))
    vpv = Mref_V_vpv_ref(i-1) + frac * (Mref_V_vpv_ref(i)-Mref_V_vpv_ref(i-1))
    vph = Mref_V_vph_ref(i-1) + frac * (Mref_V_vph_ref(i)-Mref_V_vph_ref(i-1))
    vsv = Mref_V_vsv_ref(i-1) + frac * (Mref_V_vsv_ref(i)-Mref_V_vsv_ref(i-1))
    vsh = Mref_V_vsh_ref(i-1) + frac * (Mref_V_vsh_ref(i)-Mref_V_vsh_ref(i-1))
    eta = Mref_V_eta_ref(i-1) + frac * (Mref_V_eta_ref(i)-Mref_V_eta_ref(i-1))
    Qkappa = Mref_V_Qkappa_ref(i-1) + frac * (Mref_V_Qkappa_ref(i)-Mref_V_Qkappa_ref(i-1))
    Qmu = Mref_V_Qmu_ref(i-1) + frac * (Mref_V_Qmu_ref(i)-Mref_V_Qmu_ref(i-1))
  endif

  ! make sure Vs is zero in the outer core even if roundoff errors on depth
  ! also set fictitious attenuation to a very high value (attenuation is not used in the fluid)
  if (iregion_code == IREGION_OUTER_CORE) then
    vsv = 0.d0
    vsh = 0.d0
    Qkappa = 3000.d0
    Qmu = 3000.d0
  endif

  ! non-dimensionalize
  ! time scaling (s^{-1}) is done with scaleval
  scaleval = dsqrt(PI*GRAV*RHOAV)
  rho = rho/RHOAV
  vpv = vpv/(R_PLANET*scaleval)
  vph = vph/(R_PLANET*scaleval)
  vsv = vsv/(R_PLANET*scaleval)
  vsh = vsh/(R_PLANET*scaleval)

  end subroutine model_1dref

!-------------------

  subroutine define_model_1dref(USE_EXTERNAL_CRUSTAL_MODEL)

  use constants
  use model_1dref_par

  implicit none

  logical :: USE_EXTERNAL_CRUSTAL_MODEL

  ! define the 1D REF model of Kustowski et al. (2007)
  ! table starts at inner core and moves up till surface

  Mref_V_radius_ref( 1 : 30 ) = (/ &
  0.d0, &
  6824.d0, &
  13648.d0, &
  20472.d0, &
  27296.d0, &
  34120.d0, &
  40944.d0, &
  47768.d0, &
  54592.d0, &
  61416.d0, &
  68240.d0, &
  75064.d0, &
  81888.d0, &
  88712.d0, &
  95536.d0, &
  102360.d0, &
  109184.d0, &
  116007.d0, &
  122831.d0, &
  129655.d0, &
  136479.d0, &
  143303.d0, &
  150127.d0, &
  156951.d0, &
  163775.d0, &
  170599.d0, &
  177423.d0, &
  184247.d0, &
  191071.d0, &
  197895.d0 /)

  Mref_V_radius_ref( 31 : 60 ) = (/ &
  204719.d0, &
  211543.d0, &
  218367.d0, &
  225191.d0, &
  232015.d0, &
  238839.d0, &
  245663.d0, &
  252487.d0, &
  259311.d0, &
  266135.d0, &
  272959.d0, &
  279783.d0, &
  286607.d0, &
  293431.d0, &
  300255.d0, &
  307079.d0, &
  313903.d0, &
  320727.d0, &
  327551.d0, &
  334375.d0, &
  341199.d0, &
  348022.d0, &
  354846.d0, &
  361670.d0, &
  368494.d0, &
  375318.d0, &
  382142.d0, &
  388966.d0, &
  395790.d0, &
  402614.d0 /)

  Mref_V_radius_ref( 61 : 90 ) = (/ &
  409438.d0, &
  416262.d0, &
  423086.d0, &
  429910.d0, &
  436734.d0, &
  443558.d0, &
  450382.d0, &
  457206.d0, &
  464030.d0, &
  470854.d0, &
  477678.d0, &
  484502.d0, &
  491326.d0, &
  498150.d0, &
  504974.d0, &
  511798.d0, &
  518622.d0, &
  525446.d0, &
  532270.d0, &
  539094.d0, &
  545918.d0, &
  552742.d0, &
  559566.d0, &
  566390.d0, &
  573214.d0, &
  580037.d0, &
  586861.d0, &
  593685.d0, &
  600509.d0, &
  607333.d0 /)

  Mref_V_radius_ref( 91 : 120 ) = (/ &
  614157.d0, &
  620981.d0, &
  627805.d0, &
  634629.d0, &
  641453.d0, &
  648277.d0, &
  655101.d0, &
  661925.d0, &
  668749.d0, &
  675573.d0, &
  682397.d0, &
  689221.d0, &
  696045.d0, &
  702869.d0, &
  709693.d0, &
  716517.d0, &
  723341.d0, &
  730165.d0, &
  736989.d0, &
  743813.d0, &
  750637.d0, &
  757461.d0, &
  764285.d0, &
  771109.d0, &
  777933.d0, &
  784757.d0, &
  791581.d0, &
  798405.d0, &
  805229.d0, &
  812052.d0 /)

  Mref_V_radius_ref( 121 : 150 ) = (/ &
  818876.d0, &
  825700.d0, &
  832524.d0, &
  839348.d0, &
  846172.d0, &
  852996.d0, &
  859820.d0, &
  866644.d0, &
  873468.d0, &
  880292.d0, &
  887116.d0, &
  893940.d0, &
  900764.d0, &
  907588.d0, &
  914412.d0, &
  921236.d0, &
  928060.d0, &
  934884.d0, &
  941708.d0, &
  948532.d0, &
  955356.d0, &
  962180.d0, &
  969004.d0, &
  975828.d0, &
  982652.d0, &
  989476.d0, &
  996300.d0, &
  1003124.d0, &
  1009948.d0, &
  1016772.d0 /)

  Mref_V_radius_ref( 151 : 180 ) = (/ &
  1023596.d0, &
  1030420.d0, &
  1037244.d0, &
  1044067.d0, &
  1050891.d0, &
  1057715.d0, &
  1064539.d0, &
  1071363.d0, &
  1078187.d0, &
  1085011.d0, &
  1091835.d0, &
  1098659.d0, &
  1105483.d0, &
  1112307.d0, &
  1119131.d0, &
  1125955.d0, &
  1132779.d0, &
  1139603.d0, &
  1146427.d0, &
  1153251.d0, &
  1160075.d0, &
  1166899.d0, &
  1173723.d0, &
  1180547.d0, &
  1187371.d0, &
  1194195.d0, &
  1201019.d0, &
  1207843.d0, &
  1214667.d0, &
  1221491.d0 /)

  Mref_V_radius_ref( 181 : 210 ) = (/ &
  1221491.d0, &
  1234250.d0, &
  1247010.d0, &
  1259770.d0, &
  1272530.d0, &
  1285289.d0, &
  1298049.d0, &
  1310809.d0, &
  1323568.d0, &
  1336328.d0, &
  1349088.d0, &
  1361847.d0, &
  1374607.d0, &
  1387367.d0, &
  1400127.d0, &
  1412886.d0, &
  1425646.d0, &
  1438406.d0, &
  1451165.d0, &
  1463925.d0, &
  1476685.d0, &
  1489444.d0, &
  1502204.d0, &
  1514964.d0, &
  1527724.d0, &
  1540483.d0, &
  1553243.d0, &
  1566003.d0, &
  1578762.d0, &
  1591522.d0 /)

  Mref_V_radius_ref( 211 : 240 ) = (/ &
  1604282.d0, &
  1617041.d0, &
  1629801.d0, &
  1642561.d0, &
  1655321.d0, &
  1668080.d0, &
  1680840.d0, &
  1693600.d0, &
  1706359.d0, &
  1719119.d0, &
  1731879.d0, &
  1744638.d0, &
  1757398.d0, &
  1770158.d0, &
  1782918.d0, &
  1795677.d0, &
  1808437.d0, &
  1821197.d0, &
  1833956.d0, &
  1846716.d0, &
  1859476.d0, &
  1872235.d0, &
  1884995.d0, &
  1897755.d0, &
  1910515.d0, &
  1923274.d0, &
  1936034.d0, &
  1948794.d0, &
  1961553.d0, &
  1974313.d0 /)

  Mref_V_radius_ref( 241 : 270 ) = (/ &
  1987073.d0, &
  1999832.d0, &
  2012592.d0, &
  2025352.d0, &
  2038112.d0, &
  2050871.d0, &
  2063631.d0, &
  2076391.d0, &
  2089150.d0, &
  2101910.d0, &
  2114670.d0, &
  2127429.d0, &
  2140189.d0, &
  2152949.d0, &
  2165709.d0, &
  2178468.d0, &
  2191228.d0, &
  2203988.d0, &
  2216747.d0, &
  2229507.d0, &
  2242267.d0, &
  2255026.d0, &
  2267786.d0, &
  2280546.d0, &
  2293306.d0, &
  2306065.d0, &
  2318825.d0, &
  2331585.d0, &
  2344344.d0, &
  2357104.d0 /)

  Mref_V_radius_ref( 271 : 300 ) = (/ &
  2369864.d0, &
  2382623.d0, &
  2395383.d0, &
  2408143.d0, &
  2420903.d0, &
  2433662.d0, &
  2446422.d0, &
  2459182.d0, &
  2471941.d0, &
  2484701.d0, &
  2497461.d0, &
  2510220.d0, &
  2522980.d0, &
  2535740.d0, &
  2548500.d0, &
  2561259.d0, &
  2574019.d0, &
  2586779.d0, &
  2599538.d0, &
  2612298.d0, &
  2625058.d0, &
  2637818.d0, &
  2650577.d0, &
  2663337.d0, &
  2676097.d0, &
  2688856.d0, &
  2701616.d0, &
  2714376.d0, &
  2727135.d0, &
  2739895.d0 /)

  Mref_V_radius_ref( 301 : 330 ) = (/ &
  2752655.d0, &
  2765415.d0, &
  2778174.d0, &
  2790934.d0, &
  2803694.d0, &
  2816453.d0, &
  2829213.d0, &
  2841973.d0, &
  2854732.d0, &
  2867492.d0, &
  2880252.d0, &
  2893012.d0, &
  2905771.d0, &
  2918531.d0, &
  2931291.d0, &
  2944050.d0, &
  2956810.d0, &
  2969570.d0, &
  2982329.d0, &
  2995089.d0, &
  3007849.d0, &
  3020609.d0, &
  3033368.d0, &
  3046128.d0, &
  3058888.d0, &
  3071647.d0, &
  3084407.d0, &
  3097167.d0, &
  3109926.d0, &
  3122686.d0 /)

  Mref_V_radius_ref( 331 : 360 ) = (/ &
  3135446.d0, &
  3148206.d0, &
  3160965.d0, &
  3173725.d0, &
  3186485.d0, &
  3199244.d0, &
  3212004.d0, &
  3224764.d0, &
  3237523.d0, &
  3250283.d0, &
  3263043.d0, &
  3275803.d0, &
  3288562.d0, &
  3301322.d0, &
  3314082.d0, &
  3326841.d0, &
  3339601.d0, &
  3352361.d0, &
  3365120.d0, &
  3377880.d0, &
  3390640.d0, &
  3403400.d0, &
  3416159.d0, &
  3428919.d0, &
  3441679.d0, &
  3454438.d0, &
  3467198.d0, &
  3479958.d0, &
  3479958.d0, &
  3489335.d0 /)

  Mref_V_radius_ref( 361 : 390 ) = (/ &
  3498713.d0, &
  3508091.d0, &
  3517468.d0, &
  3526846.d0, &
  3536224.d0, &
  3545601.d0, &
  3554979.d0, &
  3564357.d0, &
  3573734.d0, &
  3583112.d0, &
  3592489.d0, &
  3601867.d0, &
  3611245.d0, &
  3620622.d0, &
  3630000.d0, &
  3630000.d0, &
  3639471.d0, &
  3648942.d0, &
  3658413.d0, &
  3667885.d0, &
  3677356.d0, &
  3686827.d0, &
  3696298.d0, &
  3705769.d0, &
  3715240.d0, &
  3724712.d0, &
  3734183.d0, &
  3743654.d0, &
  3753125.d0, &
  3762596.d0 /)

  Mref_V_radius_ref( 391 : 420 ) = (/ &
  3772067.d0, &
  3781538.d0, &
  3791010.d0, &
  3800481.d0, &
  3809952.d0, &
  3819423.d0, &
  3828894.d0, &
  3838365.d0, &
  3847837.d0, &
  3857308.d0, &
  3866779.d0, &
  3876250.d0, &
  3885721.d0, &
  3895192.d0, &
  3904663.d0, &
  3914135.d0, &
  3923606.d0, &
  3933077.d0, &
  3942548.d0, &
  3952019.d0, &
  3961490.d0, &
  3970962.d0, &
  3980433.d0, &
  3989904.d0, &
  3999375.d0, &
  4008846.d0, &
  4018317.d0, &
  4027788.d0, &
  4037260.d0, &
  4046731.d0 /)

  Mref_V_radius_ref( 421 : 450 ) = (/ &
  4056202.d0, &
  4065673.d0, &
  4075144.d0, &
  4084615.d0, &
  4094087.d0, &
  4103558.d0, &
  4113029.d0, &
  4122500.d0, &
  4131971.d0, &
  4141442.d0, &
  4150913.d0, &
  4160385.d0, &
  4169856.d0, &
  4179327.d0, &
  4188798.d0, &
  4198269.d0, &
  4207740.d0, &
  4217212.d0, &
  4226683.d0, &
  4236154.d0, &
  4245625.d0, &
  4255096.d0, &
  4264567.d0, &
  4274038.d0, &
  4283510.d0, &
  4292981.d0, &
  4302452.d0, &
  4311923.d0, &
  4321394.d0, &
  4330865.d0 /)

  Mref_V_radius_ref( 451 : 480 ) = (/ &
  4340337.d0, &
  4349808.d0, &
  4359279.d0, &
  4368750.d0, &
  4378221.d0, &
  4387692.d0, &
  4397163.d0, &
  4406635.d0, &
  4416106.d0, &
  4425577.d0, &
  4435048.d0, &
  4444519.d0, &
  4453990.d0, &
  4463462.d0, &
  4472933.d0, &
  4482404.d0, &
  4491875.d0, &
  4501346.d0, &
  4510817.d0, &
  4520288.d0, &
  4529760.d0, &
  4539231.d0, &
  4548702.d0, &
  4558173.d0, &
  4567644.d0, &
  4577115.d0, &
  4586587.d0, &
  4596058.d0, &
  4605529.d0, &
  4615000.d0 /)

  Mref_V_radius_ref( 481 : 510 ) = (/ &
  4624471.d0, &
  4633942.d0, &
  4643413.d0, &
  4652885.d0, &
  4662356.d0, &
  4671827.d0, &
  4681298.d0, &
  4690769.d0, &
  4700240.d0, &
  4709712.d0, &
  4719183.d0, &
  4728654.d0, &
  4738125.d0, &
  4747596.d0, &
  4757067.d0, &
  4766538.d0, &
  4776010.d0, &
  4785481.d0, &
  4794952.d0, &
  4804423.d0, &
  4813894.d0, &
  4823365.d0, &
  4832837.d0, &
  4842308.d0, &
  4851779.d0, &
  4861250.d0, &
  4870721.d0, &
  4880192.d0, &
  4889663.d0, &
  4899135.d0 /)

  Mref_V_radius_ref( 511 : 540 ) = (/ &
  4908606.d0, &
  4918077.d0, &
  4927548.d0, &
  4937019.d0, &
  4946490.d0, &
  4955962.d0, &
  4965433.d0, &
  4974904.d0, &
  4984375.d0, &
  4993846.d0, &
  5003317.d0, &
  5012788.d0, &
  5022260.d0, &
  5031731.d0, &
  5041202.d0, &
  5050673.d0, &
  5060144.d0, &
  5069615.d0, &
  5079087.d0, &
  5088558.d0, &
  5098029.d0, &
  5107500.d0, &
  5116971.d0, &
  5126442.d0, &
  5135913.d0, &
  5145385.d0, &
  5154856.d0, &
  5164327.d0, &
  5173798.d0, &
  5183269.d0 /)

  Mref_V_radius_ref( 541 : 570 ) = (/ &
  5192740.d0, &
  5202212.d0, &
  5211683.d0, &
  5221154.d0, &
  5230625.d0, &
  5240096.d0, &
  5249567.d0, &
  5259038.d0, &
  5268510.d0, &
  5277981.d0, &
  5287452.d0, &
  5296923.d0, &
  5306394.d0, &
  5315865.d0, &
  5325337.d0, &
  5334808.d0, &
  5344279.d0, &
  5353750.d0, &
  5363221.d0, &
  5372692.d0, &
  5382163.d0, &
  5391635.d0, &
  5401106.d0, &
  5410577.d0, &
  5420048.d0, &
  5429519.d0, &
  5438990.d0, &
  5448462.d0, &
  5457933.d0, &
  5467404.d0 /)

  Mref_V_radius_ref( 571 : 600 ) = (/ &
  5476875.d0, &
  5486346.d0, &
  5495817.d0, &
  5505288.d0, &
  5514760.d0, &
  5524231.d0, &
  5533702.d0, &
  5543173.d0, &
  5552644.d0, &
  5562115.d0, &
  5571587.d0, &
  5581058.d0, &
  5590529.d0, &
  5600000.d0, &
  5600000.d0, &
  5607562.d0, &
  5615125.d0, &
  5622688.d0, &
  5630250.d0, &
  5637812.d0, &
  5645375.d0, &
  5652938.d0, &
  5660500.d0, &
  5668062.d0, &
  5675625.d0, &
  5683188.d0, &
  5690750.d0, &
  5698312.d0, &
  5705875.d0, &
  5713438.d0 /)

  Mref_V_radius_ref( 601 : 630 ) = (/ &
  5721000.d0, &
  5721000.d0, &
  5724572.d0, &
  5728143.d0, &
  5731714.d0, &
  5735286.d0, &
  5738857.d0, &
  5742428.d0, &
  5746000.d0, &
  5749572.d0, &
  5753143.d0, &
  5756714.d0, &
  5760286.d0, &
  5763857.d0, &
  5767428.d0, &
  5771000.d0, &
  5771000.d0, &
  5777334.d0, &
  5783666.d0, &
  5790000.d0, &
  5796334.d0, &
  5802666.d0, &
  5809000.d0, &
  5815334.d0, &
  5821666.d0, &
  5828000.d0, &
  5834334.d0, &
  5840666.d0, &
  5847000.d0, &
  5853334.d0 /)

  Mref_V_radius_ref( 631 : 660 ) = (/ &
  5859666.d0, &
  5866000.d0, &
  5872334.d0, &
  5878666.d0, &
  5885000.d0, &
  5891334.d0, &
  5897666.d0, &
  5904000.d0, &
  5910334.d0, &
  5916666.d0, &
  5923000.d0, &
  5929334.d0, &
  5935666.d0, &
  5942000.d0, &
  5948334.d0, &
  5954666.d0, &
  5961000.d0, &
  5961000.d0, & ! 410km depth
  5967334.d0, &
  5973666.d0, &
  5980000.d0, &
  5986334.d0, &
  5992666.d0, &
  5999000.d0, &
  6005334.d0, &
  6011666.d0, &
  6018000.d0, &
  6024334.d0, &
  6030666.d0, &
  6037000.d0 /)

  Mref_V_radius_ref( 661 : 690 ) = (/ &
  6043334.d0, &
  6049666.d0, &
  6056000.d0, &
  6062334.d0, &
  6068666.d0, &
  6075000.d0, &
  6081334.d0, &
  6087666.d0, &
  6094000.d0, &
  6100334.d0, &
  6106666.d0, &
  6113000.d0, &
  6119334.d0, &
  6125666.d0, &
  6132000.d0, &
  6138334.d0, &
  6144666.d0, &
  6151000.d0, &
  6151000.d0, & ! 220km depth
  6157087.d0, &
  6163174.d0, &
  6169261.d0, &
  6175348.d0, &
  6181435.d0, &
  6187522.d0, &
  6193609.d0, &
  6199696.d0, &
  6205783.d0, &
  6211870.d0, &
  6217957.d0 /)

  Mref_V_radius_ref( 691 : 720 ) = (/ &
  6224043.d0, &
  6230130.d0, &
  6236217.d0, &
  6242304.d0, &
  6248391.d0, &
  6254478.d0, &
  6260565.d0, &
  6266652.d0, &
  6272739.d0, &
  6278826.d0, &
  6284913.d0, &
  6291000.d0, &
  6291000.d0, & ! 80km depth
  6294971.d0, &
  6298943.d0, &
  6302914.d0, &
  6306886.d0, &
  6310857.d0, &
  6314829.d0, &
  6318800.d0, &
  6322771.d0, &
  6326743.d0, &
  6330714.d0, &
  6334686.d0, &
  6338657.d0, &
  6342629.d0, &
  6346600.d0, &
  6346600.d0, & ! crustal model: moho depth at 24.4km
  6347540.d0, &
  6348480.d0 /)

  Mref_V_radius_ref( 721 : 750 ) = (/ &
  6349420.d0, &
  6350360.d0, &
  6351300.d0, &
  6352240.d0, &
  6353180.d0, &
  6354120.d0, &
  6355060.d0, &
  6356000.d0, &
  6356000.d0, &
  6357200.d0, &
  6358400.d0, &
  6359600.d0, &
  6360800.d0, &
  6362000.d0, &
  6363200.d0, &
  6364400.d0, &
  6365600.d0, &
  6366800.d0, &
  6368000.d0, & ! 3km depth
  6368000.d0, &
  6368300.d0, &
  6368600.d0, &
  6368900.d0, &
  6369200.d0, &
  6369500.d0, &
  6369800.d0, &
  6370100.d0, &
  6370400.d0, &
  6370700.d0, &
  6371000.d0 /)

  Mref_V_density_ref( 1 : 30 ) = (/ &
  13088.48d0, &
  13088.47d0, &
  13088.44d0, &
  13088.39d0, &
  13088.32d0, &
  13088.22d0, &
  13088.11d0, &
  13087.98d0, &
  13087.83d0, &
  13087.66d0, &
  13087.46d0, &
  13087.25d0, &
  13087.02d0, &
  13086.76d0, &
  13086.49d0, &
  13086.2d0, &
  13085.88d0, &
  13085.55d0, &
  13085.19d0, &
  13084.82d0, &
  13084.42d0, &
  13084.01d0, &
  13083.57d0, &
  13083.11d0, &
  13082.64d0, &
  13082.14d0, &
  13081.62d0, &
  13081.09d0, &
  13080.53d0, &
  13079.95d0 /)

  Mref_V_density_ref( 31 : 60 ) = (/ &
  13079.35d0, &
  13078.73d0, &
  13078.09d0, &
  13077.44d0, &
  13076.76d0, &
  13076.06d0, &
  13075.34d0, &
  13074.6d0, &
  13073.84d0, &
  13073.06d0, &
  13072.25d0, &
  13071.43d0, &
  13070.59d0, &
  13069.73d0, &
  13068.85d0, &
  13067.95d0, &
  13067.02d0, &
  13066.08d0, &
  13065.12d0, &
  13064.13d0, &
  13063.13d0, &
  13062.1d0, &
  13061.06d0, &
  13060.d0, &
  13058.91d0, &
  13057.81d0, &
  13056.68d0, &
  13055.53d0, &
  13054.37d0, &
  13053.18d0 /)

  Mref_V_density_ref( 61 : 90 ) = (/ &
  13051.98d0, &
  13050.75d0, &
  13049.5d0, &
  13048.23d0, &
  13046.95d0, &
  13045.64d0, &
  13044.31d0, &
  13042.96d0, &
  13041.59d0, &
  13040.2d0, &
  13038.79d0, &
  13037.36d0, &
  13035.91d0, &
  13034.44d0, &
  13032.95d0, &
  13031.44d0, &
  13029.91d0, &
  13028.36d0, &
  13026.79d0, &
  13025.2d0, &
  13023.58d0, &
  13021.95d0, &
  13020.3d0, &
  13018.63d0, &
  13016.93d0, &
  13015.22d0, &
  13013.49d0, &
  13011.73d0, &
  13009.96d0, &
  13008.16d0 /)

  Mref_V_density_ref( 91 : 120 ) = (/ &
  13006.35d0, &
  13004.51d0, &
  13002.66d0, &
  13000.78d0, &
  12998.88d0, &
  12996.97d0, &
  12995.03d0, &
  12993.07d0, &
  12991.1d0, &
  12989.1d0, &
  12987.08d0, &
  12985.04d0, &
  12982.99d0, &
  12980.91d0, &
  12978.81d0, &
  12976.69d0, &
  12974.55d0, &
  12972.39d0, &
  12970.21d0, &
  12968.01d0, &
  12965.79d0, &
  12963.55d0, &
  12961.29d0, &
  12959.01d0, &
  12956.7d0, &
  12954.38d0, &
  12952.04d0, &
  12949.68d0, &
  12947.29d0, &
  12944.89d0 /)

  Mref_V_density_ref( 121 : 150 ) = (/ &
  12942.47d0, &
  12940.02d0, &
  12937.56d0, &
  12935.08d0, &
  12932.57d0, &
  12930.05d0, &
  12927.5d0, &
  12924.94d0, &
  12922.35d0, &
  12919.75d0, &
  12917.12d0, &
  12914.47d0, &
  12911.81d0, &
  12909.12d0, &
  12906.41d0, &
  12903.68d0, &
  12900.94d0, &
  12898.17d0, &
  12895.38d0, &
  12892.57d0, &
  12889.74d0, &
  12886.89d0, &
  12884.02d0, &
  12881.13d0, &
  12878.22d0, &
  12875.29d0, &
  12872.34d0, &
  12869.37d0, &
  12866.38d0, &
  12863.37d0 /)

  Mref_V_density_ref( 151 : 180 ) = (/ &
  12860.34d0, &
  12857.29d0, &
  12854.21d0, &
  12851.12d0, &
  12848.01d0, &
  12844.88d0, &
  12841.72d0, &
  12838.55d0, &
  12835.35d0, &
  12832.14d0, &
  12828.91d0, &
  12825.65d0, &
  12822.38d0, &
  12819.08d0, &
  12815.76d0, &
  12812.43d0, &
  12809.07d0, &
  12805.7d0, &
  12802.3d0, &
  12798.88d0, &
  12795.44d0, &
  12791.99d0, &
  12788.51d0, &
  12785.01d0, &
  12781.49d0, &
  12777.95d0, &
  12774.4d0, &
  12770.82d0, &
  12767.22d0, &
  12763.6d0 /)

  Mref_V_density_ref( 181 : 210 ) = (/ &
  12166.35d0, &
  12159.77d0, &
  12153.14d0, &
  12146.45d0, &
  12139.71d0, &
  12132.91d0, &
  12126.05d0, &
  12119.14d0, &
  12112.18d0, &
  12105.15d0, &
  12098.07d0, &
  12090.93d0, &
  12083.73d0, &
  12076.48d0, &
  12069.17d0, &
  12061.8d0, &
  12054.37d0, &
  12046.88d0, &
  12039.33d0, &
  12031.72d0, &
  12024.05d0, &
  12016.33d0, &
  12008.54d0, &
  12000.69d0, &
  11992.78d0, &
  11984.81d0, &
  11976.78d0, &
  11968.68d0, &
  11960.53d0, &
  11952.31d0 /)

  Mref_V_density_ref( 211 : 240 ) = (/ &
  11944.03d0, &
  11935.69d0, &
  11927.28d0, &
  11918.81d0, &
  11910.28d0, &
  11901.68d0, &
  11893.02d0, &
  11884.3d0, &
  11875.51d0, &
  11866.66d0, &
  11857.74d0, &
  11848.75d0, &
  11839.7d0, &
  11830.58d0, &
  11821.4d0, &
  11812.15d0, &
  11802.84d0, &
  11793.45d0, &
  11784.01d0, &
  11774.49d0, &
  11764.9d0, &
  11755.25d0, &
  11745.53d0, &
  11735.74d0, &
  11725.88d0, &
  11715.95d0, &
  11705.95d0, &
  11695.89d0, &
  11685.75d0, &
  11675.54d0 /)

  Mref_V_density_ref( 241 : 270 ) = (/ &
  11665.26d0, &
  11654.92d0, &
  11644.5d0, &
  11634.01d0, &
  11623.44d0, &
  11612.81d0, &
  11602.1d0, &
  11591.32d0, &
  11580.47d0, &
  11569.55d0, &
  11558.55d0, &
  11547.48d0, &
  11536.34d0, &
  11525.12d0, &
  11513.83d0, &
  11502.46d0, &
  11491.02d0, &
  11479.51d0, &
  11467.91d0, &
  11456.25d0, &
  11444.5d0, &
  11432.69d0, &
  11420.79d0, &
  11408.82d0, &
  11396.77d0, &
  11384.64d0, &
  11372.44d0, &
  11360.16d0, &
  11347.8d0, &
  11335.37d0 /)

  Mref_V_density_ref( 271 : 300 ) = (/ &
  11322.85d0, &
  11310.26d0, &
  11297.58d0, &
  11284.83d0, &
  11272.d0, &
  11259.09d0, &
  11246.1d0, &
  11233.03d0, &
  11219.87d0, &
  11206.64d0, &
  11193.33d0, &
  11179.93d0, &
  11166.45d0, &
  11152.89d0, &
  11139.25d0, &
  11125.53d0, &
  11111.72d0, &
  11097.83d0, &
  11083.86d0, &
  11069.8d0, &
  11055.66d0, &
  11041.44d0, &
  11027.13d0, &
  11012.74d0, &
  10998.26d0, &
  10983.7d0, &
  10969.05d0, &
  10954.32d0, &
  10939.5d0, &
  10924.59d0 /)

  Mref_V_density_ref( 301 : 330 ) = (/ &
  10909.6d0, &
  10894.52d0, &
  10879.35d0, &
  10864.1d0, &
  10848.76d0, &
  10833.33d0, &
  10817.81d0, &
  10802.21d0, &
  10786.51d0, &
  10770.73d0, &
  10754.86d0, &
  10738.9d0, &
  10722.85d0, &
  10706.71d0, &
  10690.48d0, &
  10674.16d0, &
  10657.75d0, &
  10641.24d0, &
  10624.65d0, &
  10607.96d0, &
  10591.19d0, &
  10574.32d0, &
  10557.36d0, &
  10540.3d0, &
  10523.16d0, &
  10505.92d0, &
  10488.58d0, &
  10471.15d0, &
  10453.63d0, &
  10436.02d0 /)

  Mref_V_density_ref( 331 : 360 ) = (/ &
  10418.31d0, &
  10400.51d0, &
  10382.61d0, &
  10364.61d0, &
  10346.52d0, &
  10328.34d0, &
  10310.05d0, &
  10291.68d0, &
  10273.2d0, &
  10254.63d0, &
  10235.96d0, &
  10217.2d0, &
  10198.33d0, &
  10179.37d0, &
  10160.31d0, &
  10141.15d0, &
  10121.9d0, &
  10102.54d0, &
  10083.09d0, &
  10063.53d0, &
  10043.88d0, &
  10024.12d0, &
  10004.27d0, &
  9984.32d0, &
  9964.26d0, &
  9944.1d0, &
  9923.84d0, &
  9903.48d0, &
  5566.45d0, &
  5561.75d0 /)

  Mref_V_density_ref( 361 : 390 ) = (/ &
  5557.05d0, &
  5552.36d0, &
  5547.66d0, &
  5542.97d0, &
  5538.28d0, &
  5533.59d0, &
  5528.9d0, &
  5524.21d0, &
  5519.53d0, &
  5514.85d0, &
  5510.16d0, &
  5505.48d0, &
  5500.81d0, &
  5496.13d0, &
  5491.45d0, &
  5491.45d0, &
  5486.73d0, &
  5482.01d0, &
  5477.29d0, &
  5472.57d0, &
  5467.85d0, &
  5463.13d0, &
  5458.42d0, &
  5453.7d0, &
  5448.99d0, &
  5444.27d0, &
  5439.56d0, &
  5434.85d0, &
  5430.13d0, &
  5425.42d0 /)

  Mref_V_density_ref( 391 : 420 ) = (/ &
  5420.71d0, &
  5416.d0, &
  5411.29d0, &
  5406.57d0, &
  5401.86d0, &
  5397.15d0, &
  5392.44d0, &
  5387.73d0, &
  5383.02d0, &
  5378.3d0, &
  5373.59d0, &
  5368.88d0, &
  5364.17d0, &
  5359.45d0, &
  5354.74d0, &
  5350.02d0, &
  5345.31d0, &
  5340.59d0, &
  5335.87d0, &
  5331.16d0, &
  5326.44d0, &
  5321.72d0, &
  5317.d0, &
  5312.28d0, &
  5307.55d0, &
  5302.83d0, &
  5298.1d0, &
  5293.38d0, &
  5288.65d0, &
  5283.92d0 /)

  Mref_V_density_ref( 421 : 450 ) = (/ &
  5279.19d0, &
  5274.45d0, &
  5269.72d0, &
  5264.98d0, &
  5260.25d0, &
  5255.51d0, &
  5250.77d0, &
  5246.02d0, &
  5241.28d0, &
  5236.53d0, &
  5231.78d0, &
  5227.03d0, &
  5222.27d0, &
  5217.52d0, &
  5212.76d0, &
  5208.d0, &
  5203.23d0, &
  5198.47d0, &
  5193.7d0, &
  5188.93d0, &
  5184.15d0, &
  5179.38d0, &
  5174.6d0, &
  5169.82d0, &
  5165.03d0, &
  5160.24d0, &
  5155.45d0, &
  5150.65d0, &
  5145.86d0, &
  5141.06d0 /)

  Mref_V_density_ref( 451 : 480 ) = (/ &
  5136.25d0, &
  5131.44d0, &
  5126.63d0, &
  5121.82d0, &
  5117.d0, &
  5112.18d0, &
  5107.35d0, &
  5102.52d0, &
  5097.69d0, &
  5092.85d0, &
  5088.01d0, &
  5083.16d0, &
  5078.31d0, &
  5073.46d0, &
  5068.6d0, &
  5063.74d0, &
  5058.87d0, &
  5054.d0, &
  5049.13d0, &
  5044.25d0, &
  5039.36d0, &
  5034.47d0, &
  5029.58d0, &
  5024.68d0, &
  5019.78d0, &
  5014.87d0, &
  5009.96d0, &
  5005.04d0, &
  5000.12d0, &
  4995.19d0 /)

  Mref_V_density_ref( 481 : 510 ) = (/ &
  4990.26d0, &
  4985.32d0, &
  4980.38d0, &
  4975.43d0, &
  4970.47d0, &
  4965.51d0, &
  4960.55d0, &
  4955.58d0, &
  4950.6d0, &
  4945.62d0, &
  4940.63d0, &
  4935.64d0, &
  4930.64d0, &
  4925.63d0, &
  4920.62d0, &
  4915.6d0, &
  4910.58d0, &
  4905.55d0, &
  4900.51d0, &
  4895.47d0, &
  4890.42d0, &
  4885.37d0, &
  4880.31d0, &
  4875.24d0, &
  4870.16d0, &
  4865.08d0, &
  4859.99d0, &
  4854.9d0, &
  4849.8d0, &
  4844.69d0 /)

  Mref_V_density_ref( 511 : 540 ) = (/ &
  4839.57d0, &
  4834.45d0, &
  4829.32d0, &
  4824.18d0, &
  4819.04d0, &
  4813.88d0, &
  4808.73d0, &
  4803.56d0, &
  4798.39d0, &
  4793.2d0, &
  4788.02d0, &
  4782.82d0, &
  4777.61d0, &
  4772.4d0, &
  4767.18d0, &
  4761.95d0, &
  4756.72d0, &
  4751.47d0, &
  4746.22d0, &
  4740.95d0, &
  4735.68d0, &
  4730.4d0, &
  4725.1d0, &
  4719.8d0, &
  4714.48d0, &
  4709.15d0, &
  4703.81d0, &
  4698.44d0, &
  4693.08d0, &
  4687.69d0 /)

  Mref_V_density_ref( 541 : 570 ) = (/ &
  4682.29d0, &
  4676.87d0, &
  4671.44d0, &
  4665.99d0, &
  4660.52d0, &
  4655.03d0, &
  4649.52d0, &
  4644.d0, &
  4638.46d0, &
  4632.89d0, &
  4627.31d0, &
  4621.7d0, &
  4616.08d0, &
  4610.44d0, &
  4604.76d0, &
  4599.08d0, &
  4593.36d0, &
  4587.63d0, &
  4581.86d0, &
  4576.07d0, &
  4570.26d0, &
  4564.43d0, &
  4558.56d0, &
  4552.67d0, &
  4546.76d0, &
  4540.82d0, &
  4534.84d0, &
  4528.85d0, &
  4522.81d0, &
  4516.76d0 /)

  Mref_V_density_ref( 571 : 600 ) = (/ &
  4510.67d0, &
  4504.56d0, &
  4498.41d0, &
  4492.23d0, &
  4486.02d0, &
  4479.78d0, &
  4473.51d0, &
  4467.2d0, &
  4460.87d0, &
  4454.49d0, &
  4448.08d0, &
  4441.63d0, &
  4435.13d0, &
  4428.6d0, &
  4428.59d0, &
  4423.32d0, &
  4418.01d0, &
  4412.67d0, &
  4407.3d0, &
  4401.9d0, &
  4396.45d0, &
  4390.96d0, &
  4385.4d0, &
  4379.81d0, &
  4374.16d0, &
  4368.47d0, &
  4362.72d0, &
  4356.92d0, &
  4351.08d0, &
  4345.18d0 /)

  Mref_V_density_ref( 601 : 630 ) = (/ &
  4339.24d0, &
  4047.01d0, &
  4042.5d0, &
  4037.98d0, &
  4033.48d0, &
  4028.95d0, &
  4024.45d0, &
  4019.93d0, &
  4015.42d0, &
  4010.9d0, &
  4006.38d0, &
  4001.87d0, &
  3997.35d0, &
  3992.84d0, &
  3988.32d0, &
  3983.8d0, &
  3983.8d0, &
  3975.79d0, &
  3967.77d0, &
  3959.76d0, &
  3951.75d0, &
  3943.73d0, &
  3935.71d0, &
  3927.69d0, &
  3919.67d0, &
  3911.65d0, &
  3903.61d0, &
  3895.58d0, &
  3887.56d0, &
  3879.53d0 /)

  Mref_V_density_ref( 631 : 660 ) = (/ &
  3871.5d0, &
  3863.46d0, &
  3855.42d0, &
  3847.38d0, &
  3839.33d0, &
  3831.27d0, &
  3823.22d0, &
  3815.16d0, &
  3807.09d0, &
  3799.01d0, &
  3790.94d0, &
  3782.85d0, &
  3774.78d0, &
  3766.66d0, &
  3758.56d0, &
  3750.45d0, &
  3742.34d0, &
  3554.91d0, &
  3551.d0, &
  3547.07d0, &
  3543.16d0, &
  3539.23d0, &
  3535.32d0, &
  3531.39d0, &
  3527.46d0, &
  3523.57d0, &
  3519.67d0, &
  3515.77d0, &
  3511.91d0, &
  3508.06d0 /)

  Mref_V_density_ref( 661 : 690 ) = (/ &
  3504.21d0, &
  3500.39d0, &
  3496.58d0, &
  3492.8d0, &
  3489.05d0, &
  3485.32d0, &
  3481.61d0, &
  3477.88d0, &
  3474.16d0, &
  3470.41d0, &
  3466.59d0, &
  3462.74d0, &
  3458.78d0, &
  3454.75d0, &
  3450.61d0, &
  3446.33d0, &
  3441.91d0, &
  3437.35d0, &
  3437.34d0, &
  3432.81d0, &
  3428.15d0, &
  3423.37d0, &
  3418.47d0, &
  3413.47d0, &
  3408.35d0, &
  3403.15d0, &
  3397.87d0, &
  3392.5d0, &
  3387.07d0, &
  3381.58d0 /)

  Mref_V_density_ref( 691 : 720 ) = (/ &
  3376.03d0, &
  3370.45d0, &
  3364.87d0, &
  3359.31d0, &
  3353.79d0, &
  3348.37d0, &
  3343.03d0, &
  3337.85d0, &
  3332.83d0, &
  3328.01d0, &
  3323.39d0, &
  3319.04d0, &
  3319.03d0, &
  3316.33d0, &
  3313.75d0, &
  3311.3d0, &
  3308.97d0, &
  3306.73d0, &
  3304.58d0, &
  3302.53d0, &
  3300.55d0, &
  3298.63d0, &
  3296.79d0, &
  3295.d0, &
  3293.25d0, &
  3291.54d0, &
  3289.84d0, &
  2900.d0, &
  2900.d0, &
  2900.d0 /)

  Mref_V_density_ref( 721 : 750 ) = (/ &
  2900.d0, &
  2900.d0, &
  2900.d0, &
  2900.d0, &
  2900.d0, &
  2900.d0, &
  2900.d0, &
  2900.d0, &
  2600.d0, &
  2600.d0, &
  2600.d0, &
  2600.d0, &
  2600.d0, &
  2600.d0, &
  2600.d0, &
  2600.d0, &
  2600.d0, &
  2600.d0, &
  2600.d0, &
  2600.d0, &
  2600.d0, &
  2600.d0, &
  2600.d0, &
  2600.d0, &
  2600.d0, &
  2600.d0, &
  2600.d0, &
  2600.d0, &
  2600.d0, &
  2600.d0 /)

  Mref_V_vpv_ref( 1 : 30 ) = (/ &
  11262.2d0, &
  11262.2d0, &
  11262.18d0, &
  11262.14d0, &
  11262.09d0, &
  11262.02d0, &
  11261.94d0, &
  11261.85d0, &
  11261.74d0, &
  11261.61d0, &
  11261.47d0, &
  11261.32d0, &
  11261.15d0, &
  11260.97d0, &
  11260.77d0, &
  11260.56d0, &
  11260.34d0, &
  11260.09d0, &
  11259.84d0, &
  11259.57d0, &
  11259.28d0, &
  11258.99d0, &
  11258.67d0, &
  11258.34d0, &
  11258.d0, &
  11257.64d0, &
  11257.27d0, &
  11256.88d0, &
  11256.48d0, &
  11256.06d0 /)

  Mref_V_vpv_ref( 31 : 60 ) = (/ &
  11255.63d0, &
  11255.19d0, &
  11254.73d0, &
  11254.25d0, &
  11253.76d0, &
  11253.26d0, &
  11252.74d0, &
  11252.21d0, &
  11251.66d0, &
  11251.1d0, &
  11250.52d0, &
  11249.93d0, &
  11249.33d0, &
  11248.71d0, &
  11248.07d0, &
  11247.42d0, &
  11246.76d0, &
  11246.08d0, &
  11245.38d0, &
  11244.67d0, &
  11243.95d0, &
  11243.21d0, &
  11242.46d0, &
  11241.7d0, &
  11240.91d0, &
  11240.12d0, &
  11239.31d0, &
  11238.48d0, &
  11237.64d0, &
  11236.79d0 /)

  Mref_V_vpv_ref( 61 : 90 ) = (/ &
  11235.92d0, &
  11235.04d0, &
  11234.14d0, &
  11233.23d0, &
  11232.3d0, &
  11231.36d0, &
  11230.4d0, &
  11229.43d0, &
  11228.44d0, &
  11227.44d0, &
  11226.43d0, &
  11225.4d0, &
  11224.36d0, &
  11223.3d0, &
  11222.22d0, &
  11221.14d0, &
  11220.03d0, &
  11218.92d0, &
  11217.78d0, &
  11216.64d0, &
  11215.48d0, &
  11214.3d0, &
  11213.11d0, &
  11211.91d0, &
  11210.69d0, &
  11209.45d0, &
  11208.21d0, &
  11206.94d0, &
  11205.67d0, &
  11204.37d0 /)

  Mref_V_vpv_ref( 91 : 120 ) = (/ &
  11203.07d0, &
  11201.74d0, &
  11200.41d0, &
  11199.06d0, &
  11197.69d0, &
  11196.31d0, &
  11194.92d0, &
  11193.51d0, &
  11192.09d0, &
  11190.65d0, &
  11189.19d0, &
  11187.73d0, &
  11186.24d0, &
  11184.75d0, &
  11183.24d0, &
  11181.71d0, &
  11180.17d0, &
  11178.61d0, &
  11177.04d0, &
  11175.46d0, &
  11173.86d0, &
  11172.25d0, &
  11170.62d0, &
  11168.98d0, &
  11167.32d0, &
  11165.65d0, &
  11163.96d0, &
  11162.26d0, &
  11160.54d0, &
  11158.81d0 /)

  Mref_V_vpv_ref( 121 : 150 ) = (/ &
  11157.07d0, &
  11155.31d0, &
  11153.54d0, &
  11151.75d0, &
  11149.94d0, &
  11148.13d0, &
  11146.29d0, &
  11144.45d0, &
  11142.58d0, &
  11140.71d0, &
  11138.82d0, &
  11136.91d0, &
  11134.99d0, &
  11133.06d0, &
  11131.11d0, &
  11129.14d0, &
  11127.16d0, &
  11125.17d0, &
  11123.16d0, &
  11121.14d0, &
  11119.1d0, &
  11117.05d0, &
  11114.99d0, &
  11112.9d0, &
  11110.81d0, &
  11108.7d0, &
  11106.57d0, &
  11104.44d0, &
  11102.28d0, &
  11100.11d0 /)

  Mref_V_vpv_ref( 151 : 180 ) = (/ &
  11097.93d0, &
  11095.73d0, &
  11093.52d0, &
  11091.29d0, &
  11089.05d0, &
  11086.8d0, &
  11084.53d0, &
  11082.24d0, &
  11079.94d0, &
  11077.63d0, &
  11075.3d0, &
  11072.95d0, &
  11070.59d0, &
  11068.22d0, &
  11065.83d0, &
  11063.43d0, &
  11061.02d0, &
  11058.58d0, &
  11056.14d0, &
  11053.68d0, &
  11051.2d0, &
  11048.71d0, &
  11046.21d0, &
  11043.69d0, &
  11041.16d0, &
  11038.61d0, &
  11036.05d0, &
  11033.47d0, &
  11030.88d0, &
  11028.27d0 /)

  Mref_V_vpv_ref( 181 : 210 ) = (/ &
  10355.69d0, &
  10348.28d0, &
  10340.85d0, &
  10333.39d0, &
  10325.91d0, &
  10318.4d0, &
  10310.87d0, &
  10303.3d0, &
  10295.71d0, &
  10288.09d0, &
  10280.44d0, &
  10272.76d0, &
  10265.04d0, &
  10257.3d0, &
  10249.52d0, &
  10241.71d0, &
  10233.86d0, &
  10225.98d0, &
  10218.06d0, &
  10210.11d0, &
  10202.12d0, &
  10194.1d0, &
  10186.04d0, &
  10177.94d0, &
  10169.79d0, &
  10161.61d0, &
  10153.39d0, &
  10145.13d0, &
  10136.83d0, &
  10128.48d0 /)

  Mref_V_vpv_ref( 211 : 240 ) = (/ &
  10120.09d0, &
  10111.66d0, &
  10103.18d0, &
  10094.66d0, &
  10086.09d0, &
  10077.48d0, &
  10068.82d0, &
  10060.11d0, &
  10051.35d0, &
  10042.54d0, &
  10033.69d0, &
  10024.78d0, &
  10015.82d0, &
  10006.82d0, &
  9997.75d0, &
  9988.64d0, &
  9979.47d0, &
  9970.25d0, &
  9960.97d0, &
  9951.64d0, &
  9942.25d0, &
  9932.81d0, &
  9923.31d0, &
  9913.75d0, &
  9904.13d0, &
  9894.45d0, &
  9884.71d0, &
  9874.91d0, &
  9865.05d0, &
  9855.13d0 /)

  Mref_V_vpv_ref( 241 : 270 ) = (/ &
  9845.14d0, &
  9835.09d0, &
  9824.98d0, &
  9814.8d0, &
  9804.56d0, &
  9794.25d0, &
  9783.87d0, &
  9773.43d0, &
  9762.92d0, &
  9752.34d0, &
  9741.69d0, &
  9730.97d0, &
  9720.18d0, &
  9709.32d0, &
  9698.39d0, &
  9687.38d0, &
  9676.31d0, &
  9665.15d0, &
  9653.93d0, &
  9642.63d0, &
  9631.25d0, &
  9619.8d0, &
  9608.27d0, &
  9596.66d0, &
  9584.97d0, &
  9573.2d0, &
  9561.36d0, &
  9549.43d0, &
  9537.43d0, &
  9525.34d0 /)

  Mref_V_vpv_ref( 271 : 300 ) = (/ &
  9513.17d0, &
  9500.91d0, &
  9488.57d0, &
  9476.15d0, &
  9463.64d0, &
  9451.05d0, &
  9438.37d0, &
  9425.61d0, &
  9412.75d0, &
  9399.81d0, &
  9386.78d0, &
  9373.66d0, &
  9360.45d0, &
  9347.15d0, &
  9333.76d0, &
  9320.27d0, &
  9306.7d0, &
  9293.03d0, &
  9279.26d0, &
  9265.4d0, &
  9251.45d0, &
  9237.4d0, &
  9223.25d0, &
  9209.d0, &
  9194.66d0, &
  9180.22d0, &
  9165.68d0, &
  9151.03d0, &
  9136.29d0, &
  9121.45d0 /)

  Mref_V_vpv_ref( 301 : 330 ) = (/ &
  9106.5d0, &
  9091.46d0, &
  9076.3d0, &
  9061.05d0, &
  9045.69d0, &
  9030.23d0, &
  9014.65d0, &
  8998.98d0, &
  8983.19d0, &
  8967.3d0, &
  8951.3d0, &
  8935.19d0, &
  8918.97d0, &
  8902.64d0, &
  8886.2d0, &
  8869.64d0, &
  8852.98d0, &
  8836.2d0, &
  8819.31d0, &
  8802.3d0, &
  8785.18d0, &
  8767.94d0, &
  8750.59d0, &
  8733.12d0, &
  8715.53d0, &
  8697.82d0, &
  8680.d0, &
  8662.05d0, &
  8643.99d0, &
  8625.8d0 /)

  Mref_V_vpv_ref( 331 : 360 ) = (/ &
  8607.49d0, &
  8589.06d0, &
  8570.51d0, &
  8551.83d0, &
  8533.03d0, &
  8514.1d0, &
  8495.05d0, &
  8475.87d0, &
  8456.57d0, &
  8437.14d0, &
  8417.58d0, &
  8397.89d0, &
  8378.07d0, &
  8358.12d0, &
  8338.04d0, &
  8317.83d0, &
  8297.49d0, &
  8277.01d0, &
  8256.41d0, &
  8235.66d0, &
  8214.79d0, &
  8193.77d0, &
  8172.62d0, &
  8151.34d0, &
  8129.92d0, &
  8108.36d0, &
  8086.66d0, &
  8064.82d0, &
  13716.6d0, &
  13714.29d0 /)

  Mref_V_vpv_ref( 361 : 390 ) = (/ &
  13712.d0, &
  13709.7d0, &
  13707.42d0, &
  13705.14d0, &
  13702.86d0, &
  13700.59d0, &
  13698.33d0, &
  13696.07d0, &
  13693.82d0, &
  13691.57d0, &
  13689.33d0, &
  13687.09d0, &
  13684.86d0, &
  13682.63d0, &
  13680.41d0, &
  13680.41d0, &
  13668.9d0, &
  13657.43d0, &
  13645.97d0, &
  13634.54d0, &
  13623.14d0, &
  13611.76d0, &
  13600.4d0, &
  13589.07d0, &
  13577.76d0, &
  13566.47d0, &
  13555.2d0, &
  13543.95d0, &
  13532.72d0, &
  13521.51d0 /)

  Mref_V_vpv_ref( 391 : 420 ) = (/ &
  13510.32d0, &
  13499.14d0, &
  13487.99d0, &
  13476.85d0, &
  13465.73d0, &
  13454.63d0, &
  13443.54d0, &
  13432.46d0, &
  13421.41d0, &
  13410.36d0, &
  13399.33d0, &
  13388.31d0, &
  13377.31d0, &
  13366.31d0, &
  13355.33d0, &
  13344.36d0, &
  13333.4d0, &
  13322.45d0, &
  13311.51d0, &
  13300.58d0, &
  13289.66d0, &
  13278.74d0, &
  13267.84d0, &
  13256.93d0, &
  13246.04d0, &
  13235.15d0, &
  13224.27d0, &
  13213.39d0, &
  13202.51d0, &
  13191.64d0 /)

  Mref_V_vpv_ref( 421 : 450 ) = (/ &
  13180.78d0, &
  13169.91d0, &
  13159.05d0, &
  13148.19d0, &
  13137.33d0, &
  13126.47d0, &
  13115.61d0, &
  13104.75d0, &
  13093.89d0, &
  13083.02d0, &
  13072.16d0, &
  13061.29d0, &
  13050.42d0, &
  13039.55d0, &
  13028.67d0, &
  13017.78d0, &
  13006.9d0, &
  12996.d0, &
  12985.1d0, &
  12974.19d0, &
  12963.28d0, &
  12952.36d0, &
  12941.42d0, &
  12930.48d0, &
  12919.54d0, &
  12908.58d0, &
  12897.61d0, &
  12886.63d0, &
  12875.63d0, &
  12864.63d0 /)

  Mref_V_vpv_ref( 451 : 480 ) = (/ &
  12853.61d0, &
  12842.58d0, &
  12831.54d0, &
  12820.48d0, &
  12809.41d0, &
  12798.32d0, &
  12787.22d0, &
  12776.1d0, &
  12764.96d0, &
  12753.81d0, &
  12742.63d0, &
  12731.44d0, &
  12720.24d0, &
  12709.01d0, &
  12697.76d0, &
  12686.49d0, &
  12675.2d0, &
  12663.89d0, &
  12652.56d0, &
  12641.2d0, &
  12629.82d0, &
  12618.42d0, &
  12606.99d0, &
  12595.54d0, &
  12584.06d0, &
  12572.56d0, &
  12561.03d0, &
  12549.48d0, &
  12537.89d0, &
  12526.28d0 /)

  Mref_V_vpv_ref( 481 : 510 ) = (/ &
  12514.64d0, &
  12502.98d0, &
  12491.28d0, &
  12479.55d0, &
  12467.79d0, &
  12456.01d0, &
  12444.19d0, &
  12432.33d0, &
  12420.45d0, &
  12408.53d0, &
  12396.58d0, &
  12384.6d0, &
  12372.58d0, &
  12360.52d0, &
  12348.43d0, &
  12336.3d0, &
  12324.14d0, &
  12311.94d0, &
  12299.7d0, &
  12287.42d0, &
  12275.11d0, &
  12262.75d0, &
  12250.35d0, &
  12237.92d0, &
  12225.44d0, &
  12212.92d0, &
  12200.36d0, &
  12187.76d0, &
  12175.11d0, &
  12162.43d0 /)

  Mref_V_vpv_ref( 511 : 540 ) = (/ &
  12149.69d0, &
  12136.91d0, &
  12124.09d0, &
  12111.22d0, &
  12098.31d0, &
  12085.34d0, &
  12072.34d0, &
  12059.28d0, &
  12046.17d0, &
  12033.02d0, &
  12019.82d0, &
  12006.56d0, &
  11993.26d0, &
  11979.9d0, &
  11966.5d0, &
  11953.04d0, &
  11939.53d0, &
  11925.97d0, &
  11912.35d0, &
  11898.69d0, &
  11884.96d0, &
  11871.19d0, &
  11857.37d0, &
  11843.48d0, &
  11829.55d0, &
  11815.57d0, &
  11801.53d0, &
  11787.44d0, &
  11773.3d0, &
  11759.1d0 /)

  Mref_V_vpv_ref( 541 : 570 ) = (/ &
  11744.85d0, &
  11730.55d0, &
  11716.18d0, &
  11701.78d0, &
  11687.31d0, &
  11672.8d0, &
  11658.23d0, &
  11643.6d0, &
  11628.92d0, &
  11614.19d0, &
  11599.4d0, &
  11584.57d0, &
  11569.68d0, &
  11554.72d0, &
  11539.72d0, &
  11524.67d0, &
  11509.56d0, &
  11494.39d0, &
  11479.17d0, &
  11463.89d0, &
  11448.55d0, &
  11433.17d0, &
  11417.73d0, &
  11402.23d0, &
  11386.68d0, &
  11371.07d0, &
  11355.41d0, &
  11339.69d0, &
  11323.91d0, &
  11308.09d0 /)

  Mref_V_vpv_ref( 571 : 600 ) = (/ &
  11292.2d0, &
  11276.25d0, &
  11260.25d0, &
  11244.19d0, &
  11228.08d0, &
  11211.9d0, &
  11195.67d0, &
  11179.38d0, &
  11163.04d0, &
  11146.63d0, &
  11130.18d0, &
  11113.67d0, &
  11097.11d0, &
  11080.51d0, &
  11080.51d0, &
  11063.01d0, &
  11045.22d0, &
  11026.82d0, &
  11008.47d0, &
  10989.04d0, &
  10969.63d0, &
  10948.76d0, &
  10928.02d0, &
  10907.42d0, &
  10886.94d0, &
  10866.6d0, &
  10846.41d0, &
  10826.35d0, &
  10806.42d0, &
  10786.61d0 /)

  Mref_V_vpv_ref( 601 : 630 ) = (/ &
  10766.9d0, &
  10278.88d0, &
  10261.87d0, &
  10244.84d0, &
  10227.82d0, &
  10210.8d0, &
  10193.78d0, &
  10176.77d0, &
  10159.74d0, &
  10142.72d0, &
  10125.71d0, &
  10108.7d0, &
  10091.68d0, &
  10074.68d0, &
  10057.68d0, &
  10040.64d0, &
  10040.67d0, &
  10010.52d0, &
  9980.51d0, &
  9950.64d0, &
  9920.91d0, &
  9891.35d0, &
  9861.96d0, &
  9832.79d0, &
  9803.79d0, &
  9774.98d0, &
  9746.41d0, &
  9718.08d0, &
  9689.96d0, &
  9662.1d0 /)

  Mref_V_vpv_ref( 631 : 660 ) = (/ &
  9634.47d0, &
  9607.11d0, &
  9579.97d0, &
  9553.08d0, &
  9526.38d0, &
  9499.78d0, &
  9473.25d0, &
  9446.74d0, &
  9420.19d0, &
  9393.55d0, &
  9366.75d0, &
  9339.76d0, &
  9312.5d0, &
  9284.96d0, &
  9257.04d0, &
  9228.73d0, &
  9199.94d0, &
  8940.94d0, &
  8930.61d0, &
  8920.22d0, &
  8909.68d0, &
  8898.47d0, &
  8886.28d0, &
  8873.03d0, &
  8858.58d0, &
  8842.82d0, &
  8825.64d0, &
  8806.94d0, &
  8786.67d0, &
  8764.85d0 /)

  Mref_V_vpv_ref( 661 : 690 ) = (/ &
  8741.49d0, &
  8716.63d0, &
  8690.3d0, &
  8662.5d0, &
  8633.28d0, &
  8602.66d0, &
  8570.81d0, &
  8538.06d0, &
  8504.66d0, &
  8470.92d0, &
  8437.13d0, &
  8403.52d0, &
  8370.42d0, &
  8338.11d0, &
  8306.25d0, &
  8275.42d0, &
  8241.77d0, &
  8207.37d0, &
  8207.01d0, &
  8174.32d0, &
  8141.99d0, &
  8110.4d0, &
  8079.71d0, &
  8050.15d0, &
  8021.89d0, &
  7995.08d0, &
  7969.97d0, &
  7946.7d0, &
  7925.45d0, &
  7906.44d0 /)

  Mref_V_vpv_ref( 691 : 720 ) = (/ &
  7889.8d0, &
  7875.56d0, &
  7863.64d0, &
  7853.87d0, &
  7846.17d0, &
  7840.38d0, &
  7836.39d0, &
  7834.11d0, &
  7833.38d0, &
  7834.11d0, &
  7836.11d0, &
  7839.12d0, &
  7839.37d0, &
  7841.82d0, &
  7844.77d0, &
  7848.07d0, &
  7851.72d0, &
  7855.75d0, &
  7860.14d0, &
  7864.89d0, &
  7870.01d0, &
  7875.49d0, &
  7881.33d0, &
  7887.54d0, &
  7894.13d0, &
  7901.1d0, &
  7908.24d0, &
  6800.d0, &
  6800.d0, &
  6800.d0 /)

  Mref_V_vpv_ref( 721 : 750 ) = (/ &
  6800.d0, &
  6800.d0, &
  6800.d0, &
  6800.d0, &
  6800.d0, &
  6800.d0, &
  6800.d0, &
  6800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0 /)

  Mref_V_vsv_ref( 1 : 30 ) = (/ &
  3667.8d0, &
  3667.79d0, &
  3667.78d0, &
  3667.75d0, &
  3667.72d0, &
  3667.67d0, &
  3667.62d0, &
  3667.55d0, &
  3667.47d0, &
  3667.39d0, &
  3667.29d0, &
  3667.18d0, &
  3667.06d0, &
  3666.94d0, &
  3666.8d0, &
  3666.65d0, &
  3666.49d0, &
  3666.32d0, &
  3666.15d0, &
  3665.96d0, &
  3665.76d0, &
  3665.55d0, &
  3665.33d0, &
  3665.1d0, &
  3664.86d0, &
  3664.61d0, &
  3664.35d0, &
  3664.08d0, &
  3663.8d0, &
  3663.51d0 /)

  Mref_V_vsv_ref( 31 : 60 ) = (/ &
  3663.21d0, &
  3662.9d0, &
  3662.57d0, &
  3662.24d0, &
  3661.9d0, &
  3661.55d0, &
  3661.19d0, &
  3660.81d0, &
  3660.43d0, &
  3660.04d0, &
  3659.64d0, &
  3659.22d0, &
  3658.8d0, &
  3658.36d0, &
  3657.92d0, &
  3657.47d0, &
  3657.d0, &
  3656.53d0, &
  3656.04d0, &
  3655.55d0, &
  3655.04d0, &
  3654.53d0, &
  3654.d0, &
  3653.47d0, &
  3652.92d0, &
  3652.36d0, &
  3651.8d0, &
  3651.22d0, &
  3650.63d0, &
  3650.04d0 /)

  Mref_V_vsv_ref( 61 : 90 ) = (/ &
  3649.43d0, &
  3648.81d0, &
  3648.19d0, &
  3647.55d0, &
  3646.9d0, &
  3646.24d0, &
  3645.57d0, &
  3644.89d0, &
  3644.21d0, &
  3643.51d0, &
  3642.8d0, &
  3642.08d0, &
  3641.35d0, &
  3640.61d0, &
  3639.86d0, &
  3639.1d0, &
  3638.33d0, &
  3637.55d0, &
  3636.76d0, &
  3635.96d0, &
  3635.14d0, &
  3634.32d0, &
  3633.49d0, &
  3632.65d0, &
  3631.8d0, &
  3630.93d0, &
  3630.06d0, &
  3629.18d0, &
  3628.29d0, &
  3627.38d0 /)

  Mref_V_vsv_ref( 91 : 120 ) = (/ &
  3626.47d0, &
  3625.55d0, &
  3624.61d0, &
  3623.67d0, &
  3622.71d0, &
  3621.75d0, &
  3620.78d0, &
  3619.79d0, &
  3618.8d0, &
  3617.79d0, &
  3616.78d0, &
  3615.75d0, &
  3614.71d0, &
  3613.67d0, &
  3612.61d0, &
  3611.55d0, &
  3610.47d0, &
  3609.38d0, &
  3608.28d0, &
  3607.18d0, &
  3606.06d0, &
  3604.93d0, &
  3603.79d0, &
  3602.65d0, &
  3601.49d0, &
  3600.32d0, &
  3599.14d0, &
  3597.95d0, &
  3596.75d0, &
  3595.54d0 /)

  Mref_V_vsv_ref( 121 : 150 ) = (/ &
  3594.32d0, &
  3593.1d0, &
  3591.86d0, &
  3590.61d0, &
  3589.34d0, &
  3588.07d0, &
  3586.79d0, &
  3585.5d0, &
  3584.2d0, &
  3582.89d0, &
  3581.57d0, &
  3580.24d0, &
  3578.9d0, &
  3577.54d0, &
  3576.18d0, &
  3574.81d0, &
  3573.43d0, &
  3572.03d0, &
  3570.63d0, &
  3569.22d0, &
  3567.79d0, &
  3566.36d0, &
  3564.91d0, &
  3563.46d0, &
  3562.d0, &
  3560.52d0, &
  3559.04d0, &
  3557.54d0, &
  3556.04d0, &
  3554.52d0 /)

  Mref_V_vsv_ref( 151 : 180 ) = (/ &
  3553.d0, &
  3551.46d0, &
  3549.91d0, &
  3548.36d0, &
  3546.79d0, &
  3545.21d0, &
  3543.63d0, &
  3542.03d0, &
  3540.42d0, &
  3538.81d0, &
  3537.18d0, &
  3535.54d0, &
  3533.89d0, &
  3532.23d0, &
  3530.57d0, &
  3528.89d0, &
  3527.2d0, &
  3525.5d0, &
  3523.79d0, &
  3522.07d0, &
  3520.34d0, &
  3518.6d0, &
  3516.85d0, &
  3515.09d0, &
  3513.32d0, &
  3511.54d0, &
  3509.75d0, &
  3507.95d0, &
  3506.13d0, &
  3504.31d0 /)

  Mref_V_vsv_ref( 181 : 210 ) = (/ &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0 /)

  Mref_V_vsv_ref( 211 : 240 ) = (/ &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0 /)

  Mref_V_vsv_ref( 241 : 270 ) = (/ &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0 /)

  Mref_V_vsv_ref( 271 : 300 ) = (/ &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0 /)

  Mref_V_vsv_ref( 301 : 330 ) = (/ &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0 /)

  Mref_V_vsv_ref( 331 : 360 ) = (/ &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  7264.66d0, &
  7264.75d0 /)

  Mref_V_vsv_ref( 361 : 390 ) = (/ &
  7264.85d0, &
  7264.94d0, &
  7265.03d0, &
  7265.12d0, &
  7265.21d0, &
  7265.29d0, &
  7265.38d0, &
  7265.46d0, &
  7265.54d0, &
  7265.62d0, &
  7265.69d0, &
  7265.76d0, &
  7265.84d0, &
  7265.91d0, &
  7265.97d0, &
  7265.97d0, &
  7261.63d0, &
  7257.29d0, &
  7252.97d0, &
  7248.64d0, &
  7244.33d0, &
  7240.01d0, &
  7235.71d0, &
  7231.41d0, &
  7227.12d0, &
  7222.83d0, &
  7218.55d0, &
  7214.27d0, &
  7210.d0, &
  7205.73d0 /)

  Mref_V_vsv_ref( 391 : 420 ) = (/ &
  7201.47d0, &
  7197.21d0, &
  7192.95d0, &
  7188.7d0, &
  7184.45d0, &
  7180.21d0, &
  7175.97d0, &
  7171.73d0, &
  7167.5d0, &
  7163.27d0, &
  7159.04d0, &
  7154.81d0, &
  7150.59d0, &
  7146.37d0, &
  7142.15d0, &
  7137.93d0, &
  7133.71d0, &
  7129.5d0, &
  7125.29d0, &
  7121.07d0, &
  7116.86d0, &
  7112.65d0, &
  7108.44d0, &
  7104.23d0, &
  7100.02d0, &
  7095.81d0, &
  7091.6d0, &
  7087.39d0, &
  7083.18d0, &
  7078.96d0 /)

  Mref_V_vsv_ref( 421 : 450 ) = (/ &
  7074.75d0, &
  7070.54d0, &
  7066.32d0, &
  7062.1d0, &
  7057.88d0, &
  7053.66d0, &
  7049.44d0, &
  7045.22d0, &
  7040.99d0, &
  7036.76d0, &
  7032.52d0, &
  7028.29d0, &
  7024.05d0, &
  7019.81d0, &
  7015.56d0, &
  7011.31d0, &
  7007.06d0, &
  7002.8d0, &
  6998.54d0, &
  6994.27d0, &
  6990.d0, &
  6985.72d0, &
  6981.44d0, &
  6977.15d0, &
  6972.86d0, &
  6968.57d0, &
  6964.26d0, &
  6959.95d0, &
  6955.64d0, &
  6951.32d0 /)

  Mref_V_vsv_ref( 451 : 480 ) = (/ &
  6946.99d0, &
  6942.66d0, &
  6938.31d0, &
  6933.97d0, &
  6929.61d0, &
  6925.25d0, &
  6920.88d0, &
  6916.5d0, &
  6912.11d0, &
  6907.72d0, &
  6903.32d0, &
  6898.91d0, &
  6894.49d0, &
  6890.06d0, &
  6885.62d0, &
  6881.17d0, &
  6876.72d0, &
  6872.25d0, &
  6867.78d0, &
  6863.29d0, &
  6858.8d0, &
  6854.29d0, &
  6849.78d0, &
  6845.25d0, &
  6840.71d0, &
  6836.16d0, &
  6831.6d0, &
  6827.03d0, &
  6822.45d0, &
  6817.85d0 /)

  Mref_V_vsv_ref( 481 : 510 ) = (/ &
  6813.25d0, &
  6808.63d0, &
  6804.d0, &
  6799.35d0, &
  6794.7d0, &
  6790.03d0, &
  6785.34d0, &
  6780.65d0, &
  6775.94d0, &
  6771.22d0, &
  6766.48d0, &
  6761.73d0, &
  6756.97d0, &
  6752.19d0, &
  6747.4d0, &
  6742.59d0, &
  6737.76d0, &
  6732.93d0, &
  6728.07d0, &
  6723.21d0, &
  6718.32d0, &
  6713.42d0, &
  6708.51d0, &
  6703.57d0, &
  6698.62d0, &
  6693.66d0, &
  6688.68d0, &
  6683.68d0, &
  6678.66d0, &
  6673.63d0 /)

  Mref_V_vsv_ref( 511 : 540 ) = (/ &
  6668.58d0, &
  6663.51d0, &
  6658.43d0, &
  6653.32d0, &
  6648.2d0, &
  6643.06d0, &
  6637.9d0, &
  6632.73d0, &
  6627.53d0, &
  6622.31d0, &
  6617.08d0, &
  6611.82d0, &
  6606.55d0, &
  6601.26d0, &
  6595.94d0, &
  6590.61d0, &
  6584.91d0, &
  6579.51d0, &
  6574.11d0, &
  6568.67d0, &
  6563.22d0, &
  6557.74d0, &
  6552.24d0, &
  6546.73d0, &
  6541.19d0, &
  6535.63d0, &
  6530.05d0, &
  6524.44d0, &
  6518.82d0, &
  6513.17d0 /)

  Mref_V_vsv_ref( 541 : 570 ) = (/ &
  6507.5d0, &
  6501.8d0, &
  6496.09d0, &
  6490.35d0, &
  6484.59d0, &
  6478.8d0, &
  6472.99d0, &
  6467.16d0, &
  6461.3d0, &
  6455.42d0, &
  6449.51d0, &
  6443.58d0, &
  6437.63d0, &
  6431.65d0, &
  6425.65d0, &
  6419.61d0, &
  6413.56d0, &
  6407.48d0, &
  6401.37d0, &
  6395.25d0, &
  6389.09d0, &
  6382.91d0, &
  6376.7d0, &
  6370.46d0, &
  6364.2d0, &
  6357.91d0, &
  6351.59d0, &
  6345.25d0, &
  6338.88d0, &
  6332.49d0 /)

  Mref_V_vsv_ref( 571 : 600 ) = (/ &
  6326.05d0, &
  6319.6d0, &
  6313.13d0, &
  6306.62d0, &
  6300.08d0, &
  6293.52d0, &
  6286.92d0, &
  6280.29d0, &
  6273.64d0, &
  6266.96d0, &
  6260.25d0, &
  6253.51d0, &
  6246.75d0, &
  6239.95d0, &
  6239.95d0, &
  6219.68d0, &
  6200.29d0, &
  6181.16d0, &
  6162.04d0, &
  6143.01d0, &
  6123.98d0, &
  6103.71d0, &
  6083.53d0, &
  6063.45d0, &
  6043.44d0, &
  6023.52d0, &
  6003.73d0, &
  5984.03d0, &
  5964.38d0, &
  5944.81d0 /)

  Mref_V_vsv_ref( 601 : 630 ) = (/ &
  5925.27d0, &
  5550.32d0, &
  5541.2d0, &
  5532.08d0, &
  5522.96d0, &
  5513.83d0, &
  5504.71d0, &
  5495.59d0, &
  5486.47d0, &
  5477.35d0, &
  5468.22d0, &
  5459.1d0, &
  5449.97d0, &
  5440.84d0, &
  5431.71d0, &
  5422.57d0, &
  5422.59d0, &
  5406.39d0, &
  5390.3d0, &
  5374.34d0, &
  5358.52d0, &
  5342.83d0, &
  5327.31d0, &
  5311.92d0, &
  5296.73d0, &
  5281.71d0, &
  5266.86d0, &
  5252.21d0, &
  5237.78d0, &
  5223.55d0 /)

  Mref_V_vsv_ref( 631 : 660 ) = (/ &
  5209.54d0, &
  5195.72d0, &
  5182.1d0, &
  5168.69d0, &
  5155.42d0, &
  5142.22d0, &
  5129.05d0, &
  5115.84d0, &
  5102.55d0, &
  5089.14d0, &
  5075.5d0, &
  5061.63d0, &
  5047.46d0, &
  5032.93d0, &
  5018.03d0, &
  5002.66d0, &
  4986.77d0, &
  4802.15d0, &
  4798.23d0, &
  4794.28d0, &
  4790.38d0, &
  4785.95d0, &
  4780.83d0, &
  4775.01d0, &
  4768.45d0, &
  4761.12d0, &
  4752.97d0, &
  4744.01d0, &
  4734.25d0, &
  4723.77d0 /)

  Mref_V_vsv_ref( 661 : 690 ) = (/ &
  4712.7d0, &
  4701.12d0, &
  4689.11d0, &
  4676.77d0, &
  4664.2d0, &
  4651.49d0, &
  4638.69d0, &
  4625.88d0, &
  4613.07d0, &
  4600.31d0, &
  4587.67d0, &
  4575.18d0, &
  4562.88d0, &
  4550.85d0, &
  4539.08d0, &
  4527.67d0, &
  4516.65d0, &
  4506.09d0, &
  4506.d0, &
  4496.29d0, &
  4487.d0, &
  4478.17d0, &
  4469.83d0, &
  4462.d0, &
  4454.69d0, &
  4447.94d0, &
  4441.76d0, &
  4436.18d0, &
  4431.2d0, &
  4426.83d0 /)

  Mref_V_vsv_ref( 691 : 720 ) = (/ &
  4423.12d0, &
  4420.09d0, &
  4417.81d0, &
  4416.3d0, &
  4415.67d0, &
  4415.93d0, &
  4417.15d0, &
  4419.42d0, &
  4422.78d0, &
  4427.25d0, &
  4432.88d0, &
  4439.57d0, &
  4439.74d0, &
  4444.71d0, &
  4450.28d0, &
  4456.35d0, &
  4462.89d0, &
  4469.94d0, &
  4477.4d0, &
  4485.33d0, &
  4493.69d0, &
  4502.48d0, &
  4511.66d0, &
  4521.24d0, &
  4531.23d0, &
  4541.57d0, &
  4552.08d0, &
  3900.d0, &
  3900.d0, &
  3900.d0 /)

  Mref_V_vsv_ref( 721 : 750 ) = (/ &
  3900.d0, &
  3900.d0, &
  3900.d0, &
  3900.d0, &
  3900.d0, &
  3900.d0, &
  3900.d0, &
  3900.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0 /)

  Mref_V_Qkappa_ref( 1 : 30 ) = (/ &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0 /)

  Mref_V_Qkappa_ref( 31 : 60 ) = (/ &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0 /)

  Mref_V_Qkappa_ref( 61 : 90 ) = (/ &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0 /)

  Mref_V_Qkappa_ref( 91 : 120 ) = (/ &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0 /)

  Mref_V_Qkappa_ref( 121 : 150 ) = (/ &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0 /)

  Mref_V_Qkappa_ref( 151 : 180 ) = (/ &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0, &
  1327.6d0 /)

  Mref_V_Qkappa_ref( 181 : 210 ) = (/ &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0 /)

  Mref_V_Qkappa_ref( 211 : 240 ) = (/ &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0 /)

  Mref_V_Qkappa_ref( 241 : 270 ) = (/ &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0 /)

  Mref_V_Qkappa_ref( 271 : 300 ) = (/ &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0 /)

  Mref_V_Qkappa_ref( 301 : 330 ) = (/ &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0 /)

  Mref_V_Qkappa_ref( 331 : 360 ) = (/ &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0 /)

  Mref_V_Qkappa_ref( 361 : 390 ) = (/ &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0 /)

  Mref_V_Qkappa_ref( 391 : 420 ) = (/ &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0 /)

  Mref_V_Qkappa_ref( 421 : 450 ) = (/ &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0 /)

  Mref_V_Qkappa_ref( 451 : 480 ) = (/ &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0 /)

  Mref_V_Qkappa_ref( 481 : 510 ) = (/ &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0 /)

  Mref_V_Qkappa_ref( 511 : 540 ) = (/ &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0 /)

  Mref_V_Qkappa_ref( 541 : 570 ) = (/ &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0 /)

  Mref_V_Qkappa_ref( 571 : 600 ) = (/ &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0 /)

  Mref_V_Qkappa_ref( 601 : 630 ) = (/ &
  57822.5d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0 /)

  Mref_V_Qkappa_ref( 631 : 660 ) = (/ &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0 /)

  Mref_V_Qkappa_ref( 661 : 690 ) = (/ &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0 /)

  Mref_V_Qkappa_ref( 691 : 720 ) = (/ &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  943.d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0 /)

  Mref_V_Qkappa_ref( 721 : 750 ) = (/ &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0, &
  57822.5d0 /)

  Mref_V_Qmu_ref( 1 : 30 ) = (/ &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0 /)

  Mref_V_Qmu_ref( 31 : 60 ) = (/ &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0 /)

  Mref_V_Qmu_ref( 61 : 90 ) = (/ &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0 /)

  Mref_V_Qmu_ref( 91 : 120 ) = (/ &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0 /)

  Mref_V_Qmu_ref( 121 : 150 ) = (/ &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0 /)

  Mref_V_Qmu_ref( 151 : 180 ) = (/ &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0, &
  104.d0 /)

  Mref_V_Qmu_ref( 181 : 210 ) = (/ &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0 /)

  Mref_V_Qmu_ref( 211 : 240 ) = (/ &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0 /)

  Mref_V_Qmu_ref( 241 : 270 ) = (/ &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0 /)

  Mref_V_Qmu_ref( 271 : 300 ) = (/ &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0 /)

  Mref_V_Qmu_ref( 301 : 330 ) = (/ &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0 /)

  Mref_V_Qmu_ref( 331 : 360 ) = (/ &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
 4999.d0, &
  355.d0, &
  355.d0 /)

  Mref_V_Qmu_ref( 361 : 390 ) = (/ &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0 /)

  Mref_V_Qmu_ref( 391 : 420 ) = (/ &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0 /)

  Mref_V_Qmu_ref( 421 : 450 ) = (/ &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0 /)

  Mref_V_Qmu_ref( 451 : 480 ) = (/ &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0 /)

  Mref_V_Qmu_ref( 481 : 510 ) = (/ &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0 /)

  Mref_V_Qmu_ref( 511 : 540 ) = (/ &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0 /)

  Mref_V_Qmu_ref( 541 : 570 ) = (/ &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0 /)

  Mref_V_Qmu_ref( 571 : 600 ) = (/ &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0, &
  355.d0 /)

  Mref_V_Qmu_ref( 601 : 630 ) = (/ &
  355.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0 /)

  Mref_V_Qmu_ref( 631 : 660 ) = (/ &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0 /)

  Mref_V_Qmu_ref( 661 : 690 ) = (/ &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  165.d0, &
  70.d0, &
  70.d0, &
  70.d0, &
  70.d0, &
  70.d0, &
  70.d0, &
  70.d0, &
  70.d0, &
  70.d0, &
  70.d0, &
  70.d0, &
  70.d0 /)

  Mref_V_Qmu_ref( 691 : 720 ) = (/ &
  70.d0, &
  70.d0, &
  70.d0, &
  70.d0, &
  70.d0, &
  70.d0, &
  70.d0, &
  70.d0, &
  70.d0, &
  70.d0, &
  70.d0, &
  70.d0, &
  191.d0, &
  191.d0, &
  191.d0, &
  191.d0, &
  191.d0, &
  191.d0, &
  191.d0, &
  191.d0, &
  191.d0, &
  191.d0, &
  191.d0, &
  191.d0, &
  191.d0, &
  191.d0, &
  191.d0, &
  300.d0, &
  300.d0, &
  300.d0 /)

  Mref_V_Qmu_ref( 721 : 750 ) = (/ &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0, &
  300.d0 /)

  Mref_V_vph_ref( 1 : 30 ) = (/ &
  11262.2d0, &
  11262.2d0, &
  11262.18d0, &
  11262.14d0, &
  11262.09d0, &
  11262.02d0, &
  11261.94d0, &
  11261.85d0, &
  11261.74d0, &
  11261.61d0, &
  11261.47d0, &
  11261.32d0, &
  11261.15d0, &
  11260.97d0, &
  11260.77d0, &
  11260.56d0, &
  11260.34d0, &
  11260.09d0, &
  11259.84d0, &
  11259.57d0, &
  11259.28d0, &
  11258.99d0, &
  11258.67d0, &
  11258.34d0, &
  11258.d0, &
  11257.64d0, &
  11257.27d0, &
  11256.88d0, &
  11256.48d0, &
  11256.06d0 /)

  Mref_V_vph_ref( 31 : 60 ) = (/ &
  11255.63d0, &
  11255.19d0, &
  11254.73d0, &
  11254.25d0, &
  11253.76d0, &
  11253.26d0, &
  11252.74d0, &
  11252.21d0, &
  11251.66d0, &
  11251.1d0, &
  11250.52d0, &
  11249.93d0, &
  11249.33d0, &
  11248.71d0, &
  11248.07d0, &
  11247.42d0, &
  11246.76d0, &
  11246.08d0, &
  11245.38d0, &
  11244.67d0, &
  11243.95d0, &
  11243.21d0, &
  11242.46d0, &
  11241.7d0, &
  11240.91d0, &
  11240.12d0, &
  11239.31d0, &
  11238.48d0, &
  11237.64d0, &
  11236.79d0 /)

  Mref_V_vph_ref( 61 : 90 ) = (/ &
  11235.92d0, &
  11235.04d0, &
  11234.14d0, &
  11233.23d0, &
  11232.3d0, &
  11231.36d0, &
  11230.4d0, &
  11229.43d0, &
  11228.44d0, &
  11227.44d0, &
  11226.43d0, &
  11225.4d0, &
  11224.36d0, &
  11223.3d0, &
  11222.22d0, &
  11221.14d0, &
  11220.03d0, &
  11218.92d0, &
  11217.78d0, &
  11216.64d0, &
  11215.48d0, &
  11214.3d0, &
  11213.11d0, &
  11211.91d0, &
  11210.69d0, &
  11209.45d0, &
  11208.21d0, &
  11206.94d0, &
  11205.67d0, &
  11204.37d0 /)

  Mref_V_vph_ref( 91 : 120 ) = (/ &
  11203.07d0, &
  11201.74d0, &
  11200.41d0, &
  11199.06d0, &
  11197.69d0, &
  11196.31d0, &
  11194.92d0, &
  11193.51d0, &
  11192.09d0, &
  11190.65d0, &
  11189.19d0, &
  11187.73d0, &
  11186.24d0, &
  11184.75d0, &
  11183.24d0, &
  11181.71d0, &
  11180.17d0, &
  11178.61d0, &
  11177.04d0, &
  11175.46d0, &
  11173.86d0, &
  11172.25d0, &
  11170.62d0, &
  11168.98d0, &
  11167.32d0, &
  11165.65d0, &
  11163.96d0, &
  11162.26d0, &
  11160.54d0, &
  11158.81d0 /)

  Mref_V_vph_ref( 121 : 150 ) = (/ &
  11157.07d0, &
  11155.31d0, &
  11153.54d0, &
  11151.75d0, &
  11149.94d0, &
  11148.13d0, &
  11146.29d0, &
  11144.45d0, &
  11142.58d0, &
  11140.71d0, &
  11138.82d0, &
  11136.91d0, &
  11134.99d0, &
  11133.06d0, &
  11131.11d0, &
  11129.14d0, &
  11127.16d0, &
  11125.17d0, &
  11123.16d0, &
  11121.14d0, &
  11119.1d0, &
  11117.05d0, &
  11114.99d0, &
  11112.9d0, &
  11110.81d0, &
  11108.7d0, &
  11106.57d0, &
  11104.44d0, &
  11102.28d0, &
  11100.11d0 /)

  Mref_V_vph_ref( 151 : 180 ) = (/ &
  11097.93d0, &
  11095.73d0, &
  11093.52d0, &
  11091.29d0, &
  11089.05d0, &
  11086.8d0, &
  11084.53d0, &
  11082.24d0, &
  11079.94d0, &
  11077.63d0, &
  11075.3d0, &
  11072.95d0, &
  11070.59d0, &
  11068.22d0, &
  11065.83d0, &
  11063.43d0, &
  11061.02d0, &
  11058.58d0, &
  11056.14d0, &
  11053.68d0, &
  11051.2d0, &
  11048.71d0, &
  11046.21d0, &
  11043.69d0, &
  11041.16d0, &
  11038.61d0, &
  11036.05d0, &
  11033.47d0, &
  11030.88d0, &
  11028.27d0 /)

  Mref_V_vph_ref( 181 : 210 ) = (/ &
  10355.69d0, &
  10348.28d0, &
  10340.85d0, &
  10333.39d0, &
  10325.91d0, &
  10318.4d0, &
  10310.87d0, &
  10303.3d0, &
  10295.71d0, &
  10288.09d0, &
  10280.44d0, &
  10272.76d0, &
  10265.04d0, &
  10257.3d0, &
  10249.52d0, &
  10241.71d0, &
  10233.86d0, &
  10225.98d0, &
  10218.06d0, &
  10210.11d0, &
  10202.12d0, &
  10194.1d0, &
  10186.04d0, &
  10177.94d0, &
  10169.79d0, &
  10161.61d0, &
  10153.39d0, &
  10145.13d0, &
  10136.83d0, &
  10128.48d0 /)

  Mref_V_vph_ref( 211 : 240 ) = (/ &
  10120.09d0, &
  10111.66d0, &
  10103.18d0, &
  10094.66d0, &
  10086.09d0, &
  10077.48d0, &
  10068.82d0, &
  10060.11d0, &
  10051.35d0, &
  10042.54d0, &
  10033.69d0, &
  10024.78d0, &
  10015.82d0, &
  10006.82d0, &
  9997.75d0, &
  9988.64d0, &
  9979.47d0, &
  9970.25d0, &
  9960.97d0, &
  9951.64d0, &
  9942.25d0, &
  9932.81d0, &
  9923.31d0, &
  9913.75d0, &
  9904.13d0, &
  9894.45d0, &
  9884.71d0, &
  9874.91d0, &
  9865.05d0, &
  9855.13d0 /)

  Mref_V_vph_ref( 241 : 270 ) = (/ &
  9845.14d0, &
  9835.09d0, &
  9824.98d0, &
  9814.8d0, &
  9804.56d0, &
  9794.25d0, &
  9783.87d0, &
  9773.43d0, &
  9762.92d0, &
  9752.34d0, &
  9741.69d0, &
  9730.97d0, &
  9720.18d0, &
  9709.32d0, &
  9698.39d0, &
  9687.38d0, &
  9676.31d0, &
  9665.15d0, &
  9653.93d0, &
  9642.63d0, &
  9631.25d0, &
  9619.8d0, &
  9608.27d0, &
  9596.66d0, &
  9584.97d0, &
  9573.2d0, &
  9561.36d0, &
  9549.43d0, &
  9537.43d0, &
  9525.34d0 /)

  Mref_V_vph_ref( 271 : 300 ) = (/ &
  9513.17d0, &
  9500.91d0, &
  9488.57d0, &
  9476.15d0, &
  9463.64d0, &
  9451.05d0, &
  9438.37d0, &
  9425.61d0, &
  9412.75d0, &
  9399.81d0, &
  9386.78d0, &
  9373.66d0, &
  9360.45d0, &
  9347.15d0, &
  9333.76d0, &
  9320.27d0, &
  9306.7d0, &
  9293.03d0, &
  9279.26d0, &
  9265.4d0, &
  9251.45d0, &
  9237.4d0, &
  9223.25d0, &
  9209.d0, &
  9194.66d0, &
  9180.22d0, &
  9165.68d0, &
  9151.03d0, &
  9136.29d0, &
  9121.45d0 /)

  Mref_V_vph_ref( 301 : 330 ) = (/ &
  9106.5d0, &
  9091.46d0, &
  9076.3d0, &
  9061.05d0, &
  9045.69d0, &
  9030.23d0, &
  9014.65d0, &
  8998.98d0, &
  8983.19d0, &
  8967.3d0, &
  8951.3d0, &
  8935.19d0, &
  8918.97d0, &
  8902.64d0, &
  8886.2d0, &
  8869.64d0, &
  8852.98d0, &
  8836.2d0, &
  8819.31d0, &
  8802.3d0, &
  8785.18d0, &
  8767.94d0, &
  8750.59d0, &
  8733.12d0, &
  8715.53d0, &
  8697.82d0, &
  8680.d0, &
  8662.05d0, &
  8643.99d0, &
  8625.8d0 /)

  Mref_V_vph_ref( 331 : 360 ) = (/ &
  8607.49d0, &
  8589.06d0, &
  8570.51d0, &
  8551.83d0, &
  8533.03d0, &
  8514.1d0, &
  8495.05d0, &
  8475.87d0, &
  8456.57d0, &
  8437.14d0, &
  8417.58d0, &
  8397.89d0, &
  8378.07d0, &
  8358.12d0, &
  8338.04d0, &
  8317.83d0, &
  8297.49d0, &
  8277.01d0, &
  8256.41d0, &
  8235.66d0, &
  8214.79d0, &
  8193.77d0, &
  8172.62d0, &
  8151.34d0, &
  8129.92d0, &
  8108.36d0, &
  8086.66d0, &
  8064.82d0, &
  13716.6d0, &
  13714.29d0 /)

  Mref_V_vph_ref( 361 : 390 ) = (/ &
  13712.d0, &
  13709.7d0, &
  13707.42d0, &
  13705.14d0, &
  13702.86d0, &
  13700.59d0, &
  13698.33d0, &
  13696.07d0, &
  13693.82d0, &
  13691.57d0, &
  13689.33d0, &
  13687.09d0, &
  13684.86d0, &
  13682.63d0, &
  13680.41d0, &
  13680.41d0, &
  13668.9d0, &
  13657.43d0, &
  13645.97d0, &
  13634.54d0, &
  13623.14d0, &
  13611.76d0, &
  13600.4d0, &
  13589.07d0, &
  13577.76d0, &
  13566.47d0, &
  13555.2d0, &
  13543.95d0, &
  13532.72d0, &
  13521.51d0 /)

  Mref_V_vph_ref( 391 : 420 ) = (/ &
  13510.32d0, &
  13499.14d0, &
  13487.99d0, &
  13476.85d0, &
  13465.73d0, &
  13454.63d0, &
  13443.54d0, &
  13432.46d0, &
  13421.41d0, &
  13410.36d0, &
  13399.33d0, &
  13388.31d0, &
  13377.31d0, &
  13366.31d0, &
  13355.33d0, &
  13344.36d0, &
  13333.4d0, &
  13322.45d0, &
  13311.51d0, &
  13300.58d0, &
  13289.66d0, &
  13278.74d0, &
  13267.84d0, &
  13256.93d0, &
  13246.04d0, &
  13235.15d0, &
  13224.27d0, &
  13213.39d0, &
  13202.51d0, &
  13191.64d0 /)

  Mref_V_vph_ref( 421 : 450 ) = (/ &
  13180.78d0, &
  13169.91d0, &
  13159.05d0, &
  13148.19d0, &
  13137.33d0, &
  13126.47d0, &
  13115.61d0, &
  13104.75d0, &
  13093.89d0, &
  13083.02d0, &
  13072.16d0, &
  13061.29d0, &
  13050.42d0, &
  13039.55d0, &
  13028.67d0, &
  13017.78d0, &
  13006.9d0, &
  12996.d0, &
  12985.1d0, &
  12974.19d0, &
  12963.28d0, &
  12952.36d0, &
  12941.42d0, &
  12930.48d0, &
  12919.54d0, &
  12908.58d0, &
  12897.61d0, &
  12886.63d0, &
  12875.63d0, &
  12864.63d0 /)

  Mref_V_vph_ref( 451 : 480 ) = (/ &
  12853.61d0, &
  12842.58d0, &
  12831.54d0, &
  12820.48d0, &
  12809.41d0, &
  12798.32d0, &
  12787.22d0, &
  12776.1d0, &
  12764.96d0, &
  12753.81d0, &
  12742.63d0, &
  12731.44d0, &
  12720.24d0, &
  12709.01d0, &
  12697.76d0, &
  12686.49d0, &
  12675.2d0, &
  12663.89d0, &
  12652.56d0, &
  12641.2d0, &
  12629.82d0, &
  12618.42d0, &
  12606.99d0, &
  12595.54d0, &
  12584.06d0, &
  12572.56d0, &
  12561.03d0, &
  12549.48d0, &
  12537.89d0, &
  12526.28d0 /)

  Mref_V_vph_ref( 481 : 510 ) = (/ &
  12514.64d0, &
  12502.98d0, &
  12491.28d0, &
  12479.55d0, &
  12467.79d0, &
  12456.01d0, &
  12444.19d0, &
  12432.33d0, &
  12420.45d0, &
  12408.53d0, &
  12396.58d0, &
  12384.6d0, &
  12372.58d0, &
  12360.52d0, &
  12348.43d0, &
  12336.3d0, &
  12324.14d0, &
  12311.94d0, &
  12299.7d0, &
  12287.42d0, &
  12275.11d0, &
  12262.75d0, &
  12250.35d0, &
  12237.92d0, &
  12225.44d0, &
  12212.92d0, &
  12200.36d0, &
  12187.76d0, &
  12175.11d0, &
  12162.43d0 /)

  Mref_V_vph_ref( 511 : 540 ) = (/ &
  12149.69d0, &
  12136.91d0, &
  12124.09d0, &
  12111.22d0, &
  12098.31d0, &
  12085.34d0, &
  12072.34d0, &
  12059.28d0, &
  12046.17d0, &
  12033.02d0, &
  12019.82d0, &
  12006.56d0, &
  11993.26d0, &
  11979.9d0, &
  11966.5d0, &
  11953.04d0, &
  11939.53d0, &
  11925.97d0, &
  11912.35d0, &
  11898.69d0, &
  11884.96d0, &
  11871.19d0, &
  11857.37d0, &
  11843.48d0, &
  11829.55d0, &
  11815.57d0, &
  11801.53d0, &
  11787.44d0, &
  11773.3d0, &
  11759.1d0 /)

  Mref_V_vph_ref( 541 : 570 ) = (/ &
  11744.85d0, &
  11730.55d0, &
  11716.18d0, &
  11701.78d0, &
  11687.31d0, &
  11672.8d0, &
  11658.23d0, &
  11643.6d0, &
  11628.92d0, &
  11614.19d0, &
  11599.4d0, &
  11584.57d0, &
  11569.68d0, &
  11554.72d0, &
  11539.72d0, &
  11524.67d0, &
  11509.56d0, &
  11494.39d0, &
  11479.17d0, &
  11463.89d0, &
  11448.55d0, &
  11433.17d0, &
  11417.73d0, &
  11402.23d0, &
  11386.68d0, &
  11371.07d0, &
  11355.41d0, &
  11339.69d0, &
  11323.91d0, &
  11308.09d0 /)

  Mref_V_vph_ref( 571 : 600 ) = (/ &
  11292.2d0, &
  11276.25d0, &
  11260.25d0, &
  11244.19d0, &
  11228.08d0, &
  11211.9d0, &
  11195.67d0, &
  11179.38d0, &
  11163.04d0, &
  11146.63d0, &
  11130.18d0, &
  11113.67d0, &
  11097.11d0, &
  11080.51d0, &
  11080.51d0, &
  11063.01d0, &
  11045.22d0, &
  11026.82d0, &
  11008.47d0, &
  10989.04d0, &
  10969.63d0, &
  10948.76d0, &
  10928.02d0, &
  10907.42d0, &
  10886.94d0, &
  10866.6d0, &
  10846.41d0, &
  10826.35d0, &
  10806.42d0, &
  10786.61d0 /)

  Mref_V_vph_ref( 601 : 630 ) = (/ &
  10766.9d0, &
  10278.88d0, &
  10261.87d0, &
  10244.84d0, &
  10227.82d0, &
  10210.8d0, &
  10193.78d0, &
  10176.77d0, &
  10159.74d0, &
  10142.72d0, &
  10125.71d0, &
  10108.7d0, &
  10091.68d0, &
  10074.68d0, &
  10057.68d0, &
  10040.64d0, &
  10040.67d0, &
  10010.52d0, &
  9980.51d0, &
  9950.64d0, &
  9920.91d0, &
  9891.35d0, &
  9861.96d0, &
  9832.79d0, &
  9803.79d0, &
  9774.98d0, &
  9746.41d0, &
  9718.08d0, &
  9689.96d0, &
  9662.1d0 /)

  Mref_V_vph_ref( 631 : 660 ) = (/ &
  9634.47d0, &
  9607.11d0, &
  9579.97d0, &
  9553.08d0, &
  9526.38d0, &
  9499.78d0, &
  9473.25d0, &
  9446.74d0, &
  9420.19d0, &
  9393.55d0, &
  9366.75d0, &
  9339.76d0, &
  9312.5d0, &
  9284.96d0, &
  9257.04d0, &
  9228.73d0, &
  9199.94d0, &
  8940.94d0, &
  8930.61d0, &
  8920.22d0, &
  8909.68d0, &
  8898.47d0, &
  8886.28d0, &
  8873.03d0, &
  8858.58d0, &
  8842.82d0, &
  8825.64d0, &
  8806.94d0, &
  8786.67d0, &
  8764.85d0 /)

  Mref_V_vph_ref( 661 : 690 ) = (/ &
  8741.49d0, &
  8716.63d0, &
  8690.3d0, &
  8662.5d0, &
  8633.28d0, &
  8602.66d0, &
  8570.81d0, &
  8538.06d0, &
  8504.66d0, &
  8470.92d0, &
  8437.13d0, &
  8403.52d0, &
  8370.42d0, &
  8338.11d0, &
  8307.42d0, &
  8278.36d0, &
  8255.33d0, &
  8236.9d0, &
  8236.81d0, &
  8222.27d0, &
  8210.47d0, &
  8201.14d0, &
  8193.99d0, &
  8188.67d0, &
  8184.92d0, &
  8182.39d0, &
  8180.79d0, &
  8179.83d0, &
  8179.17d0, &
  8178.54d0 /)

  Mref_V_vph_ref( 691 : 720 ) = (/ &
  8177.64d0, &
  8176.3d0, &
  8174.55d0, &
  8172.42d0, &
  8169.91d0, &
  8167.05d0, &
  8163.88d0, &
  8160.37d0, &
  8156.58d0, &
  8152.57d0, &
  8148.41d0, &
  8144.2d0, &
  8144.32d0, &
  8141.6d0, &
  8139.01d0, &
  8136.5d0, &
  8134.11d0, &
  8131.82d0, &
  8129.66d0, &
  8127.6d0, &
  8125.65d0, &
  8123.87d0, &
  8122.23d0, &
  8120.74d0, &
  8119.38d0, &
  8118.22d0, &
  8117.13d0, &
  6800.d0, &
  6800.d0, &
  6800.d0 /)

  Mref_V_vph_ref( 721 : 750 ) = (/ &
  6800.d0, &
  6800.d0, &
  6800.d0, &
  6800.d0, &
  6800.d0, &
  6800.d0, &
  6800.d0, &
  6800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0, &
  5800.d0 /)

  Mref_V_vsh_ref( 1 : 30 ) = (/ &
  3667.8d0, &
  3667.79d0, &
  3667.78d0, &
  3667.75d0, &
  3667.72d0, &
  3667.67d0, &
  3667.62d0, &
  3667.55d0, &
  3667.47d0, &
  3667.39d0, &
  3667.29d0, &
  3667.18d0, &
  3667.06d0, &
  3666.94d0, &
  3666.8d0, &
  3666.65d0, &
  3666.49d0, &
  3666.32d0, &
  3666.15d0, &
  3665.96d0, &
  3665.76d0, &
  3665.55d0, &
  3665.33d0, &
  3665.1d0, &
  3664.86d0, &
  3664.61d0, &
  3664.35d0, &
  3664.08d0, &
  3663.8d0, &
  3663.51d0 /)

  Mref_V_vsh_ref( 31 : 60 ) = (/ &
  3663.21d0, &
  3662.9d0, &
  3662.57d0, &
  3662.24d0, &
  3661.9d0, &
  3661.55d0, &
  3661.19d0, &
  3660.81d0, &
  3660.43d0, &
  3660.04d0, &
  3659.64d0, &
  3659.22d0, &
  3658.8d0, &
  3658.36d0, &
  3657.92d0, &
  3657.47d0, &
  3657.d0, &
  3656.53d0, &
  3656.04d0, &
  3655.55d0, &
  3655.04d0, &
  3654.53d0, &
  3654.d0, &
  3653.47d0, &
  3652.92d0, &
  3652.36d0, &
  3651.8d0, &
  3651.22d0, &
  3650.63d0, &
  3650.04d0 /)

  Mref_V_vsh_ref( 61 : 90 ) = (/ &
  3649.43d0, &
  3648.81d0, &
  3648.19d0, &
  3647.55d0, &
  3646.9d0, &
  3646.24d0, &
  3645.57d0, &
  3644.89d0, &
  3644.21d0, &
  3643.51d0, &
  3642.8d0, &
  3642.08d0, &
  3641.35d0, &
  3640.61d0, &
  3639.86d0, &
  3639.1d0, &
  3638.33d0, &
  3637.55d0, &
  3636.76d0, &
  3635.96d0, &
  3635.14d0, &
  3634.32d0, &
  3633.49d0, &
  3632.65d0, &
  3631.8d0, &
  3630.93d0, &
  3630.06d0, &
  3629.18d0, &
  3628.29d0, &
  3627.38d0 /)

  Mref_V_vsh_ref( 91 : 120 ) = (/ &
  3626.47d0, &
  3625.55d0, &
  3624.61d0, &
  3623.67d0, &
  3622.71d0, &
  3621.75d0, &
  3620.78d0, &
  3619.79d0, &
  3618.8d0, &
  3617.79d0, &
  3616.78d0, &
  3615.75d0, &
  3614.71d0, &
  3613.67d0, &
  3612.61d0, &
  3611.55d0, &
  3610.47d0, &
  3609.38d0, &
  3608.28d0, &
  3607.18d0, &
  3606.06d0, &
  3604.93d0, &
  3603.79d0, &
  3602.65d0, &
  3601.49d0, &
  3600.32d0, &
  3599.14d0, &
  3597.95d0, &
  3596.75d0, &
  3595.54d0 /)

  Mref_V_vsh_ref( 121 : 150 ) = (/ &
  3594.32d0, &
  3593.1d0, &
  3591.86d0, &
  3590.61d0, &
  3589.34d0, &
  3588.07d0, &
  3586.79d0, &
  3585.5d0, &
  3584.2d0, &
  3582.89d0, &
  3581.57d0, &
  3580.24d0, &
  3578.9d0, &
  3577.54d0, &
  3576.18d0, &
  3574.81d0, &
  3573.43d0, &
  3572.03d0, &
  3570.63d0, &
  3569.22d0, &
  3567.79d0, &
  3566.36d0, &
  3564.91d0, &
  3563.46d0, &
  3562.d0, &
  3560.52d0, &
  3559.04d0, &
  3557.54d0, &
  3556.04d0, &
  3554.52d0 /)

  Mref_V_vsh_ref( 151 : 180 ) = (/ &
  3553.d0, &
  3551.46d0, &
  3549.91d0, &
  3548.36d0, &
  3546.79d0, &
  3545.21d0, &
  3543.63d0, &
  3542.03d0, &
  3540.42d0, &
  3538.81d0, &
  3537.18d0, &
  3535.54d0, &
  3533.89d0, &
  3532.23d0, &
  3530.57d0, &
  3528.89d0, &
  3527.2d0, &
  3525.5d0, &
  3523.79d0, &
  3522.07d0, &
  3520.34d0, &
  3518.6d0, &
  3516.85d0, &
  3515.09d0, &
  3513.32d0, &
  3511.54d0, &
  3509.75d0, &
  3507.95d0, &
  3506.13d0, &
  3504.31d0 /)

  Mref_V_vsh_ref( 181 : 210 ) = (/ &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0 /)

  Mref_V_vsh_ref( 211 : 240 ) = (/ &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0 /)

  Mref_V_vsh_ref( 241 : 270 ) = (/ &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0 /)

  Mref_V_vsh_ref( 271 : 300 ) = (/ &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0 /)

  Mref_V_vsh_ref( 301 : 330 ) = (/ &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0 /)

  Mref_V_vsh_ref( 331 : 360 ) = (/ &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  0.d0, &
  7264.66d0, &
  7264.75d0 /)

  Mref_V_vsh_ref( 361 : 390 ) = (/ &
  7264.85d0, &
  7264.94d0, &
  7265.03d0, &
  7265.12d0, &
  7265.21d0, &
  7265.29d0, &
  7265.38d0, &
  7265.46d0, &
  7265.54d0, &
  7265.62d0, &
  7265.69d0, &
  7265.76d0, &
  7265.84d0, &
  7265.91d0, &
  7265.97d0, &
  7265.97d0, &
  7261.63d0, &
  7257.29d0, &
  7252.97d0, &
  7248.64d0, &
  7244.33d0, &
  7240.01d0, &
  7235.71d0, &
  7231.41d0, &
  7227.12d0, &
  7222.83d0, &
  7218.55d0, &
  7214.27d0, &
  7210.d0, &
  7205.73d0 /)

  Mref_V_vsh_ref( 391 : 420 ) = (/ &
  7201.47d0, &
  7197.21d0, &
  7192.95d0, &
  7188.7d0, &
  7184.45d0, &
  7180.21d0, &
  7175.97d0, &
  7171.73d0, &
  7167.5d0, &
  7163.27d0, &
  7159.04d0, &
  7154.81d0, &
  7150.59d0, &
  7146.37d0, &
  7142.15d0, &
  7137.93d0, &
  7133.71d0, &
  7129.5d0, &
  7125.29d0, &
  7121.07d0, &
  7116.86d0, &
  7112.65d0, &
  7108.44d0, &
  7104.23d0, &
  7100.02d0, &
  7095.81d0, &
  7091.6d0, &
  7087.39d0, &
  7083.18d0, &
  7078.96d0 /)

  Mref_V_vsh_ref( 421 : 450 ) = (/ &
  7074.75d0, &
  7070.54d0, &
  7066.32d0, &
  7062.1d0, &
  7057.88d0, &
  7053.66d0, &
  7049.44d0, &
  7045.22d0, &
  7040.99d0, &
  7036.76d0, &
  7032.52d0, &
  7028.29d0, &
  7024.05d0, &
  7019.81d0, &
  7015.56d0, &
  7011.31d0, &
  7007.06d0, &
  7002.8d0, &
  6998.54d0, &
  6994.27d0, &
  6990.d0, &
  6985.72d0, &
  6981.44d0, &
  6977.15d0, &
  6972.86d0, &
  6968.57d0, &
  6964.26d0, &
  6959.95d0, &
  6955.64d0, &
  6951.32d0 /)

  Mref_V_vsh_ref( 451 : 480 ) = (/ &
  6946.99d0, &
  6942.66d0, &
  6938.31d0, &
  6933.97d0, &
  6929.61d0, &
  6925.25d0, &
  6920.88d0, &
  6916.5d0, &
  6912.11d0, &
  6907.72d0, &
  6903.32d0, &
  6898.91d0, &
  6894.49d0, &
  6890.06d0, &
  6885.62d0, &
  6881.17d0, &
  6876.72d0, &
  6872.25d0, &
  6867.78d0, &
  6863.29d0, &
  6858.8d0, &
  6854.29d0, &
  6849.78d0, &
  6845.25d0, &
  6840.71d0, &
  6836.16d0, &
  6831.6d0, &
  6827.03d0, &
  6822.45d0, &
  6817.85d0 /)

  Mref_V_vsh_ref( 481 : 510 ) = (/ &
  6813.25d0, &
  6808.63d0, &
  6804.d0, &
  6799.35d0, &
  6794.7d0, &
  6790.03d0, &
  6785.34d0, &
  6780.65d0, &
  6775.94d0, &
  6771.22d0, &
  6766.48d0, &
  6761.73d0, &
  6756.97d0, &
  6752.19d0, &
  6747.4d0, &
  6742.59d0, &
  6737.76d0, &
  6732.93d0, &
  6728.07d0, &
  6723.21d0, &
  6718.32d0, &
  6713.42d0, &
  6708.51d0, &
  6703.57d0, &
  6698.62d0, &
  6693.66d0, &
  6688.68d0, &
  6683.68d0, &
  6678.66d0, &
  6673.63d0 /)

  Mref_V_vsh_ref( 511 : 540 ) = (/ &
  6668.58d0, &
  6663.51d0, &
  6658.43d0, &
  6653.32d0, &
  6648.2d0, &
  6643.06d0, &
  6637.9d0, &
  6632.73d0, &
  6627.53d0, &
  6622.31d0, &
  6617.08d0, &
  6611.82d0, &
  6606.55d0, &
  6601.26d0, &
  6595.94d0, &
  6590.61d0, &
  6584.91d0, &
  6579.51d0, &
  6574.11d0, &
  6568.67d0, &
  6563.22d0, &
  6557.74d0, &
  6552.24d0, &
  6546.73d0, &
  6541.19d0, &
  6535.63d0, &
  6530.05d0, &
  6524.44d0, &
  6518.82d0, &
  6513.17d0 /)

  Mref_V_vsh_ref( 541 : 570 ) = (/ &
  6507.5d0, &
  6501.8d0, &
  6496.09d0, &
  6490.35d0, &
  6484.59d0, &
  6478.8d0, &
  6472.99d0, &
  6467.16d0, &
  6461.3d0, &
  6455.42d0, &
  6449.51d0, &
  6443.58d0, &
  6437.63d0, &
  6431.65d0, &
  6425.65d0, &
  6419.61d0, &
  6413.56d0, &
  6407.48d0, &
  6401.37d0, &
  6395.25d0, &
  6389.09d0, &
  6382.91d0, &
  6376.7d0, &
  6370.46d0, &
  6364.2d0, &
  6357.91d0, &
  6351.59d0, &
  6345.25d0, &
  6338.88d0, &
  6332.49d0 /)

  Mref_V_vsh_ref( 571 : 600 ) = (/ &
  6326.05d0, &
  6319.6d0, &
  6313.13d0, &
  6306.62d0, &
  6300.08d0, &
  6293.52d0, &
  6286.92d0, &
  6280.29d0, &
  6273.64d0, &
  6266.96d0, &
  6260.25d0, &
  6253.51d0, &
  6246.75d0, &
  6239.95d0, &
  6239.95d0, &
  6219.68d0, &
  6200.29d0, &
  6181.16d0, &
  6162.04d0, &
  6143.01d0, &
  6123.98d0, &
  6103.71d0, &
  6083.53d0, &
  6063.45d0, &
  6043.44d0, &
  6023.52d0, &
  6003.73d0, &
  5984.03d0, &
  5964.38d0, &
  5944.81d0 /)

  Mref_V_vsh_ref( 601 : 630 ) = (/ &
  5925.27d0, &
  5550.32d0, &
  5541.2d0, &
  5532.08d0, &
  5522.96d0, &
  5513.83d0, &
  5504.71d0, &
  5495.59d0, &
  5486.47d0, &
  5477.35d0, &
  5468.22d0, &
  5459.1d0, &
  5449.97d0, &
  5440.84d0, &
  5431.71d0, &
  5422.57d0, &
  5422.59d0, &
  5406.39d0, &
  5390.3d0, &
  5374.34d0, &
  5358.52d0, &
  5342.83d0, &
  5327.31d0, &
  5311.92d0, &
  5296.73d0, &
  5281.71d0, &
  5266.86d0, &
  5252.21d0, &
  5237.78d0, &
  5223.55d0 /)

  Mref_V_vsh_ref( 631 : 660 ) = (/ &
  5209.54d0, &
  5195.72d0, &
  5182.1d0, &
  5168.69d0, &
  5155.42d0, &
  5142.22d0, &
  5129.05d0, &
  5115.84d0, &
  5102.55d0, &
  5089.14d0, &
  5075.5d0, &
  5061.63d0, &
  5047.46d0, &
  5032.93d0, &
  5018.03d0, &
  5002.66d0, &
  4986.77d0, &
  4803.78d0, &
  4800.54d0, &
  4797.28d0, &
  4793.96d0, &
  4790.18d0, &
  4785.78d0, &
  4780.71d0, &
  4775.d0, &
  4768.58d0, &
  4761.41d0, &
  4753.51d0, &
  4744.86d0, &
  4735.64d0 /)

  Mref_V_vsh_ref( 661 : 690 ) = (/ &
  4725.88d0, &
  4715.76d0, &
  4705.34d0, &
  4694.74d0, &
  4684.08d0, &
  4673.46d0, &
  4662.94d0, &
  4652.61d0, &
  4642.55d0, &
  4632.81d0, &
  4623.51d0, &
  4614.68d0, &
  4606.39d0, &
  4598.73d0, &
  4591.76d0, &
  4585.56d0, &
  4580.21d0, &
  4575.75d0, &
  4575.74d0, &
  4572.27d0, &
  4569.53d0, &
  4567.46d0, &
  4566.02d0, &
  4565.1d0, &
  4564.66d0, &
  4564.65d0, &
  4564.99d0, &
  4565.62d0, &
  4566.47d0, &
  4567.46d0 /)

  Mref_V_vsh_ref( 691 : 720 ) = (/ &
  4568.58d0, &
  4569.7d0, &
  4570.85d0, &
  4571.91d0, &
  4572.83d0, &
  4573.6d0, &
  4574.16d0, &
  4574.44d0, &
  4574.42d0, &
  4574.04d0, &
  4573.36d0, &
  4572.41d0, &
  4572.46d0, &
  4571.71d0, &
  4570.93d0, &
  4570.06d0, &
  4569.16d0, &
  4568.21d0, &
  4567.22d0, &
  4566.21d0, &
  4565.16d0, &
  4564.11d0, &
  4563.05d0, &
  4562.d0, &
  4560.94d0, &
  4559.94d0, &
  4558.94d0, &
  3900.d0, &
  3900.d0, &
  3900.d0 /)

  Mref_V_vsh_ref( 721 : 750 ) = (/ &
  3900.d0, &
  3900.d0, &
  3900.d0, &
  3900.d0, &
  3900.d0, &
  3900.d0, &
  3900.d0, &
  3900.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0, &
  3200.d0 /)

! anisotropy defined between layer
!   5980000.0   = depth at 391km
! to
!   6346600.0   = depth at 24.4km (PREM moho depth)

  Mref_V_eta_ref( 1 : 30 ) = (/ &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0 /)

  Mref_V_eta_ref( 31 : 60 ) = (/ &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0 /)

  Mref_V_eta_ref( 61 : 90 ) = (/ &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0 /)

  Mref_V_eta_ref( 91 : 120 ) = (/ &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0 /)

  Mref_V_eta_ref( 121 : 150 ) = (/ &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0 /)

  Mref_V_eta_ref( 151 : 180 ) = (/ &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0 /)

  Mref_V_eta_ref( 181 : 210 ) = (/ &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0 /)

  Mref_V_eta_ref( 211 : 240 ) = (/ &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0 /)

  Mref_V_eta_ref( 241 : 270 ) = (/ &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0 /)

  Mref_V_eta_ref( 271 : 300 ) = (/ &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0 /)

  Mref_V_eta_ref( 301 : 330 ) = (/ &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0 /)

  Mref_V_eta_ref( 331 : 360 ) = (/ &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0 /)

  Mref_V_eta_ref( 361 : 390 ) = (/ &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0 /)

  Mref_V_eta_ref( 391 : 420 ) = (/ &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0 /)

  Mref_V_eta_ref( 421 : 450 ) = (/ &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0 /)

  Mref_V_eta_ref( 451 : 480 ) = (/ &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0 /)

  Mref_V_eta_ref( 481 : 510 ) = (/ &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0 /)

  Mref_V_eta_ref( 511 : 540 ) = (/ &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0 /)

  Mref_V_eta_ref( 541 : 570 ) = (/ &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0 /)

  Mref_V_eta_ref( 571 : 600 ) = (/ &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0 /)

  Mref_V_eta_ref( 601 : 630 ) = (/ &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0 /)

  Mref_V_eta_ref( 631 : 660 ) = (/ &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &  ! 410km depth
  1.d0, &
  1.d0, &
  1.d0, &      ! 391km depth
  0.99999d0, & ! 384.666km depth, at r = 5986.334km
  0.99997d0, &
  0.99995d0, &
  0.99994d0, &
  0.9999d0, &
  0.99986d0, &
  0.9998d0, &
  0.99974d0, &
  0.99966d0 /)

  Mref_V_eta_ref( 661 : 690 ) = (/ &
  0.99957d0, &
  0.99947d0, &
  0.99934d0, &
  0.9992d0, &
  0.99904d0, &
  0.99886d0, &
  0.99864d0, &
  0.99832d0, &
  0.9979d0, &
  0.99732d0, &
  0.99654d0, &
  0.99553d0, &
  0.99426d0, &
  0.99268d0, &
  0.99075d0, &
  0.98843d0, &
  0.98571d0, &
  0.98255d0, &
  0.9825d0, &
  0.97907d0, &
  0.97531d0, &
  0.97128d0, &
  0.96704d0, &
  0.96268d0, &
  0.95823d0, &
  0.95378d0, &
  0.94938d0, &
  0.94509d0, &
  0.94098d0, &
  0.93712d0 /)

  Mref_V_eta_ref( 691 : 720 ) = (/ &
  0.93356d0, &
  0.93034d0, &
  0.92743d0, &
  0.92483d0, &
  0.92251d0, &
  0.92046d0, &
  0.91867d0, &
  0.91711d0, &
  0.91577d0, &
  0.91465d0, &
  0.91371d0, &
  0.91296d0, &
  0.91294d0, &
  0.91254d0, &
  0.91221d0, &
  0.91193d0, &
  0.91171d0, &
  0.91155d0, &
  0.91142d0, &
  0.91134d0, &
  0.9113d0, &
  0.91129d0, &
  0.9113d0, &
  0.91135d0, &
  0.9114d0, &
  0.91147d0, &
  0.91155d0, &       ! index 717
  1.d0, &            ! index 718 - crustal model: moho depth at 24.4km
  1.d0, &
  1.d0 /)

  Mref_V_eta_ref( 721 : 750 ) = (/ &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0, &
  1.d0 /)

  ! strip the crust and replace it by mantle
  if (SUPPRESS_CRUSTAL_MESH .or. USE_EXTERNAL_CRUSTAL_MODEL) then
    ! sets values for depths less than 24.4 km to mantle values below
    Mref_V_density_ref(718:750) = Mref_V_density_ref(717)
    Mref_V_vpv_ref(718:750) = Mref_V_vpv_ref(717)
    Mref_V_vph_ref(718:750) = Mref_V_vph_ref(717)
    Mref_V_vsv_ref(718:750) = Mref_V_vsv_ref(717)
    Mref_V_vsh_ref(718:750) = Mref_V_vsh_ref(717)
    Mref_V_Qmu_ref(718:750) = Mref_V_Qmu_ref(717)
    Mref_V_Qkappa_ref(718:750) = Mref_V_Qkappa_ref(717)
    ! daniel: eta missing? Mref_V_eta_ref
  endif

  end subroutine define_model_1dref

