!=====================================================================
!
!          S p e c f e m 3 D  G l o b e  V e r s i o n  7 . 0
!          --------------------------------------------------
!
!     Main historical authors: Dimitri Komatitsch and Jeroen Tromp
!                        Princeton University, USA
!                and CNRS / University of Marseille, France
!                 (there are currently many more authors!)
! (c) Princeton University and CNRS / University of Marseille, April 2014
!
! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 3 of the License, or
! (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
!
!=====================================================================

!--------------------------------------------------------------------------------------------------
! case65TAY
!
! 1D mean model from thermal simulations for Mars by Plesa et al. (2020, GRL, InSight special issue)
!
! geothermal case 65, Taylor composition
!
! uses PREM attenuation
!--------------------------------------------------------------------------------------------------

  module model_case65tay_par

  ! number of layers in DATA/case65TAY/case65TAY.dat
  integer, parameter :: NR_case65TAY = 681

  ! model_case65TAY_variables
  double precision, dimension(:),allocatable :: &
    Mcase65TAY_V_radius,Mcase65TAY_V_density, &
    Mcase65TAY_V_vp,Mcase65TAY_V_vs, &
    Mcase65TAY_V_Qkappa,Mcase65TAY_V_Qmu

  end module model_case65tay_par

!
!--------------------------------------------------------------------------------------------------
!

  subroutine model_case65TAY_broadcast(CRUSTAL)

! standard routine to setup model

  use constants, only: myrank
  use model_case65tay_par

  implicit none

  logical :: CRUSTAL

  ! local parameters
  integer :: ier

  ! allocates model arrays
  allocate(Mcase65TAY_V_radius(NR_case65TAY), &
           Mcase65TAY_V_density(NR_case65TAY), &
           Mcase65TAY_V_vp(NR_case65TAY), &
           Mcase65TAY_V_vs(NR_case65TAY), &
           Mcase65TAY_V_Qkappa(NR_case65TAY), &
           Mcase65TAY_V_Qmu(NR_case65TAY), &
           stat=ier)
  if (ier /= 0 ) call exit_MPI(myrank,'Error allocating Mcase65TAY_V arrays')
  Mcase65TAY_V_radius(:) = 0.d0
  Mcase65TAY_V_density(:) = 0.d0
  Mcase65TAY_V_vp(:) = 0.d0
  Mcase65TAY_V_vs(:) = 0.d0
  Mcase65TAY_V_Qkappa(:) = 9999.d0
  Mcase65TAY_V_Qmu(:) = 9999.d0

  ! all processes will define same parameters
  call define_model_case65TAY(CRUSTAL)

  end subroutine model_case65TAY_broadcast

!
!-------------------------------------------------------------------------------------------------
!

  subroutine model_case65TAY(x,rho,vp,vs,Qkappa,Qmu,iregion_code)

  use constants
  use shared_parameters, only: R_PLANET,RHOAV

  use model_case65tay_par

  implicit none

! input:
! radius r: meters

! output:
! density rho: kg/m^3
! compressional wave speed vp: km/s
! shear wave speed vs: km/s

  double precision :: x,rho,vp,vs,Qmu,Qkappa
  integer :: iregion_code

  ! local parameters
  double precision :: r,frac,scaleval
  integer :: i

  ! compute real physical radius in meters
  r = x * R_PLANET

  i = 1
  do while(r >= Mcase65TAY_V_radius(i) .and. i /= NR_case65TAY)
    i = i + 1
  enddo

  ! make sure we stay in the right region and never take a point above
  ! and a point below the ICB or the CMB and interpolate between them,
  ! which would lead to a wrong value (keeping in mind that we interpolate
  ! between points i-1 and i below)

  ! see table entry below
  ! inner core range [1:104]
  if (iregion_code == IREGION_INNER_CORE .and. i > 104) i = 104

  ! outer core index range [105:370]
  ! due to the interpolation below, we add +1 to the bottom
  if (iregion_code == IREGION_OUTER_CORE .and. i <= 105) i = 105 + 1
  if (iregion_code == IREGION_OUTER_CORE .and. i > 370) i = 370

  ! crust/mantle index range [371:681]
  ! due to the interpolation below, we add +1 to the bottom
  if (iregion_code == IREGION_CRUST_MANTLE .and. i <= 371) i = 371 + 1
  if (iregion_code == IREGION_CRUST_MANTLE .and. i > NR_case65TAY) i = NR_case65TAY

  ! model value interpolation
  if (i == 1) then
    rho = Mcase65TAY_V_density(i)
    vp = Mcase65TAY_V_vp(i)
    vs = Mcase65TAY_V_vs(i)
    Qmu = Mcase65TAY_V_Qmu(i)
    Qkappa = Mcase65TAY_V_Qkappa(i)
  else
    ! interpolate from radius(i-1) to r using the values at i-1 and i
    frac = (r-Mcase65TAY_V_radius(i-1))/(Mcase65TAY_V_radius(i)-Mcase65TAY_V_radius(i-1))

    rho = Mcase65TAY_V_density(i-1) + frac * (Mcase65TAY_V_density(i)-Mcase65TAY_V_density(i-1))
    vp = Mcase65TAY_V_vp(i-1) + frac * (Mcase65TAY_V_vp(i)-Mcase65TAY_V_vp(i-1))
    vs = Mcase65TAY_V_vs(i-1) + frac * (Mcase65TAY_V_vs(i)-Mcase65TAY_V_vs(i-1))
    Qmu = Mcase65TAY_V_Qmu(i-1) + frac * (Mcase65TAY_V_Qmu(i)-Mcase65TAY_V_Qmu(i-1))
    Qkappa = Mcase65TAY_V_Qkappa(i-1) + frac * (Mcase65TAY_V_Qkappa(i)-Mcase65TAY_V_Qkappa(i-1))
  endif

  ! make sure Vs is zero in the outer core even if roundoff errors on depth
  ! also set fictitious attenuation to a very high value (attenuation is not used in the fluid)
  if (iregion_code == IREGION_OUTER_CORE) then
    vs = 0.d0
    Qkappa = 3000.d0
    Qmu = 3000.d0
  endif

  ! non-dimensionalize
  ! time scaling (s^{-1}) is done with scaleval
  scaleval = dsqrt(PI*GRAV*RHOAV)
  rho = rho*1000.0d0/RHOAV
  vp = vp*1000.0d0/(R_PLANET*scaleval)
  vs = vs*1000.0d0/(R_PLANET*scaleval)

  end subroutine model_case65TAY

!-------------------

  subroutine define_model_case65TAY(USE_EXTERNAL_CRUSTAL_MODEL)

  use constants, only: myrank,SUPPRESS_CRUSTAL_MESH
  use shared_parameters, only: RICB,RCMB,ROCEAN
  use model_case65tay_par

  implicit none

  logical,intent(in) :: USE_EXTERNAL_CRUSTAL_MODEL

  ! local parameters
  integer :: i
  double precision, parameter :: TOL = 1.d-9

  ! define all the values in the model
  Mcase65TAY_V_radius(1) = 0.d0
  Mcase65TAY_V_radius(2) = 5000.d0
  Mcase65TAY_V_radius(3) = 10000.d0
  Mcase65TAY_V_radius(4) = 15000.d0
  Mcase65TAY_V_radius(5) = 20000.d0
  Mcase65TAY_V_radius(6) = 25000.d0
  Mcase65TAY_V_radius(7) = 30000.d0
  Mcase65TAY_V_radius(8) = 35000.d0
  Mcase65TAY_V_radius(9) = 40000.d0
  Mcase65TAY_V_radius(10) = 45000.d0
  Mcase65TAY_V_radius(11) = 50000.d0
  Mcase65TAY_V_radius(12) = 55000.d0
  Mcase65TAY_V_radius(13) = 60000.d0
  Mcase65TAY_V_radius(14) = 65000.d0
  Mcase65TAY_V_radius(15) = 70000.d0
  Mcase65TAY_V_radius(16) = 75000.d0
  Mcase65TAY_V_radius(17) = 80000.d0
  Mcase65TAY_V_radius(18) = 85000.d0
  Mcase65TAY_V_radius(19) = 90000.d0
  Mcase65TAY_V_radius(20) = 95000.d0
  Mcase65TAY_V_radius(21) = 100000.d0
  Mcase65TAY_V_radius(22) = 105000.d0
  Mcase65TAY_V_radius(23) = 110000.d0
  Mcase65TAY_V_radius(24) = 115000.d0
  Mcase65TAY_V_radius(25) = 120000.d0
  Mcase65TAY_V_radius(26) = 125000.d0
  Mcase65TAY_V_radius(27) = 130000.d0
  Mcase65TAY_V_radius(28) = 135000.d0
  Mcase65TAY_V_radius(29) = 140000.d0
  Mcase65TAY_V_radius(30) = 145000.d0
  Mcase65TAY_V_radius(31) = 150000.d0
  Mcase65TAY_V_radius(32) = 155000.d0
  Mcase65TAY_V_radius(33) = 160000.d0
  Mcase65TAY_V_radius(34) = 165000.d0
  Mcase65TAY_V_radius(35) = 170000.d0
  Mcase65TAY_V_radius(36) = 175000.d0
  Mcase65TAY_V_radius(37) = 180000.d0
  Mcase65TAY_V_radius(38) = 185000.d0
  Mcase65TAY_V_radius(39) = 190000.d0
  Mcase65TAY_V_radius(40) = 195000.d0
  Mcase65TAY_V_radius(41) = 200000.d0
  Mcase65TAY_V_radius(42) = 205000.d0
  Mcase65TAY_V_radius(43) = 210000.d0
  Mcase65TAY_V_radius(44) = 215000.d0
  Mcase65TAY_V_radius(45) = 220000.d0
  Mcase65TAY_V_radius(46) = 225000.d0
  Mcase65TAY_V_radius(47) = 230000.d0
  Mcase65TAY_V_radius(48) = 235000.d0
  Mcase65TAY_V_radius(49) = 240000.d0
  Mcase65TAY_V_radius(50) = 245000.d0
  Mcase65TAY_V_radius(51) = 250000.d0
  Mcase65TAY_V_radius(52) = 255000.d0
  Mcase65TAY_V_radius(53) = 260000.d0
  Mcase65TAY_V_radius(54) = 265000.d0
  Mcase65TAY_V_radius(55) = 270000.d0
  Mcase65TAY_V_radius(56) = 275000.d0
  Mcase65TAY_V_radius(57) = 280000.d0
  Mcase65TAY_V_radius(58) = 285000.d0
  Mcase65TAY_V_radius(59) = 290000.d0
  Mcase65TAY_V_radius(60) = 295000.d0
  Mcase65TAY_V_radius(61) = 300000.d0
  Mcase65TAY_V_radius(62) = 305000.d0
  Mcase65TAY_V_radius(63) = 310000.d0
  Mcase65TAY_V_radius(64) = 315000.d0
  Mcase65TAY_V_radius(65) = 320000.d0
  Mcase65TAY_V_radius(66) = 325000.d0
  Mcase65TAY_V_radius(67) = 330000.d0
  Mcase65TAY_V_radius(68) = 335000.d0
  Mcase65TAY_V_radius(69) = 340000.d0
  Mcase65TAY_V_radius(70) = 345000.d0
  Mcase65TAY_V_radius(71) = 350000.d0
  Mcase65TAY_V_radius(72) = 355000.d0
  Mcase65TAY_V_radius(73) = 360000.d0
  Mcase65TAY_V_radius(74) = 365000.d0
  Mcase65TAY_V_radius(75) = 370000.d0
  Mcase65TAY_V_radius(76) = 375000.d0
  Mcase65TAY_V_radius(77) = 380000.d0
  Mcase65TAY_V_radius(78) = 385000.d0
  Mcase65TAY_V_radius(79) = 390000.d0
  Mcase65TAY_V_radius(80) = 395000.d0
  Mcase65TAY_V_radius(81) = 400000.d0
  Mcase65TAY_V_radius(82) = 405000.d0
  Mcase65TAY_V_radius(83) = 410000.d0
  Mcase65TAY_V_radius(84) = 415000.d0
  Mcase65TAY_V_radius(85) = 420000.d0
  Mcase65TAY_V_radius(86) = 425000.d0
  Mcase65TAY_V_radius(87) = 430000.d0
  Mcase65TAY_V_radius(88) = 435000.d0
  Mcase65TAY_V_radius(89) = 440000.d0
  Mcase65TAY_V_radius(90) = 445000.d0
  Mcase65TAY_V_radius(91) = 450000.d0
  Mcase65TAY_V_radius(92) = 455000.d0
  Mcase65TAY_V_radius(93) = 460000.d0
  Mcase65TAY_V_radius(94) = 465000.d0
  Mcase65TAY_V_radius(95) = 470000.d0
  Mcase65TAY_V_radius(96) = 475000.d0
  Mcase65TAY_V_radius(97) = 480000.d0
  Mcase65TAY_V_radius(98) = 485000.d0
  Mcase65TAY_V_radius(99) = 490000.d0
  Mcase65TAY_V_radius(100) = 495000.d0
  Mcase65TAY_V_radius(101) = 500000.d0
  Mcase65TAY_V_radius(102) = 505000.d0
  Mcase65TAY_V_radius(103) = 510000.d0
  Mcase65TAY_V_radius(104) = 515000.d0  ! ICB inner core
  Mcase65TAY_V_radius(105) = 515000.d0  ! ICB outer core
  Mcase65TAY_V_radius(106) = 520000.d0
  Mcase65TAY_V_radius(107) = 525000.d0
  Mcase65TAY_V_radius(108) = 530000.d0
  Mcase65TAY_V_radius(109) = 535000.d0
  Mcase65TAY_V_radius(110) = 540000.d0
  Mcase65TAY_V_radius(111) = 545000.d0
  Mcase65TAY_V_radius(112) = 550000.d0
  Mcase65TAY_V_radius(113) = 555000.d0
  Mcase65TAY_V_radius(114) = 560000.d0
  Mcase65TAY_V_radius(115) = 565000.d0
  Mcase65TAY_V_radius(116) = 570000.d0
  Mcase65TAY_V_radius(117) = 575000.d0
  Mcase65TAY_V_radius(118) = 580000.d0
  Mcase65TAY_V_radius(119) = 585000.d0
  Mcase65TAY_V_radius(120) = 590000.d0
  Mcase65TAY_V_radius(121) = 595000.d0
  Mcase65TAY_V_radius(122) = 600000.d0
  Mcase65TAY_V_radius(123) = 605000.d0
  Mcase65TAY_V_radius(124) = 610000.d0
  Mcase65TAY_V_radius(125) = 615000.d0
  Mcase65TAY_V_radius(126) = 620000.d0
  Mcase65TAY_V_radius(127) = 625000.d0
  Mcase65TAY_V_radius(128) = 630000.d0
  Mcase65TAY_V_radius(129) = 635000.d0
  Mcase65TAY_V_radius(130) = 640000.d0
  Mcase65TAY_V_radius(131) = 645000.d0
  Mcase65TAY_V_radius(132) = 650000.d0
  Mcase65TAY_V_radius(133) = 655000.d0
  Mcase65TAY_V_radius(134) = 660000.d0
  Mcase65TAY_V_radius(135) = 665000.d0
  Mcase65TAY_V_radius(136) = 670000.d0
  Mcase65TAY_V_radius(137) = 675000.d0
  Mcase65TAY_V_radius(138) = 680000.d0
  Mcase65TAY_V_radius(139) = 685000.d0
  Mcase65TAY_V_radius(140) = 690000.d0
  Mcase65TAY_V_radius(141) = 695000.d0
  Mcase65TAY_V_radius(142) = 700000.d0
  Mcase65TAY_V_radius(143) = 705000.d0
  Mcase65TAY_V_radius(144) = 710000.d0
  Mcase65TAY_V_radius(145) = 715000.d0
  Mcase65TAY_V_radius(146) = 720000.d0
  Mcase65TAY_V_radius(147) = 725000.d0
  Mcase65TAY_V_radius(148) = 730000.d0
  Mcase65TAY_V_radius(149) = 735000.d0
  Mcase65TAY_V_radius(150) = 740000.d0
  Mcase65TAY_V_radius(151) = 745000.d0
  Mcase65TAY_V_radius(152) = 750000.d0
  Mcase65TAY_V_radius(153) = 755000.d0
  Mcase65TAY_V_radius(154) = 760000.d0
  Mcase65TAY_V_radius(155) = 765000.d0
  Mcase65TAY_V_radius(156) = 770000.d0
  Mcase65TAY_V_radius(157) = 775000.d0
  Mcase65TAY_V_radius(158) = 780000.d0
  Mcase65TAY_V_radius(159) = 785000.d0
  Mcase65TAY_V_radius(160) = 790000.d0
  Mcase65TAY_V_radius(161) = 795000.d0
  Mcase65TAY_V_radius(162) = 800000.d0
  Mcase65TAY_V_radius(163) = 805000.d0
  Mcase65TAY_V_radius(164) = 810000.d0
  Mcase65TAY_V_radius(165) = 815000.d0
  Mcase65TAY_V_radius(166) = 820000.d0
  Mcase65TAY_V_radius(167) = 825000.d0
  Mcase65TAY_V_radius(168) = 830000.d0
  Mcase65TAY_V_radius(169) = 835000.d0
  Mcase65TAY_V_radius(170) = 840000.d0
  Mcase65TAY_V_radius(171) = 845000.d0
  Mcase65TAY_V_radius(172) = 850000.d0
  Mcase65TAY_V_radius(173) = 855000.d0
  Mcase65TAY_V_radius(174) = 860000.d0
  Mcase65TAY_V_radius(175) = 865000.d0
  Mcase65TAY_V_radius(176) = 870000.d0
  Mcase65TAY_V_radius(177) = 875000.d0
  Mcase65TAY_V_radius(178) = 880000.d0
  Mcase65TAY_V_radius(179) = 885000.d0
  Mcase65TAY_V_radius(180) = 890000.d0
  Mcase65TAY_V_radius(181) = 895000.d0
  Mcase65TAY_V_radius(182) = 900000.d0
  Mcase65TAY_V_radius(183) = 905000.d0
  Mcase65TAY_V_radius(184) = 910000.d0
  Mcase65TAY_V_radius(185) = 915000.d0
  Mcase65TAY_V_radius(186) = 920000.d0
  Mcase65TAY_V_radius(187) = 925000.d0
  Mcase65TAY_V_radius(188) = 930000.d0
  Mcase65TAY_V_radius(189) = 935000.d0
  Mcase65TAY_V_radius(190) = 940000.d0
  Mcase65TAY_V_radius(191) = 945000.d0
  Mcase65TAY_V_radius(192) = 950000.d0
  Mcase65TAY_V_radius(193) = 955000.d0
  Mcase65TAY_V_radius(194) = 960000.d0
  Mcase65TAY_V_radius(195) = 965000.d0
  Mcase65TAY_V_radius(196) = 970000.d0
  Mcase65TAY_V_radius(197) = 975000.d0
  Mcase65TAY_V_radius(198) = 980000.d0
  Mcase65TAY_V_radius(199) = 985000.d0
  Mcase65TAY_V_radius(200) = 990000.d0
  Mcase65TAY_V_radius(201) = 995000.d0
  Mcase65TAY_V_radius(202) = 1000000.d0
  Mcase65TAY_V_radius(203) = 1005000.d0
  Mcase65TAY_V_radius(204) = 1010000.d0
  Mcase65TAY_V_radius(205) = 1015000.d0
  Mcase65TAY_V_radius(206) = 1020000.d0
  Mcase65TAY_V_radius(207) = 1025000.d0
  Mcase65TAY_V_radius(208) = 1030000.d0
  Mcase65TAY_V_radius(209) = 1035000.d0
  Mcase65TAY_V_radius(210) = 1040000.d0
  Mcase65TAY_V_radius(211) = 1045000.d0
  Mcase65TAY_V_radius(212) = 1050000.d0
  Mcase65TAY_V_radius(213) = 1055000.d0
  Mcase65TAY_V_radius(214) = 1060000.d0
  Mcase65TAY_V_radius(215) = 1065000.d0
  Mcase65TAY_V_radius(216) = 1070000.d0
  Mcase65TAY_V_radius(217) = 1075000.d0
  Mcase65TAY_V_radius(218) = 1080000.d0
  Mcase65TAY_V_radius(219) = 1085000.d0
  Mcase65TAY_V_radius(220) = 1090000.d0
  Mcase65TAY_V_radius(221) = 1095000.d0
  Mcase65TAY_V_radius(222) = 1100000.d0
  Mcase65TAY_V_radius(223) = 1105000.d0
  Mcase65TAY_V_radius(224) = 1110000.d0
  Mcase65TAY_V_radius(225) = 1115000.d0
  Mcase65TAY_V_radius(226) = 1120000.d0
  Mcase65TAY_V_radius(227) = 1125000.d0
  Mcase65TAY_V_radius(228) = 1130000.d0
  Mcase65TAY_V_radius(229) = 1135000.d0
  Mcase65TAY_V_radius(230) = 1140000.d0
  Mcase65TAY_V_radius(231) = 1145000.d0
  Mcase65TAY_V_radius(232) = 1150000.d0
  Mcase65TAY_V_radius(233) = 1155000.d0
  Mcase65TAY_V_radius(234) = 1160000.d0
  Mcase65TAY_V_radius(235) = 1165000.d0
  Mcase65TAY_V_radius(236) = 1170000.d0
  Mcase65TAY_V_radius(237) = 1175000.d0
  Mcase65TAY_V_radius(238) = 1180000.d0
  Mcase65TAY_V_radius(239) = 1185000.d0
  Mcase65TAY_V_radius(240) = 1190000.d0
  Mcase65TAY_V_radius(241) = 1195000.d0
  Mcase65TAY_V_radius(242) = 1200000.d0
  Mcase65TAY_V_radius(243) = 1205000.d0
  Mcase65TAY_V_radius(244) = 1210000.d0
  Mcase65TAY_V_radius(245) = 1215000.d0
  Mcase65TAY_V_radius(246) = 1220000.d0
  Mcase65TAY_V_radius(247) = 1225000.d0
  Mcase65TAY_V_radius(248) = 1230000.d0
  Mcase65TAY_V_radius(249) = 1235000.d0
  Mcase65TAY_V_radius(250) = 1240000.d0
  Mcase65TAY_V_radius(251) = 1245000.d0
  Mcase65TAY_V_radius(252) = 1250000.d0
  Mcase65TAY_V_radius(253) = 1255000.d0
  Mcase65TAY_V_radius(254) = 1260000.d0
  Mcase65TAY_V_radius(255) = 1265000.d0
  Mcase65TAY_V_radius(256) = 1270000.d0
  Mcase65TAY_V_radius(257) = 1275000.d0
  Mcase65TAY_V_radius(258) = 1280000.d0
  Mcase65TAY_V_radius(259) = 1285000.d0
  Mcase65TAY_V_radius(260) = 1290000.d0
  Mcase65TAY_V_radius(261) = 1295000.d0
  Mcase65TAY_V_radius(262) = 1300000.d0
  Mcase65TAY_V_radius(263) = 1305000.d0
  Mcase65TAY_V_radius(264) = 1310000.d0
  Mcase65TAY_V_radius(265) = 1315000.d0
  Mcase65TAY_V_radius(266) = 1320000.d0
  Mcase65TAY_V_radius(267) = 1325000.d0
  Mcase65TAY_V_radius(268) = 1330000.d0
  Mcase65TAY_V_radius(269) = 1335000.d0
  Mcase65TAY_V_radius(270) = 1340000.d0
  Mcase65TAY_V_radius(271) = 1345000.d0
  Mcase65TAY_V_radius(272) = 1350000.d0
  Mcase65TAY_V_radius(273) = 1355000.d0
  Mcase65TAY_V_radius(274) = 1360000.d0
  Mcase65TAY_V_radius(275) = 1365000.d0
  Mcase65TAY_V_radius(276) = 1370000.d0
  Mcase65TAY_V_radius(277) = 1375000.d0
  Mcase65TAY_V_radius(278) = 1380000.d0
  Mcase65TAY_V_radius(279) = 1385000.d0
  Mcase65TAY_V_radius(280) = 1390000.d0
  Mcase65TAY_V_radius(281) = 1395000.d0
  Mcase65TAY_V_radius(282) = 1400000.d0
  Mcase65TAY_V_radius(283) = 1405000.d0
  Mcase65TAY_V_radius(284) = 1410000.d0
  Mcase65TAY_V_radius(285) = 1415000.d0
  Mcase65TAY_V_radius(286) = 1420000.d0
  Mcase65TAY_V_radius(287) = 1425000.d0
  Mcase65TAY_V_radius(288) = 1430000.d0
  Mcase65TAY_V_radius(289) = 1435000.d0
  Mcase65TAY_V_radius(290) = 1440000.d0
  Mcase65TAY_V_radius(291) = 1445000.d0
  Mcase65TAY_V_radius(292) = 1450000.d0
  Mcase65TAY_V_radius(293) = 1455000.d0
  Mcase65TAY_V_radius(294) = 1460000.d0
  Mcase65TAY_V_radius(295) = 1465000.d0
  Mcase65TAY_V_radius(296) = 1470000.d0
  Mcase65TAY_V_radius(297) = 1475000.d0
  Mcase65TAY_V_radius(298) = 1480000.d0
  Mcase65TAY_V_radius(299) = 1485000.d0
  Mcase65TAY_V_radius(300) = 1490000.d0
  Mcase65TAY_V_radius(301) = 1495000.d0
  Mcase65TAY_V_radius(302) = 1500000.d0
  Mcase65TAY_V_radius(303) = 1505000.d0
  Mcase65TAY_V_radius(304) = 1510000.d0
  Mcase65TAY_V_radius(305) = 1515000.d0
  Mcase65TAY_V_radius(306) = 1520000.d0
  Mcase65TAY_V_radius(307) = 1525000.d0
  Mcase65TAY_V_radius(308) = 1530000.d0
  Mcase65TAY_V_radius(309) = 1535000.d0
  Mcase65TAY_V_radius(310) = 1540000.d0
  Mcase65TAY_V_radius(311) = 1545000.d0
  Mcase65TAY_V_radius(312) = 1550000.d0
  Mcase65TAY_V_radius(313) = 1555000.d0
  Mcase65TAY_V_radius(314) = 1560000.d0
  Mcase65TAY_V_radius(315) = 1565000.d0
  Mcase65TAY_V_radius(316) = 1570000.d0
  Mcase65TAY_V_radius(317) = 1575000.d0
  Mcase65TAY_V_radius(318) = 1580000.d0
  Mcase65TAY_V_radius(319) = 1585000.d0
  Mcase65TAY_V_radius(320) = 1590000.d0
  Mcase65TAY_V_radius(321) = 1595000.d0
  Mcase65TAY_V_radius(322) = 1600000.d0
  Mcase65TAY_V_radius(323) = 1605000.d0
  Mcase65TAY_V_radius(324) = 1610000.d0
  Mcase65TAY_V_radius(325) = 1615000.d0
  Mcase65TAY_V_radius(326) = 1620000.d0
  Mcase65TAY_V_radius(327) = 1625000.d0
  Mcase65TAY_V_radius(328) = 1630000.d0
  Mcase65TAY_V_radius(329) = 1635000.d0
  Mcase65TAY_V_radius(330) = 1640000.d0
  Mcase65TAY_V_radius(331) = 1645000.d0
  Mcase65TAY_V_radius(332) = 1650000.d0
  Mcase65TAY_V_radius(333) = 1655000.d0
  Mcase65TAY_V_radius(334) = 1660000.d0
  Mcase65TAY_V_radius(335) = 1665000.d0
  Mcase65TAY_V_radius(336) = 1670000.d0
  Mcase65TAY_V_radius(337) = 1675000.d0
  Mcase65TAY_V_radius(338) = 1680000.d0
  Mcase65TAY_V_radius(339) = 1685000.d0
  Mcase65TAY_V_radius(340) = 1690000.d0
  Mcase65TAY_V_radius(341) = 1695000.d0
  Mcase65TAY_V_radius(342) = 1700000.d0
  Mcase65TAY_V_radius(343) = 1705000.d0
  Mcase65TAY_V_radius(344) = 1710000.d0
  Mcase65TAY_V_radius(345) = 1715000.d0
  Mcase65TAY_V_radius(346) = 1720000.d0
  Mcase65TAY_V_radius(347) = 1725000.d0
  Mcase65TAY_V_radius(348) = 1730000.d0
  Mcase65TAY_V_radius(349) = 1735000.d0
  Mcase65TAY_V_radius(350) = 1740000.d0
  Mcase65TAY_V_radius(351) = 1745000.d0
  Mcase65TAY_V_radius(352) = 1750000.d0
  Mcase65TAY_V_radius(353) = 1755000.d0
  Mcase65TAY_V_radius(354) = 1760000.d0
  Mcase65TAY_V_radius(355) = 1765000.d0
  Mcase65TAY_V_radius(356) = 1770000.d0
  Mcase65TAY_V_radius(357) = 1775000.d0
  Mcase65TAY_V_radius(358) = 1780000.d0
  Mcase65TAY_V_radius(359) = 1785000.d0
  Mcase65TAY_V_radius(360) = 1790000.d0
  Mcase65TAY_V_radius(361) = 1795000.d0
  Mcase65TAY_V_radius(362) = 1800000.d0
  Mcase65TAY_V_radius(363) = 1805000.d0
  Mcase65TAY_V_radius(364) = 1810000.d0
  Mcase65TAY_V_radius(365) = 1815000.d0
  Mcase65TAY_V_radius(366) = 1820000.d0
  Mcase65TAY_V_radius(367) = 1825000.d0
  Mcase65TAY_V_radius(368) = 1830000.d0
  Mcase65TAY_V_radius(369) = 1835000.d0
  Mcase65TAY_V_radius(370) = 1840000.d0 ! CMB outer core
  Mcase65TAY_V_radius(371) = 1840000.d0 ! CMB mantle
  Mcase65TAY_V_radius(372) = 1845000.d0
  Mcase65TAY_V_radius(373) = 1850000.d0
  Mcase65TAY_V_radius(374) = 1855000.d0
  Mcase65TAY_V_radius(375) = 1860000.d0
  Mcase65TAY_V_radius(376) = 1865000.d0
  Mcase65TAY_V_radius(377) = 1870000.d0
  Mcase65TAY_V_radius(378) = 1875000.d0
  Mcase65TAY_V_radius(379) = 1880000.d0
  Mcase65TAY_V_radius(380) = 1885000.d0
  Mcase65TAY_V_radius(381) = 1890000.d0
  Mcase65TAY_V_radius(382) = 1895000.d0
  Mcase65TAY_V_radius(383) = 1900000.d0
  Mcase65TAY_V_radius(384) = 1905000.d0
  Mcase65TAY_V_radius(385) = 1910000.d0
  Mcase65TAY_V_radius(386) = 1915000.d0
  Mcase65TAY_V_radius(387) = 1920000.d0
  Mcase65TAY_V_radius(388) = 1925000.d0
  Mcase65TAY_V_radius(389) = 1930000.d0
  Mcase65TAY_V_radius(390) = 1935000.d0
  Mcase65TAY_V_radius(391) = 1940000.d0
  Mcase65TAY_V_radius(392) = 1945000.d0
  Mcase65TAY_V_radius(393) = 1950000.d0
  Mcase65TAY_V_radius(394) = 1955000.d0
  Mcase65TAY_V_radius(395) = 1960000.d0
  Mcase65TAY_V_radius(396) = 1965000.d0
  Mcase65TAY_V_radius(397) = 1970000.d0
  Mcase65TAY_V_radius(398) = 1975000.d0
  Mcase65TAY_V_radius(399) = 1980000.d0
  Mcase65TAY_V_radius(400) = 1985000.d0
  Mcase65TAY_V_radius(401) = 1990000.d0
  Mcase65TAY_V_radius(402) = 1995000.d0
  Mcase65TAY_V_radius(403) = 2000000.d0
  Mcase65TAY_V_radius(404) = 2005000.d0
  Mcase65TAY_V_radius(405) = 2010000.d0
  Mcase65TAY_V_radius(406) = 2015000.d0
  Mcase65TAY_V_radius(407) = 2020000.d0
  Mcase65TAY_V_radius(408) = 2025000.d0
  Mcase65TAY_V_radius(409) = 2030000.d0
  Mcase65TAY_V_radius(410) = 2035000.d0
  Mcase65TAY_V_radius(411) = 2040000.d0
  Mcase65TAY_V_radius(412) = 2045000.d0
  Mcase65TAY_V_radius(413) = 2050000.d0
  Mcase65TAY_V_radius(414) = 2055000.d0
  Mcase65TAY_V_radius(415) = 2060000.d0
  Mcase65TAY_V_radius(416) = 2065000.d0
  Mcase65TAY_V_radius(417) = 2070000.d0
  Mcase65TAY_V_radius(418) = 2075000.d0
  Mcase65TAY_V_radius(419) = 2080000.d0
  Mcase65TAY_V_radius(420) = 2085000.d0
  Mcase65TAY_V_radius(421) = 2090000.d0
  Mcase65TAY_V_radius(422) = 2095000.d0
  Mcase65TAY_V_radius(423) = 2100000.d0
  Mcase65TAY_V_radius(424) = 2105000.d0
  Mcase65TAY_V_radius(425) = 2110000.d0
  Mcase65TAY_V_radius(426) = 2115000.d0
  Mcase65TAY_V_radius(427) = 2120000.d0
  Mcase65TAY_V_radius(428) = 2125000.d0
  Mcase65TAY_V_radius(429) = 2130000.d0
  Mcase65TAY_V_radius(430) = 2135000.d0
  Mcase65TAY_V_radius(431) = 2140000.d0
  Mcase65TAY_V_radius(432) = 2145000.d0
  Mcase65TAY_V_radius(433) = 2150000.d0
  Mcase65TAY_V_radius(434) = 2155000.d0
  Mcase65TAY_V_radius(435) = 2160000.d0
  Mcase65TAY_V_radius(436) = 2165000.d0
  Mcase65TAY_V_radius(437) = 2170000.d0
  Mcase65TAY_V_radius(438) = 2175000.d0
  Mcase65TAY_V_radius(439) = 2180000.d0
  Mcase65TAY_V_radius(440) = 2185000.d0
  Mcase65TAY_V_radius(441) = 2190000.d0
  Mcase65TAY_V_radius(442) = 2195000.d0
  Mcase65TAY_V_radius(443) = 2200000.d0
  Mcase65TAY_V_radius(444) = 2205000.d0
  Mcase65TAY_V_radius(445) = 2210000.d0
  Mcase65TAY_V_radius(446) = 2215000.d0
  Mcase65TAY_V_radius(447) = 2220000.d0
  Mcase65TAY_V_radius(448) = 2225000.d0
  Mcase65TAY_V_radius(449) = 2230000.d0
  Mcase65TAY_V_radius(450) = 2235000.d0
  Mcase65TAY_V_radius(451) = 2240000.d0
  Mcase65TAY_V_radius(452) = 2245000.d0
  Mcase65TAY_V_radius(453) = 2250000.d0
  Mcase65TAY_V_radius(454) = 2255000.d0
  Mcase65TAY_V_radius(455) = 2260000.d0
  Mcase65TAY_V_radius(456) = 2265000.d0
  Mcase65TAY_V_radius(457) = 2270000.d0
  Mcase65TAY_V_radius(458) = 2275000.d0
  Mcase65TAY_V_radius(459) = 2280000.d0
  Mcase65TAY_V_radius(460) = 2285000.d0
  Mcase65TAY_V_radius(461) = 2290000.d0
  Mcase65TAY_V_radius(462) = 2295000.d0
  Mcase65TAY_V_radius(463) = 2300000.d0
  Mcase65TAY_V_radius(464) = 2305000.d0
  Mcase65TAY_V_radius(465) = 2310000.d0
  Mcase65TAY_V_radius(466) = 2315000.d0
  Mcase65TAY_V_radius(467) = 2320000.d0
  Mcase65TAY_V_radius(468) = 2325000.d0
  Mcase65TAY_V_radius(469) = 2330000.d0
  Mcase65TAY_V_radius(470) = 2335000.d0
  Mcase65TAY_V_radius(471) = 2340000.d0
  Mcase65TAY_V_radius(472) = 2345000.d0
  Mcase65TAY_V_radius(473) = 2350000.d0
  Mcase65TAY_V_radius(474) = 2355000.d0
  Mcase65TAY_V_radius(475) = 2360000.d0
  Mcase65TAY_V_radius(476) = 2365000.d0
  Mcase65TAY_V_radius(477) = 2370000.d0
  Mcase65TAY_V_radius(478) = 2375000.d0
  Mcase65TAY_V_radius(479) = 2380000.d0
  Mcase65TAY_V_radius(480) = 2385000.d0
  Mcase65TAY_V_radius(481) = 2390000.d0
  Mcase65TAY_V_radius(482) = 2395000.d0
  Mcase65TAY_V_radius(483) = 2400000.d0
  Mcase65TAY_V_radius(484) = 2405000.d0
  Mcase65TAY_V_radius(485) = 2410000.d0
  Mcase65TAY_V_radius(486) = 2415000.d0
  Mcase65TAY_V_radius(487) = 2420000.d0
  Mcase65TAY_V_radius(488) = 2425000.d0
  Mcase65TAY_V_radius(489) = 2430000.d0
  Mcase65TAY_V_radius(490) = 2435000.d0
  Mcase65TAY_V_radius(491) = 2440000.d0
  Mcase65TAY_V_radius(492) = 2445000.d0
  Mcase65TAY_V_radius(493) = 2450000.d0
  Mcase65TAY_V_radius(494) = 2455000.d0
  Mcase65TAY_V_radius(495) = 2460000.d0
  Mcase65TAY_V_radius(496) = 2465000.d0
  Mcase65TAY_V_radius(497) = 2470000.d0
  Mcase65TAY_V_radius(498) = 2475000.d0
  Mcase65TAY_V_radius(499) = 2480000.d0
  Mcase65TAY_V_radius(500) = 2485000.d0
  Mcase65TAY_V_radius(501) = 2490000.d0
  Mcase65TAY_V_radius(502) = 2495000.d0
  Mcase65TAY_V_radius(503) = 2500000.d0
  Mcase65TAY_V_radius(504) = 2505000.d0
  Mcase65TAY_V_radius(505) = 2510000.d0
  Mcase65TAY_V_radius(506) = 2515000.d0
  Mcase65TAY_V_radius(507) = 2520000.d0
  Mcase65TAY_V_radius(508) = 2525000.d0
  Mcase65TAY_V_radius(509) = 2530000.d0
  Mcase65TAY_V_radius(510) = 2535000.d0
  Mcase65TAY_V_radius(511) = 2540000.d0
  Mcase65TAY_V_radius(512) = 2545000.d0
  Mcase65TAY_V_radius(513) = 2550000.d0
  Mcase65TAY_V_radius(514) = 2555000.d0
  Mcase65TAY_V_radius(515) = 2560000.d0
  Mcase65TAY_V_radius(516) = 2565000.d0
  Mcase65TAY_V_radius(517) = 2570000.d0
  Mcase65TAY_V_radius(518) = 2575000.d0
  Mcase65TAY_V_radius(519) = 2580000.d0
  Mcase65TAY_V_radius(520) = 2585000.d0
  Mcase65TAY_V_radius(521) = 2590000.d0
  Mcase65TAY_V_radius(522) = 2595000.d0
  Mcase65TAY_V_radius(523) = 2600000.d0
  Mcase65TAY_V_radius(524) = 2605000.d0
  Mcase65TAY_V_radius(525) = 2610000.d0
  Mcase65TAY_V_radius(526) = 2615000.d0
  Mcase65TAY_V_radius(527) = 2620000.d0
  Mcase65TAY_V_radius(528) = 2625000.d0
  Mcase65TAY_V_radius(529) = 2630000.d0
  Mcase65TAY_V_radius(530) = 2635000.d0
  Mcase65TAY_V_radius(531) = 2640000.d0
  Mcase65TAY_V_radius(532) = 2645000.d0
  Mcase65TAY_V_radius(533) = 2650000.d0
  Mcase65TAY_V_radius(534) = 2655000.d0
  Mcase65TAY_V_radius(535) = 2660000.d0
  Mcase65TAY_V_radius(536) = 2665000.d0
  Mcase65TAY_V_radius(537) = 2670000.d0
  Mcase65TAY_V_radius(538) = 2675000.d0
  Mcase65TAY_V_radius(539) = 2680000.d0
  Mcase65TAY_V_radius(540) = 2685000.d0
  Mcase65TAY_V_radius(541) = 2690000.d0
  Mcase65TAY_V_radius(542) = 2695000.d0
  Mcase65TAY_V_radius(543) = 2700000.d0
  Mcase65TAY_V_radius(544) = 2705000.d0
  Mcase65TAY_V_radius(545) = 2710000.d0
  Mcase65TAY_V_radius(546) = 2715000.d0
  Mcase65TAY_V_radius(547) = 2720000.d0
  Mcase65TAY_V_radius(548) = 2725000.d0
  Mcase65TAY_V_radius(549) = 2730000.d0
  Mcase65TAY_V_radius(550) = 2735000.d0
  Mcase65TAY_V_radius(551) = 2740000.d0
  Mcase65TAY_V_radius(552) = 2745000.d0
  Mcase65TAY_V_radius(553) = 2750000.d0
  Mcase65TAY_V_radius(554) = 2755000.d0
  Mcase65TAY_V_radius(555) = 2760000.d0
  Mcase65TAY_V_radius(556) = 2765000.d0
  Mcase65TAY_V_radius(557) = 2770000.d0
  Mcase65TAY_V_radius(558) = 2775000.d0
  Mcase65TAY_V_radius(559) = 2780000.d0
  Mcase65TAY_V_radius(560) = 2785000.d0
  Mcase65TAY_V_radius(561) = 2790000.d0
  Mcase65TAY_V_radius(562) = 2795000.d0
  Mcase65TAY_V_radius(563) = 2800000.d0
  Mcase65TAY_V_radius(564) = 2805000.d0
  Mcase65TAY_V_radius(565) = 2810000.d0
  Mcase65TAY_V_radius(566) = 2815000.d0
  Mcase65TAY_V_radius(567) = 2820000.d0
  Mcase65TAY_V_radius(568) = 2825000.d0
  Mcase65TAY_V_radius(569) = 2830000.d0
  Mcase65TAY_V_radius(570) = 2835000.d0
  Mcase65TAY_V_radius(571) = 2840000.d0
  Mcase65TAY_V_radius(572) = 2845000.d0
  Mcase65TAY_V_radius(573) = 2850000.d0
  Mcase65TAY_V_radius(574) = 2855000.d0
  Mcase65TAY_V_radius(575) = 2860000.d0
  Mcase65TAY_V_radius(576) = 2865000.d0
  Mcase65TAY_V_radius(577) = 2870000.d0
  Mcase65TAY_V_radius(578) = 2875000.d0
  Mcase65TAY_V_radius(579) = 2880000.d0
  Mcase65TAY_V_radius(580) = 2885000.d0
  Mcase65TAY_V_radius(581) = 2890000.d0
  Mcase65TAY_V_radius(582) = 2895000.d0
  Mcase65TAY_V_radius(583) = 2900000.d0
  Mcase65TAY_V_radius(584) = 2905000.d0
  Mcase65TAY_V_radius(585) = 2910000.d0
  Mcase65TAY_V_radius(586) = 2915000.d0
  Mcase65TAY_V_radius(587) = 2920000.d0
  Mcase65TAY_V_radius(588) = 2925000.d0
  Mcase65TAY_V_radius(589) = 2930000.d0
  Mcase65TAY_V_radius(590) = 2935000.d0
  Mcase65TAY_V_radius(591) = 2940000.d0
  Mcase65TAY_V_radius(592) = 2945000.d0
  Mcase65TAY_V_radius(593) = 2950000.d0
  Mcase65TAY_V_radius(594) = 2955000.d0
  Mcase65TAY_V_radius(595) = 2960000.d0
  Mcase65TAY_V_radius(596) = 2965000.d0
  Mcase65TAY_V_radius(597) = 2970000.d0
  Mcase65TAY_V_radius(598) = 2975000.d0
  Mcase65TAY_V_radius(599) = 2980000.d0
  Mcase65TAY_V_radius(600) = 2985000.d0
  Mcase65TAY_V_radius(601) = 2990000.d0
  Mcase65TAY_V_radius(602) = 2995000.d0
  Mcase65TAY_V_radius(603) = 3000000.d0
  Mcase65TAY_V_radius(604) = 3005000.d0
  Mcase65TAY_V_radius(605) = 3010000.d0
  Mcase65TAY_V_radius(606) = 3015000.d0
  Mcase65TAY_V_radius(607) = 3020000.d0
  Mcase65TAY_V_radius(608) = 3025000.d0
  Mcase65TAY_V_radius(609) = 3030000.d0
  Mcase65TAY_V_radius(610) = 3035000.d0
  Mcase65TAY_V_radius(611) = 3040000.d0
  Mcase65TAY_V_radius(612) = 3045000.d0
  Mcase65TAY_V_radius(613) = 3050000.d0
  Mcase65TAY_V_radius(614) = 3055000.d0
  Mcase65TAY_V_radius(615) = 3060000.d0
  Mcase65TAY_V_radius(616) = 3065000.d0
  Mcase65TAY_V_radius(617) = 3070000.d0
  Mcase65TAY_V_radius(618) = 3075000.d0
  Mcase65TAY_V_radius(619) = 3080000.d0
  Mcase65TAY_V_radius(620) = 3085000.d0
  Mcase65TAY_V_radius(621) = 3090000.d0
  Mcase65TAY_V_radius(622) = 3095000.d0
  Mcase65TAY_V_radius(623) = 3100000.d0
  Mcase65TAY_V_radius(624) = 3105000.d0
  Mcase65TAY_V_radius(625) = 3110000.d0
  Mcase65TAY_V_radius(626) = 3115000.d0
  Mcase65TAY_V_radius(627) = 3120000.d0
  Mcase65TAY_V_radius(628) = 3125000.d0
  Mcase65TAY_V_radius(629) = 3130000.d0
  Mcase65TAY_V_radius(630) = 3135000.d0
  Mcase65TAY_V_radius(631) = 3140000.d0
  Mcase65TAY_V_radius(632) = 3145000.d0
  Mcase65TAY_V_radius(633) = 3150000.d0
  Mcase65TAY_V_radius(634) = 3155000.d0
  Mcase65TAY_V_radius(635) = 3160000.d0
  Mcase65TAY_V_radius(636) = 3165000.d0
  Mcase65TAY_V_radius(637) = 3170000.d0
  Mcase65TAY_V_radius(638) = 3175000.d0
  Mcase65TAY_V_radius(639) = 3180000.d0
  Mcase65TAY_V_radius(640) = 3185000.d0
  Mcase65TAY_V_radius(641) = 3190000.d0
  Mcase65TAY_V_radius(642) = 3195000.d0
  Mcase65TAY_V_radius(643) = 3200000.d0
  Mcase65TAY_V_radius(644) = 3205000.d0
  Mcase65TAY_V_radius(645) = 3210000.d0
  Mcase65TAY_V_radius(646) = 3215000.d0
  Mcase65TAY_V_radius(647) = 3220000.d0
  Mcase65TAY_V_radius(648) = 3225000.d0
  Mcase65TAY_V_radius(649) = 3230000.d0
  Mcase65TAY_V_radius(650) = 3235000.d0
  Mcase65TAY_V_radius(651) = 3240000.d0
  Mcase65TAY_V_radius(652) = 3245000.d0
  Mcase65TAY_V_radius(653) = 3250000.d0
  Mcase65TAY_V_radius(654) = 3255000.d0
  Mcase65TAY_V_radius(655) = 3260000.d0
  Mcase65TAY_V_radius(656) = 3265000.d0
  Mcase65TAY_V_radius(657) = 3270000.d0
  Mcase65TAY_V_radius(658) = 3275000.d0
  Mcase65TAY_V_radius(659) = 3280000.d0   ! MOHO (smooth transition to crustal values)
  Mcase65TAY_V_radius(660) = 3285000.d0
  Mcase65TAY_V_radius(661) = 3290000.d0
  Mcase65TAY_V_radius(662) = 3295000.d0
  Mcase65TAY_V_radius(663) = 3300000.d0
  Mcase65TAY_V_radius(664) = 3305000.d0
  Mcase65TAY_V_radius(665) = 3310000.d0
  Mcase65TAY_V_radius(666) = 3315000.d0
  Mcase65TAY_V_radius(667) = 3320000.d0
  Mcase65TAY_V_radius(668) = 3325000.d0
  Mcase65TAY_V_radius(669) = 3330000.d0
  Mcase65TAY_V_radius(670) = 3335000.d0
  Mcase65TAY_V_radius(671) = 3340000.d0
  Mcase65TAY_V_radius(672) = 3345000.d0
  Mcase65TAY_V_radius(673) = 3350000.d0
  Mcase65TAY_V_radius(674) = 3355000.d0
  Mcase65TAY_V_radius(675) = 3360000.d0
  Mcase65TAY_V_radius(676) = 3365000.d0
  Mcase65TAY_V_radius(677) = 3370000.d0
  Mcase65TAY_V_radius(678) = 3375000.d0
  Mcase65TAY_V_radius(679) = 3380000.d0
  Mcase65TAY_V_radius(680) = 3385000.d0
  Mcase65TAY_V_radius(681) = 3390000.d0

  Mcase65TAY_V_density(1) = 8.6085d0
  Mcase65TAY_V_density(2) = 8.6085d0
  Mcase65TAY_V_density(3) = 8.6085d0
  Mcase65TAY_V_density(4) = 8.6085d0
  Mcase65TAY_V_density(5) = 8.6085d0
  Mcase65TAY_V_density(6) = 8.6085d0
  Mcase65TAY_V_density(7) = 8.6085d0
  Mcase65TAY_V_density(8) = 8.6085d0
  Mcase65TAY_V_density(9) = 8.6085d0
  Mcase65TAY_V_density(10) = 8.6085d0
  Mcase65TAY_V_density(11) = 8.6085d0
  Mcase65TAY_V_density(12) = 8.6085d0
  Mcase65TAY_V_density(13) = 8.6085d0
  Mcase65TAY_V_density(14) = 8.6085d0
  Mcase65TAY_V_density(15) = 8.6085d0
  Mcase65TAY_V_density(16) = 8.6085d0
  Mcase65TAY_V_density(17) = 8.6085d0
  Mcase65TAY_V_density(18) = 8.6085d0
  Mcase65TAY_V_density(19) = 8.6085d0
  Mcase65TAY_V_density(20) = 8.6085d0
  Mcase65TAY_V_density(21) = 8.6085d0
  Mcase65TAY_V_density(22) = 8.6085d0
  Mcase65TAY_V_density(23) = 8.6085d0
  Mcase65TAY_V_density(24) = 8.6085d0
  Mcase65TAY_V_density(25) = 8.6085d0
  Mcase65TAY_V_density(26) = 8.6085d0
  Mcase65TAY_V_density(27) = 8.6085d0
  Mcase65TAY_V_density(28) = 8.6085d0
  Mcase65TAY_V_density(29) = 8.6085d0
  Mcase65TAY_V_density(30) = 8.6085d0
  Mcase65TAY_V_density(31) = 8.6085d0
  Mcase65TAY_V_density(32) = 8.6085d0
  Mcase65TAY_V_density(33) = 8.6085d0
  Mcase65TAY_V_density(34) = 8.6085d0
  Mcase65TAY_V_density(35) = 8.6085d0
  Mcase65TAY_V_density(36) = 8.6085d0
  Mcase65TAY_V_density(37) = 8.6085d0
  Mcase65TAY_V_density(38) = 8.6085d0
  Mcase65TAY_V_density(39) = 8.6085d0
  Mcase65TAY_V_density(40) = 8.6085d0
  Mcase65TAY_V_density(41) = 8.6085d0
  Mcase65TAY_V_density(42) = 8.6085d0
  Mcase65TAY_V_density(43) = 8.6085d0
  Mcase65TAY_V_density(44) = 8.6085d0
  Mcase65TAY_V_density(45) = 8.6085d0
  Mcase65TAY_V_density(46) = 8.6085d0
  Mcase65TAY_V_density(47) = 8.6085d0
  Mcase65TAY_V_density(48) = 8.6085d0
  Mcase65TAY_V_density(49) = 8.6085d0
  Mcase65TAY_V_density(50) = 8.6085d0
  Mcase65TAY_V_density(51) = 8.6085d0
  Mcase65TAY_V_density(52) = 8.6085d0
  Mcase65TAY_V_density(53) = 8.6085d0
  Mcase65TAY_V_density(54) = 8.6085d0
  Mcase65TAY_V_density(55) = 8.6085d0
  Mcase65TAY_V_density(56) = 8.6085d0
  Mcase65TAY_V_density(57) = 8.6085d0
  Mcase65TAY_V_density(58) = 8.6085d0
  Mcase65TAY_V_density(59) = 8.6085d0
  Mcase65TAY_V_density(60) = 8.6085d0
  Mcase65TAY_V_density(61) = 8.6085d0
  Mcase65TAY_V_density(62) = 8.6085d0
  Mcase65TAY_V_density(63) = 8.6085d0
  Mcase65TAY_V_density(64) = 8.6085d0
  Mcase65TAY_V_density(65) = 8.6085d0
  Mcase65TAY_V_density(66) = 8.6085d0
  Mcase65TAY_V_density(67) = 8.6085d0
  Mcase65TAY_V_density(68) = 8.6085d0
  Mcase65TAY_V_density(69) = 8.6085d0
  Mcase65TAY_V_density(70) = 8.6085d0
  Mcase65TAY_V_density(71) = 8.6085d0
  Mcase65TAY_V_density(72) = 8.6085d0
  Mcase65TAY_V_density(73) = 8.6085d0
  Mcase65TAY_V_density(74) = 8.6085d0
  Mcase65TAY_V_density(75) = 8.6085d0
  Mcase65TAY_V_density(76) = 8.6085d0
  Mcase65TAY_V_density(77) = 8.6085d0
  Mcase65TAY_V_density(78) = 8.6085d0
  Mcase65TAY_V_density(79) = 8.6085d0
  Mcase65TAY_V_density(80) = 8.6085d0
  Mcase65TAY_V_density(81) = 8.6085d0
  Mcase65TAY_V_density(82) = 8.6085d0
  Mcase65TAY_V_density(83) = 8.6085d0
  Mcase65TAY_V_density(84) = 8.6085d0
  Mcase65TAY_V_density(85) = 8.6085d0
  Mcase65TAY_V_density(86) = 8.6085d0
  Mcase65TAY_V_density(87) = 8.6085d0
  Mcase65TAY_V_density(88) = 8.6085d0
  Mcase65TAY_V_density(89) = 8.6085d0
  Mcase65TAY_V_density(90) = 8.6085d0
  Mcase65TAY_V_density(91) = 8.6085d0
  Mcase65TAY_V_density(92) = 8.6085d0
  Mcase65TAY_V_density(93) = 8.6085d0
  Mcase65TAY_V_density(94) = 8.6085d0
  Mcase65TAY_V_density(95) = 8.6085d0
  Mcase65TAY_V_density(96) = 8.6085d0
  Mcase65TAY_V_density(97) = 8.6085d0
  Mcase65TAY_V_density(98) = 8.6085d0
  Mcase65TAY_V_density(99) = 8.6085d0
  Mcase65TAY_V_density(100) = 8.6085d0
  Mcase65TAY_V_density(101) = 8.6085d0
  Mcase65TAY_V_density(102) = 8.6085d0
  Mcase65TAY_V_density(103) = 8.6085d0
  Mcase65TAY_V_density(104) = 8.6085d0 ! ICB inner core
  Mcase65TAY_V_density(105) = 5.9159d0 ! ICB outer core
  Mcase65TAY_V_density(106) = 5.915d0
  Mcase65TAY_V_density(107) = 5.914d0
  Mcase65TAY_V_density(108) = 5.9131d0
  Mcase65TAY_V_density(109) = 5.9121d0
  Mcase65TAY_V_density(110) = 5.9112d0
  Mcase65TAY_V_density(111) = 5.9102d0
  Mcase65TAY_V_density(112) = 5.9092d0
  Mcase65TAY_V_density(113) = 5.9082d0
  Mcase65TAY_V_density(114) = 5.9072d0
  Mcase65TAY_V_density(115) = 5.9062d0
  Mcase65TAY_V_density(116) = 5.9052d0
  Mcase65TAY_V_density(117) = 5.9041d0
  Mcase65TAY_V_density(118) = 5.9031d0
  Mcase65TAY_V_density(119) = 5.902d0
  Mcase65TAY_V_density(120) = 5.901d0
  Mcase65TAY_V_density(121) = 5.8999d0
  Mcase65TAY_V_density(122) = 5.8988d0
  Mcase65TAY_V_density(123) = 5.8978d0
  Mcase65TAY_V_density(124) = 5.8967d0
  Mcase65TAY_V_density(125) = 5.8955d0
  Mcase65TAY_V_density(126) = 5.8944d0
  Mcase65TAY_V_density(127) = 5.8933d0
  Mcase65TAY_V_density(128) = 5.8922d0
  Mcase65TAY_V_density(129) = 5.891d0
  Mcase65TAY_V_density(130) = 5.8899d0
  Mcase65TAY_V_density(131) = 5.8887d0
  Mcase65TAY_V_density(132) = 5.8875d0
  Mcase65TAY_V_density(133) = 5.8864d0
  Mcase65TAY_V_density(134) = 5.8852d0
  Mcase65TAY_V_density(135) = 5.884d0
  Mcase65TAY_V_density(136) = 5.8828d0
  Mcase65TAY_V_density(137) = 5.8815d0
  Mcase65TAY_V_density(138) = 5.8803d0
  Mcase65TAY_V_density(139) = 5.8791d0
  Mcase65TAY_V_density(140) = 5.8778d0
  Mcase65TAY_V_density(141) = 5.8766d0
  Mcase65TAY_V_density(142) = 5.8753d0
  Mcase65TAY_V_density(143) = 5.8741d0
  Mcase65TAY_V_density(144) = 5.8728d0
  Mcase65TAY_V_density(145) = 5.8715d0
  Mcase65TAY_V_density(146) = 5.8702d0
  Mcase65TAY_V_density(147) = 5.8689d0
  Mcase65TAY_V_density(148) = 5.8675d0
  Mcase65TAY_V_density(149) = 5.8662d0
  Mcase65TAY_V_density(150) = 5.8649d0
  Mcase65TAY_V_density(151) = 5.8635d0
  Mcase65TAY_V_density(152) = 5.8622d0
  Mcase65TAY_V_density(153) = 5.8608d0
  Mcase65TAY_V_density(154) = 5.8594d0
  Mcase65TAY_V_density(155) = 5.858d0
  Mcase65TAY_V_density(156) = 5.8567d0
  Mcase65TAY_V_density(157) = 5.8553d0
  Mcase65TAY_V_density(158) = 5.8538d0
  Mcase65TAY_V_density(159) = 5.8524d0
  Mcase65TAY_V_density(160) = 5.851d0
  Mcase65TAY_V_density(161) = 5.8495d0
  Mcase65TAY_V_density(162) = 5.8481d0
  Mcase65TAY_V_density(163) = 5.8466d0
  Mcase65TAY_V_density(164) = 5.8452d0
  Mcase65TAY_V_density(165) = 5.8437d0
  Mcase65TAY_V_density(166) = 5.8422d0
  Mcase65TAY_V_density(167) = 5.8407d0
  Mcase65TAY_V_density(168) = 5.8392d0
  Mcase65TAY_V_density(169) = 5.8377d0
  Mcase65TAY_V_density(170) = 5.8361d0
  Mcase65TAY_V_density(171) = 5.8346d0
  Mcase65TAY_V_density(172) = 5.833d0
  Mcase65TAY_V_density(173) = 5.8315d0
  Mcase65TAY_V_density(174) = 5.8299d0
  Mcase65TAY_V_density(175) = 5.8284d0
  Mcase65TAY_V_density(176) = 5.8268d0
  Mcase65TAY_V_density(177) = 5.8252d0
  Mcase65TAY_V_density(178) = 5.8236d0
  Mcase65TAY_V_density(179) = 5.822d0
  Mcase65TAY_V_density(180) = 5.8203d0
  Mcase65TAY_V_density(181) = 5.8187d0
  Mcase65TAY_V_density(182) = 5.8171d0
  Mcase65TAY_V_density(183) = 5.8154d0
  Mcase65TAY_V_density(184) = 5.8137d0
  Mcase65TAY_V_density(185) = 5.8121d0
  Mcase65TAY_V_density(186) = 5.8104d0
  Mcase65TAY_V_density(187) = 5.8087d0
  Mcase65TAY_V_density(188) = 5.807d0
  Mcase65TAY_V_density(189) = 5.8053d0
  Mcase65TAY_V_density(190) = 5.8036d0
  Mcase65TAY_V_density(191) = 5.8018d0
  Mcase65TAY_V_density(192) = 5.8001d0
  Mcase65TAY_V_density(193) = 5.7983d0
  Mcase65TAY_V_density(194) = 5.7966d0
  Mcase65TAY_V_density(195) = 5.7948d0
  Mcase65TAY_V_density(196) = 5.793d0
  Mcase65TAY_V_density(197) = 5.7913d0
  Mcase65TAY_V_density(198) = 5.7894d0
  Mcase65TAY_V_density(199) = 5.7876d0
  Mcase65TAY_V_density(200) = 5.7858d0
  Mcase65TAY_V_density(201) = 5.784d0
  Mcase65TAY_V_density(202) = 5.7821d0
  Mcase65TAY_V_density(203) = 5.7803d0
  Mcase65TAY_V_density(204) = 5.7785d0
  Mcase65TAY_V_density(205) = 5.7766d0
  Mcase65TAY_V_density(206) = 5.7747d0
  Mcase65TAY_V_density(207) = 5.7728d0
  Mcase65TAY_V_density(208) = 5.7709d0
  Mcase65TAY_V_density(209) = 5.769d0
  Mcase65TAY_V_density(210) = 5.7671d0
  Mcase65TAY_V_density(211) = 5.7652d0
  Mcase65TAY_V_density(212) = 5.7632d0
  Mcase65TAY_V_density(213) = 5.7613d0
  Mcase65TAY_V_density(214) = 5.7593d0
  Mcase65TAY_V_density(215) = 5.7574d0
  Mcase65TAY_V_density(216) = 5.7554d0
  Mcase65TAY_V_density(217) = 5.7534d0
  Mcase65TAY_V_density(218) = 5.7514d0
  Mcase65TAY_V_density(219) = 5.7494d0
  Mcase65TAY_V_density(220) = 5.7474d0
  Mcase65TAY_V_density(221) = 5.7453d0
  Mcase65TAY_V_density(222) = 5.7433d0
  Mcase65TAY_V_density(223) = 5.7413d0
  Mcase65TAY_V_density(224) = 5.7392d0
  Mcase65TAY_V_density(225) = 5.7371d0
  Mcase65TAY_V_density(226) = 5.7351d0
  Mcase65TAY_V_density(227) = 5.733d0
  Mcase65TAY_V_density(228) = 5.7309d0
  Mcase65TAY_V_density(229) = 5.7288d0
  Mcase65TAY_V_density(230) = 5.7267d0
  Mcase65TAY_V_density(231) = 5.7245d0
  Mcase65TAY_V_density(232) = 5.7224d0
  Mcase65TAY_V_density(233) = 5.7202d0
  Mcase65TAY_V_density(234) = 5.7181d0
  Mcase65TAY_V_density(235) = 5.7159d0
  Mcase65TAY_V_density(236) = 5.7137d0
  Mcase65TAY_V_density(237) = 5.7115d0
  Mcase65TAY_V_density(238) = 5.7093d0
  Mcase65TAY_V_density(239) = 5.7071d0
  Mcase65TAY_V_density(240) = 5.7049d0
  Mcase65TAY_V_density(241) = 5.7027d0
  Mcase65TAY_V_density(242) = 5.7004d0
  Mcase65TAY_V_density(243) = 5.6982d0
  Mcase65TAY_V_density(244) = 5.6959d0
  Mcase65TAY_V_density(245) = 5.6937d0
  Mcase65TAY_V_density(246) = 5.6914d0
  Mcase65TAY_V_density(247) = 5.6891d0
  Mcase65TAY_V_density(248) = 5.6868d0
  Mcase65TAY_V_density(249) = 5.6845d0
  Mcase65TAY_V_density(250) = 5.6821d0
  Mcase65TAY_V_density(251) = 5.6798d0
  Mcase65TAY_V_density(252) = 5.6775d0
  Mcase65TAY_V_density(253) = 5.6751d0
  Mcase65TAY_V_density(254) = 5.6727d0
  Mcase65TAY_V_density(255) = 5.6704d0
  Mcase65TAY_V_density(256) = 5.668d0
  Mcase65TAY_V_density(257) = 5.6656d0
  Mcase65TAY_V_density(258) = 5.6632d0
  Mcase65TAY_V_density(259) = 5.6607d0
  Mcase65TAY_V_density(260) = 5.6583d0
  Mcase65TAY_V_density(261) = 5.6559d0
  Mcase65TAY_V_density(262) = 5.6534d0
  Mcase65TAY_V_density(263) = 5.651d0
  Mcase65TAY_V_density(264) = 5.6485d0
  Mcase65TAY_V_density(265) = 5.646d0
  Mcase65TAY_V_density(266) = 5.6435d0
  Mcase65TAY_V_density(267) = 5.641d0
  Mcase65TAY_V_density(268) = 5.6385d0
  Mcase65TAY_V_density(269) = 5.636d0
  Mcase65TAY_V_density(270) = 5.6334d0
  Mcase65TAY_V_density(271) = 5.6309d0
  Mcase65TAY_V_density(272) = 5.6283d0
  Mcase65TAY_V_density(273) = 5.6257d0
  Mcase65TAY_V_density(274) = 5.6232d0
  Mcase65TAY_V_density(275) = 5.6206d0
  Mcase65TAY_V_density(276) = 5.618d0
  Mcase65TAY_V_density(277) = 5.6154d0
  Mcase65TAY_V_density(278) = 5.6127d0
  Mcase65TAY_V_density(279) = 5.6101d0
  Mcase65TAY_V_density(280) = 5.6074d0
  Mcase65TAY_V_density(281) = 5.6048d0
  Mcase65TAY_V_density(282) = 5.6021d0
  Mcase65TAY_V_density(283) = 5.5994d0
  Mcase65TAY_V_density(284) = 5.5968d0
  Mcase65TAY_V_density(285) = 5.5941d0
  Mcase65TAY_V_density(286) = 5.5913d0
  Mcase65TAY_V_density(287) = 5.5886d0
  Mcase65TAY_V_density(288) = 5.5859d0
  Mcase65TAY_V_density(289) = 5.5831d0
  Mcase65TAY_V_density(290) = 5.5804d0
  Mcase65TAY_V_density(291) = 5.5776d0
  Mcase65TAY_V_density(292) = 5.5748d0
  Mcase65TAY_V_density(293) = 5.572d0
  Mcase65TAY_V_density(294) = 5.5692d0
  Mcase65TAY_V_density(295) = 5.5664d0
  Mcase65TAY_V_density(296) = 5.5636d0
  Mcase65TAY_V_density(297) = 5.5608d0
  Mcase65TAY_V_density(298) = 5.5579d0
  Mcase65TAY_V_density(299) = 5.555d0
  Mcase65TAY_V_density(300) = 5.5522d0
  Mcase65TAY_V_density(301) = 5.5493d0
  Mcase65TAY_V_density(302) = 5.5464d0
  Mcase65TAY_V_density(303) = 5.5435d0
  Mcase65TAY_V_density(304) = 5.5406d0
  Mcase65TAY_V_density(305) = 5.5376d0
  Mcase65TAY_V_density(306) = 5.5347d0
  Mcase65TAY_V_density(307) = 5.5318d0
  Mcase65TAY_V_density(308) = 5.5288d0
  Mcase65TAY_V_density(309) = 5.5258d0
  Mcase65TAY_V_density(310) = 5.5228d0
  Mcase65TAY_V_density(311) = 5.5198d0
  Mcase65TAY_V_density(312) = 5.5168d0
  Mcase65TAY_V_density(313) = 5.5138d0
  Mcase65TAY_V_density(314) = 5.5108d0
  Mcase65TAY_V_density(315) = 5.5077d0
  Mcase65TAY_V_density(316) = 5.5046d0
  Mcase65TAY_V_density(317) = 5.5016d0
  Mcase65TAY_V_density(318) = 5.4985d0
  Mcase65TAY_V_density(319) = 5.4954d0
  Mcase65TAY_V_density(320) = 5.4923d0
  Mcase65TAY_V_density(321) = 5.4892d0
  Mcase65TAY_V_density(322) = 5.4861d0
  Mcase65TAY_V_density(323) = 5.4829d0
  Mcase65TAY_V_density(324) = 5.4797d0
  Mcase65TAY_V_density(325) = 5.4766d0
  Mcase65TAY_V_density(326) = 5.4734d0
  Mcase65TAY_V_density(327) = 5.4702d0
  Mcase65TAY_V_density(328) = 5.467d0
  Mcase65TAY_V_density(329) = 5.4638d0
  Mcase65TAY_V_density(330) = 5.4606d0
  Mcase65TAY_V_density(331) = 5.4573d0
  Mcase65TAY_V_density(332) = 5.4541d0
  Mcase65TAY_V_density(333) = 5.4508d0
  Mcase65TAY_V_density(334) = 5.4475d0
  Mcase65TAY_V_density(335) = 5.4442d0
  Mcase65TAY_V_density(336) = 5.4409d0
  Mcase65TAY_V_density(337) = 5.4376d0
  Mcase65TAY_V_density(338) = 5.4343d0
  Mcase65TAY_V_density(339) = 5.4309d0
  Mcase65TAY_V_density(340) = 5.4276d0
  Mcase65TAY_V_density(341) = 5.4242d0
  Mcase65TAY_V_density(342) = 5.4208d0
  Mcase65TAY_V_density(343) = 5.4174d0
  Mcase65TAY_V_density(344) = 5.4141d0
  Mcase65TAY_V_density(345) = 5.4106d0
  Mcase65TAY_V_density(346) = 5.4072d0
  Mcase65TAY_V_density(347) = 5.4037d0
  Mcase65TAY_V_density(348) = 5.4003d0
  Mcase65TAY_V_density(349) = 5.3968d0
  Mcase65TAY_V_density(350) = 5.3933d0
  Mcase65TAY_V_density(351) = 5.3899d0
  Mcase65TAY_V_density(352) = 5.3863d0
  Mcase65TAY_V_density(353) = 5.3828d0
  Mcase65TAY_V_density(354) = 5.3793d0
  Mcase65TAY_V_density(355) = 5.3757d0
  Mcase65TAY_V_density(356) = 5.3722d0
  Mcase65TAY_V_density(357) = 5.3686d0
  Mcase65TAY_V_density(358) = 5.365d0
  Mcase65TAY_V_density(359) = 5.3614d0
  Mcase65TAY_V_density(360) = 5.3578d0
  Mcase65TAY_V_density(361) = 5.3542d0
  Mcase65TAY_V_density(362) = 5.3505d0
  Mcase65TAY_V_density(363) = 5.3469d0
  Mcase65TAY_V_density(364) = 5.3432d0
  Mcase65TAY_V_density(365) = 5.3395d0
  Mcase65TAY_V_density(366) = 5.3358d0
  Mcase65TAY_V_density(367) = 5.3321d0
  Mcase65TAY_V_density(368) = 5.3284d0
  Mcase65TAY_V_density(369) = 5.3246d0
  Mcase65TAY_V_density(370) = 5.3246d0 ! CMB outer core
  Mcase65TAY_V_density(371) = 4.0417d0 ! CMB mantle
  Mcase65TAY_V_density(372) = 4.0409d0
  Mcase65TAY_V_density(373) = 4.04d0
  Mcase65TAY_V_density(374) = 4.0392d0
  Mcase65TAY_V_density(375) = 4.0383d0
  Mcase65TAY_V_density(376) = 4.0374d0
  Mcase65TAY_V_density(377) = 4.0366d0
  Mcase65TAY_V_density(378) = 4.0357d0
  Mcase65TAY_V_density(379) = 4.0343d0
  Mcase65TAY_V_density(380) = 4.0327d0
  Mcase65TAY_V_density(381) = 4.0311d0
  Mcase65TAY_V_density(382) = 4.0294d0
  Mcase65TAY_V_density(383) = 4.0278d0
  Mcase65TAY_V_density(384) = 4.0259d0
  Mcase65TAY_V_density(385) = 4.0239d0
  Mcase65TAY_V_density(386) = 4.0218d0
  Mcase65TAY_V_density(387) = 4.0198d0
  Mcase65TAY_V_density(388) = 4.0178d0
  Mcase65TAY_V_density(389) = 4.0153d0
  Mcase65TAY_V_density(390) = 4.0128d0
  Mcase65TAY_V_density(391) = 4.0102d0
  Mcase65TAY_V_density(392) = 4.0077d0
  Mcase65TAY_V_density(393) = 4.0051d0
  Mcase65TAY_V_density(394) = 4.0022d0
  Mcase65TAY_V_density(395) = 3.9993d0
  Mcase65TAY_V_density(396) = 3.9964d0
  Mcase65TAY_V_density(397) = 3.9935d0
  Mcase65TAY_V_density(398) = 3.9906d0
  Mcase65TAY_V_density(399) = 3.9878d0
  Mcase65TAY_V_density(400) = 3.985d0
  Mcase65TAY_V_density(401) = 3.9822d0
  Mcase65TAY_V_density(402) = 3.9795d0
  Mcase65TAY_V_density(403) = 3.9767d0
  Mcase65TAY_V_density(404) = 3.9739d0
  Mcase65TAY_V_density(405) = 3.9711d0
  Mcase65TAY_V_density(406) = 3.9684d0
  Mcase65TAY_V_density(407) = 3.9656d0
  Mcase65TAY_V_density(408) = 3.9629d0
  Mcase65TAY_V_density(409) = 3.9601d0
  Mcase65TAY_V_density(410) = 3.9573d0
  Mcase65TAY_V_density(411) = 3.9546d0
  Mcase65TAY_V_density(412) = 3.9518d0
  Mcase65TAY_V_density(413) = 3.9492d0
  Mcase65TAY_V_density(414) = 3.9469d0
  Mcase65TAY_V_density(415) = 3.9446d0
  Mcase65TAY_V_density(416) = 3.9423d0
  Mcase65TAY_V_density(417) = 3.9401d0
  Mcase65TAY_V_density(418) = 3.938d0
  Mcase65TAY_V_density(419) = 3.9362d0
  Mcase65TAY_V_density(420) = 3.9344d0
  Mcase65TAY_V_density(421) = 3.9327d0
  Mcase65TAY_V_density(422) = 3.9309d0
  Mcase65TAY_V_density(423) = 3.9292d0
  Mcase65TAY_V_density(424) = 3.9277d0
  Mcase65TAY_V_density(425) = 3.9261d0
  Mcase65TAY_V_density(426) = 3.9245d0
  Mcase65TAY_V_density(427) = 3.9229d0
  Mcase65TAY_V_density(428) = 3.9214d0
  Mcase65TAY_V_density(429) = 3.9198d0
  Mcase65TAY_V_density(430) = 3.9182d0
  Mcase65TAY_V_density(431) = 3.9166d0
  Mcase65TAY_V_density(432) = 3.915d0
  Mcase65TAY_V_density(433) = 3.9134d0
  Mcase65TAY_V_density(434) = 3.9118d0
  Mcase65TAY_V_density(435) = 3.9102d0
  Mcase65TAY_V_density(436) = 3.9086d0
  Mcase65TAY_V_density(437) = 3.907d0
  Mcase65TAY_V_density(438) = 3.9054d0
  Mcase65TAY_V_density(439) = 3.9038d0
  Mcase65TAY_V_density(440) = 3.9023d0
  Mcase65TAY_V_density(441) = 3.9007d0
  Mcase65TAY_V_density(442) = 3.8991d0
  Mcase65TAY_V_density(443) = 3.8976d0
  Mcase65TAY_V_density(444) = 3.896d0
  Mcase65TAY_V_density(445) = 3.8945d0
  Mcase65TAY_V_density(446) = 3.8929d0
  Mcase65TAY_V_density(447) = 3.8914d0
  Mcase65TAY_V_density(448) = 3.8897d0
  Mcase65TAY_V_density(449) = 3.888d0
  Mcase65TAY_V_density(450) = 3.8863d0
  Mcase65TAY_V_density(451) = 3.8847d0
  Mcase65TAY_V_density(452) = 3.883d0
  Mcase65TAY_V_density(453) = 3.8813d0
  Mcase65TAY_V_density(454) = 3.8796d0
  Mcase65TAY_V_density(455) = 3.8779d0
  Mcase65TAY_V_density(456) = 3.8762d0
  Mcase65TAY_V_density(457) = 3.8746d0
  Mcase65TAY_V_density(458) = 3.8719d0
  Mcase65TAY_V_density(459) = 3.8693d0
  Mcase65TAY_V_density(460) = 3.8666d0
  Mcase65TAY_V_density(461) = 3.8639d0
  Mcase65TAY_V_density(462) = 3.8609d0
  Mcase65TAY_V_density(463) = 3.8541d0
  Mcase65TAY_V_density(464) = 3.8472d0
  Mcase65TAY_V_density(465) = 3.8403d0
  Mcase65TAY_V_density(466) = 3.8334d0
  Mcase65TAY_V_density(467) = 3.8261d0
  Mcase65TAY_V_density(468) = 3.8163d0
  Mcase65TAY_V_density(469) = 3.8066d0
  Mcase65TAY_V_density(470) = 3.7968d0
  Mcase65TAY_V_density(471) = 3.7871d0
  Mcase65TAY_V_density(472) = 3.7777d0
  Mcase65TAY_V_density(473) = 3.7694d0
  Mcase65TAY_V_density(474) = 3.7611d0
  Mcase65TAY_V_density(475) = 3.7528d0
  Mcase65TAY_V_density(476) = 3.7445d0
  Mcase65TAY_V_density(477) = 3.7372d0
  Mcase65TAY_V_density(478) = 3.7322d0
  Mcase65TAY_V_density(479) = 3.7272d0
  Mcase65TAY_V_density(480) = 3.7221d0
  Mcase65TAY_V_density(481) = 3.7171d0
  Mcase65TAY_V_density(482) = 3.7131d0
  Mcase65TAY_V_density(483) = 3.7105d0
  Mcase65TAY_V_density(484) = 3.7079d0
  Mcase65TAY_V_density(485) = 3.7054d0
  Mcase65TAY_V_density(486) = 3.7028d0
  Mcase65TAY_V_density(487) = 3.7006d0
  Mcase65TAY_V_density(488) = 3.6988d0
  Mcase65TAY_V_density(489) = 3.697d0
  Mcase65TAY_V_density(490) = 3.6952d0
  Mcase65TAY_V_density(491) = 3.6933d0
  Mcase65TAY_V_density(492) = 3.6916d0
  Mcase65TAY_V_density(493) = 3.69d0
  Mcase65TAY_V_density(494) = 3.6883d0
  Mcase65TAY_V_density(495) = 3.6867d0
  Mcase65TAY_V_density(496) = 3.685d0
  Mcase65TAY_V_density(497) = 3.6834d0
  Mcase65TAY_V_density(498) = 3.6818d0
  Mcase65TAY_V_density(499) = 3.6802d0
  Mcase65TAY_V_density(500) = 3.6786d0
  Mcase65TAY_V_density(501) = 3.6771d0
  Mcase65TAY_V_density(502) = 3.6755d0
  Mcase65TAY_V_density(503) = 3.6739d0
  Mcase65TAY_V_density(504) = 3.6723d0
  Mcase65TAY_V_density(505) = 3.6707d0
  Mcase65TAY_V_density(506) = 3.6691d0
  Mcase65TAY_V_density(507) = 3.6675d0
  Mcase65TAY_V_density(508) = 3.6658d0
  Mcase65TAY_V_density(509) = 3.6642d0
  Mcase65TAY_V_density(510) = 3.6626d0
  Mcase65TAY_V_density(511) = 3.661d0
  Mcase65TAY_V_density(512) = 3.6595d0
  Mcase65TAY_V_density(513) = 3.658d0
  Mcase65TAY_V_density(514) = 3.6564d0
  Mcase65TAY_V_density(515) = 3.6549d0
  Mcase65TAY_V_density(516) = 3.6534d0
  Mcase65TAY_V_density(517) = 3.6519d0
  Mcase65TAY_V_density(518) = 3.6504d0
  Mcase65TAY_V_density(519) = 3.649d0
  Mcase65TAY_V_density(520) = 3.6475d0
  Mcase65TAY_V_density(521) = 3.646d0
  Mcase65TAY_V_density(522) = 3.6446d0
  Mcase65TAY_V_density(523) = 3.6432d0
  Mcase65TAY_V_density(524) = 3.6418d0
  Mcase65TAY_V_density(525) = 3.6404d0
  Mcase65TAY_V_density(526) = 3.639d0
  Mcase65TAY_V_density(527) = 3.6376d0
  Mcase65TAY_V_density(528) = 3.6362d0
  Mcase65TAY_V_density(529) = 3.6348d0
  Mcase65TAY_V_density(530) = 3.6334d0
  Mcase65TAY_V_density(531) = 3.632d0
  Mcase65TAY_V_density(532) = 3.6305d0
  Mcase65TAY_V_density(533) = 3.6291d0
  Mcase65TAY_V_density(534) = 3.6276d0
  Mcase65TAY_V_density(535) = 3.6262d0
  Mcase65TAY_V_density(536) = 3.6248d0
  Mcase65TAY_V_density(537) = 3.6234d0
  Mcase65TAY_V_density(538) = 3.6221d0
  Mcase65TAY_V_density(539) = 3.6207d0
  Mcase65TAY_V_density(540) = 3.6194d0
  Mcase65TAY_V_density(541) = 3.618d0
  Mcase65TAY_V_density(542) = 3.6166d0
  Mcase65TAY_V_density(543) = 3.6152d0
  Mcase65TAY_V_density(544) = 3.6138d0
  Mcase65TAY_V_density(545) = 3.6124d0
  Mcase65TAY_V_density(546) = 3.611d0
  Mcase65TAY_V_density(547) = 3.6096d0
  Mcase65TAY_V_density(548) = 3.6082d0
  Mcase65TAY_V_density(549) = 3.6068d0
  Mcase65TAY_V_density(550) = 3.6054d0
  Mcase65TAY_V_density(551) = 3.6041d0
  Mcase65TAY_V_density(552) = 3.6028d0
  Mcase65TAY_V_density(553) = 3.6015d0
  Mcase65TAY_V_density(554) = 3.6002d0
  Mcase65TAY_V_density(555) = 3.5989d0
  Mcase65TAY_V_density(556) = 3.5977d0
  Mcase65TAY_V_density(557) = 3.5965d0
  Mcase65TAY_V_density(558) = 3.5953d0
  Mcase65TAY_V_density(559) = 3.594d0
  Mcase65TAY_V_density(560) = 3.5928d0
  Mcase65TAY_V_density(561) = 3.5916d0
  Mcase65TAY_V_density(562) = 3.5904d0
  Mcase65TAY_V_density(563) = 3.5892d0
  Mcase65TAY_V_density(564) = 3.588d0
  Mcase65TAY_V_density(565) = 3.5868d0
  Mcase65TAY_V_density(566) = 3.5856d0
  Mcase65TAY_V_density(567) = 3.5844d0
  Mcase65TAY_V_density(568) = 3.5833d0
  Mcase65TAY_V_density(569) = 3.5821d0
  Mcase65TAY_V_density(570) = 3.5809d0
  Mcase65TAY_V_density(571) = 3.5798d0
  Mcase65TAY_V_density(572) = 3.5786d0
  Mcase65TAY_V_density(573) = 3.5775d0
  Mcase65TAY_V_density(574) = 3.5763d0
  Mcase65TAY_V_density(575) = 3.5752d0
  Mcase65TAY_V_density(576) = 3.5741d0
  Mcase65TAY_V_density(577) = 3.5729d0
  Mcase65TAY_V_density(578) = 3.5718d0
  Mcase65TAY_V_density(579) = 3.5707d0
  Mcase65TAY_V_density(580) = 3.5696d0
  Mcase65TAY_V_density(581) = 3.5684d0
  Mcase65TAY_V_density(582) = 3.5673d0
  Mcase65TAY_V_density(583) = 3.5662d0
  Mcase65TAY_V_density(584) = 3.5651d0
  Mcase65TAY_V_density(585) = 3.564d0
  Mcase65TAY_V_density(586) = 3.563d0
  Mcase65TAY_V_density(587) = 3.5619d0
  Mcase65TAY_V_density(588) = 3.5608d0
  Mcase65TAY_V_density(589) = 3.5598d0
  Mcase65TAY_V_density(590) = 3.5587d0
  Mcase65TAY_V_density(591) = 3.5576d0
  Mcase65TAY_V_density(592) = 3.5565d0
  Mcase65TAY_V_density(593) = 3.5555d0
  Mcase65TAY_V_density(594) = 3.5544d0
  Mcase65TAY_V_density(595) = 3.5533d0
  Mcase65TAY_V_density(596) = 3.5523d0
  Mcase65TAY_V_density(597) = 3.5514d0
  Mcase65TAY_V_density(598) = 3.5504d0
  Mcase65TAY_V_density(599) = 3.5494d0
  Mcase65TAY_V_density(600) = 3.5484d0
  Mcase65TAY_V_density(601) = 3.5473d0
  Mcase65TAY_V_density(602) = 3.5463d0
  Mcase65TAY_V_density(603) = 3.5452d0
  Mcase65TAY_V_density(604) = 3.5442d0
  Mcase65TAY_V_density(605) = 3.5432d0
  Mcase65TAY_V_density(606) = 3.5422d0
  Mcase65TAY_V_density(607) = 3.5413d0
  Mcase65TAY_V_density(608) = 3.5403d0
  Mcase65TAY_V_density(609) = 3.5393d0
  Mcase65TAY_V_density(610) = 3.5384d0
  Mcase65TAY_V_density(611) = 3.5374d0
  Mcase65TAY_V_density(612) = 3.5364d0
  Mcase65TAY_V_density(613) = 3.5355d0
  Mcase65TAY_V_density(614) = 3.5345d0
  Mcase65TAY_V_density(615) = 3.5335d0
  Mcase65TAY_V_density(616) = 3.5326d0
  Mcase65TAY_V_density(617) = 3.5316d0
  Mcase65TAY_V_density(618) = 3.5307d0
  Mcase65TAY_V_density(619) = 3.5298d0
  Mcase65TAY_V_density(620) = 3.5288d0
  Mcase65TAY_V_density(621) = 3.5279d0
  Mcase65TAY_V_density(622) = 3.5269d0
  Mcase65TAY_V_density(623) = 3.526d0
  Mcase65TAY_V_density(624) = 3.5251d0
  Mcase65TAY_V_density(625) = 3.5242d0
  Mcase65TAY_V_density(626) = 3.5233d0
  Mcase65TAY_V_density(627) = 3.5224d0
  Mcase65TAY_V_density(628) = 3.5215d0
  Mcase65TAY_V_density(629) = 3.5207d0
  Mcase65TAY_V_density(630) = 3.5197d0
  Mcase65TAY_V_density(631) = 3.5188d0
  Mcase65TAY_V_density(632) = 3.5178d0
  Mcase65TAY_V_density(633) = 3.5169d0
  Mcase65TAY_V_density(634) = 3.5159d0
  Mcase65TAY_V_density(635) = 3.5151d0
  Mcase65TAY_V_density(636) = 3.5143d0
  Mcase65TAY_V_density(637) = 3.5134d0
  Mcase65TAY_V_density(638) = 3.5126d0
  Mcase65TAY_V_density(639) = 3.5117d0
  Mcase65TAY_V_density(640) = 3.5107d0
  Mcase65TAY_V_density(641) = 3.5096d0
  Mcase65TAY_V_density(642) = 3.5086d0
  Mcase65TAY_V_density(643) = 3.5075d0
  Mcase65TAY_V_density(644) = 3.5063d0
  Mcase65TAY_V_density(645) = 3.5047d0
  Mcase65TAY_V_density(646) = 3.5032d0
  Mcase65TAY_V_density(647) = 3.5016d0
  Mcase65TAY_V_density(648) = 3.5d0
  Mcase65TAY_V_density(649) = 3.4979d0
  Mcase65TAY_V_density(650) = 3.4946d0
  Mcase65TAY_V_density(651) = 3.4913d0
  Mcase65TAY_V_density(652) = 3.488d0
  Mcase65TAY_V_density(653) = 3.4847d0
  Mcase65TAY_V_density(654) = 3.4763d0
  Mcase65TAY_V_density(655) = 3.4606d0
  Mcase65TAY_V_density(656) = 3.4449d0
  Mcase65TAY_V_density(657) = 3.4291d0
  Mcase65TAY_V_density(658) = 3.4134d0
  Mcase65TAY_V_density(659) = 3.3945d0
  Mcase65TAY_V_density(660) = 3.3726d0
  Mcase65TAY_V_density(661) = 3.3506d0
  Mcase65TAY_V_density(662) = 3.3337d0
  Mcase65TAY_V_density(663) = 3.3218d0
  Mcase65TAY_V_density(664) = 3.3099d0
  Mcase65TAY_V_density(665) = 3.2974d0
  Mcase65TAY_V_density(666) = 3.2841d0
  Mcase65TAY_V_density(667) = 3.2708d0
  Mcase65TAY_V_density(668) = 3.2533d0
  Mcase65TAY_V_density(669) = 3.2313d0
  Mcase65TAY_V_density(670) = 3.2094d0
  Mcase65TAY_V_density(671) = 3.1933d0
  Mcase65TAY_V_density(672) = 3.183d0
  Mcase65TAY_V_density(673) = 3.1727d0
  Mcase65TAY_V_density(674) = 3.1654d0
  Mcase65TAY_V_density(675) = 3.1612d0
  Mcase65TAY_V_density(676) = 3.157d0
  Mcase65TAY_V_density(677) = 3.1541d0
  Mcase65TAY_V_density(678) = 3.1523d0
  Mcase65TAY_V_density(679) = 3.1505d0
  Mcase65TAY_V_density(680) = 3.1497d0
  Mcase65TAY_V_density(681) = 3.1499d0

  Mcase65TAY_V_vp(1) = 6.5896d0
  Mcase65TAY_V_vp(2) = 6.5896d0
  Mcase65TAY_V_vp(3) = 6.5896d0
  Mcase65TAY_V_vp(4) = 6.5896d0
  Mcase65TAY_V_vp(5) = 6.5896d0
  Mcase65TAY_V_vp(6) = 6.5896d0
  Mcase65TAY_V_vp(7) = 6.5896d0
  Mcase65TAY_V_vp(8) = 6.5896d0
  Mcase65TAY_V_vp(9) = 6.5896d0
  Mcase65TAY_V_vp(10) = 6.5896d0
  Mcase65TAY_V_vp(11) = 6.5896d0
  Mcase65TAY_V_vp(12) = 6.5896d0
  Mcase65TAY_V_vp(13) = 6.5896d0
  Mcase65TAY_V_vp(14) = 6.5896d0
  Mcase65TAY_V_vp(15) = 6.5896d0
  Mcase65TAY_V_vp(16) = 6.5896d0
  Mcase65TAY_V_vp(17) = 6.5896d0
  Mcase65TAY_V_vp(18) = 6.5896d0
  Mcase65TAY_V_vp(19) = 6.5896d0
  Mcase65TAY_V_vp(20) = 6.5896d0
  Mcase65TAY_V_vp(21) = 6.5896d0
  Mcase65TAY_V_vp(22) = 6.5896d0
  Mcase65TAY_V_vp(23) = 6.5896d0
  Mcase65TAY_V_vp(24) = 6.5896d0
  Mcase65TAY_V_vp(25) = 6.5896d0
  Mcase65TAY_V_vp(26) = 6.5896d0
  Mcase65TAY_V_vp(27) = 6.5896d0
  Mcase65TAY_V_vp(28) = 6.5896d0
  Mcase65TAY_V_vp(29) = 6.5896d0
  Mcase65TAY_V_vp(30) = 6.5896d0
  Mcase65TAY_V_vp(31) = 6.5896d0
  Mcase65TAY_V_vp(32) = 6.5896d0
  Mcase65TAY_V_vp(33) = 6.5896d0
  Mcase65TAY_V_vp(34) = 6.5896d0
  Mcase65TAY_V_vp(35) = 6.5896d0
  Mcase65TAY_V_vp(36) = 6.5896d0
  Mcase65TAY_V_vp(37) = 6.5896d0
  Mcase65TAY_V_vp(38) = 6.5896d0
  Mcase65TAY_V_vp(39) = 6.5896d0
  Mcase65TAY_V_vp(40) = 6.5896d0
  Mcase65TAY_V_vp(41) = 6.5896d0
  Mcase65TAY_V_vp(42) = 6.5896d0
  Mcase65TAY_V_vp(43) = 6.5896d0
  Mcase65TAY_V_vp(44) = 6.5896d0
  Mcase65TAY_V_vp(45) = 6.5896d0
  Mcase65TAY_V_vp(46) = 6.5896d0
  Mcase65TAY_V_vp(47) = 6.5896d0
  Mcase65TAY_V_vp(48) = 6.5896d0
  Mcase65TAY_V_vp(49) = 6.5896d0
  Mcase65TAY_V_vp(50) = 6.5896d0
  Mcase65TAY_V_vp(51) = 6.5896d0
  Mcase65TAY_V_vp(52) = 6.5896d0
  Mcase65TAY_V_vp(53) = 6.5896d0
  Mcase65TAY_V_vp(54) = 6.5896d0
  Mcase65TAY_V_vp(55) = 6.5896d0
  Mcase65TAY_V_vp(56) = 6.5896d0
  Mcase65TAY_V_vp(57) = 6.5896d0
  Mcase65TAY_V_vp(58) = 6.5896d0
  Mcase65TAY_V_vp(59) = 6.5896d0
  Mcase65TAY_V_vp(60) = 6.5896d0
  Mcase65TAY_V_vp(61) = 6.5896d0
  Mcase65TAY_V_vp(62) = 6.5896d0
  Mcase65TAY_V_vp(63) = 6.5896d0
  Mcase65TAY_V_vp(64) = 6.5896d0
  Mcase65TAY_V_vp(65) = 6.5896d0
  Mcase65TAY_V_vp(66) = 6.5896d0
  Mcase65TAY_V_vp(67) = 6.5896d0
  Mcase65TAY_V_vp(68) = 6.5896d0
  Mcase65TAY_V_vp(69) = 6.5896d0
  Mcase65TAY_V_vp(70) = 6.5896d0
  Mcase65TAY_V_vp(71) = 6.5896d0
  Mcase65TAY_V_vp(72) = 6.5896d0
  Mcase65TAY_V_vp(73) = 6.5896d0
  Mcase65TAY_V_vp(74) = 6.5896d0
  Mcase65TAY_V_vp(75) = 6.5896d0
  Mcase65TAY_V_vp(76) = 6.5896d0
  Mcase65TAY_V_vp(77) = 6.5896d0
  Mcase65TAY_V_vp(78) = 6.5896d0
  Mcase65TAY_V_vp(79) = 6.5896d0
  Mcase65TAY_V_vp(80) = 6.5896d0
  Mcase65TAY_V_vp(81) = 6.5896d0
  Mcase65TAY_V_vp(82) = 6.5896d0
  Mcase65TAY_V_vp(83) = 6.5896d0
  Mcase65TAY_V_vp(84) = 6.5896d0
  Mcase65TAY_V_vp(85) = 6.5896d0
  Mcase65TAY_V_vp(86) = 6.5896d0
  Mcase65TAY_V_vp(87) = 6.5896d0
  Mcase65TAY_V_vp(88) = 6.5896d0
  Mcase65TAY_V_vp(89) = 6.5896d0
  Mcase65TAY_V_vp(90) = 6.5896d0
  Mcase65TAY_V_vp(91) = 6.5896d0
  Mcase65TAY_V_vp(92) = 6.5896d0
  Mcase65TAY_V_vp(93) = 6.5896d0
  Mcase65TAY_V_vp(94) = 6.5896d0
  Mcase65TAY_V_vp(95) = 6.5896d0
  Mcase65TAY_V_vp(96) = 6.5896d0
  Mcase65TAY_V_vp(97) = 6.5896d0
  Mcase65TAY_V_vp(98) = 6.5896d0
  Mcase65TAY_V_vp(99) = 6.5896d0
  Mcase65TAY_V_vp(100) = 6.5896d0
  Mcase65TAY_V_vp(101) = 6.5896d0
  Mcase65TAY_V_vp(102) = 6.5896d0
  Mcase65TAY_V_vp(103) = 6.5896d0
  Mcase65TAY_V_vp(104) = 6.5896d0 ! ICB inner core
  Mcase65TAY_V_vp(105) = 5.2641d0 ! ICB outer core
  Mcase65TAY_V_vp(106) = 5.2631d0
  Mcase65TAY_V_vp(107) = 5.262d0
  Mcase65TAY_V_vp(108) = 5.2609d0
  Mcase65TAY_V_vp(109) = 5.2598d0
  Mcase65TAY_V_vp(110) = 5.2587d0
  Mcase65TAY_V_vp(111) = 5.2576d0
  Mcase65TAY_V_vp(112) = 5.2565d0
  Mcase65TAY_V_vp(113) = 5.2554d0
  Mcase65TAY_V_vp(114) = 5.2542d0
  Mcase65TAY_V_vp(115) = 5.2531d0
  Mcase65TAY_V_vp(116) = 5.252d0
  Mcase65TAY_V_vp(117) = 5.2508d0
  Mcase65TAY_V_vp(118) = 5.2496d0
  Mcase65TAY_V_vp(119) = 5.2484d0
  Mcase65TAY_V_vp(120) = 5.2472d0
  Mcase65TAY_V_vp(121) = 5.246d0
  Mcase65TAY_V_vp(122) = 5.2448d0
  Mcase65TAY_V_vp(123) = 5.2435d0
  Mcase65TAY_V_vp(124) = 5.2423d0
  Mcase65TAY_V_vp(125) = 5.241d0
  Mcase65TAY_V_vp(126) = 5.2398d0
  Mcase65TAY_V_vp(127) = 5.2385d0
  Mcase65TAY_V_vp(128) = 5.2372d0
  Mcase65TAY_V_vp(129) = 5.2359d0
  Mcase65TAY_V_vp(130) = 5.2346d0
  Mcase65TAY_V_vp(131) = 5.2333d0
  Mcase65TAY_V_vp(132) = 5.232d0
  Mcase65TAY_V_vp(133) = 5.2306d0
  Mcase65TAY_V_vp(134) = 5.2293d0
  Mcase65TAY_V_vp(135) = 5.2279d0
  Mcase65TAY_V_vp(136) = 5.2265d0
  Mcase65TAY_V_vp(137) = 5.2251d0
  Mcase65TAY_V_vp(138) = 5.2238d0
  Mcase65TAY_V_vp(139) = 5.2224d0
  Mcase65TAY_V_vp(140) = 5.2209d0
  Mcase65TAY_V_vp(141) = 5.2195d0
  Mcase65TAY_V_vp(142) = 5.2181d0
  Mcase65TAY_V_vp(143) = 5.2166d0
  Mcase65TAY_V_vp(144) = 5.2152d0
  Mcase65TAY_V_vp(145) = 5.2137d0
  Mcase65TAY_V_vp(146) = 5.2122d0
  Mcase65TAY_V_vp(147) = 5.2107d0
  Mcase65TAY_V_vp(148) = 5.2092d0
  Mcase65TAY_V_vp(149) = 5.2077d0
  Mcase65TAY_V_vp(150) = 5.2062d0
  Mcase65TAY_V_vp(151) = 5.2046d0
  Mcase65TAY_V_vp(152) = 5.2031d0
  Mcase65TAY_V_vp(153) = 5.2015d0
  Mcase65TAY_V_vp(154) = 5.2d0
  Mcase65TAY_V_vp(155) = 5.1984d0
  Mcase65TAY_V_vp(156) = 5.1968d0
  Mcase65TAY_V_vp(157) = 5.1952d0
  Mcase65TAY_V_vp(158) = 5.1936d0
  Mcase65TAY_V_vp(159) = 5.192d0
  Mcase65TAY_V_vp(160) = 5.1903d0
  Mcase65TAY_V_vp(161) = 5.1887d0
  Mcase65TAY_V_vp(162) = 5.187d0
  Mcase65TAY_V_vp(163) = 5.1853d0
  Mcase65TAY_V_vp(164) = 5.1837d0
  Mcase65TAY_V_vp(165) = 5.182d0
  Mcase65TAY_V_vp(166) = 5.1803d0
  Mcase65TAY_V_vp(167) = 5.1786d0
  Mcase65TAY_V_vp(168) = 5.1768d0
  Mcase65TAY_V_vp(169) = 5.1751d0
  Mcase65TAY_V_vp(170) = 5.1734d0
  Mcase65TAY_V_vp(171) = 5.1716d0
  Mcase65TAY_V_vp(172) = 5.1698d0
  Mcase65TAY_V_vp(173) = 5.168d0
  Mcase65TAY_V_vp(174) = 5.1663d0
  Mcase65TAY_V_vp(175) = 5.1645d0
  Mcase65TAY_V_vp(176) = 5.1626d0
  Mcase65TAY_V_vp(177) = 5.1608d0
  Mcase65TAY_V_vp(178) = 5.159d0
  Mcase65TAY_V_vp(179) = 5.1571d0
  Mcase65TAY_V_vp(180) = 5.1553d0
  Mcase65TAY_V_vp(181) = 5.1534d0
  Mcase65TAY_V_vp(182) = 5.1515d0
  Mcase65TAY_V_vp(183) = 5.1496d0
  Mcase65TAY_V_vp(184) = 5.1477d0
  Mcase65TAY_V_vp(185) = 5.1458d0
  Mcase65TAY_V_vp(186) = 5.1439d0
  Mcase65TAY_V_vp(187) = 5.1419d0
  Mcase65TAY_V_vp(188) = 5.1399d0
  Mcase65TAY_V_vp(189) = 5.138d0
  Mcase65TAY_V_vp(190) = 5.136d0
  Mcase65TAY_V_vp(191) = 5.134d0
  Mcase65TAY_V_vp(192) = 5.132d0
  Mcase65TAY_V_vp(193) = 5.13d0
  Mcase65TAY_V_vp(194) = 5.128d0
  Mcase65TAY_V_vp(195) = 5.1259d0
  Mcase65TAY_V_vp(196) = 5.1239d0
  Mcase65TAY_V_vp(197) = 5.1219d0
  Mcase65TAY_V_vp(198) = 5.1198d0
  Mcase65TAY_V_vp(199) = 5.1177d0
  Mcase65TAY_V_vp(200) = 5.1156d0
  Mcase65TAY_V_vp(201) = 5.1135d0
  Mcase65TAY_V_vp(202) = 5.1114d0
  Mcase65TAY_V_vp(203) = 5.1092d0
  Mcase65TAY_V_vp(204) = 5.1071d0
  Mcase65TAY_V_vp(205) = 5.105d0
  Mcase65TAY_V_vp(206) = 5.1028d0
  Mcase65TAY_V_vp(207) = 5.1006d0
  Mcase65TAY_V_vp(208) = 5.0984d0
  Mcase65TAY_V_vp(209) = 5.0962d0
  Mcase65TAY_V_vp(210) = 5.094d0
  Mcase65TAY_V_vp(211) = 5.0918d0
  Mcase65TAY_V_vp(212) = 5.0895d0
  Mcase65TAY_V_vp(213) = 5.0873d0
  Mcase65TAY_V_vp(214) = 5.085d0
  Mcase65TAY_V_vp(215) = 5.0828d0
  Mcase65TAY_V_vp(216) = 5.0805d0
  Mcase65TAY_V_vp(217) = 5.0782d0
  Mcase65TAY_V_vp(218) = 5.0759d0
  Mcase65TAY_V_vp(219) = 5.0735d0
  Mcase65TAY_V_vp(220) = 5.0712d0
  Mcase65TAY_V_vp(221) = 5.0688d0
  Mcase65TAY_V_vp(222) = 5.0665d0
  Mcase65TAY_V_vp(223) = 5.0641d0
  Mcase65TAY_V_vp(224) = 5.0617d0
  Mcase65TAY_V_vp(225) = 5.0593d0
  Mcase65TAY_V_vp(226) = 5.0569d0
  Mcase65TAY_V_vp(227) = 5.0545d0
  Mcase65TAY_V_vp(228) = 5.0521d0
  Mcase65TAY_V_vp(229) = 5.0496d0
  Mcase65TAY_V_vp(230) = 5.0472d0
  Mcase65TAY_V_vp(231) = 5.0447d0
  Mcase65TAY_V_vp(232) = 5.0422d0
  Mcase65TAY_V_vp(233) = 5.0397d0
  Mcase65TAY_V_vp(234) = 5.0372d0
  Mcase65TAY_V_vp(235) = 5.0347d0
  Mcase65TAY_V_vp(236) = 5.0321d0
  Mcase65TAY_V_vp(237) = 5.0296d0
  Mcase65TAY_V_vp(238) = 5.027d0
  Mcase65TAY_V_vp(239) = 5.0244d0
  Mcase65TAY_V_vp(240) = 5.0219d0
  Mcase65TAY_V_vp(241) = 5.0193d0
  Mcase65TAY_V_vp(242) = 5.0167d0
  Mcase65TAY_V_vp(243) = 5.014d0
  Mcase65TAY_V_vp(244) = 5.0114d0
  Mcase65TAY_V_vp(245) = 5.0087d0
  Mcase65TAY_V_vp(246) = 5.0061d0
  Mcase65TAY_V_vp(247) = 5.0034d0
  Mcase65TAY_V_vp(248) = 5.0007d0
  Mcase65TAY_V_vp(249) = 4.998d0
  Mcase65TAY_V_vp(250) = 4.9953d0
  Mcase65TAY_V_vp(251) = 4.9926d0
  Mcase65TAY_V_vp(252) = 4.9898d0
  Mcase65TAY_V_vp(253) = 4.9871d0
  Mcase65TAY_V_vp(254) = 4.9843d0
  Mcase65TAY_V_vp(255) = 4.9815d0
  Mcase65TAY_V_vp(256) = 4.9787d0
  Mcase65TAY_V_vp(257) = 4.9759d0
  Mcase65TAY_V_vp(258) = 4.9731d0
  Mcase65TAY_V_vp(259) = 4.9702d0
  Mcase65TAY_V_vp(260) = 4.9674d0
  Mcase65TAY_V_vp(261) = 4.9645d0
  Mcase65TAY_V_vp(262) = 4.9616d0
  Mcase65TAY_V_vp(263) = 4.9588d0
  Mcase65TAY_V_vp(264) = 4.9559d0
  Mcase65TAY_V_vp(265) = 4.9529d0
  Mcase65TAY_V_vp(266) = 4.95d0
  Mcase65TAY_V_vp(267) = 4.9471d0
  Mcase65TAY_V_vp(268) = 4.9441d0
  Mcase65TAY_V_vp(269) = 4.9411d0
  Mcase65TAY_V_vp(270) = 4.9381d0
  Mcase65TAY_V_vp(271) = 4.9351d0
  Mcase65TAY_V_vp(272) = 4.9321d0
  Mcase65TAY_V_vp(273) = 4.9291d0
  Mcase65TAY_V_vp(274) = 4.9261d0
  Mcase65TAY_V_vp(275) = 4.923d0
  Mcase65TAY_V_vp(276) = 4.9199d0
  Mcase65TAY_V_vp(277) = 4.9168d0
  Mcase65TAY_V_vp(278) = 4.9138d0
  Mcase65TAY_V_vp(279) = 4.9106d0
  Mcase65TAY_V_vp(280) = 4.9075d0
  Mcase65TAY_V_vp(281) = 4.9044d0
  Mcase65TAY_V_vp(282) = 4.9012d0
  Mcase65TAY_V_vp(283) = 4.898d0
  Mcase65TAY_V_vp(284) = 4.8949d0
  Mcase65TAY_V_vp(285) = 4.8917d0
  Mcase65TAY_V_vp(286) = 4.8884d0
  Mcase65TAY_V_vp(287) = 4.8852d0
  Mcase65TAY_V_vp(288) = 4.882d0
  Mcase65TAY_V_vp(289) = 4.8787d0
  Mcase65TAY_V_vp(290) = 4.8754d0
  Mcase65TAY_V_vp(291) = 4.8722d0
  Mcase65TAY_V_vp(292) = 4.8689d0
  Mcase65TAY_V_vp(293) = 4.8656d0
  Mcase65TAY_V_vp(294) = 4.8622d0
  Mcase65TAY_V_vp(295) = 4.8589d0
  Mcase65TAY_V_vp(296) = 4.8555d0
  Mcase65TAY_V_vp(297) = 4.8521d0
  Mcase65TAY_V_vp(298) = 4.8487d0
  Mcase65TAY_V_vp(299) = 4.8453d0
  Mcase65TAY_V_vp(300) = 4.8419d0
  Mcase65TAY_V_vp(301) = 4.8385d0
  Mcase65TAY_V_vp(302) = 4.835d0
  Mcase65TAY_V_vp(303) = 4.8316d0
  Mcase65TAY_V_vp(304) = 4.8281d0
  Mcase65TAY_V_vp(305) = 4.8246d0
  Mcase65TAY_V_vp(306) = 4.8211d0
  Mcase65TAY_V_vp(307) = 4.8176d0
  Mcase65TAY_V_vp(308) = 4.814d0
  Mcase65TAY_V_vp(309) = 4.8105d0
  Mcase65TAY_V_vp(310) = 4.8069d0
  Mcase65TAY_V_vp(311) = 4.8033d0
  Mcase65TAY_V_vp(312) = 4.7997d0
  Mcase65TAY_V_vp(313) = 4.7961d0
  Mcase65TAY_V_vp(314) = 4.7924d0
  Mcase65TAY_V_vp(315) = 4.7888d0
  Mcase65TAY_V_vp(316) = 4.7851d0
  Mcase65TAY_V_vp(317) = 4.7814d0
  Mcase65TAY_V_vp(318) = 4.7778d0
  Mcase65TAY_V_vp(319) = 4.774d0
  Mcase65TAY_V_vp(320) = 4.7703d0
  Mcase65TAY_V_vp(321) = 4.7665d0
  Mcase65TAY_V_vp(322) = 4.7628d0
  Mcase65TAY_V_vp(323) = 4.759d0
  Mcase65TAY_V_vp(324) = 4.7552d0
  Mcase65TAY_V_vp(325) = 4.7514d0
  Mcase65TAY_V_vp(326) = 4.7476d0
  Mcase65TAY_V_vp(327) = 4.7437d0
  Mcase65TAY_V_vp(328) = 4.7399d0
  Mcase65TAY_V_vp(329) = 4.736d0
  Mcase65TAY_V_vp(330) = 4.7321d0
  Mcase65TAY_V_vp(331) = 4.7282d0
  Mcase65TAY_V_vp(332) = 4.7242d0
  Mcase65TAY_V_vp(333) = 4.7203d0
  Mcase65TAY_V_vp(334) = 4.7163d0
  Mcase65TAY_V_vp(335) = 4.7123d0
  Mcase65TAY_V_vp(336) = 4.7084d0
  Mcase65TAY_V_vp(337) = 4.7044d0
  Mcase65TAY_V_vp(338) = 4.7003d0
  Mcase65TAY_V_vp(339) = 4.6963d0
  Mcase65TAY_V_vp(340) = 4.6922d0
  Mcase65TAY_V_vp(341) = 4.6881d0
  Mcase65TAY_V_vp(342) = 4.684d0
  Mcase65TAY_V_vp(343) = 4.6799d0
  Mcase65TAY_V_vp(344) = 4.6758d0
  Mcase65TAY_V_vp(345) = 4.6716d0
  Mcase65TAY_V_vp(346) = 4.6674d0
  Mcase65TAY_V_vp(347) = 4.6632d0
  Mcase65TAY_V_vp(348) = 4.659d0
  Mcase65TAY_V_vp(349) = 4.6548d0
  Mcase65TAY_V_vp(350) = 4.6506d0
  Mcase65TAY_V_vp(351) = 4.6463d0
  Mcase65TAY_V_vp(352) = 4.642d0
  Mcase65TAY_V_vp(353) = 4.6377d0
  Mcase65TAY_V_vp(354) = 4.6334d0
  Mcase65TAY_V_vp(355) = 4.6291d0
  Mcase65TAY_V_vp(356) = 4.6247d0
  Mcase65TAY_V_vp(357) = 4.6203d0
  Mcase65TAY_V_vp(358) = 4.616d0
  Mcase65TAY_V_vp(359) = 4.6116d0
  Mcase65TAY_V_vp(360) = 4.6071d0
  Mcase65TAY_V_vp(361) = 4.6027d0
  Mcase65TAY_V_vp(362) = 4.5982d0
  Mcase65TAY_V_vp(363) = 4.5937d0
  Mcase65TAY_V_vp(364) = 4.5892d0
  Mcase65TAY_V_vp(365) = 4.5847d0
  Mcase65TAY_V_vp(366) = 4.5802d0
  Mcase65TAY_V_vp(367) = 4.5756d0
  Mcase65TAY_V_vp(368) = 4.571d0
  Mcase65TAY_V_vp(369) = 4.5664d0
  Mcase65TAY_V_vp(370) = 4.5664d0 ! CMB outer core
  Mcase65TAY_V_vp(371) = 9.4952d0 ! CMB mantle
  Mcase65TAY_V_vp(372) = 9.4978d0
  Mcase65TAY_V_vp(373) = 9.5004d0
  Mcase65TAY_V_vp(374) = 9.5024d0
  Mcase65TAY_V_vp(375) = 9.504d0
  Mcase65TAY_V_vp(376) = 9.5057d0
  Mcase65TAY_V_vp(377) = 9.5073d0
  Mcase65TAY_V_vp(378) = 9.509d0
  Mcase65TAY_V_vp(379) = 9.5099d0
  Mcase65TAY_V_vp(380) = 9.5105d0
  Mcase65TAY_V_vp(381) = 9.5111d0
  Mcase65TAY_V_vp(382) = 9.5117d0
  Mcase65TAY_V_vp(383) = 9.5123d0
  Mcase65TAY_V_vp(384) = 9.5093d0
  Mcase65TAY_V_vp(385) = 9.5048d0
  Mcase65TAY_V_vp(386) = 9.5003d0
  Mcase65TAY_V_vp(387) = 9.4959d0
  Mcase65TAY_V_vp(388) = 9.4914d0
  Mcase65TAY_V_vp(389) = 9.4836d0
  Mcase65TAY_V_vp(390) = 9.475d0
  Mcase65TAY_V_vp(391) = 9.4664d0
  Mcase65TAY_V_vp(392) = 9.4577d0
  Mcase65TAY_V_vp(393) = 9.4491d0
  Mcase65TAY_V_vp(394) = 9.4388d0
  Mcase65TAY_V_vp(395) = 9.4284d0
  Mcase65TAY_V_vp(396) = 9.4179d0
  Mcase65TAY_V_vp(397) = 9.4075d0
  Mcase65TAY_V_vp(398) = 9.397d0
  Mcase65TAY_V_vp(399) = 9.3868d0
  Mcase65TAY_V_vp(400) = 9.3766d0
  Mcase65TAY_V_vp(401) = 9.3664d0
  Mcase65TAY_V_vp(402) = 9.3562d0
  Mcase65TAY_V_vp(403) = 9.346d0
  Mcase65TAY_V_vp(404) = 9.3358d0
  Mcase65TAY_V_vp(405) = 9.3255d0
  Mcase65TAY_V_vp(406) = 9.3153d0
  Mcase65TAY_V_vp(407) = 9.305d0
  Mcase65TAY_V_vp(408) = 9.2948d0
  Mcase65TAY_V_vp(409) = 9.2845d0
  Mcase65TAY_V_vp(410) = 9.2742d0
  Mcase65TAY_V_vp(411) = 9.2639d0
  Mcase65TAY_V_vp(412) = 9.2536d0
  Mcase65TAY_V_vp(413) = 9.2438d0
  Mcase65TAY_V_vp(414) = 9.2358d0
  Mcase65TAY_V_vp(415) = 9.2277d0
  Mcase65TAY_V_vp(416) = 9.2196d0
  Mcase65TAY_V_vp(417) = 9.2116d0
  Mcase65TAY_V_vp(418) = 9.2043d0
  Mcase65TAY_V_vp(419) = 9.1988d0
  Mcase65TAY_V_vp(420) = 9.1932d0
  Mcase65TAY_V_vp(421) = 9.1876d0
  Mcase65TAY_V_vp(422) = 9.1821d0
  Mcase65TAY_V_vp(423) = 9.1769d0
  Mcase65TAY_V_vp(424) = 9.1724d0
  Mcase65TAY_V_vp(425) = 9.1679d0
  Mcase65TAY_V_vp(426) = 9.1633d0
  Mcase65TAY_V_vp(427) = 9.1588d0
  Mcase65TAY_V_vp(428) = 9.1543d0
  Mcase65TAY_V_vp(429) = 9.15d0
  Mcase65TAY_V_vp(430) = 9.1457d0
  Mcase65TAY_V_vp(431) = 9.1414d0
  Mcase65TAY_V_vp(432) = 9.1371d0
  Mcase65TAY_V_vp(433) = 9.1327d0
  Mcase65TAY_V_vp(434) = 9.1285d0
  Mcase65TAY_V_vp(435) = 9.1242d0
  Mcase65TAY_V_vp(436) = 9.1199d0
  Mcase65TAY_V_vp(437) = 9.1156d0
  Mcase65TAY_V_vp(438) = 9.1113d0
  Mcase65TAY_V_vp(439) = 9.107d0
  Mcase65TAY_V_vp(440) = 9.1028d0
  Mcase65TAY_V_vp(441) = 9.0985d0
  Mcase65TAY_V_vp(442) = 9.0943d0
  Mcase65TAY_V_vp(443) = 9.09d0
  Mcase65TAY_V_vp(444) = 9.0858d0
  Mcase65TAY_V_vp(445) = 9.0815d0
  Mcase65TAY_V_vp(446) = 9.0773d0
  Mcase65TAY_V_vp(447) = 9.073d0
  Mcase65TAY_V_vp(448) = 9.0685d0
  Mcase65TAY_V_vp(449) = 9.0639d0
  Mcase65TAY_V_vp(450) = 9.0594d0
  Mcase65TAY_V_vp(451) = 9.0548d0
  Mcase65TAY_V_vp(452) = 9.0503d0
  Mcase65TAY_V_vp(453) = 9.0457d0
  Mcase65TAY_V_vp(454) = 9.0412d0
  Mcase65TAY_V_vp(455) = 9.0367d0
  Mcase65TAY_V_vp(456) = 9.0321d0
  Mcase65TAY_V_vp(457) = 9.0276d0
  Mcase65TAY_V_vp(458) = 9.0196d0
  Mcase65TAY_V_vp(459) = 9.0116d0
  Mcase65TAY_V_vp(460) = 9.0036d0
  Mcase65TAY_V_vp(461) = 8.9956d0
  Mcase65TAY_V_vp(462) = 8.9864d0
  Mcase65TAY_V_vp(463) = 8.963d0
  Mcase65TAY_V_vp(464) = 8.9395d0
  Mcase65TAY_V_vp(465) = 8.9161d0
  Mcase65TAY_V_vp(466) = 8.8927d0
  Mcase65TAY_V_vp(467) = 8.8674d0
  Mcase65TAY_V_vp(468) = 8.8326d0
  Mcase65TAY_V_vp(469) = 8.7979d0
  Mcase65TAY_V_vp(470) = 8.7632d0
  Mcase65TAY_V_vp(471) = 8.7284d0
  Mcase65TAY_V_vp(472) = 8.6952d0
  Mcase65TAY_V_vp(473) = 8.6663d0
  Mcase65TAY_V_vp(474) = 8.6373d0
  Mcase65TAY_V_vp(475) = 8.6084d0
  Mcase65TAY_V_vp(476) = 8.5795d0
  Mcase65TAY_V_vp(477) = 8.5549d0
  Mcase65TAY_V_vp(478) = 8.5387d0
  Mcase65TAY_V_vp(479) = 8.5226d0
  Mcase65TAY_V_vp(480) = 8.5065d0
  Mcase65TAY_V_vp(481) = 8.4904d0
  Mcase65TAY_V_vp(482) = 8.4781d0
  Mcase65TAY_V_vp(483) = 8.4711d0
  Mcase65TAY_V_vp(484) = 8.4642d0
  Mcase65TAY_V_vp(485) = 8.4572d0
  Mcase65TAY_V_vp(486) = 8.4503d0
  Mcase65TAY_V_vp(487) = 8.4443d0
  Mcase65TAY_V_vp(488) = 8.4393d0
  Mcase65TAY_V_vp(489) = 8.4343d0
  Mcase65TAY_V_vp(490) = 8.4293d0
  Mcase65TAY_V_vp(491) = 8.4243d0
  Mcase65TAY_V_vp(492) = 8.4192d0
  Mcase65TAY_V_vp(493) = 8.4141d0
  Mcase65TAY_V_vp(494) = 8.4091d0
  Mcase65TAY_V_vp(495) = 8.404d0
  Mcase65TAY_V_vp(496) = 8.3989d0
  Mcase65TAY_V_vp(497) = 8.394d0
  Mcase65TAY_V_vp(498) = 8.3892d0
  Mcase65TAY_V_vp(499) = 8.3843d0
  Mcase65TAY_V_vp(500) = 8.3795d0
  Mcase65TAY_V_vp(501) = 8.3746d0
  Mcase65TAY_V_vp(502) = 8.37d0
  Mcase65TAY_V_vp(503) = 8.3654d0
  Mcase65TAY_V_vp(504) = 8.3608d0
  Mcase65TAY_V_vp(505) = 8.3562d0
  Mcase65TAY_V_vp(506) = 8.3516d0
  Mcase65TAY_V_vp(507) = 8.3471d0
  Mcase65TAY_V_vp(508) = 8.3425d0
  Mcase65TAY_V_vp(509) = 8.338d0
  Mcase65TAY_V_vp(510) = 8.3335d0
  Mcase65TAY_V_vp(511) = 8.329d0
  Mcase65TAY_V_vp(512) = 8.3245d0
  Mcase65TAY_V_vp(513) = 8.3199d0
  Mcase65TAY_V_vp(514) = 8.3154d0
  Mcase65TAY_V_vp(515) = 8.3109d0
  Mcase65TAY_V_vp(516) = 8.3064d0
  Mcase65TAY_V_vp(517) = 8.302d0
  Mcase65TAY_V_vp(518) = 8.2976d0
  Mcase65TAY_V_vp(519) = 8.2932d0
  Mcase65TAY_V_vp(520) = 8.2887d0
  Mcase65TAY_V_vp(521) = 8.2843d0
  Mcase65TAY_V_vp(522) = 8.28d0
  Mcase65TAY_V_vp(523) = 8.2757d0
  Mcase65TAY_V_vp(524) = 8.2714d0
  Mcase65TAY_V_vp(525) = 8.267d0
  Mcase65TAY_V_vp(526) = 8.2627d0
  Mcase65TAY_V_vp(527) = 8.2584d0
  Mcase65TAY_V_vp(528) = 8.2541d0
  Mcase65TAY_V_vp(529) = 8.2499d0
  Mcase65TAY_V_vp(530) = 8.2456d0
  Mcase65TAY_V_vp(531) = 8.2411d0
  Mcase65TAY_V_vp(532) = 8.2362d0
  Mcase65TAY_V_vp(533) = 8.2313d0
  Mcase65TAY_V_vp(534) = 8.2264d0
  Mcase65TAY_V_vp(535) = 8.2215d0
  Mcase65TAY_V_vp(536) = 8.217d0
  Mcase65TAY_V_vp(537) = 8.213d0
  Mcase65TAY_V_vp(538) = 8.209d0
  Mcase65TAY_V_vp(539) = 8.205d0
  Mcase65TAY_V_vp(540) = 8.201d0
  Mcase65TAY_V_vp(541) = 8.1967d0
  Mcase65TAY_V_vp(542) = 8.1921d0
  Mcase65TAY_V_vp(543) = 8.1875d0
  Mcase65TAY_V_vp(544) = 8.183d0
  Mcase65TAY_V_vp(545) = 8.1784d0
  Mcase65TAY_V_vp(546) = 8.1739d0
  Mcase65TAY_V_vp(547) = 8.1694d0
  Mcase65TAY_V_vp(548) = 8.1649d0
  Mcase65TAY_V_vp(549) = 8.1605d0
  Mcase65TAY_V_vp(550) = 8.156d0
  Mcase65TAY_V_vp(551) = 8.1521d0
  Mcase65TAY_V_vp(552) = 8.1485d0
  Mcase65TAY_V_vp(553) = 8.1449d0
  Mcase65TAY_V_vp(554) = 8.1413d0
  Mcase65TAY_V_vp(555) = 8.1377d0
  Mcase65TAY_V_vp(556) = 8.1343d0
  Mcase65TAY_V_vp(557) = 8.131d0
  Mcase65TAY_V_vp(558) = 8.1277d0
  Mcase65TAY_V_vp(559) = 8.1244d0
  Mcase65TAY_V_vp(560) = 8.1211d0
  Mcase65TAY_V_vp(561) = 8.118d0
  Mcase65TAY_V_vp(562) = 8.1148d0
  Mcase65TAY_V_vp(563) = 8.1117d0
  Mcase65TAY_V_vp(564) = 8.1086d0
  Mcase65TAY_V_vp(565) = 8.1055d0
  Mcase65TAY_V_vp(566) = 8.1025d0
  Mcase65TAY_V_vp(567) = 8.0995d0
  Mcase65TAY_V_vp(568) = 8.0966d0
  Mcase65TAY_V_vp(569) = 8.0936d0
  Mcase65TAY_V_vp(570) = 8.0906d0
  Mcase65TAY_V_vp(571) = 8.0877d0
  Mcase65TAY_V_vp(572) = 8.0849d0
  Mcase65TAY_V_vp(573) = 8.082d0
  Mcase65TAY_V_vp(574) = 8.0791d0
  Mcase65TAY_V_vp(575) = 8.0763d0
  Mcase65TAY_V_vp(576) = 8.0735d0
  Mcase65TAY_V_vp(577) = 8.0708d0
  Mcase65TAY_V_vp(578) = 8.068d0
  Mcase65TAY_V_vp(579) = 8.0653d0
  Mcase65TAY_V_vp(580) = 8.0626d0
  Mcase65TAY_V_vp(581) = 8.0599d0
  Mcase65TAY_V_vp(582) = 8.0573d0
  Mcase65TAY_V_vp(583) = 8.0546d0
  Mcase65TAY_V_vp(584) = 8.052d0
  Mcase65TAY_V_vp(585) = 8.0494d0
  Mcase65TAY_V_vp(586) = 8.0469d0
  Mcase65TAY_V_vp(587) = 8.0445d0
  Mcase65TAY_V_vp(588) = 8.0421d0
  Mcase65TAY_V_vp(589) = 8.0396d0
  Mcase65TAY_V_vp(590) = 8.0372d0
  Mcase65TAY_V_vp(591) = 8.0347d0
  Mcase65TAY_V_vp(592) = 8.0323d0
  Mcase65TAY_V_vp(593) = 8.0298d0
  Mcase65TAY_V_vp(594) = 8.0274d0
  Mcase65TAY_V_vp(595) = 8.025d0
  Mcase65TAY_V_vp(596) = 8.023d0
  Mcase65TAY_V_vp(597) = 8.0209d0
  Mcase65TAY_V_vp(598) = 8.0188d0
  Mcase65TAY_V_vp(599) = 8.0167d0
  Mcase65TAY_V_vp(600) = 8.0145d0
  Mcase65TAY_V_vp(601) = 8.0123d0
  Mcase65TAY_V_vp(602) = 8.0101d0
  Mcase65TAY_V_vp(603) = 8.0078d0
  Mcase65TAY_V_vp(604) = 8.0056d0
  Mcase65TAY_V_vp(605) = 8.0036d0
  Mcase65TAY_V_vp(606) = 8.0017d0
  Mcase65TAY_V_vp(607) = 7.9998d0
  Mcase65TAY_V_vp(608) = 7.9979d0
  Mcase65TAY_V_vp(609) = 7.996d0
  Mcase65TAY_V_vp(610) = 7.9942d0
  Mcase65TAY_V_vp(611) = 7.9923d0
  Mcase65TAY_V_vp(612) = 7.9905d0
  Mcase65TAY_V_vp(613) = 7.9886d0
  Mcase65TAY_V_vp(614) = 7.9868d0
  Mcase65TAY_V_vp(615) = 7.9851d0
  Mcase65TAY_V_vp(616) = 7.9834d0
  Mcase65TAY_V_vp(617) = 7.9817d0
  Mcase65TAY_V_vp(618) = 7.98d0
  Mcase65TAY_V_vp(619) = 7.9784d0
  Mcase65TAY_V_vp(620) = 7.9767d0
  Mcase65TAY_V_vp(621) = 7.9751d0
  Mcase65TAY_V_vp(622) = 7.9735d0
  Mcase65TAY_V_vp(623) = 7.9719d0
  Mcase65TAY_V_vp(624) = 7.9703d0
  Mcase65TAY_V_vp(625) = 7.969d0
  Mcase65TAY_V_vp(626) = 7.9677d0
  Mcase65TAY_V_vp(627) = 7.9663d0
  Mcase65TAY_V_vp(628) = 7.965d0
  Mcase65TAY_V_vp(629) = 7.9637d0
  Mcase65TAY_V_vp(630) = 7.9621d0
  Mcase65TAY_V_vp(631) = 7.9606d0
  Mcase65TAY_V_vp(632) = 7.9591d0
  Mcase65TAY_V_vp(633) = 7.9575d0
  Mcase65TAY_V_vp(634) = 7.956d0
  Mcase65TAY_V_vp(635) = 7.955d0
  Mcase65TAY_V_vp(636) = 7.954d0
  Mcase65TAY_V_vp(637) = 7.953d0
  Mcase65TAY_V_vp(638) = 7.952d0
  Mcase65TAY_V_vp(639) = 7.9508d0
  Mcase65TAY_V_vp(640) = 7.9487d0
  Mcase65TAY_V_vp(641) = 7.9466d0
  Mcase65TAY_V_vp(642) = 7.9445d0
  Mcase65TAY_V_vp(643) = 7.9423d0
  Mcase65TAY_V_vp(644) = 7.9396d0
  Mcase65TAY_V_vp(645) = 7.9348d0
  Mcase65TAY_V_vp(646) = 7.93d0
  Mcase65TAY_V_vp(647) = 7.9253d0
  Mcase65TAY_V_vp(648) = 7.9205d0
  Mcase65TAY_V_vp(649) = 7.9128d0
  Mcase65TAY_V_vp(650) = 7.899d0
  Mcase65TAY_V_vp(651) = 7.8853d0
  Mcase65TAY_V_vp(652) = 7.8716d0
  Mcase65TAY_V_vp(653) = 7.8578d0
  Mcase65TAY_V_vp(654) = 7.8161d0
  Mcase65TAY_V_vp(655) = 7.7347d0
  Mcase65TAY_V_vp(656) = 7.6532d0
  Mcase65TAY_V_vp(657) = 7.5718d0
  Mcase65TAY_V_vp(658) = 7.4903d0
  Mcase65TAY_V_vp(659) = 7.3916d0
  Mcase65TAY_V_vp(660) = 7.2756d0
  Mcase65TAY_V_vp(661) = 7.1597d0
  Mcase65TAY_V_vp(662) = 7.0704d0
  Mcase65TAY_V_vp(663) = 7.008d0
  Mcase65TAY_V_vp(664) = 6.9456d0
  Mcase65TAY_V_vp(665) = 6.8792d0
  Mcase65TAY_V_vp(666) = 6.8088d0
  Mcase65TAY_V_vp(667) = 6.7385d0
  Mcase65TAY_V_vp(668) = 6.6507d0
  Mcase65TAY_V_vp(669) = 6.5457d0
  Mcase65TAY_V_vp(670) = 6.4406d0
  Mcase65TAY_V_vp(671) = 6.3555d0
  Mcase65TAY_V_vp(672) = 6.2905d0
  Mcase65TAY_V_vp(673) = 6.2255d0
  Mcase65TAY_V_vp(674) = 6.1803d0
  Mcase65TAY_V_vp(675) = 6.155d0
  Mcase65TAY_V_vp(676) = 6.1296d0
  Mcase65TAY_V_vp(677) = 6.1112d0
  Mcase65TAY_V_vp(678) = 6.0996d0
  Mcase65TAY_V_vp(679) = 6.088d0
  Mcase65TAY_V_vp(680) = 6.0833d0
  Mcase65TAY_V_vp(681) = 6.0854d0

  Mcase65TAY_V_vs(1) = 3.1655d0
  Mcase65TAY_V_vs(2) = 3.1655d0
  Mcase65TAY_V_vs(3) = 3.1655d0
  Mcase65TAY_V_vs(4) = 3.1655d0
  Mcase65TAY_V_vs(5) = 3.1655d0
  Mcase65TAY_V_vs(6) = 3.1655d0
  Mcase65TAY_V_vs(7) = 3.1655d0
  Mcase65TAY_V_vs(8) = 3.1655d0
  Mcase65TAY_V_vs(9) = 3.1655d0
  Mcase65TAY_V_vs(10) = 3.1655d0
  Mcase65TAY_V_vs(11) = 3.1655d0
  Mcase65TAY_V_vs(12) = 3.1655d0
  Mcase65TAY_V_vs(13) = 3.1655d0
  Mcase65TAY_V_vs(14) = 3.1655d0
  Mcase65TAY_V_vs(15) = 3.1655d0
  Mcase65TAY_V_vs(16) = 3.1655d0
  Mcase65TAY_V_vs(17) = 3.1655d0
  Mcase65TAY_V_vs(18) = 3.1655d0
  Mcase65TAY_V_vs(19) = 3.1655d0
  Mcase65TAY_V_vs(20) = 3.1655d0
  Mcase65TAY_V_vs(21) = 3.1655d0
  Mcase65TAY_V_vs(22) = 3.1655d0
  Mcase65TAY_V_vs(23) = 3.1655d0
  Mcase65TAY_V_vs(24) = 3.1655d0
  Mcase65TAY_V_vs(25) = 3.1655d0
  Mcase65TAY_V_vs(26) = 3.1655d0
  Mcase65TAY_V_vs(27) = 3.1655d0
  Mcase65TAY_V_vs(28) = 3.1655d0
  Mcase65TAY_V_vs(29) = 3.1655d0
  Mcase65TAY_V_vs(30) = 3.1655d0
  Mcase65TAY_V_vs(31) = 3.1655d0
  Mcase65TAY_V_vs(32) = 3.1655d0
  Mcase65TAY_V_vs(33) = 3.1655d0
  Mcase65TAY_V_vs(34) = 3.1655d0
  Mcase65TAY_V_vs(35) = 3.1655d0
  Mcase65TAY_V_vs(36) = 3.1655d0
  Mcase65TAY_V_vs(37) = 3.1655d0
  Mcase65TAY_V_vs(38) = 3.1655d0
  Mcase65TAY_V_vs(39) = 3.1655d0
  Mcase65TAY_V_vs(40) = 3.1655d0
  Mcase65TAY_V_vs(41) = 3.1655d0
  Mcase65TAY_V_vs(42) = 3.1655d0
  Mcase65TAY_V_vs(43) = 3.1655d0
  Mcase65TAY_V_vs(44) = 3.1655d0
  Mcase65TAY_V_vs(45) = 3.1655d0
  Mcase65TAY_V_vs(46) = 3.1655d0
  Mcase65TAY_V_vs(47) = 3.1655d0
  Mcase65TAY_V_vs(48) = 3.1655d0
  Mcase65TAY_V_vs(49) = 3.1655d0
  Mcase65TAY_V_vs(50) = 3.1655d0
  Mcase65TAY_V_vs(51) = 3.1655d0
  Mcase65TAY_V_vs(52) = 3.1655d0
  Mcase65TAY_V_vs(53) = 3.1655d0
  Mcase65TAY_V_vs(54) = 3.1655d0
  Mcase65TAY_V_vs(55) = 3.1655d0
  Mcase65TAY_V_vs(56) = 3.1655d0
  Mcase65TAY_V_vs(57) = 3.1655d0
  Mcase65TAY_V_vs(58) = 3.1655d0
  Mcase65TAY_V_vs(59) = 3.1655d0
  Mcase65TAY_V_vs(60) = 3.1655d0
  Mcase65TAY_V_vs(61) = 3.1655d0
  Mcase65TAY_V_vs(62) = 3.1655d0
  Mcase65TAY_V_vs(63) = 3.1655d0
  Mcase65TAY_V_vs(64) = 3.1655d0
  Mcase65TAY_V_vs(65) = 3.1655d0
  Mcase65TAY_V_vs(66) = 3.1655d0
  Mcase65TAY_V_vs(67) = 3.1655d0
  Mcase65TAY_V_vs(68) = 3.1655d0
  Mcase65TAY_V_vs(69) = 3.1655d0
  Mcase65TAY_V_vs(70) = 3.1655d0
  Mcase65TAY_V_vs(71) = 3.1655d0
  Mcase65TAY_V_vs(72) = 3.1655d0
  Mcase65TAY_V_vs(73) = 3.1655d0
  Mcase65TAY_V_vs(74) = 3.1655d0
  Mcase65TAY_V_vs(75) = 3.1655d0
  Mcase65TAY_V_vs(76) = 3.1655d0
  Mcase65TAY_V_vs(77) = 3.1655d0
  Mcase65TAY_V_vs(78) = 3.1655d0
  Mcase65TAY_V_vs(79) = 3.1655d0
  Mcase65TAY_V_vs(80) = 3.1655d0
  Mcase65TAY_V_vs(81) = 3.1655d0
  Mcase65TAY_V_vs(82) = 3.1655d0
  Mcase65TAY_V_vs(83) = 3.1655d0
  Mcase65TAY_V_vs(84) = 3.1655d0
  Mcase65TAY_V_vs(85) = 3.1655d0
  Mcase65TAY_V_vs(86) = 3.1655d0
  Mcase65TAY_V_vs(87) = 3.1655d0
  Mcase65TAY_V_vs(88) = 3.1655d0
  Mcase65TAY_V_vs(89) = 3.1655d0
  Mcase65TAY_V_vs(90) = 3.1655d0
  Mcase65TAY_V_vs(91) = 3.1655d0
  Mcase65TAY_V_vs(92) = 3.1655d0
  Mcase65TAY_V_vs(93) = 3.1655d0
  Mcase65TAY_V_vs(94) = 3.1655d0
  Mcase65TAY_V_vs(95) = 3.1655d0
  Mcase65TAY_V_vs(96) = 3.1655d0
  Mcase65TAY_V_vs(97) = 3.1655d0
  Mcase65TAY_V_vs(98) = 3.1655d0
  Mcase65TAY_V_vs(99) = 3.1655d0
  Mcase65TAY_V_vs(100) = 3.1655d0
  Mcase65TAY_V_vs(101) = 3.1655d0
  Mcase65TAY_V_vs(102) = 3.1655d0
  Mcase65TAY_V_vs(103) = 3.1655d0
  Mcase65TAY_V_vs(104) = 3.1655d0  ! ICB inner core
  Mcase65TAY_V_vs(105) = 0.d0      ! ICB outer core
  Mcase65TAY_V_vs(106) = 0.d0
  Mcase65TAY_V_vs(107) = 0.d0
  Mcase65TAY_V_vs(108) = 0.d0
  Mcase65TAY_V_vs(109) = 0.d0
  Mcase65TAY_V_vs(110) = 0.d0
  Mcase65TAY_V_vs(111) = 0.d0
  Mcase65TAY_V_vs(112) = 0.d0
  Mcase65TAY_V_vs(113) = 0.d0
  Mcase65TAY_V_vs(114) = 0.d0
  Mcase65TAY_V_vs(115) = 0.d0
  Mcase65TAY_V_vs(116) = 0.d0
  Mcase65TAY_V_vs(117) = 0.d0
  Mcase65TAY_V_vs(118) = 0.d0
  Mcase65TAY_V_vs(119) = 0.d0
  Mcase65TAY_V_vs(120) = 0.d0
  Mcase65TAY_V_vs(121) = 0.d0
  Mcase65TAY_V_vs(122) = 0.d0
  Mcase65TAY_V_vs(123) = 0.d0
  Mcase65TAY_V_vs(124) = 0.d0
  Mcase65TAY_V_vs(125) = 0.d0
  Mcase65TAY_V_vs(126) = 0.d0
  Mcase65TAY_V_vs(127) = 0.d0
  Mcase65TAY_V_vs(128) = 0.d0
  Mcase65TAY_V_vs(129) = 0.d0
  Mcase65TAY_V_vs(130) = 0.d0
  Mcase65TAY_V_vs(131) = 0.d0
  Mcase65TAY_V_vs(132) = 0.d0
  Mcase65TAY_V_vs(133) = 0.d0
  Mcase65TAY_V_vs(134) = 0.d0
  Mcase65TAY_V_vs(135) = 0.d0
  Mcase65TAY_V_vs(136) = 0.d0
  Mcase65TAY_V_vs(137) = 0.d0
  Mcase65TAY_V_vs(138) = 0.d0
  Mcase65TAY_V_vs(139) = 0.d0
  Mcase65TAY_V_vs(140) = 0.d0
  Mcase65TAY_V_vs(141) = 0.d0
  Mcase65TAY_V_vs(142) = 0.d0
  Mcase65TAY_V_vs(143) = 0.d0
  Mcase65TAY_V_vs(144) = 0.d0
  Mcase65TAY_V_vs(145) = 0.d0
  Mcase65TAY_V_vs(146) = 0.d0
  Mcase65TAY_V_vs(147) = 0.d0
  Mcase65TAY_V_vs(148) = 0.d0
  Mcase65TAY_V_vs(149) = 0.d0
  Mcase65TAY_V_vs(150) = 0.d0
  Mcase65TAY_V_vs(151) = 0.d0
  Mcase65TAY_V_vs(152) = 0.d0
  Mcase65TAY_V_vs(153) = 0.d0
  Mcase65TAY_V_vs(154) = 0.d0
  Mcase65TAY_V_vs(155) = 0.d0
  Mcase65TAY_V_vs(156) = 0.d0
  Mcase65TAY_V_vs(157) = 0.d0
  Mcase65TAY_V_vs(158) = 0.d0
  Mcase65TAY_V_vs(159) = 0.d0
  Mcase65TAY_V_vs(160) = 0.d0
  Mcase65TAY_V_vs(161) = 0.d0
  Mcase65TAY_V_vs(162) = 0.d0
  Mcase65TAY_V_vs(163) = 0.d0
  Mcase65TAY_V_vs(164) = 0.d0
  Mcase65TAY_V_vs(165) = 0.d0
  Mcase65TAY_V_vs(166) = 0.d0
  Mcase65TAY_V_vs(167) = 0.d0
  Mcase65TAY_V_vs(168) = 0.d0
  Mcase65TAY_V_vs(169) = 0.d0
  Mcase65TAY_V_vs(170) = 0.d0
  Mcase65TAY_V_vs(171) = 0.d0
  Mcase65TAY_V_vs(172) = 0.d0
  Mcase65TAY_V_vs(173) = 0.d0
  Mcase65TAY_V_vs(174) = 0.d0
  Mcase65TAY_V_vs(175) = 0.d0
  Mcase65TAY_V_vs(176) = 0.d0
  Mcase65TAY_V_vs(177) = 0.d0
  Mcase65TAY_V_vs(178) = 0.d0
  Mcase65TAY_V_vs(179) = 0.d0
  Mcase65TAY_V_vs(180) = 0.d0
  Mcase65TAY_V_vs(181) = 0.d0
  Mcase65TAY_V_vs(182) = 0.d0
  Mcase65TAY_V_vs(183) = 0.d0
  Mcase65TAY_V_vs(184) = 0.d0
  Mcase65TAY_V_vs(185) = 0.d0
  Mcase65TAY_V_vs(186) = 0.d0
  Mcase65TAY_V_vs(187) = 0.d0
  Mcase65TAY_V_vs(188) = 0.d0
  Mcase65TAY_V_vs(189) = 0.d0
  Mcase65TAY_V_vs(190) = 0.d0
  Mcase65TAY_V_vs(191) = 0.d0
  Mcase65TAY_V_vs(192) = 0.d0
  Mcase65TAY_V_vs(193) = 0.d0
  Mcase65TAY_V_vs(194) = 0.d0
  Mcase65TAY_V_vs(195) = 0.d0
  Mcase65TAY_V_vs(196) = 0.d0
  Mcase65TAY_V_vs(197) = 0.d0
  Mcase65TAY_V_vs(198) = 0.d0
  Mcase65TAY_V_vs(199) = 0.d0
  Mcase65TAY_V_vs(200) = 0.d0
  Mcase65TAY_V_vs(201) = 0.d0
  Mcase65TAY_V_vs(202) = 0.d0
  Mcase65TAY_V_vs(203) = 0.d0
  Mcase65TAY_V_vs(204) = 0.d0
  Mcase65TAY_V_vs(205) = 0.d0
  Mcase65TAY_V_vs(206) = 0.d0
  Mcase65TAY_V_vs(207) = 0.d0
  Mcase65TAY_V_vs(208) = 0.d0
  Mcase65TAY_V_vs(209) = 0.d0
  Mcase65TAY_V_vs(210) = 0.d0
  Mcase65TAY_V_vs(211) = 0.d0
  Mcase65TAY_V_vs(212) = 0.d0
  Mcase65TAY_V_vs(213) = 0.d0
  Mcase65TAY_V_vs(214) = 0.d0
  Mcase65TAY_V_vs(215) = 0.d0
  Mcase65TAY_V_vs(216) = 0.d0
  Mcase65TAY_V_vs(217) = 0.d0
  Mcase65TAY_V_vs(218) = 0.d0
  Mcase65TAY_V_vs(219) = 0.d0
  Mcase65TAY_V_vs(220) = 0.d0
  Mcase65TAY_V_vs(221) = 0.d0
  Mcase65TAY_V_vs(222) = 0.d0
  Mcase65TAY_V_vs(223) = 0.d0
  Mcase65TAY_V_vs(224) = 0.d0
  Mcase65TAY_V_vs(225) = 0.d0
  Mcase65TAY_V_vs(226) = 0.d0
  Mcase65TAY_V_vs(227) = 0.d0
  Mcase65TAY_V_vs(228) = 0.d0
  Mcase65TAY_V_vs(229) = 0.d0
  Mcase65TAY_V_vs(230) = 0.d0
  Mcase65TAY_V_vs(231) = 0.d0
  Mcase65TAY_V_vs(232) = 0.d0
  Mcase65TAY_V_vs(233) = 0.d0
  Mcase65TAY_V_vs(234) = 0.d0
  Mcase65TAY_V_vs(235) = 0.d0
  Mcase65TAY_V_vs(236) = 0.d0
  Mcase65TAY_V_vs(237) = 0.d0
  Mcase65TAY_V_vs(238) = 0.d0
  Mcase65TAY_V_vs(239) = 0.d0
  Mcase65TAY_V_vs(240) = 0.d0
  Mcase65TAY_V_vs(241) = 0.d0
  Mcase65TAY_V_vs(242) = 0.d0
  Mcase65TAY_V_vs(243) = 0.d0
  Mcase65TAY_V_vs(244) = 0.d0
  Mcase65TAY_V_vs(245) = 0.d0
  Mcase65TAY_V_vs(246) = 0.d0
  Mcase65TAY_V_vs(247) = 0.d0
  Mcase65TAY_V_vs(248) = 0.d0
  Mcase65TAY_V_vs(249) = 0.d0
  Mcase65TAY_V_vs(250) = 0.d0
  Mcase65TAY_V_vs(251) = 0.d0
  Mcase65TAY_V_vs(252) = 0.d0
  Mcase65TAY_V_vs(253) = 0.d0
  Mcase65TAY_V_vs(254) = 0.d0
  Mcase65TAY_V_vs(255) = 0.d0
  Mcase65TAY_V_vs(256) = 0.d0
  Mcase65TAY_V_vs(257) = 0.d0
  Mcase65TAY_V_vs(258) = 0.d0
  Mcase65TAY_V_vs(259) = 0.d0
  Mcase65TAY_V_vs(260) = 0.d0
  Mcase65TAY_V_vs(261) = 0.d0
  Mcase65TAY_V_vs(262) = 0.d0
  Mcase65TAY_V_vs(263) = 0.d0
  Mcase65TAY_V_vs(264) = 0.d0
  Mcase65TAY_V_vs(265) = 0.d0
  Mcase65TAY_V_vs(266) = 0.d0
  Mcase65TAY_V_vs(267) = 0.d0
  Mcase65TAY_V_vs(268) = 0.d0
  Mcase65TAY_V_vs(269) = 0.d0
  Mcase65TAY_V_vs(270) = 0.d0
  Mcase65TAY_V_vs(271) = 0.d0
  Mcase65TAY_V_vs(272) = 0.d0
  Mcase65TAY_V_vs(273) = 0.d0
  Mcase65TAY_V_vs(274) = 0.d0
  Mcase65TAY_V_vs(275) = 0.d0
  Mcase65TAY_V_vs(276) = 0.d0
  Mcase65TAY_V_vs(277) = 0.d0
  Mcase65TAY_V_vs(278) = 0.d0
  Mcase65TAY_V_vs(279) = 0.d0
  Mcase65TAY_V_vs(280) = 0.d0
  Mcase65TAY_V_vs(281) = 0.d0
  Mcase65TAY_V_vs(282) = 0.d0
  Mcase65TAY_V_vs(283) = 0.d0
  Mcase65TAY_V_vs(284) = 0.d0
  Mcase65TAY_V_vs(285) = 0.d0
  Mcase65TAY_V_vs(286) = 0.d0
  Mcase65TAY_V_vs(287) = 0.d0
  Mcase65TAY_V_vs(288) = 0.d0
  Mcase65TAY_V_vs(289) = 0.d0
  Mcase65TAY_V_vs(290) = 0.d0
  Mcase65TAY_V_vs(291) = 0.d0
  Mcase65TAY_V_vs(292) = 0.d0
  Mcase65TAY_V_vs(293) = 0.d0
  Mcase65TAY_V_vs(294) = 0.d0
  Mcase65TAY_V_vs(295) = 0.d0
  Mcase65TAY_V_vs(296) = 0.d0
  Mcase65TAY_V_vs(297) = 0.d0
  Mcase65TAY_V_vs(298) = 0.d0
  Mcase65TAY_V_vs(299) = 0.d0
  Mcase65TAY_V_vs(300) = 0.d0
  Mcase65TAY_V_vs(301) = 0.d0
  Mcase65TAY_V_vs(302) = 0.d0
  Mcase65TAY_V_vs(303) = 0.d0
  Mcase65TAY_V_vs(304) = 0.d0
  Mcase65TAY_V_vs(305) = 0.d0
  Mcase65TAY_V_vs(306) = 0.d0
  Mcase65TAY_V_vs(307) = 0.d0
  Mcase65TAY_V_vs(308) = 0.d0
  Mcase65TAY_V_vs(309) = 0.d0
  Mcase65TAY_V_vs(310) = 0.d0
  Mcase65TAY_V_vs(311) = 0.d0
  Mcase65TAY_V_vs(312) = 0.d0
  Mcase65TAY_V_vs(313) = 0.d0
  Mcase65TAY_V_vs(314) = 0.d0
  Mcase65TAY_V_vs(315) = 0.d0
  Mcase65TAY_V_vs(316) = 0.d0
  Mcase65TAY_V_vs(317) = 0.d0
  Mcase65TAY_V_vs(318) = 0.d0
  Mcase65TAY_V_vs(319) = 0.d0
  Mcase65TAY_V_vs(320) = 0.d0
  Mcase65TAY_V_vs(321) = 0.d0
  Mcase65TAY_V_vs(322) = 0.d0
  Mcase65TAY_V_vs(323) = 0.d0
  Mcase65TAY_V_vs(324) = 0.d0
  Mcase65TAY_V_vs(325) = 0.d0
  Mcase65TAY_V_vs(326) = 0.d0
  Mcase65TAY_V_vs(327) = 0.d0
  Mcase65TAY_V_vs(328) = 0.d0
  Mcase65TAY_V_vs(329) = 0.d0
  Mcase65TAY_V_vs(330) = 0.d0
  Mcase65TAY_V_vs(331) = 0.d0
  Mcase65TAY_V_vs(332) = 0.d0
  Mcase65TAY_V_vs(333) = 0.d0
  Mcase65TAY_V_vs(334) = 0.d0
  Mcase65TAY_V_vs(335) = 0.d0
  Mcase65TAY_V_vs(336) = 0.d0
  Mcase65TAY_V_vs(337) = 0.d0
  Mcase65TAY_V_vs(338) = 0.d0
  Mcase65TAY_V_vs(339) = 0.d0
  Mcase65TAY_V_vs(340) = 0.d0
  Mcase65TAY_V_vs(341) = 0.d0
  Mcase65TAY_V_vs(342) = 0.d0
  Mcase65TAY_V_vs(343) = 0.d0
  Mcase65TAY_V_vs(344) = 0.d0
  Mcase65TAY_V_vs(345) = 0.d0
  Mcase65TAY_V_vs(346) = 0.d0
  Mcase65TAY_V_vs(347) = 0.d0
  Mcase65TAY_V_vs(348) = 0.d0
  Mcase65TAY_V_vs(349) = 0.d0
  Mcase65TAY_V_vs(350) = 0.d0
  Mcase65TAY_V_vs(351) = 0.d0
  Mcase65TAY_V_vs(352) = 0.d0
  Mcase65TAY_V_vs(353) = 0.d0
  Mcase65TAY_V_vs(354) = 0.d0
  Mcase65TAY_V_vs(355) = 0.d0
  Mcase65TAY_V_vs(356) = 0.d0
  Mcase65TAY_V_vs(357) = 0.d0
  Mcase65TAY_V_vs(358) = 0.d0
  Mcase65TAY_V_vs(359) = 0.d0
  Mcase65TAY_V_vs(360) = 0.d0
  Mcase65TAY_V_vs(361) = 0.d0
  Mcase65TAY_V_vs(362) = 0.d0
  Mcase65TAY_V_vs(363) = 0.d0
  Mcase65TAY_V_vs(364) = 0.d0
  Mcase65TAY_V_vs(365) = 0.d0
  Mcase65TAY_V_vs(366) = 0.d0
  Mcase65TAY_V_vs(367) = 0.d0
  Mcase65TAY_V_vs(368) = 0.d0
  Mcase65TAY_V_vs(369) = 0.d0
  Mcase65TAY_V_vs(370) = 0.d0     ! CMB outer core
  Mcase65TAY_V_vs(371) = 5.0478d0 ! CMB mantle
  Mcase65TAY_V_vs(372) = 5.051d0
  Mcase65TAY_V_vs(373) = 5.0541d0
  Mcase65TAY_V_vs(374) = 5.0569d0
  Mcase65TAY_V_vs(375) = 5.0594d0
  Mcase65TAY_V_vs(376) = 5.0619d0
  Mcase65TAY_V_vs(377) = 5.0644d0
  Mcase65TAY_V_vs(378) = 5.0669d0
  Mcase65TAY_V_vs(379) = 5.0688d0
  Mcase65TAY_V_vs(380) = 5.0705d0
  Mcase65TAY_V_vs(381) = 5.0721d0
  Mcase65TAY_V_vs(382) = 5.0737d0
  Mcase65TAY_V_vs(383) = 5.0753d0
  Mcase65TAY_V_vs(384) = 5.0744d0
  Mcase65TAY_V_vs(385) = 5.0724d0
  Mcase65TAY_V_vs(386) = 5.0705d0
  Mcase65TAY_V_vs(387) = 5.0685d0
  Mcase65TAY_V_vs(388) = 5.0665d0
  Mcase65TAY_V_vs(389) = 5.0624d0
  Mcase65TAY_V_vs(390) = 5.0577d0
  Mcase65TAY_V_vs(391) = 5.0529d0
  Mcase65TAY_V_vs(392) = 5.0482d0
  Mcase65TAY_V_vs(393) = 5.0435d0
  Mcase65TAY_V_vs(394) = 5.0378d0
  Mcase65TAY_V_vs(395) = 5.032d0
  Mcase65TAY_V_vs(396) = 5.0261d0
  Mcase65TAY_V_vs(397) = 5.0203d0
  Mcase65TAY_V_vs(398) = 5.0145d0
  Mcase65TAY_V_vs(399) = 5.0089d0
  Mcase65TAY_V_vs(400) = 5.0034d0
  Mcase65TAY_V_vs(401) = 4.9979d0
  Mcase65TAY_V_vs(402) = 4.9923d0
  Mcase65TAY_V_vs(403) = 4.9868d0
  Mcase65TAY_V_vs(404) = 4.9813d0
  Mcase65TAY_V_vs(405) = 4.9758d0
  Mcase65TAY_V_vs(406) = 4.9703d0
  Mcase65TAY_V_vs(407) = 4.9648d0
  Mcase65TAY_V_vs(408) = 4.9593d0
  Mcase65TAY_V_vs(409) = 4.9539d0
  Mcase65TAY_V_vs(410) = 4.9484d0
  Mcase65TAY_V_vs(411) = 4.943d0
  Mcase65TAY_V_vs(412) = 4.9375d0
  Mcase65TAY_V_vs(413) = 4.9324d0
  Mcase65TAY_V_vs(414) = 4.9284d0
  Mcase65TAY_V_vs(415) = 4.9244d0
  Mcase65TAY_V_vs(416) = 4.9205d0
  Mcase65TAY_V_vs(417) = 4.9165d0
  Mcase65TAY_V_vs(418) = 4.913d0
  Mcase65TAY_V_vs(419) = 4.9106d0
  Mcase65TAY_V_vs(420) = 4.9083d0
  Mcase65TAY_V_vs(421) = 4.906d0
  Mcase65TAY_V_vs(422) = 4.9036d0
  Mcase65TAY_V_vs(423) = 4.9016d0
  Mcase65TAY_V_vs(424) = 4.8999d0
  Mcase65TAY_V_vs(425) = 4.8983d0
  Mcase65TAY_V_vs(426) = 4.8966d0
  Mcase65TAY_V_vs(427) = 4.895d0
  Mcase65TAY_V_vs(428) = 4.8935d0
  Mcase65TAY_V_vs(429) = 4.892d0
  Mcase65TAY_V_vs(430) = 4.8906d0
  Mcase65TAY_V_vs(431) = 4.8892d0
  Mcase65TAY_V_vs(432) = 4.8877d0
  Mcase65TAY_V_vs(433) = 4.8864d0
  Mcase65TAY_V_vs(434) = 4.8851d0
  Mcase65TAY_V_vs(435) = 4.8838d0
  Mcase65TAY_V_vs(436) = 4.8826d0
  Mcase65TAY_V_vs(437) = 4.8813d0
  Mcase65TAY_V_vs(438) = 4.88d0
  Mcase65TAY_V_vs(439) = 4.8786d0
  Mcase65TAY_V_vs(440) = 4.8773d0
  Mcase65TAY_V_vs(441) = 4.876d0
  Mcase65TAY_V_vs(442) = 4.8746d0
  Mcase65TAY_V_vs(443) = 4.8733d0
  Mcase65TAY_V_vs(444) = 4.8719d0
  Mcase65TAY_V_vs(445) = 4.8706d0
  Mcase65TAY_V_vs(446) = 4.8692d0
  Mcase65TAY_V_vs(447) = 4.8679d0
  Mcase65TAY_V_vs(448) = 4.8665d0
  Mcase65TAY_V_vs(449) = 4.8651d0
  Mcase65TAY_V_vs(450) = 4.8637d0
  Mcase65TAY_V_vs(451) = 4.8623d0
  Mcase65TAY_V_vs(452) = 4.8609d0
  Mcase65TAY_V_vs(453) = 4.8596d0
  Mcase65TAY_V_vs(454) = 4.8583d0
  Mcase65TAY_V_vs(455) = 4.857d0
  Mcase65TAY_V_vs(456) = 4.8557d0
  Mcase65TAY_V_vs(457) = 4.8544d0
  Mcase65TAY_V_vs(458) = 4.851d0
  Mcase65TAY_V_vs(459) = 4.8476d0
  Mcase65TAY_V_vs(460) = 4.8442d0
  Mcase65TAY_V_vs(461) = 4.8409d0
  Mcase65TAY_V_vs(462) = 4.8367d0
  Mcase65TAY_V_vs(463) = 4.8235d0
  Mcase65TAY_V_vs(464) = 4.8102d0
  Mcase65TAY_V_vs(465) = 4.797d0
  Mcase65TAY_V_vs(466) = 4.7838d0
  Mcase65TAY_V_vs(467) = 4.7692d0
  Mcase65TAY_V_vs(468) = 4.7482d0
  Mcase65TAY_V_vs(469) = 4.7272d0
  Mcase65TAY_V_vs(470) = 4.7062d0
  Mcase65TAY_V_vs(471) = 4.6852d0
  Mcase65TAY_V_vs(472) = 4.6651d0
  Mcase65TAY_V_vs(473) = 4.6479d0
  Mcase65TAY_V_vs(474) = 4.6306d0
  Mcase65TAY_V_vs(475) = 4.6134d0
  Mcase65TAY_V_vs(476) = 4.5962d0
  Mcase65TAY_V_vs(477) = 4.5819d0
  Mcase65TAY_V_vs(478) = 4.5733d0
  Mcase65TAY_V_vs(479) = 4.5647d0
  Mcase65TAY_V_vs(480) = 4.5561d0
  Mcase65TAY_V_vs(481) = 4.5475d0
  Mcase65TAY_V_vs(482) = 4.5414d0
  Mcase65TAY_V_vs(483) = 4.539d0
  Mcase65TAY_V_vs(484) = 4.5365d0
  Mcase65TAY_V_vs(485) = 4.534d0
  Mcase65TAY_V_vs(486) = 4.5316d0
  Mcase65TAY_V_vs(487) = 4.5296d0
  Mcase65TAY_V_vs(488) = 4.5281d0
  Mcase65TAY_V_vs(489) = 4.5266d0
  Mcase65TAY_V_vs(490) = 4.5251d0
  Mcase65TAY_V_vs(491) = 4.5236d0
  Mcase65TAY_V_vs(492) = 4.522d0
  Mcase65TAY_V_vs(493) = 4.5202d0
  Mcase65TAY_V_vs(494) = 4.5184d0
  Mcase65TAY_V_vs(495) = 4.5166d0
  Mcase65TAY_V_vs(496) = 4.5149d0
  Mcase65TAY_V_vs(497) = 4.5132d0
  Mcase65TAY_V_vs(498) = 4.5116d0
  Mcase65TAY_V_vs(499) = 4.51d0
  Mcase65TAY_V_vs(500) = 4.5084d0
  Mcase65TAY_V_vs(501) = 4.5068d0
  Mcase65TAY_V_vs(502) = 4.5055d0
  Mcase65TAY_V_vs(503) = 4.5043d0
  Mcase65TAY_V_vs(504) = 4.503d0
  Mcase65TAY_V_vs(505) = 4.5017d0
  Mcase65TAY_V_vs(506) = 4.5005d0
  Mcase65TAY_V_vs(507) = 4.4993d0
  Mcase65TAY_V_vs(508) = 4.4981d0
  Mcase65TAY_V_vs(509) = 4.497d0
  Mcase65TAY_V_vs(510) = 4.4958d0
  Mcase65TAY_V_vs(511) = 4.4946d0
  Mcase65TAY_V_vs(512) = 4.4934d0
  Mcase65TAY_V_vs(513) = 4.4921d0
  Mcase65TAY_V_vs(514) = 4.4909d0
  Mcase65TAY_V_vs(515) = 4.4897d0
  Mcase65TAY_V_vs(516) = 4.4884d0
  Mcase65TAY_V_vs(517) = 4.4872d0
  Mcase65TAY_V_vs(518) = 4.486d0
  Mcase65TAY_V_vs(519) = 4.4847d0
  Mcase65TAY_V_vs(520) = 4.4835d0
  Mcase65TAY_V_vs(521) = 4.4823d0
  Mcase65TAY_V_vs(522) = 4.4811d0
  Mcase65TAY_V_vs(523) = 4.4799d0
  Mcase65TAY_V_vs(524) = 4.4787d0
  Mcase65TAY_V_vs(525) = 4.4775d0
  Mcase65TAY_V_vs(526) = 4.4763d0
  Mcase65TAY_V_vs(527) = 4.4751d0
  Mcase65TAY_V_vs(528) = 4.4739d0
  Mcase65TAY_V_vs(529) = 4.4727d0
  Mcase65TAY_V_vs(530) = 4.4715d0
  Mcase65TAY_V_vs(531) = 4.4702d0
  Mcase65TAY_V_vs(532) = 4.4685d0
  Mcase65TAY_V_vs(533) = 4.4667d0
  Mcase65TAY_V_vs(534) = 4.465d0
  Mcase65TAY_V_vs(535) = 4.4633d0
  Mcase65TAY_V_vs(536) = 4.4618d0
  Mcase65TAY_V_vs(537) = 4.4609d0
  Mcase65TAY_V_vs(538) = 4.4601d0
  Mcase65TAY_V_vs(539) = 4.4592d0
  Mcase65TAY_V_vs(540) = 4.4583d0
  Mcase65TAY_V_vs(541) = 4.4571d0
  Mcase65TAY_V_vs(542) = 4.4557d0
  Mcase65TAY_V_vs(543) = 4.4543d0
  Mcase65TAY_V_vs(544) = 4.4528d0
  Mcase65TAY_V_vs(545) = 4.4514d0
  Mcase65TAY_V_vs(546) = 4.45d0
  Mcase65TAY_V_vs(547) = 4.4487d0
  Mcase65TAY_V_vs(548) = 4.4473d0
  Mcase65TAY_V_vs(549) = 4.446d0
  Mcase65TAY_V_vs(550) = 4.4446d0
  Mcase65TAY_V_vs(551) = 4.4438d0
  Mcase65TAY_V_vs(552) = 4.4433d0
  Mcase65TAY_V_vs(553) = 4.4428d0
  Mcase65TAY_V_vs(554) = 4.4423d0
  Mcase65TAY_V_vs(555) = 4.4417d0
  Mcase65TAY_V_vs(556) = 4.4414d0
  Mcase65TAY_V_vs(557) = 4.441d0
  Mcase65TAY_V_vs(558) = 4.4407d0
  Mcase65TAY_V_vs(559) = 4.4404d0
  Mcase65TAY_V_vs(560) = 4.44d0
  Mcase65TAY_V_vs(561) = 4.4399d0
  Mcase65TAY_V_vs(562) = 4.4397d0
  Mcase65TAY_V_vs(563) = 4.4396d0
  Mcase65TAY_V_vs(564) = 4.4395d0
  Mcase65TAY_V_vs(565) = 4.4393d0
  Mcase65TAY_V_vs(566) = 4.4393d0
  Mcase65TAY_V_vs(567) = 4.4393d0
  Mcase65TAY_V_vs(568) = 4.4392d0
  Mcase65TAY_V_vs(569) = 4.4392d0
  Mcase65TAY_V_vs(570) = 4.4392d0
  Mcase65TAY_V_vs(571) = 4.4392d0
  Mcase65TAY_V_vs(572) = 4.4393d0
  Mcase65TAY_V_vs(573) = 4.4394d0
  Mcase65TAY_V_vs(574) = 4.4394d0
  Mcase65TAY_V_vs(575) = 4.4395d0
  Mcase65TAY_V_vs(576) = 4.4396d0
  Mcase65TAY_V_vs(577) = 4.4398d0
  Mcase65TAY_V_vs(578) = 4.44d0
  Mcase65TAY_V_vs(579) = 4.4401d0
  Mcase65TAY_V_vs(580) = 4.4403d0
  Mcase65TAY_V_vs(581) = 4.4405d0
  Mcase65TAY_V_vs(582) = 4.4408d0
  Mcase65TAY_V_vs(583) = 4.441d0
  Mcase65TAY_V_vs(584) = 4.4413d0
  Mcase65TAY_V_vs(585) = 4.4416d0
  Mcase65TAY_V_vs(586) = 4.4419d0
  Mcase65TAY_V_vs(587) = 4.4423d0
  Mcase65TAY_V_vs(588) = 4.4427d0
  Mcase65TAY_V_vs(589) = 4.443d0
  Mcase65TAY_V_vs(590) = 4.4434d0
  Mcase65TAY_V_vs(591) = 4.4439d0
  Mcase65TAY_V_vs(592) = 4.4443d0
  Mcase65TAY_V_vs(593) = 4.4447d0
  Mcase65TAY_V_vs(594) = 4.4451d0
  Mcase65TAY_V_vs(595) = 4.4456d0
  Mcase65TAY_V_vs(596) = 4.4462d0
  Mcase65TAY_V_vs(597) = 4.4468d0
  Mcase65TAY_V_vs(598) = 4.4475d0
  Mcase65TAY_V_vs(599) = 4.4481d0
  Mcase65TAY_V_vs(600) = 4.4487d0
  Mcase65TAY_V_vs(601) = 4.4493d0
  Mcase65TAY_V_vs(602) = 4.4499d0
  Mcase65TAY_V_vs(603) = 4.4505d0
  Mcase65TAY_V_vs(604) = 4.4512d0
  Mcase65TAY_V_vs(605) = 4.4519d0
  Mcase65TAY_V_vs(606) = 4.4527d0
  Mcase65TAY_V_vs(607) = 4.4535d0
  Mcase65TAY_V_vs(608) = 4.4542d0
  Mcase65TAY_V_vs(609) = 4.455d0
  Mcase65TAY_V_vs(610) = 4.4559d0
  Mcase65TAY_V_vs(611) = 4.4567d0
  Mcase65TAY_V_vs(612) = 4.4576d0
  Mcase65TAY_V_vs(613) = 4.4585d0
  Mcase65TAY_V_vs(614) = 4.4593d0
  Mcase65TAY_V_vs(615) = 4.4603d0
  Mcase65TAY_V_vs(616) = 4.4613d0
  Mcase65TAY_V_vs(617) = 4.4622d0
  Mcase65TAY_V_vs(618) = 4.4632d0
  Mcase65TAY_V_vs(619) = 4.4642d0
  Mcase65TAY_V_vs(620) = 4.4652d0
  Mcase65TAY_V_vs(621) = 4.4663d0
  Mcase65TAY_V_vs(622) = 4.4673d0
  Mcase65TAY_V_vs(623) = 4.4684d0
  Mcase65TAY_V_vs(624) = 4.4695d0
  Mcase65TAY_V_vs(625) = 4.4706d0
  Mcase65TAY_V_vs(626) = 4.4718d0
  Mcase65TAY_V_vs(627) = 4.473d0
  Mcase65TAY_V_vs(628) = 4.4742d0
  Mcase65TAY_V_vs(629) = 4.4754d0
  Mcase65TAY_V_vs(630) = 4.4766d0
  Mcase65TAY_V_vs(631) = 4.4778d0
  Mcase65TAY_V_vs(632) = 4.479d0
  Mcase65TAY_V_vs(633) = 4.4802d0
  Mcase65TAY_V_vs(634) = 4.4814d0
  Mcase65TAY_V_vs(635) = 4.4827d0
  Mcase65TAY_V_vs(636) = 4.4841d0
  Mcase65TAY_V_vs(637) = 4.4855d0
  Mcase65TAY_V_vs(638) = 4.4869d0
  Mcase65TAY_V_vs(639) = 4.4881d0
  Mcase65TAY_V_vs(640) = 4.4885d0
  Mcase65TAY_V_vs(641) = 4.4889d0
  Mcase65TAY_V_vs(642) = 4.4893d0
  Mcase65TAY_V_vs(643) = 4.4897d0
  Mcase65TAY_V_vs(644) = 4.4893d0
  Mcase65TAY_V_vs(645) = 4.4862d0
  Mcase65TAY_V_vs(646) = 4.4831d0
  Mcase65TAY_V_vs(647) = 4.48d0
  Mcase65TAY_V_vs(648) = 4.4769d0
  Mcase65TAY_V_vs(649) = 4.4707d0
  Mcase65TAY_V_vs(650) = 4.4582d0
  Mcase65TAY_V_vs(651) = 4.4457d0
  Mcase65TAY_V_vs(652) = 4.4332d0
  Mcase65TAY_V_vs(653) = 4.4207d0
  Mcase65TAY_V_vs(654) = 4.3761d0
  Mcase65TAY_V_vs(655) = 4.2861d0
  Mcase65TAY_V_vs(656) = 4.196d0
  Mcase65TAY_V_vs(657) = 4.106d0
  Mcase65TAY_V_vs(658) = 4.016d0
  Mcase65TAY_V_vs(659) = 3.9072d0
  Mcase65TAY_V_vs(660) = 3.7796d0
  Mcase65TAY_V_vs(661) = 3.652d0
  Mcase65TAY_V_vs(662) = 3.554d0
  Mcase65TAY_V_vs(663) = 3.4858d0
  Mcase65TAY_V_vs(664) = 3.4175d0
  Mcase65TAY_V_vs(665) = 3.3453d0
  Mcase65TAY_V_vs(666) = 3.269d0
  Mcase65TAY_V_vs(667) = 3.1927d0
  Mcase65TAY_V_vs(668) = 3.0997d0
  Mcase65TAY_V_vs(669) = 2.9899d0
  Mcase65TAY_V_vs(670) = 2.8802d0
  Mcase65TAY_V_vs(671) = 2.7879d0
  Mcase65TAY_V_vs(672) = 2.7129d0
  Mcase65TAY_V_vs(673) = 2.638d0
  Mcase65TAY_V_vs(674) = 2.5861d0
  Mcase65TAY_V_vs(675) = 2.5574d0
  Mcase65TAY_V_vs(676) = 2.5287d0
  Mcase65TAY_V_vs(677) = 2.5078d0
  Mcase65TAY_V_vs(678) = 2.4948d0
  Mcase65TAY_V_vs(679) = 2.4818d0
  Mcase65TAY_V_vs(680) = 2.477d0
  Mcase65TAY_V_vs(681) = 2.4804d0

  Mcase65TAY_V_Qkappa(1) = 1327.7d0
  Mcase65TAY_V_Qkappa(2) = 1327.7d0
  Mcase65TAY_V_Qkappa(3) = 1327.7d0
  Mcase65TAY_V_Qkappa(4) = 1327.7d0
  Mcase65TAY_V_Qkappa(5) = 1327.7d0
  Mcase65TAY_V_Qkappa(6) = 1327.7d0
  Mcase65TAY_V_Qkappa(7) = 1327.7d0
  Mcase65TAY_V_Qkappa(8) = 1327.7d0
  Mcase65TAY_V_Qkappa(9) = 1327.7d0
  Mcase65TAY_V_Qkappa(10) = 1327.7d0
  Mcase65TAY_V_Qkappa(11) = 1327.7d0
  Mcase65TAY_V_Qkappa(12) = 1327.7d0
  Mcase65TAY_V_Qkappa(13) = 1327.7d0
  Mcase65TAY_V_Qkappa(14) = 1327.7d0
  Mcase65TAY_V_Qkappa(15) = 1327.7d0
  Mcase65TAY_V_Qkappa(16) = 1327.7d0
  Mcase65TAY_V_Qkappa(17) = 1327.7d0
  Mcase65TAY_V_Qkappa(18) = 1327.7d0
  Mcase65TAY_V_Qkappa(19) = 1327.7d0
  Mcase65TAY_V_Qkappa(20) = 1327.7d0
  Mcase65TAY_V_Qkappa(21) = 1327.7d0
  Mcase65TAY_V_Qkappa(22) = 1327.7d0
  Mcase65TAY_V_Qkappa(23) = 1327.7d0
  Mcase65TAY_V_Qkappa(24) = 1327.7d0
  Mcase65TAY_V_Qkappa(25) = 1327.7d0
  Mcase65TAY_V_Qkappa(26) = 1327.7d0
  Mcase65TAY_V_Qkappa(27) = 1327.7d0
  Mcase65TAY_V_Qkappa(28) = 1327.7d0
  Mcase65TAY_V_Qkappa(29) = 1327.7d0
  Mcase65TAY_V_Qkappa(30) = 1327.7d0
  Mcase65TAY_V_Qkappa(31) = 1327.7d0
  Mcase65TAY_V_Qkappa(32) = 1327.7d0
  Mcase65TAY_V_Qkappa(33) = 1327.7d0
  Mcase65TAY_V_Qkappa(34) = 1327.7d0
  Mcase65TAY_V_Qkappa(35) = 1327.7d0
  Mcase65TAY_V_Qkappa(36) = 1327.7d0
  Mcase65TAY_V_Qkappa(37) = 1327.7d0
  Mcase65TAY_V_Qkappa(38) = 1327.7d0
  Mcase65TAY_V_Qkappa(39) = 1327.7d0
  Mcase65TAY_V_Qkappa(40) = 1327.7d0
  Mcase65TAY_V_Qkappa(41) = 1327.7d0
  Mcase65TAY_V_Qkappa(42) = 1327.7d0
  Mcase65TAY_V_Qkappa(43) = 1327.7d0
  Mcase65TAY_V_Qkappa(44) = 1327.7d0
  Mcase65TAY_V_Qkappa(45) = 1327.7d0
  Mcase65TAY_V_Qkappa(46) = 1327.7d0
  Mcase65TAY_V_Qkappa(47) = 1327.7d0
  Mcase65TAY_V_Qkappa(48) = 1327.7d0
  Mcase65TAY_V_Qkappa(49) = 1327.7d0
  Mcase65TAY_V_Qkappa(50) = 1327.7d0
  Mcase65TAY_V_Qkappa(51) = 1327.7d0
  Mcase65TAY_V_Qkappa(52) = 1327.7d0
  Mcase65TAY_V_Qkappa(53) = 1327.7d0
  Mcase65TAY_V_Qkappa(54) = 1327.7d0
  Mcase65TAY_V_Qkappa(55) = 1327.7d0
  Mcase65TAY_V_Qkappa(56) = 1327.7d0
  Mcase65TAY_V_Qkappa(57) = 1327.7d0
  Mcase65TAY_V_Qkappa(58) = 1327.7d0
  Mcase65TAY_V_Qkappa(59) = 1327.7d0
  Mcase65TAY_V_Qkappa(60) = 1327.7d0
  Mcase65TAY_V_Qkappa(61) = 1327.7d0
  Mcase65TAY_V_Qkappa(62) = 1327.7d0
  Mcase65TAY_V_Qkappa(63) = 1327.7d0
  Mcase65TAY_V_Qkappa(64) = 1327.7d0
  Mcase65TAY_V_Qkappa(65) = 1327.7d0
  Mcase65TAY_V_Qkappa(66) = 1327.7d0
  Mcase65TAY_V_Qkappa(67) = 1327.7d0
  Mcase65TAY_V_Qkappa(68) = 1327.7d0
  Mcase65TAY_V_Qkappa(69) = 1327.7d0
  Mcase65TAY_V_Qkappa(70) = 1327.7d0
  Mcase65TAY_V_Qkappa(71) = 1327.7d0
  Mcase65TAY_V_Qkappa(72) = 1327.7d0
  Mcase65TAY_V_Qkappa(73) = 1327.7d0
  Mcase65TAY_V_Qkappa(74) = 1327.7d0
  Mcase65TAY_V_Qkappa(75) = 1327.7d0
  Mcase65TAY_V_Qkappa(76) = 1327.7d0
  Mcase65TAY_V_Qkappa(77) = 1327.7d0
  Mcase65TAY_V_Qkappa(78) = 1327.7d0
  Mcase65TAY_V_Qkappa(79) = 1327.7d0
  Mcase65TAY_V_Qkappa(80) = 1327.7d0
  Mcase65TAY_V_Qkappa(81) = 1327.7d0
  Mcase65TAY_V_Qkappa(82) = 1327.7d0
  Mcase65TAY_V_Qkappa(83) = 1327.7d0
  Mcase65TAY_V_Qkappa(84) = 1327.7d0
  Mcase65TAY_V_Qkappa(85) = 1327.7d0
  Mcase65TAY_V_Qkappa(86) = 1327.7d0
  Mcase65TAY_V_Qkappa(87) = 1327.7d0
  Mcase65TAY_V_Qkappa(88) = 1327.7d0
  Mcase65TAY_V_Qkappa(89) = 1327.7d0
  Mcase65TAY_V_Qkappa(90) = 1327.7d0
  Mcase65TAY_V_Qkappa(91) = 1327.7d0
  Mcase65TAY_V_Qkappa(92) = 1327.7d0
  Mcase65TAY_V_Qkappa(93) = 1327.7d0
  Mcase65TAY_V_Qkappa(94) = 1327.7d0
  Mcase65TAY_V_Qkappa(95) = 1327.7d0
  Mcase65TAY_V_Qkappa(96) = 1327.7d0
  Mcase65TAY_V_Qkappa(97) = 1327.7d0
  Mcase65TAY_V_Qkappa(98) = 1327.7d0
  Mcase65TAY_V_Qkappa(99) = 1327.7d0
  Mcase65TAY_V_Qkappa(100) = 1327.7d0
  Mcase65TAY_V_Qkappa(101) = 1327.7d0
  Mcase65TAY_V_Qkappa(102) = 1327.7d0
  Mcase65TAY_V_Qkappa(103) = 1327.7d0
  Mcase65TAY_V_Qkappa(104) = 1327.7d0  ! ICB inner core
  Mcase65TAY_V_Qkappa(105) = 57827.0d0 ! ICB outer core
  Mcase65TAY_V_Qkappa(106) = 57827.0d0
  Mcase65TAY_V_Qkappa(107) = 57827.0d0
  Mcase65TAY_V_Qkappa(108) = 57827.0d0
  Mcase65TAY_V_Qkappa(109) = 57827.0d0
  Mcase65TAY_V_Qkappa(110) = 57827.0d0
  Mcase65TAY_V_Qkappa(111) = 57827.0d0
  Mcase65TAY_V_Qkappa(112) = 57827.0d0
  Mcase65TAY_V_Qkappa(113) = 57827.0d0
  Mcase65TAY_V_Qkappa(114) = 57827.0d0
  Mcase65TAY_V_Qkappa(115) = 57827.0d0
  Mcase65TAY_V_Qkappa(116) = 57827.0d0
  Mcase65TAY_V_Qkappa(117) = 57827.0d0
  Mcase65TAY_V_Qkappa(118) = 57827.0d0
  Mcase65TAY_V_Qkappa(119) = 57827.0d0
  Mcase65TAY_V_Qkappa(120) = 57827.0d0
  Mcase65TAY_V_Qkappa(121) = 57827.0d0
  Mcase65TAY_V_Qkappa(122) = 57827.0d0
  Mcase65TAY_V_Qkappa(123) = 57827.0d0
  Mcase65TAY_V_Qkappa(124) = 57827.0d0
  Mcase65TAY_V_Qkappa(125) = 57827.0d0
  Mcase65TAY_V_Qkappa(126) = 57827.0d0
  Mcase65TAY_V_Qkappa(127) = 57827.0d0
  Mcase65TAY_V_Qkappa(128) = 57827.0d0
  Mcase65TAY_V_Qkappa(129) = 57827.0d0
  Mcase65TAY_V_Qkappa(130) = 57827.0d0
  Mcase65TAY_V_Qkappa(131) = 57827.0d0
  Mcase65TAY_V_Qkappa(132) = 57827.0d0
  Mcase65TAY_V_Qkappa(133) = 57827.0d0
  Mcase65TAY_V_Qkappa(134) = 57827.0d0
  Mcase65TAY_V_Qkappa(135) = 57827.0d0
  Mcase65TAY_V_Qkappa(136) = 57827.0d0
  Mcase65TAY_V_Qkappa(137) = 57827.0d0
  Mcase65TAY_V_Qkappa(138) = 57827.0d0
  Mcase65TAY_V_Qkappa(139) = 57827.0d0
  Mcase65TAY_V_Qkappa(140) = 57827.0d0
  Mcase65TAY_V_Qkappa(141) = 57827.0d0
  Mcase65TAY_V_Qkappa(142) = 57827.0d0
  Mcase65TAY_V_Qkappa(143) = 57827.0d0
  Mcase65TAY_V_Qkappa(144) = 57827.0d0
  Mcase65TAY_V_Qkappa(145) = 57827.0d0
  Mcase65TAY_V_Qkappa(146) = 57827.0d0
  Mcase65TAY_V_Qkappa(147) = 57827.0d0
  Mcase65TAY_V_Qkappa(148) = 57827.0d0
  Mcase65TAY_V_Qkappa(149) = 57827.0d0
  Mcase65TAY_V_Qkappa(150) = 57827.0d0
  Mcase65TAY_V_Qkappa(151) = 57827.0d0
  Mcase65TAY_V_Qkappa(152) = 57827.0d0
  Mcase65TAY_V_Qkappa(153) = 57827.0d0
  Mcase65TAY_V_Qkappa(154) = 57827.0d0
  Mcase65TAY_V_Qkappa(155) = 57827.0d0
  Mcase65TAY_V_Qkappa(156) = 57827.0d0
  Mcase65TAY_V_Qkappa(157) = 57827.0d0
  Mcase65TAY_V_Qkappa(158) = 57827.0d0
  Mcase65TAY_V_Qkappa(159) = 57827.0d0
  Mcase65TAY_V_Qkappa(160) = 57827.0d0
  Mcase65TAY_V_Qkappa(161) = 57827.0d0
  Mcase65TAY_V_Qkappa(162) = 57827.0d0
  Mcase65TAY_V_Qkappa(163) = 57827.0d0
  Mcase65TAY_V_Qkappa(164) = 57827.0d0
  Mcase65TAY_V_Qkappa(165) = 57827.0d0
  Mcase65TAY_V_Qkappa(166) = 57827.0d0
  Mcase65TAY_V_Qkappa(167) = 57827.0d0
  Mcase65TAY_V_Qkappa(168) = 57827.0d0
  Mcase65TAY_V_Qkappa(169) = 57827.0d0
  Mcase65TAY_V_Qkappa(170) = 57827.0d0
  Mcase65TAY_V_Qkappa(171) = 57827.0d0
  Mcase65TAY_V_Qkappa(172) = 57827.0d0
  Mcase65TAY_V_Qkappa(173) = 57827.0d0
  Mcase65TAY_V_Qkappa(174) = 57827.0d0
  Mcase65TAY_V_Qkappa(175) = 57827.0d0
  Mcase65TAY_V_Qkappa(176) = 57827.0d0
  Mcase65TAY_V_Qkappa(177) = 57827.0d0
  Mcase65TAY_V_Qkappa(178) = 57827.0d0
  Mcase65TAY_V_Qkappa(179) = 57827.0d0
  Mcase65TAY_V_Qkappa(180) = 57827.0d0
  Mcase65TAY_V_Qkappa(181) = 57827.0d0
  Mcase65TAY_V_Qkappa(182) = 57827.0d0
  Mcase65TAY_V_Qkappa(183) = 57827.0d0
  Mcase65TAY_V_Qkappa(184) = 57827.0d0
  Mcase65TAY_V_Qkappa(185) = 57827.0d0
  Mcase65TAY_V_Qkappa(186) = 57827.0d0
  Mcase65TAY_V_Qkappa(187) = 57827.0d0
  Mcase65TAY_V_Qkappa(188) = 57827.0d0
  Mcase65TAY_V_Qkappa(189) = 57827.0d0
  Mcase65TAY_V_Qkappa(190) = 57827.0d0
  Mcase65TAY_V_Qkappa(191) = 57827.0d0
  Mcase65TAY_V_Qkappa(192) = 57827.0d0
  Mcase65TAY_V_Qkappa(193) = 57827.0d0
  Mcase65TAY_V_Qkappa(194) = 57827.0d0
  Mcase65TAY_V_Qkappa(195) = 57827.0d0
  Mcase65TAY_V_Qkappa(196) = 57827.0d0
  Mcase65TAY_V_Qkappa(197) = 57827.0d0
  Mcase65TAY_V_Qkappa(198) = 57827.0d0
  Mcase65TAY_V_Qkappa(199) = 57827.0d0
  Mcase65TAY_V_Qkappa(200) = 57827.0d0
  Mcase65TAY_V_Qkappa(201) = 57827.0d0
  Mcase65TAY_V_Qkappa(202) = 57827.0d0
  Mcase65TAY_V_Qkappa(203) = 57827.0d0
  Mcase65TAY_V_Qkappa(204) = 57827.0d0
  Mcase65TAY_V_Qkappa(205) = 57827.0d0
  Mcase65TAY_V_Qkappa(206) = 57827.0d0
  Mcase65TAY_V_Qkappa(207) = 57827.0d0
  Mcase65TAY_V_Qkappa(208) = 57827.0d0
  Mcase65TAY_V_Qkappa(209) = 57827.0d0
  Mcase65TAY_V_Qkappa(210) = 57827.0d0
  Mcase65TAY_V_Qkappa(211) = 57827.0d0
  Mcase65TAY_V_Qkappa(212) = 57827.0d0
  Mcase65TAY_V_Qkappa(213) = 57827.0d0
  Mcase65TAY_V_Qkappa(214) = 57827.0d0
  Mcase65TAY_V_Qkappa(215) = 57827.0d0
  Mcase65TAY_V_Qkappa(216) = 57827.0d0
  Mcase65TAY_V_Qkappa(217) = 57827.0d0
  Mcase65TAY_V_Qkappa(218) = 57827.0d0
  Mcase65TAY_V_Qkappa(219) = 57827.0d0
  Mcase65TAY_V_Qkappa(220) = 57827.0d0
  Mcase65TAY_V_Qkappa(221) = 57827.0d0
  Mcase65TAY_V_Qkappa(222) = 57827.0d0
  Mcase65TAY_V_Qkappa(223) = 57827.0d0
  Mcase65TAY_V_Qkappa(224) = 57827.0d0
  Mcase65TAY_V_Qkappa(225) = 57827.0d0
  Mcase65TAY_V_Qkappa(226) = 57827.0d0
  Mcase65TAY_V_Qkappa(227) = 57827.0d0
  Mcase65TAY_V_Qkappa(228) = 57827.0d0
  Mcase65TAY_V_Qkappa(229) = 57827.0d0
  Mcase65TAY_V_Qkappa(230) = 57827.0d0
  Mcase65TAY_V_Qkappa(231) = 57827.0d0
  Mcase65TAY_V_Qkappa(232) = 57827.0d0
  Mcase65TAY_V_Qkappa(233) = 57827.0d0
  Mcase65TAY_V_Qkappa(234) = 57827.0d0
  Mcase65TAY_V_Qkappa(235) = 57827.0d0
  Mcase65TAY_V_Qkappa(236) = 57827.0d0
  Mcase65TAY_V_Qkappa(237) = 57827.0d0
  Mcase65TAY_V_Qkappa(238) = 57827.0d0
  Mcase65TAY_V_Qkappa(239) = 57827.0d0
  Mcase65TAY_V_Qkappa(240) = 57827.0d0
  Mcase65TAY_V_Qkappa(241) = 57827.0d0
  Mcase65TAY_V_Qkappa(242) = 57827.0d0
  Mcase65TAY_V_Qkappa(243) = 57827.0d0
  Mcase65TAY_V_Qkappa(244) = 57827.0d0
  Mcase65TAY_V_Qkappa(245) = 57827.0d0
  Mcase65TAY_V_Qkappa(246) = 57827.0d0
  Mcase65TAY_V_Qkappa(247) = 57827.0d0
  Mcase65TAY_V_Qkappa(248) = 57827.0d0
  Mcase65TAY_V_Qkappa(249) = 57827.0d0
  Mcase65TAY_V_Qkappa(250) = 57827.0d0
  Mcase65TAY_V_Qkappa(251) = 57827.0d0
  Mcase65TAY_V_Qkappa(252) = 57827.0d0
  Mcase65TAY_V_Qkappa(253) = 57827.0d0
  Mcase65TAY_V_Qkappa(254) = 57827.0d0
  Mcase65TAY_V_Qkappa(255) = 57827.0d0
  Mcase65TAY_V_Qkappa(256) = 57827.0d0
  Mcase65TAY_V_Qkappa(257) = 57827.0d0
  Mcase65TAY_V_Qkappa(258) = 57827.0d0
  Mcase65TAY_V_Qkappa(259) = 57827.0d0
  Mcase65TAY_V_Qkappa(260) = 57827.0d0
  Mcase65TAY_V_Qkappa(261) = 57827.0d0
  Mcase65TAY_V_Qkappa(262) = 57827.0d0
  Mcase65TAY_V_Qkappa(263) = 57827.0d0
  Mcase65TAY_V_Qkappa(264) = 57827.0d0
  Mcase65TAY_V_Qkappa(265) = 57827.0d0
  Mcase65TAY_V_Qkappa(266) = 57827.0d0
  Mcase65TAY_V_Qkappa(267) = 57827.0d0
  Mcase65TAY_V_Qkappa(268) = 57827.0d0
  Mcase65TAY_V_Qkappa(269) = 57827.0d0
  Mcase65TAY_V_Qkappa(270) = 57827.0d0
  Mcase65TAY_V_Qkappa(271) = 57827.0d0
  Mcase65TAY_V_Qkappa(272) = 57827.0d0
  Mcase65TAY_V_Qkappa(273) = 57827.0d0
  Mcase65TAY_V_Qkappa(274) = 57827.0d0
  Mcase65TAY_V_Qkappa(275) = 57827.0d0
  Mcase65TAY_V_Qkappa(276) = 57827.0d0
  Mcase65TAY_V_Qkappa(277) = 57827.0d0
  Mcase65TAY_V_Qkappa(278) = 57827.0d0
  Mcase65TAY_V_Qkappa(279) = 57827.0d0
  Mcase65TAY_V_Qkappa(280) = 57827.0d0
  Mcase65TAY_V_Qkappa(281) = 57827.0d0
  Mcase65TAY_V_Qkappa(282) = 57827.0d0
  Mcase65TAY_V_Qkappa(283) = 57827.0d0
  Mcase65TAY_V_Qkappa(284) = 57827.0d0
  Mcase65TAY_V_Qkappa(285) = 57827.0d0
  Mcase65TAY_V_Qkappa(286) = 57827.0d0
  Mcase65TAY_V_Qkappa(287) = 57827.0d0
  Mcase65TAY_V_Qkappa(288) = 57827.0d0
  Mcase65TAY_V_Qkappa(289) = 57827.0d0
  Mcase65TAY_V_Qkappa(290) = 57827.0d0
  Mcase65TAY_V_Qkappa(291) = 57827.0d0
  Mcase65TAY_V_Qkappa(292) = 57827.0d0
  Mcase65TAY_V_Qkappa(293) = 57827.0d0
  Mcase65TAY_V_Qkappa(294) = 57827.0d0
  Mcase65TAY_V_Qkappa(295) = 57827.0d0
  Mcase65TAY_V_Qkappa(296) = 57827.0d0
  Mcase65TAY_V_Qkappa(297) = 57827.0d0
  Mcase65TAY_V_Qkappa(298) = 57827.0d0
  Mcase65TAY_V_Qkappa(299) = 57827.0d0
  Mcase65TAY_V_Qkappa(300) = 57827.0d0
  Mcase65TAY_V_Qkappa(301) = 57827.0d0
  Mcase65TAY_V_Qkappa(302) = 57827.0d0
  Mcase65TAY_V_Qkappa(303) = 57827.0d0
  Mcase65TAY_V_Qkappa(304) = 57827.0d0
  Mcase65TAY_V_Qkappa(305) = 57827.0d0
  Mcase65TAY_V_Qkappa(306) = 57827.0d0
  Mcase65TAY_V_Qkappa(307) = 57827.0d0
  Mcase65TAY_V_Qkappa(308) = 57827.0d0
  Mcase65TAY_V_Qkappa(309) = 57827.0d0
  Mcase65TAY_V_Qkappa(310) = 57827.0d0
  Mcase65TAY_V_Qkappa(311) = 57827.0d0
  Mcase65TAY_V_Qkappa(312) = 57827.0d0
  Mcase65TAY_V_Qkappa(313) = 57827.0d0
  Mcase65TAY_V_Qkappa(314) = 57827.0d0
  Mcase65TAY_V_Qkappa(315) = 57827.0d0
  Mcase65TAY_V_Qkappa(316) = 57827.0d0
  Mcase65TAY_V_Qkappa(317) = 57827.0d0
  Mcase65TAY_V_Qkappa(318) = 57827.0d0
  Mcase65TAY_V_Qkappa(319) = 57827.0d0
  Mcase65TAY_V_Qkappa(320) = 57827.0d0
  Mcase65TAY_V_Qkappa(321) = 57827.0d0
  Mcase65TAY_V_Qkappa(322) = 57827.0d0
  Mcase65TAY_V_Qkappa(323) = 57827.0d0
  Mcase65TAY_V_Qkappa(324) = 57827.0d0
  Mcase65TAY_V_Qkappa(325) = 57827.0d0
  Mcase65TAY_V_Qkappa(326) = 57827.0d0
  Mcase65TAY_V_Qkappa(327) = 57827.0d0
  Mcase65TAY_V_Qkappa(328) = 57827.0d0
  Mcase65TAY_V_Qkappa(329) = 57827.0d0
  Mcase65TAY_V_Qkappa(330) = 57827.0d0
  Mcase65TAY_V_Qkappa(331) = 57827.0d0
  Mcase65TAY_V_Qkappa(332) = 57827.0d0
  Mcase65TAY_V_Qkappa(333) = 57827.0d0
  Mcase65TAY_V_Qkappa(334) = 57827.0d0
  Mcase65TAY_V_Qkappa(335) = 57827.0d0
  Mcase65TAY_V_Qkappa(336) = 57827.0d0
  Mcase65TAY_V_Qkappa(337) = 57827.0d0
  Mcase65TAY_V_Qkappa(338) = 57827.0d0
  Mcase65TAY_V_Qkappa(339) = 57827.0d0
  Mcase65TAY_V_Qkappa(340) = 57827.0d0
  Mcase65TAY_V_Qkappa(341) = 57827.0d0
  Mcase65TAY_V_Qkappa(342) = 57827.0d0
  Mcase65TAY_V_Qkappa(343) = 57827.0d0
  Mcase65TAY_V_Qkappa(344) = 57827.0d0
  Mcase65TAY_V_Qkappa(345) = 57827.0d0
  Mcase65TAY_V_Qkappa(346) = 57827.0d0
  Mcase65TAY_V_Qkappa(347) = 57827.0d0
  Mcase65TAY_V_Qkappa(348) = 57827.0d0
  Mcase65TAY_V_Qkappa(349) = 57827.0d0
  Mcase65TAY_V_Qkappa(350) = 57827.0d0
  Mcase65TAY_V_Qkappa(351) = 57827.0d0
  Mcase65TAY_V_Qkappa(352) = 57827.0d0
  Mcase65TAY_V_Qkappa(353) = 57827.0d0
  Mcase65TAY_V_Qkappa(354) = 57827.0d0
  Mcase65TAY_V_Qkappa(355) = 57827.0d0
  Mcase65TAY_V_Qkappa(356) = 57827.0d0
  Mcase65TAY_V_Qkappa(357) = 57827.0d0
  Mcase65TAY_V_Qkappa(358) = 57827.0d0
  Mcase65TAY_V_Qkappa(359) = 57827.0d0
  Mcase65TAY_V_Qkappa(360) = 57827.0d0
  Mcase65TAY_V_Qkappa(361) = 57827.0d0
  Mcase65TAY_V_Qkappa(362) = 57827.0d0
  Mcase65TAY_V_Qkappa(363) = 57827.0d0
  Mcase65TAY_V_Qkappa(364) = 57827.0d0
  Mcase65TAY_V_Qkappa(365) = 57827.0d0
  Mcase65TAY_V_Qkappa(366) = 57827.0d0
  Mcase65TAY_V_Qkappa(367) = 57827.0d0
  Mcase65TAY_V_Qkappa(368) = 57827.0d0
  Mcase65TAY_V_Qkappa(369) = 57827.0d0
  Mcase65TAY_V_Qkappa(370) = 57827.0d0 ! CMB outer core
  Mcase65TAY_V_Qkappa(371) = 57827.0d0 ! CMB mantle
  Mcase65TAY_V_Qkappa(372) = 57827.0d0
  Mcase65TAY_V_Qkappa(373) = 57827.0d0
  Mcase65TAY_V_Qkappa(374) = 57827.0d0
  Mcase65TAY_V_Qkappa(375) = 57827.0d0
  Mcase65TAY_V_Qkappa(376) = 57827.0d0
  Mcase65TAY_V_Qkappa(377) = 57827.0d0
  Mcase65TAY_V_Qkappa(378) = 57827.0d0
  Mcase65TAY_V_Qkappa(379) = 57827.0d0
  Mcase65TAY_V_Qkappa(380) = 57827.0d0
  Mcase65TAY_V_Qkappa(381) = 57827.0d0
  Mcase65TAY_V_Qkappa(382) = 57827.0d0
  Mcase65TAY_V_Qkappa(383) = 57827.0d0
  Mcase65TAY_V_Qkappa(384) = 57827.0d0
  Mcase65TAY_V_Qkappa(385) = 57827.0d0
  Mcase65TAY_V_Qkappa(386) = 57827.0d0
  Mcase65TAY_V_Qkappa(387) = 57827.0d0
  Mcase65TAY_V_Qkappa(388) = 57827.0d0
  Mcase65TAY_V_Qkappa(389) = 57827.0d0
  Mcase65TAY_V_Qkappa(390) = 57827.0d0
  Mcase65TAY_V_Qkappa(391) = 57827.0d0
  Mcase65TAY_V_Qkappa(392) = 57827.0d0
  Mcase65TAY_V_Qkappa(393) = 57827.0d0
  Mcase65TAY_V_Qkappa(394) = 57827.0d0
  Mcase65TAY_V_Qkappa(395) = 57827.0d0
  Mcase65TAY_V_Qkappa(396) = 57827.0d0
  Mcase65TAY_V_Qkappa(397) = 57827.0d0
  Mcase65TAY_V_Qkappa(398) = 57827.0d0
  Mcase65TAY_V_Qkappa(399) = 57827.0d0
  Mcase65TAY_V_Qkappa(400) = 57827.0d0
  Mcase65TAY_V_Qkappa(401) = 57827.0d0
  Mcase65TAY_V_Qkappa(402) = 57827.0d0
  Mcase65TAY_V_Qkappa(403) = 57827.0d0
  Mcase65TAY_V_Qkappa(404) = 57827.0d0
  Mcase65TAY_V_Qkappa(405) = 57827.0d0
  Mcase65TAY_V_Qkappa(406) = 57827.0d0
  Mcase65TAY_V_Qkappa(407) = 57827.0d0
  Mcase65TAY_V_Qkappa(408) = 57827.0d0
  Mcase65TAY_V_Qkappa(409) = 57827.0d0
  Mcase65TAY_V_Qkappa(410) = 57827.0d0
  Mcase65TAY_V_Qkappa(411) = 57827.0d0
  Mcase65TAY_V_Qkappa(412) = 57827.0d0
  Mcase65TAY_V_Qkappa(413) = 57827.0d0
  Mcase65TAY_V_Qkappa(414) = 57827.0d0
  Mcase65TAY_V_Qkappa(415) = 57827.0d0
  Mcase65TAY_V_Qkappa(416) = 57827.0d0
  Mcase65TAY_V_Qkappa(417) = 57827.0d0
  Mcase65TAY_V_Qkappa(418) = 57827.0d0
  Mcase65TAY_V_Qkappa(419) = 57827.0d0
  Mcase65TAY_V_Qkappa(420) = 57827.0d0
  Mcase65TAY_V_Qkappa(421) = 57827.0d0
  Mcase65TAY_V_Qkappa(422) = 57827.0d0
  Mcase65TAY_V_Qkappa(423) = 57827.0d0
  Mcase65TAY_V_Qkappa(424) = 57827.0d0
  Mcase65TAY_V_Qkappa(425) = 57827.0d0
  Mcase65TAY_V_Qkappa(426) = 57827.0d0
  Mcase65TAY_V_Qkappa(427) = 57827.0d0
  Mcase65TAY_V_Qkappa(428) = 57827.0d0
  Mcase65TAY_V_Qkappa(429) = 57827.0d0
  Mcase65TAY_V_Qkappa(430) = 57827.0d0
  Mcase65TAY_V_Qkappa(431) = 57827.0d0
  Mcase65TAY_V_Qkappa(432) = 57827.0d0
  Mcase65TAY_V_Qkappa(433) = 57827.0d0
  Mcase65TAY_V_Qkappa(434) = 57827.0d0
  Mcase65TAY_V_Qkappa(435) = 57827.0d0
  Mcase65TAY_V_Qkappa(436) = 57827.0d0
  Mcase65TAY_V_Qkappa(437) = 57827.0d0
  Mcase65TAY_V_Qkappa(438) = 57827.0d0
  Mcase65TAY_V_Qkappa(439) = 57827.0d0
  Mcase65TAY_V_Qkappa(440) = 57827.0d0
  Mcase65TAY_V_Qkappa(441) = 57827.0d0
  Mcase65TAY_V_Qkappa(442) = 57827.0d0
  Mcase65TAY_V_Qkappa(443) = 57827.0d0
  Mcase65TAY_V_Qkappa(444) = 57827.0d0
  Mcase65TAY_V_Qkappa(445) = 57827.0d0
  Mcase65TAY_V_Qkappa(446) = 57827.0d0
  Mcase65TAY_V_Qkappa(447) = 57827.0d0
  Mcase65TAY_V_Qkappa(448) = 57827.0d0
  Mcase65TAY_V_Qkappa(449) = 57827.0d0
  Mcase65TAY_V_Qkappa(450) = 57827.0d0
  Mcase65TAY_V_Qkappa(451) = 57827.0d0
  Mcase65TAY_V_Qkappa(452) = 57827.0d0
  Mcase65TAY_V_Qkappa(453) = 57827.0d0
  Mcase65TAY_V_Qkappa(454) = 57827.0d0
  Mcase65TAY_V_Qkappa(455) = 57827.0d0
  Mcase65TAY_V_Qkappa(456) = 57827.0d0
  Mcase65TAY_V_Qkappa(457) = 57827.0d0
  Mcase65TAY_V_Qkappa(458) = 57827.0d0
  Mcase65TAY_V_Qkappa(459) = 57827.0d0
  Mcase65TAY_V_Qkappa(460) = 57827.0d0
  Mcase65TAY_V_Qkappa(461) = 57827.0d0
  Mcase65TAY_V_Qkappa(462) = 57827.0d0
  Mcase65TAY_V_Qkappa(463) = 57827.0d0
  Mcase65TAY_V_Qkappa(464) = 57827.0d0
  Mcase65TAY_V_Qkappa(465) = 57827.0d0
  Mcase65TAY_V_Qkappa(466) = 57827.0d0
  Mcase65TAY_V_Qkappa(467) = 57827.0d0
  Mcase65TAY_V_Qkappa(468) = 57827.0d0
  Mcase65TAY_V_Qkappa(469) = 57827.0d0
  Mcase65TAY_V_Qkappa(470) = 57827.0d0
  Mcase65TAY_V_Qkappa(471) = 57827.0d0
  Mcase65TAY_V_Qkappa(472) = 57827.0d0
  Mcase65TAY_V_Qkappa(473) = 57827.0d0
  Mcase65TAY_V_Qkappa(474) = 57827.0d0
  Mcase65TAY_V_Qkappa(475) = 57827.0d0
  Mcase65TAY_V_Qkappa(476) = 57827.0d0
  Mcase65TAY_V_Qkappa(477) = 57827.0d0
  Mcase65TAY_V_Qkappa(478) = 57827.0d0
  Mcase65TAY_V_Qkappa(479) = 57827.0d0
  Mcase65TAY_V_Qkappa(480) = 57827.0d0
  Mcase65TAY_V_Qkappa(481) = 57827.0d0
  Mcase65TAY_V_Qkappa(482) = 57827.0d0
  Mcase65TAY_V_Qkappa(483) = 57827.0d0
  Mcase65TAY_V_Qkappa(484) = 57827.0d0
  Mcase65TAY_V_Qkappa(485) = 57827.0d0
  Mcase65TAY_V_Qkappa(486) = 57827.0d0
  Mcase65TAY_V_Qkappa(487) = 57827.0d0
  Mcase65TAY_V_Qkappa(488) = 57827.0d0
  Mcase65TAY_V_Qkappa(489) = 57827.0d0
  Mcase65TAY_V_Qkappa(490) = 57827.0d0
  Mcase65TAY_V_Qkappa(491) = 57827.0d0
  Mcase65TAY_V_Qkappa(492) = 57827.0d0
  Mcase65TAY_V_Qkappa(493) = 57827.0d0
  Mcase65TAY_V_Qkappa(494) = 57827.0d0
  Mcase65TAY_V_Qkappa(495) = 57827.0d0
  Mcase65TAY_V_Qkappa(496) = 57827.0d0
  Mcase65TAY_V_Qkappa(497) = 57827.0d0
  Mcase65TAY_V_Qkappa(498) = 57827.0d0
  Mcase65TAY_V_Qkappa(499) = 57827.0d0
  Mcase65TAY_V_Qkappa(500) = 57827.0d0
  Mcase65TAY_V_Qkappa(501) = 57827.0d0
  Mcase65TAY_V_Qkappa(502) = 57827.0d0
  Mcase65TAY_V_Qkappa(503) = 57827.0d0
  Mcase65TAY_V_Qkappa(504) = 57827.0d0
  Mcase65TAY_V_Qkappa(505) = 57827.0d0
  Mcase65TAY_V_Qkappa(506) = 57827.0d0
  Mcase65TAY_V_Qkappa(507) = 57827.0d0
  Mcase65TAY_V_Qkappa(508) = 57827.0d0
  Mcase65TAY_V_Qkappa(509) = 57827.0d0
  Mcase65TAY_V_Qkappa(510) = 57827.0d0
  Mcase65TAY_V_Qkappa(511) = 57827.0d0
  Mcase65TAY_V_Qkappa(512) = 57827.0d0
  Mcase65TAY_V_Qkappa(513) = 57827.0d0
  Mcase65TAY_V_Qkappa(514) = 57827.0d0
  Mcase65TAY_V_Qkappa(515) = 57827.0d0
  Mcase65TAY_V_Qkappa(516) = 57827.0d0
  Mcase65TAY_V_Qkappa(517) = 57827.0d0
  Mcase65TAY_V_Qkappa(518) = 57827.0d0
  Mcase65TAY_V_Qkappa(519) = 57827.0d0
  Mcase65TAY_V_Qkappa(520) = 57827.0d0
  Mcase65TAY_V_Qkappa(521) = 57827.0d0
  Mcase65TAY_V_Qkappa(522) = 57827.0d0
  Mcase65TAY_V_Qkappa(523) = 57827.0d0
  Mcase65TAY_V_Qkappa(524) = 57827.0d0
  Mcase65TAY_V_Qkappa(525) = 57827.0d0
  Mcase65TAY_V_Qkappa(526) = 57827.0d0
  Mcase65TAY_V_Qkappa(527) = 57827.0d0
  Mcase65TAY_V_Qkappa(528) = 57827.0d0
  Mcase65TAY_V_Qkappa(529) = 57827.0d0
  Mcase65TAY_V_Qkappa(530) = 57827.0d0
  Mcase65TAY_V_Qkappa(531) = 57827.0d0
  Mcase65TAY_V_Qkappa(532) = 57827.0d0
  Mcase65TAY_V_Qkappa(533) = 57827.0d0
  Mcase65TAY_V_Qkappa(534) = 57827.0d0
  Mcase65TAY_V_Qkappa(535) = 57827.0d0
  Mcase65TAY_V_Qkappa(536) = 57827.0d0
  Mcase65TAY_V_Qkappa(537) = 57827.0d0
  Mcase65TAY_V_Qkappa(538) = 57827.0d0
  Mcase65TAY_V_Qkappa(539) = 57827.0d0
  Mcase65TAY_V_Qkappa(540) = 57827.0d0
  Mcase65TAY_V_Qkappa(541) = 57827.0d0
  Mcase65TAY_V_Qkappa(542) = 57827.0d0
  Mcase65TAY_V_Qkappa(543) = 57827.0d0
  Mcase65TAY_V_Qkappa(544) = 57827.0d0
  Mcase65TAY_V_Qkappa(545) = 57827.0d0
  Mcase65TAY_V_Qkappa(546) = 57827.0d0
  Mcase65TAY_V_Qkappa(547) = 57827.0d0
  Mcase65TAY_V_Qkappa(548) = 57827.0d0
  Mcase65TAY_V_Qkappa(549) = 57827.0d0
  Mcase65TAY_V_Qkappa(550) = 57827.0d0
  Mcase65TAY_V_Qkappa(551) = 57827.0d0
  Mcase65TAY_V_Qkappa(552) = 57827.0d0
  Mcase65TAY_V_Qkappa(553) = 57827.0d0
  Mcase65TAY_V_Qkappa(554) = 57827.0d0
  Mcase65TAY_V_Qkappa(555) = 57827.0d0
  Mcase65TAY_V_Qkappa(556) = 57827.0d0
  Mcase65TAY_V_Qkappa(557) = 57827.0d0
  Mcase65TAY_V_Qkappa(558) = 57827.0d0
  Mcase65TAY_V_Qkappa(559) = 57827.0d0
  Mcase65TAY_V_Qkappa(560) = 57827.0d0
  Mcase65TAY_V_Qkappa(561) = 57827.0d0
  Mcase65TAY_V_Qkappa(562) = 57827.0d0
  Mcase65TAY_V_Qkappa(563) = 57827.0d0
  Mcase65TAY_V_Qkappa(564) = 57827.0d0
  Mcase65TAY_V_Qkappa(565) = 57827.0d0
  Mcase65TAY_V_Qkappa(566) = 57827.0d0
  Mcase65TAY_V_Qkappa(567) = 57827.0d0
  Mcase65TAY_V_Qkappa(568) = 57827.0d0
  Mcase65TAY_V_Qkappa(569) = 57827.0d0
  Mcase65TAY_V_Qkappa(570) = 57827.0d0
  Mcase65TAY_V_Qkappa(571) = 57827.0d0
  Mcase65TAY_V_Qkappa(572) = 57827.0d0
  Mcase65TAY_V_Qkappa(573) = 57827.0d0
  Mcase65TAY_V_Qkappa(574) = 57827.0d0
  Mcase65TAY_V_Qkappa(575) = 57827.0d0
  Mcase65TAY_V_Qkappa(576) = 57827.0d0
  Mcase65TAY_V_Qkappa(577) = 57827.0d0
  Mcase65TAY_V_Qkappa(578) = 57827.0d0
  Mcase65TAY_V_Qkappa(579) = 57827.0d0
  Mcase65TAY_V_Qkappa(580) = 57827.0d0
  Mcase65TAY_V_Qkappa(581) = 57827.0d0
  Mcase65TAY_V_Qkappa(582) = 57827.0d0
  Mcase65TAY_V_Qkappa(583) = 57827.0d0
  Mcase65TAY_V_Qkappa(584) = 57827.0d0
  Mcase65TAY_V_Qkappa(585) = 57827.0d0
  Mcase65TAY_V_Qkappa(586) = 57827.0d0
  Mcase65TAY_V_Qkappa(587) = 57827.0d0
  Mcase65TAY_V_Qkappa(588) = 57827.0d0
  Mcase65TAY_V_Qkappa(589) = 57827.0d0
  Mcase65TAY_V_Qkappa(590) = 57827.0d0
  Mcase65TAY_V_Qkappa(591) = 57827.0d0
  Mcase65TAY_V_Qkappa(592) = 57827.0d0
  Mcase65TAY_V_Qkappa(593) = 57827.0d0
  Mcase65TAY_V_Qkappa(594) = 57827.0d0
  Mcase65TAY_V_Qkappa(595) = 57827.0d0
  Mcase65TAY_V_Qkappa(596) = 57827.0d0
  Mcase65TAY_V_Qkappa(597) = 57827.0d0
  Mcase65TAY_V_Qkappa(598) = 57827.0d0
  Mcase65TAY_V_Qkappa(599) = 57827.0d0
  Mcase65TAY_V_Qkappa(600) = 57827.0d0
  Mcase65TAY_V_Qkappa(601) = 57827.0d0
  Mcase65TAY_V_Qkappa(602) = 57827.0d0
  Mcase65TAY_V_Qkappa(603) = 57827.0d0
  Mcase65TAY_V_Qkappa(604) = 57827.0d0
  Mcase65TAY_V_Qkappa(605) = 57827.0d0
  Mcase65TAY_V_Qkappa(606) = 57827.0d0
  Mcase65TAY_V_Qkappa(607) = 57827.0d0
  Mcase65TAY_V_Qkappa(608) = 57827.0d0
  Mcase65TAY_V_Qkappa(609) = 57827.0d0
  Mcase65TAY_V_Qkappa(610) = 57827.0d0
  Mcase65TAY_V_Qkappa(611) = 57827.0d0
  Mcase65TAY_V_Qkappa(612) = 57827.0d0
  Mcase65TAY_V_Qkappa(613) = 57827.0d0
  Mcase65TAY_V_Qkappa(614) = 57827.0d0
  Mcase65TAY_V_Qkappa(615) = 57827.0d0
  Mcase65TAY_V_Qkappa(616) = 57827.0d0
  Mcase65TAY_V_Qkappa(617) = 57827.0d0
  Mcase65TAY_V_Qkappa(618) = 57827.0d0
  Mcase65TAY_V_Qkappa(619) = 57827.0d0
  Mcase65TAY_V_Qkappa(620) = 57827.0d0
  Mcase65TAY_V_Qkappa(621) = 57827.0d0
  Mcase65TAY_V_Qkappa(622) = 57827.0d0
  Mcase65TAY_V_Qkappa(623) = 57827.0d0
  Mcase65TAY_V_Qkappa(624) = 57827.0d0
  Mcase65TAY_V_Qkappa(625) = 57827.0d0
  Mcase65TAY_V_Qkappa(626) = 57827.0d0
  Mcase65TAY_V_Qkappa(627) = 57827.0d0
  Mcase65TAY_V_Qkappa(628) = 57827.0d0
  Mcase65TAY_V_Qkappa(629) = 57827.0d0
  Mcase65TAY_V_Qkappa(630) = 57827.0d0
  Mcase65TAY_V_Qkappa(631) = 57827.0d0
  Mcase65TAY_V_Qkappa(632) = 57827.0d0
  Mcase65TAY_V_Qkappa(633) = 57827.0d0
  Mcase65TAY_V_Qkappa(634) = 57827.0d0
  Mcase65TAY_V_Qkappa(635) = 57827.0d0
  Mcase65TAY_V_Qkappa(636) = 57827.0d0
  Mcase65TAY_V_Qkappa(637) = 57827.0d0
  Mcase65TAY_V_Qkappa(638) = 57827.0d0
  Mcase65TAY_V_Qkappa(639) = 57827.0d0
  Mcase65TAY_V_Qkappa(640) = 57827.0d0
  Mcase65TAY_V_Qkappa(641) = 57827.0d0
  Mcase65TAY_V_Qkappa(642) = 57827.0d0
  Mcase65TAY_V_Qkappa(643) = 57827.0d0
  Mcase65TAY_V_Qkappa(644) = 57827.0d0
  Mcase65TAY_V_Qkappa(645) = 57827.0d0
  Mcase65TAY_V_Qkappa(646) = 57827.0d0
  Mcase65TAY_V_Qkappa(647) = 57827.0d0
  Mcase65TAY_V_Qkappa(648) = 57827.0d0
  Mcase65TAY_V_Qkappa(649) = 57827.0d0
  Mcase65TAY_V_Qkappa(650) = 57827.0d0
  Mcase65TAY_V_Qkappa(651) = 57827.0d0
  Mcase65TAY_V_Qkappa(652) = 57827.0d0
  Mcase65TAY_V_Qkappa(653) = 57827.0d0
  Mcase65TAY_V_Qkappa(654) = 57827.0d0
  Mcase65TAY_V_Qkappa(655) = 57827.0d0
  Mcase65TAY_V_Qkappa(656) = 57827.0d0
  Mcase65TAY_V_Qkappa(657) = 57827.0d0
  Mcase65TAY_V_Qkappa(658) = 57827.0d0
  Mcase65TAY_V_Qkappa(659) = 57827.0d0
  Mcase65TAY_V_Qkappa(660) = 57827.0d0
  Mcase65TAY_V_Qkappa(661) = 57827.0d0
  Mcase65TAY_V_Qkappa(662) = 57827.0d0
  Mcase65TAY_V_Qkappa(663) = 57827.0d0
  Mcase65TAY_V_Qkappa(664) = 57827.0d0
  Mcase65TAY_V_Qkappa(665) = 57827.0d0
  Mcase65TAY_V_Qkappa(666) = 57827.0d0
  Mcase65TAY_V_Qkappa(667) = 57827.0d0
  Mcase65TAY_V_Qkappa(668) = 57827.0d0
  Mcase65TAY_V_Qkappa(669) = 57827.0d0
  Mcase65TAY_V_Qkappa(670) = 57827.0d0
  Mcase65TAY_V_Qkappa(671) = 57827.0d0
  Mcase65TAY_V_Qkappa(672) = 57827.0d0
  Mcase65TAY_V_Qkappa(673) = 57827.0d0
  Mcase65TAY_V_Qkappa(674) = 57827.0d0
  Mcase65TAY_V_Qkappa(675) = 57827.0d0
  Mcase65TAY_V_Qkappa(676) = 57827.0d0
  Mcase65TAY_V_Qkappa(677) = 57827.0d0
  Mcase65TAY_V_Qkappa(678) = 57827.0d0
  Mcase65TAY_V_Qkappa(679) = 57827.0d0
  Mcase65TAY_V_Qkappa(680) = 57827.0d0
  Mcase65TAY_V_Qkappa(681) = 57827.0d0

  Mcase65TAY_V_Qmu(1) = 84.6d0
  Mcase65TAY_V_Qmu(2) = 84.6d0
  Mcase65TAY_V_Qmu(3) = 84.6d0
  Mcase65TAY_V_Qmu(4) = 84.6d0
  Mcase65TAY_V_Qmu(5) = 84.6d0
  Mcase65TAY_V_Qmu(6) = 84.6d0
  Mcase65TAY_V_Qmu(7) = 84.6d0
  Mcase65TAY_V_Qmu(8) = 84.6d0
  Mcase65TAY_V_Qmu(9) = 84.6d0
  Mcase65TAY_V_Qmu(10) = 84.6d0
  Mcase65TAY_V_Qmu(11) = 84.6d0
  Mcase65TAY_V_Qmu(12) = 84.6d0
  Mcase65TAY_V_Qmu(13) = 84.6d0
  Mcase65TAY_V_Qmu(14) = 84.6d0
  Mcase65TAY_V_Qmu(15) = 84.6d0
  Mcase65TAY_V_Qmu(16) = 84.6d0
  Mcase65TAY_V_Qmu(17) = 84.6d0
  Mcase65TAY_V_Qmu(18) = 84.6d0
  Mcase65TAY_V_Qmu(19) = 84.6d0
  Mcase65TAY_V_Qmu(20) = 84.6d0
  Mcase65TAY_V_Qmu(21) = 84.6d0
  Mcase65TAY_V_Qmu(22) = 84.6d0
  Mcase65TAY_V_Qmu(23) = 84.6d0
  Mcase65TAY_V_Qmu(24) = 84.6d0
  Mcase65TAY_V_Qmu(25) = 84.6d0
  Mcase65TAY_V_Qmu(26) = 84.6d0
  Mcase65TAY_V_Qmu(27) = 84.6d0
  Mcase65TAY_V_Qmu(28) = 84.6d0
  Mcase65TAY_V_Qmu(29) = 84.6d0
  Mcase65TAY_V_Qmu(30) = 84.6d0
  Mcase65TAY_V_Qmu(31) = 84.6d0
  Mcase65TAY_V_Qmu(32) = 84.6d0
  Mcase65TAY_V_Qmu(33) = 84.6d0
  Mcase65TAY_V_Qmu(34) = 84.6d0
  Mcase65TAY_V_Qmu(35) = 84.6d0
  Mcase65TAY_V_Qmu(36) = 84.6d0
  Mcase65TAY_V_Qmu(37) = 84.6d0
  Mcase65TAY_V_Qmu(38) = 84.6d0
  Mcase65TAY_V_Qmu(39) = 84.6d0
  Mcase65TAY_V_Qmu(40) = 84.6d0
  Mcase65TAY_V_Qmu(41) = 84.6d0
  Mcase65TAY_V_Qmu(42) = 84.6d0
  Mcase65TAY_V_Qmu(43) = 84.6d0
  Mcase65TAY_V_Qmu(44) = 84.6d0
  Mcase65TAY_V_Qmu(45) = 84.6d0
  Mcase65TAY_V_Qmu(46) = 84.6d0
  Mcase65TAY_V_Qmu(47) = 84.6d0
  Mcase65TAY_V_Qmu(48) = 84.6d0
  Mcase65TAY_V_Qmu(49) = 84.6d0
  Mcase65TAY_V_Qmu(50) = 84.6d0
  Mcase65TAY_V_Qmu(51) = 84.6d0
  Mcase65TAY_V_Qmu(52) = 84.6d0
  Mcase65TAY_V_Qmu(53) = 84.6d0
  Mcase65TAY_V_Qmu(54) = 84.6d0
  Mcase65TAY_V_Qmu(55) = 84.6d0
  Mcase65TAY_V_Qmu(56) = 84.6d0
  Mcase65TAY_V_Qmu(57) = 84.6d0
  Mcase65TAY_V_Qmu(58) = 84.6d0
  Mcase65TAY_V_Qmu(59) = 84.6d0
  Mcase65TAY_V_Qmu(60) = 84.6d0
  Mcase65TAY_V_Qmu(61) = 84.6d0
  Mcase65TAY_V_Qmu(62) = 84.6d0
  Mcase65TAY_V_Qmu(63) = 84.6d0
  Mcase65TAY_V_Qmu(64) = 84.6d0
  Mcase65TAY_V_Qmu(65) = 84.6d0
  Mcase65TAY_V_Qmu(66) = 84.6d0
  Mcase65TAY_V_Qmu(67) = 84.6d0
  Mcase65TAY_V_Qmu(68) = 84.6d0
  Mcase65TAY_V_Qmu(69) = 84.6d0
  Mcase65TAY_V_Qmu(70) = 84.6d0
  Mcase65TAY_V_Qmu(71) = 84.6d0
  Mcase65TAY_V_Qmu(72) = 84.6d0
  Mcase65TAY_V_Qmu(73) = 84.6d0
  Mcase65TAY_V_Qmu(74) = 84.6d0
  Mcase65TAY_V_Qmu(75) = 84.6d0
  Mcase65TAY_V_Qmu(76) = 84.6d0
  Mcase65TAY_V_Qmu(77) = 84.6d0
  Mcase65TAY_V_Qmu(78) = 84.6d0
  Mcase65TAY_V_Qmu(79) = 84.6d0
  Mcase65TAY_V_Qmu(80) = 84.6d0
  Mcase65TAY_V_Qmu(81) = 84.6d0
  Mcase65TAY_V_Qmu(82) = 84.6d0
  Mcase65TAY_V_Qmu(83) = 84.6d0
  Mcase65TAY_V_Qmu(84) = 84.6d0
  Mcase65TAY_V_Qmu(85) = 84.6d0
  Mcase65TAY_V_Qmu(86) = 84.6d0
  Mcase65TAY_V_Qmu(87) = 84.6d0
  Mcase65TAY_V_Qmu(88) = 84.6d0
  Mcase65TAY_V_Qmu(89) = 84.6d0
  Mcase65TAY_V_Qmu(90) = 84.6d0
  Mcase65TAY_V_Qmu(91) = 84.6d0
  Mcase65TAY_V_Qmu(92) = 84.6d0
  Mcase65TAY_V_Qmu(93) = 84.6d0
  Mcase65TAY_V_Qmu(94) = 84.6d0
  Mcase65TAY_V_Qmu(95) = 84.6d0
  Mcase65TAY_V_Qmu(96) = 84.6d0
  Mcase65TAY_V_Qmu(97) = 84.6d0
  Mcase65TAY_V_Qmu(98) = 84.6d0
  Mcase65TAY_V_Qmu(99) = 84.6d0
  Mcase65TAY_V_Qmu(100) = 84.6d0
  Mcase65TAY_V_Qmu(101) = 84.6d0
  Mcase65TAY_V_Qmu(102) = 84.6d0
  Mcase65TAY_V_Qmu(103) = 84.6d0
  Mcase65TAY_V_Qmu(104) = 84.6d0  ! ICB inner core
  Mcase65TAY_V_Qmu(105) = 0.d0    ! ICB outer core
  Mcase65TAY_V_Qmu(106) = 0.d0
  Mcase65TAY_V_Qmu(107) = 0.d0
  Mcase65TAY_V_Qmu(108) = 0.d0
  Mcase65TAY_V_Qmu(109) = 0.d0
  Mcase65TAY_V_Qmu(110) = 0.d0
  Mcase65TAY_V_Qmu(111) = 0.d0
  Mcase65TAY_V_Qmu(112) = 0.d0
  Mcase65TAY_V_Qmu(113) = 0.d0
  Mcase65TAY_V_Qmu(114) = 0.d0
  Mcase65TAY_V_Qmu(115) = 0.d0
  Mcase65TAY_V_Qmu(116) = 0.d0
  Mcase65TAY_V_Qmu(117) = 0.d0
  Mcase65TAY_V_Qmu(118) = 0.d0
  Mcase65TAY_V_Qmu(119) = 0.d0
  Mcase65TAY_V_Qmu(120) = 0.d0
  Mcase65TAY_V_Qmu(121) = 0.d0
  Mcase65TAY_V_Qmu(122) = 0.d0
  Mcase65TAY_V_Qmu(123) = 0.d0
  Mcase65TAY_V_Qmu(124) = 0.d0
  Mcase65TAY_V_Qmu(125) = 0.d0
  Mcase65TAY_V_Qmu(126) = 0.d0
  Mcase65TAY_V_Qmu(127) = 0.d0
  Mcase65TAY_V_Qmu(128) = 0.d0
  Mcase65TAY_V_Qmu(129) = 0.d0
  Mcase65TAY_V_Qmu(130) = 0.d0
  Mcase65TAY_V_Qmu(131) = 0.d0
  Mcase65TAY_V_Qmu(132) = 0.d0
  Mcase65TAY_V_Qmu(133) = 0.d0
  Mcase65TAY_V_Qmu(134) = 0.d0
  Mcase65TAY_V_Qmu(135) = 0.d0
  Mcase65TAY_V_Qmu(136) = 0.d0
  Mcase65TAY_V_Qmu(137) = 0.d0
  Mcase65TAY_V_Qmu(138) = 0.d0
  Mcase65TAY_V_Qmu(139) = 0.d0
  Mcase65TAY_V_Qmu(140) = 0.d0
  Mcase65TAY_V_Qmu(141) = 0.d0
  Mcase65TAY_V_Qmu(142) = 0.d0
  Mcase65TAY_V_Qmu(143) = 0.d0
  Mcase65TAY_V_Qmu(144) = 0.d0
  Mcase65TAY_V_Qmu(145) = 0.d0
  Mcase65TAY_V_Qmu(146) = 0.d0
  Mcase65TAY_V_Qmu(147) = 0.d0
  Mcase65TAY_V_Qmu(148) = 0.d0
  Mcase65TAY_V_Qmu(149) = 0.d0
  Mcase65TAY_V_Qmu(150) = 0.d0
  Mcase65TAY_V_Qmu(151) = 0.d0
  Mcase65TAY_V_Qmu(152) = 0.d0
  Mcase65TAY_V_Qmu(153) = 0.d0
  Mcase65TAY_V_Qmu(154) = 0.d0
  Mcase65TAY_V_Qmu(155) = 0.d0
  Mcase65TAY_V_Qmu(156) = 0.d0
  Mcase65TAY_V_Qmu(157) = 0.d0
  Mcase65TAY_V_Qmu(158) = 0.d0
  Mcase65TAY_V_Qmu(159) = 0.d0
  Mcase65TAY_V_Qmu(160) = 0.d0
  Mcase65TAY_V_Qmu(161) = 0.d0
  Mcase65TAY_V_Qmu(162) = 0.d0
  Mcase65TAY_V_Qmu(163) = 0.d0
  Mcase65TAY_V_Qmu(164) = 0.d0
  Mcase65TAY_V_Qmu(165) = 0.d0
  Mcase65TAY_V_Qmu(166) = 0.d0
  Mcase65TAY_V_Qmu(167) = 0.d0
  Mcase65TAY_V_Qmu(168) = 0.d0
  Mcase65TAY_V_Qmu(169) = 0.d0
  Mcase65TAY_V_Qmu(170) = 0.d0
  Mcase65TAY_V_Qmu(171) = 0.d0
  Mcase65TAY_V_Qmu(172) = 0.d0
  Mcase65TAY_V_Qmu(173) = 0.d0
  Mcase65TAY_V_Qmu(174) = 0.d0
  Mcase65TAY_V_Qmu(175) = 0.d0
  Mcase65TAY_V_Qmu(176) = 0.d0
  Mcase65TAY_V_Qmu(177) = 0.d0
  Mcase65TAY_V_Qmu(178) = 0.d0
  Mcase65TAY_V_Qmu(179) = 0.d0
  Mcase65TAY_V_Qmu(180) = 0.d0
  Mcase65TAY_V_Qmu(181) = 0.d0
  Mcase65TAY_V_Qmu(182) = 0.d0
  Mcase65TAY_V_Qmu(183) = 0.d0
  Mcase65TAY_V_Qmu(184) = 0.d0
  Mcase65TAY_V_Qmu(185) = 0.d0
  Mcase65TAY_V_Qmu(186) = 0.d0
  Mcase65TAY_V_Qmu(187) = 0.d0
  Mcase65TAY_V_Qmu(188) = 0.d0
  Mcase65TAY_V_Qmu(189) = 0.d0
  Mcase65TAY_V_Qmu(190) = 0.d0
  Mcase65TAY_V_Qmu(191) = 0.d0
  Mcase65TAY_V_Qmu(192) = 0.d0
  Mcase65TAY_V_Qmu(193) = 0.d0
  Mcase65TAY_V_Qmu(194) = 0.d0
  Mcase65TAY_V_Qmu(195) = 0.d0
  Mcase65TAY_V_Qmu(196) = 0.d0
  Mcase65TAY_V_Qmu(197) = 0.d0
  Mcase65TAY_V_Qmu(198) = 0.d0
  Mcase65TAY_V_Qmu(199) = 0.d0
  Mcase65TAY_V_Qmu(200) = 0.d0
  Mcase65TAY_V_Qmu(201) = 0.d0
  Mcase65TAY_V_Qmu(202) = 0.d0
  Mcase65TAY_V_Qmu(203) = 0.d0
  Mcase65TAY_V_Qmu(204) = 0.d0
  Mcase65TAY_V_Qmu(205) = 0.d0
  Mcase65TAY_V_Qmu(206) = 0.d0
  Mcase65TAY_V_Qmu(207) = 0.d0
  Mcase65TAY_V_Qmu(208) = 0.d0
  Mcase65TAY_V_Qmu(209) = 0.d0
  Mcase65TAY_V_Qmu(210) = 0.d0
  Mcase65TAY_V_Qmu(211) = 0.d0
  Mcase65TAY_V_Qmu(212) = 0.d0
  Mcase65TAY_V_Qmu(213) = 0.d0
  Mcase65TAY_V_Qmu(214) = 0.d0
  Mcase65TAY_V_Qmu(215) = 0.d0
  Mcase65TAY_V_Qmu(216) = 0.d0
  Mcase65TAY_V_Qmu(217) = 0.d0
  Mcase65TAY_V_Qmu(218) = 0.d0
  Mcase65TAY_V_Qmu(219) = 0.d0
  Mcase65TAY_V_Qmu(220) = 0.d0
  Mcase65TAY_V_Qmu(221) = 0.d0
  Mcase65TAY_V_Qmu(222) = 0.d0
  Mcase65TAY_V_Qmu(223) = 0.d0
  Mcase65TAY_V_Qmu(224) = 0.d0
  Mcase65TAY_V_Qmu(225) = 0.d0
  Mcase65TAY_V_Qmu(226) = 0.d0
  Mcase65TAY_V_Qmu(227) = 0.d0
  Mcase65TAY_V_Qmu(228) = 0.d0
  Mcase65TAY_V_Qmu(229) = 0.d0
  Mcase65TAY_V_Qmu(230) = 0.d0
  Mcase65TAY_V_Qmu(231) = 0.d0
  Mcase65TAY_V_Qmu(232) = 0.d0
  Mcase65TAY_V_Qmu(233) = 0.d0
  Mcase65TAY_V_Qmu(234) = 0.d0
  Mcase65TAY_V_Qmu(235) = 0.d0
  Mcase65TAY_V_Qmu(236) = 0.d0
  Mcase65TAY_V_Qmu(237) = 0.d0
  Mcase65TAY_V_Qmu(238) = 0.d0
  Mcase65TAY_V_Qmu(239) = 0.d0
  Mcase65TAY_V_Qmu(240) = 0.d0
  Mcase65TAY_V_Qmu(241) = 0.d0
  Mcase65TAY_V_Qmu(242) = 0.d0
  Mcase65TAY_V_Qmu(243) = 0.d0
  Mcase65TAY_V_Qmu(244) = 0.d0
  Mcase65TAY_V_Qmu(245) = 0.d0
  Mcase65TAY_V_Qmu(246) = 0.d0
  Mcase65TAY_V_Qmu(247) = 0.d0
  Mcase65TAY_V_Qmu(248) = 0.d0
  Mcase65TAY_V_Qmu(249) = 0.d0
  Mcase65TAY_V_Qmu(250) = 0.d0
  Mcase65TAY_V_Qmu(251) = 0.d0
  Mcase65TAY_V_Qmu(252) = 0.d0
  Mcase65TAY_V_Qmu(253) = 0.d0
  Mcase65TAY_V_Qmu(254) = 0.d0
  Mcase65TAY_V_Qmu(255) = 0.d0
  Mcase65TAY_V_Qmu(256) = 0.d0
  Mcase65TAY_V_Qmu(257) = 0.d0
  Mcase65TAY_V_Qmu(258) = 0.d0
  Mcase65TAY_V_Qmu(259) = 0.d0
  Mcase65TAY_V_Qmu(260) = 0.d0
  Mcase65TAY_V_Qmu(261) = 0.d0
  Mcase65TAY_V_Qmu(262) = 0.d0
  Mcase65TAY_V_Qmu(263) = 0.d0
  Mcase65TAY_V_Qmu(264) = 0.d0
  Mcase65TAY_V_Qmu(265) = 0.d0
  Mcase65TAY_V_Qmu(266) = 0.d0
  Mcase65TAY_V_Qmu(267) = 0.d0
  Mcase65TAY_V_Qmu(268) = 0.d0
  Mcase65TAY_V_Qmu(269) = 0.d0
  Mcase65TAY_V_Qmu(270) = 0.d0
  Mcase65TAY_V_Qmu(271) = 0.d0
  Mcase65TAY_V_Qmu(272) = 0.d0
  Mcase65TAY_V_Qmu(273) = 0.d0
  Mcase65TAY_V_Qmu(274) = 0.d0
  Mcase65TAY_V_Qmu(275) = 0.d0
  Mcase65TAY_V_Qmu(276) = 0.d0
  Mcase65TAY_V_Qmu(277) = 0.d0
  Mcase65TAY_V_Qmu(278) = 0.d0
  Mcase65TAY_V_Qmu(279) = 0.d0
  Mcase65TAY_V_Qmu(280) = 0.d0
  Mcase65TAY_V_Qmu(281) = 0.d0
  Mcase65TAY_V_Qmu(282) = 0.d0
  Mcase65TAY_V_Qmu(283) = 0.d0
  Mcase65TAY_V_Qmu(284) = 0.d0
  Mcase65TAY_V_Qmu(285) = 0.d0
  Mcase65TAY_V_Qmu(286) = 0.d0
  Mcase65TAY_V_Qmu(287) = 0.d0
  Mcase65TAY_V_Qmu(288) = 0.d0
  Mcase65TAY_V_Qmu(289) = 0.d0
  Mcase65TAY_V_Qmu(290) = 0.d0
  Mcase65TAY_V_Qmu(291) = 0.d0
  Mcase65TAY_V_Qmu(292) = 0.d0
  Mcase65TAY_V_Qmu(293) = 0.d0
  Mcase65TAY_V_Qmu(294) = 0.d0
  Mcase65TAY_V_Qmu(295) = 0.d0
  Mcase65TAY_V_Qmu(296) = 0.d0
  Mcase65TAY_V_Qmu(297) = 0.d0
  Mcase65TAY_V_Qmu(298) = 0.d0
  Mcase65TAY_V_Qmu(299) = 0.d0
  Mcase65TAY_V_Qmu(300) = 0.d0
  Mcase65TAY_V_Qmu(301) = 0.d0
  Mcase65TAY_V_Qmu(302) = 0.d0
  Mcase65TAY_V_Qmu(303) = 0.d0
  Mcase65TAY_V_Qmu(304) = 0.d0
  Mcase65TAY_V_Qmu(305) = 0.d0
  Mcase65TAY_V_Qmu(306) = 0.d0
  Mcase65TAY_V_Qmu(307) = 0.d0
  Mcase65TAY_V_Qmu(308) = 0.d0
  Mcase65TAY_V_Qmu(309) = 0.d0
  Mcase65TAY_V_Qmu(310) = 0.d0
  Mcase65TAY_V_Qmu(311) = 0.d0
  Mcase65TAY_V_Qmu(312) = 0.d0
  Mcase65TAY_V_Qmu(313) = 0.d0
  Mcase65TAY_V_Qmu(314) = 0.d0
  Mcase65TAY_V_Qmu(315) = 0.d0
  Mcase65TAY_V_Qmu(316) = 0.d0
  Mcase65TAY_V_Qmu(317) = 0.d0
  Mcase65TAY_V_Qmu(318) = 0.d0
  Mcase65TAY_V_Qmu(319) = 0.d0
  Mcase65TAY_V_Qmu(320) = 0.d0
  Mcase65TAY_V_Qmu(321) = 0.d0
  Mcase65TAY_V_Qmu(322) = 0.d0
  Mcase65TAY_V_Qmu(323) = 0.d0
  Mcase65TAY_V_Qmu(324) = 0.d0
  Mcase65TAY_V_Qmu(325) = 0.d0
  Mcase65TAY_V_Qmu(326) = 0.d0
  Mcase65TAY_V_Qmu(327) = 0.d0
  Mcase65TAY_V_Qmu(328) = 0.d0
  Mcase65TAY_V_Qmu(329) = 0.d0
  Mcase65TAY_V_Qmu(330) = 0.d0
  Mcase65TAY_V_Qmu(331) = 0.d0
  Mcase65TAY_V_Qmu(332) = 0.d0
  Mcase65TAY_V_Qmu(333) = 0.d0
  Mcase65TAY_V_Qmu(334) = 0.d0
  Mcase65TAY_V_Qmu(335) = 0.d0
  Mcase65TAY_V_Qmu(336) = 0.d0
  Mcase65TAY_V_Qmu(337) = 0.d0
  Mcase65TAY_V_Qmu(338) = 0.d0
  Mcase65TAY_V_Qmu(339) = 0.d0
  Mcase65TAY_V_Qmu(340) = 0.d0
  Mcase65TAY_V_Qmu(341) = 0.d0
  Mcase65TAY_V_Qmu(342) = 0.d0
  Mcase65TAY_V_Qmu(343) = 0.d0
  Mcase65TAY_V_Qmu(344) = 0.d0
  Mcase65TAY_V_Qmu(345) = 0.d0
  Mcase65TAY_V_Qmu(346) = 0.d0
  Mcase65TAY_V_Qmu(347) = 0.d0
  Mcase65TAY_V_Qmu(348) = 0.d0
  Mcase65TAY_V_Qmu(349) = 0.d0
  Mcase65TAY_V_Qmu(350) = 0.d0
  Mcase65TAY_V_Qmu(351) = 0.d0
  Mcase65TAY_V_Qmu(352) = 0.d0
  Mcase65TAY_V_Qmu(353) = 0.d0
  Mcase65TAY_V_Qmu(354) = 0.d0
  Mcase65TAY_V_Qmu(355) = 0.d0
  Mcase65TAY_V_Qmu(356) = 0.d0
  Mcase65TAY_V_Qmu(357) = 0.d0
  Mcase65TAY_V_Qmu(358) = 0.d0
  Mcase65TAY_V_Qmu(359) = 0.d0
  Mcase65TAY_V_Qmu(360) = 0.d0
  Mcase65TAY_V_Qmu(361) = 0.d0
  Mcase65TAY_V_Qmu(362) = 0.d0
  Mcase65TAY_V_Qmu(363) = 0.d0
  Mcase65TAY_V_Qmu(364) = 0.d0
  Mcase65TAY_V_Qmu(365) = 0.d0
  Mcase65TAY_V_Qmu(366) = 0.d0
  Mcase65TAY_V_Qmu(367) = 0.d0
  Mcase65TAY_V_Qmu(368) = 0.d0
  Mcase65TAY_V_Qmu(369) = 0.d0
  Mcase65TAY_V_Qmu(370) = 0.d0   ! CMB outer core
  Mcase65TAY_V_Qmu(371) = 312.d0 ! CMB mantle
  Mcase65TAY_V_Qmu(372) = 312.d0
  Mcase65TAY_V_Qmu(373) = 312.d0
  Mcase65TAY_V_Qmu(374) = 312.d0
  Mcase65TAY_V_Qmu(375) = 312.d0
  Mcase65TAY_V_Qmu(376) = 312.d0
  Mcase65TAY_V_Qmu(377) = 312.d0
  Mcase65TAY_V_Qmu(378) = 312.d0
  Mcase65TAY_V_Qmu(379) = 312.d0
  Mcase65TAY_V_Qmu(380) = 312.d0
  Mcase65TAY_V_Qmu(381) = 312.d0
  Mcase65TAY_V_Qmu(382) = 312.d0
  Mcase65TAY_V_Qmu(383) = 312.d0
  Mcase65TAY_V_Qmu(384) = 312.d0
  Mcase65TAY_V_Qmu(385) = 312.d0
  Mcase65TAY_V_Qmu(386) = 312.d0
  Mcase65TAY_V_Qmu(387) = 312.d0
  Mcase65TAY_V_Qmu(388) = 312.d0
  Mcase65TAY_V_Qmu(389) = 312.d0
  Mcase65TAY_V_Qmu(390) = 312.d0
  Mcase65TAY_V_Qmu(391) = 312.d0
  Mcase65TAY_V_Qmu(392) = 312.d0
  Mcase65TAY_V_Qmu(393) = 312.d0
  Mcase65TAY_V_Qmu(394) = 312.d0
  Mcase65TAY_V_Qmu(395) = 312.d0
  Mcase65TAY_V_Qmu(396) = 312.d0
  Mcase65TAY_V_Qmu(397) = 312.d0
  Mcase65TAY_V_Qmu(398) = 312.d0
  Mcase65TAY_V_Qmu(399) = 312.d0
  Mcase65TAY_V_Qmu(400) = 312.d0
  Mcase65TAY_V_Qmu(401) = 312.d0
  Mcase65TAY_V_Qmu(402) = 312.d0
  Mcase65TAY_V_Qmu(403) = 312.d0
  Mcase65TAY_V_Qmu(404) = 312.d0
  Mcase65TAY_V_Qmu(405) = 312.d0
  Mcase65TAY_V_Qmu(406) = 312.d0
  Mcase65TAY_V_Qmu(407) = 312.d0
  Mcase65TAY_V_Qmu(408) = 312.d0
  Mcase65TAY_V_Qmu(409) = 312.d0
  Mcase65TAY_V_Qmu(410) = 312.d0
  Mcase65TAY_V_Qmu(411) = 312.d0
  Mcase65TAY_V_Qmu(412) = 312.d0
  Mcase65TAY_V_Qmu(413) = 312.d0
  Mcase65TAY_V_Qmu(414) = 312.d0
  Mcase65TAY_V_Qmu(415) = 312.d0
  Mcase65TAY_V_Qmu(416) = 312.d0
  Mcase65TAY_V_Qmu(417) = 312.d0
  Mcase65TAY_V_Qmu(418) = 312.d0
  Mcase65TAY_V_Qmu(419) = 312.d0
  Mcase65TAY_V_Qmu(420) = 312.d0
  Mcase65TAY_V_Qmu(421) = 312.d0
  Mcase65TAY_V_Qmu(422) = 312.d0
  Mcase65TAY_V_Qmu(423) = 312.d0
  Mcase65TAY_V_Qmu(424) = 312.d0
  Mcase65TAY_V_Qmu(425) = 312.d0
  Mcase65TAY_V_Qmu(426) = 312.d0
  Mcase65TAY_V_Qmu(427) = 312.d0
  Mcase65TAY_V_Qmu(428) = 312.d0
  Mcase65TAY_V_Qmu(429) = 312.d0
  Mcase65TAY_V_Qmu(430) = 312.d0
  Mcase65TAY_V_Qmu(431) = 312.d0
  Mcase65TAY_V_Qmu(432) = 312.d0
  Mcase65TAY_V_Qmu(433) = 312.d0
  Mcase65TAY_V_Qmu(434) = 312.d0
  Mcase65TAY_V_Qmu(435) = 312.d0
  Mcase65TAY_V_Qmu(436) = 312.d0
  Mcase65TAY_V_Qmu(437) = 312.d0
  Mcase65TAY_V_Qmu(438) = 312.d0
  Mcase65TAY_V_Qmu(439) = 312.d0
  Mcase65TAY_V_Qmu(440) = 312.d0
  Mcase65TAY_V_Qmu(441) = 312.d0
  Mcase65TAY_V_Qmu(442) = 312.d0
  Mcase65TAY_V_Qmu(443) = 312.d0
  Mcase65TAY_V_Qmu(444) = 312.d0
  Mcase65TAY_V_Qmu(445) = 312.d0
  Mcase65TAY_V_Qmu(446) = 312.d0
  Mcase65TAY_V_Qmu(447) = 312.d0
  Mcase65TAY_V_Qmu(448) = 312.d0
  Mcase65TAY_V_Qmu(449) = 312.d0
  Mcase65TAY_V_Qmu(450) = 312.d0
  Mcase65TAY_V_Qmu(451) = 312.d0
  Mcase65TAY_V_Qmu(452) = 312.d0
  Mcase65TAY_V_Qmu(453) = 312.d0
  Mcase65TAY_V_Qmu(454) = 312.d0
  Mcase65TAY_V_Qmu(455) = 312.d0
  Mcase65TAY_V_Qmu(456) = 312.d0
  Mcase65TAY_V_Qmu(457) = 312.d0
  Mcase65TAY_V_Qmu(458) = 312.d0
  Mcase65TAY_V_Qmu(459) = 312.d0
  Mcase65TAY_V_Qmu(460) = 312.d0
  Mcase65TAY_V_Qmu(461) = 312.d0
  Mcase65TAY_V_Qmu(462) = 312.d0
  Mcase65TAY_V_Qmu(463) = 312.d0
  Mcase65TAY_V_Qmu(464) = 312.d0
  Mcase65TAY_V_Qmu(465) = 312.d0
  Mcase65TAY_V_Qmu(466) = 312.d0
  Mcase65TAY_V_Qmu(467) = 312.d0
  Mcase65TAY_V_Qmu(468) = 312.d0
  Mcase65TAY_V_Qmu(469) = 312.d0
  Mcase65TAY_V_Qmu(470) = 312.d0
  Mcase65TAY_V_Qmu(471) = 312.d0
  Mcase65TAY_V_Qmu(472) = 312.d0
  Mcase65TAY_V_Qmu(473) = 312.d0
  Mcase65TAY_V_Qmu(474) = 312.d0
  Mcase65TAY_V_Qmu(475) = 143.d0
  Mcase65TAY_V_Qmu(476) = 143.d0
  Mcase65TAY_V_Qmu(477) = 143.d0
  Mcase65TAY_V_Qmu(478) = 143.d0
  Mcase65TAY_V_Qmu(479) = 143.d0
  Mcase65TAY_V_Qmu(480) = 143.d0
  Mcase65TAY_V_Qmu(481) = 143.d0
  Mcase65TAY_V_Qmu(482) = 143.d0
  Mcase65TAY_V_Qmu(483) = 143.d0
  Mcase65TAY_V_Qmu(484) = 143.d0
  Mcase65TAY_V_Qmu(485) = 143.d0
  Mcase65TAY_V_Qmu(486) = 143.d0
  Mcase65TAY_V_Qmu(487) = 143.d0
  Mcase65TAY_V_Qmu(488) = 143.d0
  Mcase65TAY_V_Qmu(489) = 143.d0
  Mcase65TAY_V_Qmu(490) = 143.d0
  Mcase65TAY_V_Qmu(491) = 143.d0
  Mcase65TAY_V_Qmu(492) = 143.d0
  Mcase65TAY_V_Qmu(493) = 143.d0
  Mcase65TAY_V_Qmu(494) = 143.d0
  Mcase65TAY_V_Qmu(495) = 143.d0
  Mcase65TAY_V_Qmu(496) = 143.d0
  Mcase65TAY_V_Qmu(497) = 143.d0
  Mcase65TAY_V_Qmu(498) = 143.d0
  Mcase65TAY_V_Qmu(499) = 143.d0
  Mcase65TAY_V_Qmu(500) = 143.d0
  Mcase65TAY_V_Qmu(501) = 143.d0
  Mcase65TAY_V_Qmu(502) = 143.d0
  Mcase65TAY_V_Qmu(503) = 143.d0
  Mcase65TAY_V_Qmu(504) = 143.d0
  Mcase65TAY_V_Qmu(505) = 143.d0
  Mcase65TAY_V_Qmu(506) = 143.d0
  Mcase65TAY_V_Qmu(507) = 143.d0
  Mcase65TAY_V_Qmu(508) = 143.d0
  Mcase65TAY_V_Qmu(509) = 143.d0
  Mcase65TAY_V_Qmu(510) = 143.d0
  Mcase65TAY_V_Qmu(511) = 143.d0
  Mcase65TAY_V_Qmu(512) = 143.d0
  Mcase65TAY_V_Qmu(513) = 143.d0
  Mcase65TAY_V_Qmu(514) = 143.d0
  Mcase65TAY_V_Qmu(515) = 143.d0
  Mcase65TAY_V_Qmu(516) = 143.d0
  Mcase65TAY_V_Qmu(517) = 143.d0
  Mcase65TAY_V_Qmu(518) = 143.d0
  Mcase65TAY_V_Qmu(519) = 143.d0
  Mcase65TAY_V_Qmu(520) = 143.d0
  Mcase65TAY_V_Qmu(521) = 143.d0
  Mcase65TAY_V_Qmu(522) = 143.d0
  Mcase65TAY_V_Qmu(523) = 143.d0
  Mcase65TAY_V_Qmu(524) = 143.d0
  Mcase65TAY_V_Qmu(525) = 143.d0
  Mcase65TAY_V_Qmu(526) = 143.d0
  Mcase65TAY_V_Qmu(527) = 143.d0
  Mcase65TAY_V_Qmu(528) = 143.d0
  Mcase65TAY_V_Qmu(529) = 143.d0
  Mcase65TAY_V_Qmu(530) = 143.d0
  Mcase65TAY_V_Qmu(531) = 143.d0
  Mcase65TAY_V_Qmu(532) = 143.d0
  Mcase65TAY_V_Qmu(533) = 143.d0
  Mcase65TAY_V_Qmu(534) = 80.d0
  Mcase65TAY_V_Qmu(535) = 80.d0
  Mcase65TAY_V_Qmu(536) = 80.d0
  Mcase65TAY_V_Qmu(537) = 80.d0
  Mcase65TAY_V_Qmu(538) = 80.d0
  Mcase65TAY_V_Qmu(539) = 80.d0
  Mcase65TAY_V_Qmu(540) = 80.d0
  Mcase65TAY_V_Qmu(541) = 80.d0
  Mcase65TAY_V_Qmu(542) = 80.d0
  Mcase65TAY_V_Qmu(543) = 80.d0
  Mcase65TAY_V_Qmu(544) = 80.d0
  Mcase65TAY_V_Qmu(545) = 80.d0
  Mcase65TAY_V_Qmu(546) = 80.d0
  Mcase65TAY_V_Qmu(547) = 80.d0
  Mcase65TAY_V_Qmu(548) = 80.d0
  Mcase65TAY_V_Qmu(549) = 80.d0
  Mcase65TAY_V_Qmu(550) = 80.d0
  Mcase65TAY_V_Qmu(551) = 80.d0
  Mcase65TAY_V_Qmu(552) = 80.d0
  Mcase65TAY_V_Qmu(553) = 80.d0
  Mcase65TAY_V_Qmu(554) = 80.d0
  Mcase65TAY_V_Qmu(555) = 80.d0
  Mcase65TAY_V_Qmu(556) = 80.d0
  Mcase65TAY_V_Qmu(557) = 80.d0
  Mcase65TAY_V_Qmu(558) = 80.d0
  Mcase65TAY_V_Qmu(559) = 80.d0
  Mcase65TAY_V_Qmu(560) = 80.d0
  Mcase65TAY_V_Qmu(561) = 80.d0
  Mcase65TAY_V_Qmu(562) = 80.d0
  Mcase65TAY_V_Qmu(563) = 80.d0
  Mcase65TAY_V_Qmu(564) = 80.d0
  Mcase65TAY_V_Qmu(565) = 80.d0
  Mcase65TAY_V_Qmu(566) = 80.d0
  Mcase65TAY_V_Qmu(567) = 80.d0
  Mcase65TAY_V_Qmu(568) = 80.d0
  Mcase65TAY_V_Qmu(569) = 80.d0
  Mcase65TAY_V_Qmu(570) = 80.d0
  Mcase65TAY_V_Qmu(571) = 80.d0
  Mcase65TAY_V_Qmu(572) = 80.d0
  Mcase65TAY_V_Qmu(573) = 80.d0
  Mcase65TAY_V_Qmu(574) = 80.d0
  Mcase65TAY_V_Qmu(575) = 80.d0
  Mcase65TAY_V_Qmu(576) = 80.d0
  Mcase65TAY_V_Qmu(577) = 80.d0
  Mcase65TAY_V_Qmu(578) = 80.d0
  Mcase65TAY_V_Qmu(579) = 80.d0
  Mcase65TAY_V_Qmu(580) = 80.d0
  Mcase65TAY_V_Qmu(581) = 80.d0
  Mcase65TAY_V_Qmu(582) = 80.d0
  Mcase65TAY_V_Qmu(583) = 80.d0
  Mcase65TAY_V_Qmu(584) = 80.d0
  Mcase65TAY_V_Qmu(585) = 80.d0
  Mcase65TAY_V_Qmu(586) = 80.d0
  Mcase65TAY_V_Qmu(587) = 80.d0
  Mcase65TAY_V_Qmu(588) = 80.d0
  Mcase65TAY_V_Qmu(589) = 80.d0
  Mcase65TAY_V_Qmu(590) = 80.d0
  Mcase65TAY_V_Qmu(591) = 80.d0
  Mcase65TAY_V_Qmu(592) = 80.d0
  Mcase65TAY_V_Qmu(593) = 80.d0
  Mcase65TAY_V_Qmu(594) = 80.d0
  Mcase65TAY_V_Qmu(595) = 80.d0
  Mcase65TAY_V_Qmu(596) = 80.d0
  Mcase65TAY_V_Qmu(597) = 80.d0
  Mcase65TAY_V_Qmu(598) = 80.d0
  Mcase65TAY_V_Qmu(599) = 80.d0
  Mcase65TAY_V_Qmu(600) = 80.d0
  Mcase65TAY_V_Qmu(601) = 80.d0
  Mcase65TAY_V_Qmu(602) = 80.d0
  Mcase65TAY_V_Qmu(603) = 80.d0
  Mcase65TAY_V_Qmu(604) = 80.d0
  Mcase65TAY_V_Qmu(605) = 80.d0
  Mcase65TAY_V_Qmu(606) = 80.d0
  Mcase65TAY_V_Qmu(607) = 80.d0
  Mcase65TAY_V_Qmu(608) = 80.d0
  Mcase65TAY_V_Qmu(609) = 80.d0
  Mcase65TAY_V_Qmu(610) = 80.d0
  Mcase65TAY_V_Qmu(611) = 80.d0
  Mcase65TAY_V_Qmu(612) = 80.d0
  Mcase65TAY_V_Qmu(613) = 80.d0
  Mcase65TAY_V_Qmu(614) = 80.d0
  Mcase65TAY_V_Qmu(615) = 600.d0
  Mcase65TAY_V_Qmu(616) = 600.d0
  Mcase65TAY_V_Qmu(617) = 600.d0
  Mcase65TAY_V_Qmu(618) = 600.d0
  Mcase65TAY_V_Qmu(619) = 600.d0
  Mcase65TAY_V_Qmu(620) = 600.d0
  Mcase65TAY_V_Qmu(621) = 600.d0
  Mcase65TAY_V_Qmu(622) = 600.d0
  Mcase65TAY_V_Qmu(623) = 600.d0
  Mcase65TAY_V_Qmu(624) = 600.d0
  Mcase65TAY_V_Qmu(625) = 600.d0
  Mcase65TAY_V_Qmu(626) = 600.d0
  Mcase65TAY_V_Qmu(627) = 600.d0
  Mcase65TAY_V_Qmu(628) = 600.d0
  Mcase65TAY_V_Qmu(629) = 600.d0
  Mcase65TAY_V_Qmu(630) = 600.d0
  Mcase65TAY_V_Qmu(631) = 600.d0
  Mcase65TAY_V_Qmu(632) = 600.d0
  Mcase65TAY_V_Qmu(633) = 600.d0
  Mcase65TAY_V_Qmu(634) = 600.d0
  Mcase65TAY_V_Qmu(635) = 600.d0
  Mcase65TAY_V_Qmu(636) = 600.d0
  Mcase65TAY_V_Qmu(637) = 600.d0
  Mcase65TAY_V_Qmu(638) = 600.d0
  Mcase65TAY_V_Qmu(639) = 600.d0
  Mcase65TAY_V_Qmu(640) = 600.d0
  Mcase65TAY_V_Qmu(641) = 600.d0
  Mcase65TAY_V_Qmu(642) = 600.d0
  Mcase65TAY_V_Qmu(643) = 600.d0
  Mcase65TAY_V_Qmu(644) = 600.d0
  Mcase65TAY_V_Qmu(645) = 600.d0
  Mcase65TAY_V_Qmu(646) = 600.d0
  Mcase65TAY_V_Qmu(647) = 600.d0
  Mcase65TAY_V_Qmu(648) = 600.d0
  Mcase65TAY_V_Qmu(649) = 600.d0
  Mcase65TAY_V_Qmu(650) = 600.d0
  Mcase65TAY_V_Qmu(651) = 600.d0
  Mcase65TAY_V_Qmu(652) = 600.d0
  Mcase65TAY_V_Qmu(653) = 600.d0
  Mcase65TAY_V_Qmu(654) = 600.d0
  Mcase65TAY_V_Qmu(655) = 600.d0
  Mcase65TAY_V_Qmu(656) = 600.d0
  Mcase65TAY_V_Qmu(657) = 600.d0
  Mcase65TAY_V_Qmu(658) = 600.d0
  Mcase65TAY_V_Qmu(659) = 600.d0
  Mcase65TAY_V_Qmu(660) = 600.d0
  Mcase65TAY_V_Qmu(661) = 600.d0
  Mcase65TAY_V_Qmu(662) = 600.d0
  Mcase65TAY_V_Qmu(663) = 600.d0
  Mcase65TAY_V_Qmu(664) = 600.d0
  Mcase65TAY_V_Qmu(665) = 600.d0
  Mcase65TAY_V_Qmu(666) = 600.d0
  Mcase65TAY_V_Qmu(667) = 600.d0
  Mcase65TAY_V_Qmu(668) = 600.d0
  Mcase65TAY_V_Qmu(669) = 600.d0
  Mcase65TAY_V_Qmu(670) = 600.d0
  Mcase65TAY_V_Qmu(671) = 600.d0
  Mcase65TAY_V_Qmu(672) = 600.d0
  Mcase65TAY_V_Qmu(673) = 600.d0
  Mcase65TAY_V_Qmu(674) = 600.d0
  Mcase65TAY_V_Qmu(675) = 600.d0
  Mcase65TAY_V_Qmu(676) = 600.d0
  Mcase65TAY_V_Qmu(677) = 600.d0
  Mcase65TAY_V_Qmu(678) = 600.d0
  Mcase65TAY_V_Qmu(679) = 600.d0
  Mcase65TAY_V_Qmu(680) = 600.d0
  Mcase65TAY_V_Qmu(681) = 600.d0


  ! strip the crust and replace it by mantle if we use an external crustal model
  !
  ! note: for this model, there is a smooth transition from mantle to crustal models.
  !       the RMOHO is set at 3280000.d0, that is at a depth of 110 km.
  !       this would correspond to index 659 -> NRcase65TAY = 681 - 22
  !
  ! change in case a sharp transition is wanted, otherwise we keep the smooth transition
  if (.false.) then
      if (SUPPRESS_CRUSTAL_MESH .or. USE_EXTERNAL_CRUSTAL_MODEL) then
      do i = NR_case65TAY - 22,NR_case65TAY
        Mcase65TAY_V_density(i) = Mcase65TAY_V_density(NR_case65TAY-23)
        Mcase65TAY_V_vp(i) = Mcase65TAY_V_vp(NR_case65TAY-23)
        Mcase65TAY_V_vs(i) = Mcase65TAY_V_vs(NR_case65TAY-23)
        Mcase65TAY_V_Qkappa(i) = Mcase65TAY_V_Qkappa(NR_case65TAY-23)
        Mcase65TAY_V_Qmu(i) = Mcase65TAY_V_Qmu(NR_case65TAY-23)
      enddo
    endif
  endif

  ! checks
  ! ICB at radius(104)
  if (abs(RICB - Mcase65TAY_V_radius(104)) > TOL) &
    call exit_MPI(myrank,'Error: radius RICB and model radius index not matching')
  if (abs(RCMB - Mcase65TAY_V_radius(370)) > TOL) &
    call exit_MPI(myrank,'Error: radius RCMB and model radius index not matching')
  if (abs(ROCEAN - Mcase65TAY_V_radius(681)) > TOL) &
    call exit_MPI(myrank,'Error: radius ROCEAN and model radius index not matching')

  end subroutine define_model_case65TAY

